#include "dime.h"

#ifdef STANDALONE
struct user_bdy{ double unused; };
struct user_node {
	double qold, qnew, r;
	double eigen;
	int nneigh, rep;
};
struct user_elmt { 
	int up;
	int dummy;
};

user_main(user_menu, argc, argv)
MNUPTR user_menu;
int argc;
char **argv;
{
	int Ising(), Makelb(), Params(), Eigen(), showeigen(), showgroups();
	DIME_PREAMBLE;
	menu_add_function(user_menu, "Eigen", Eigen);
	menu_add_function(user_menu, "Showeigen", showeigen);
	menu_add_function(user_menu, "Showgroups", showgroups);
	menu_add_function(user_menu, "Params", Params);
	menu_add_function(user_menu, "Makelb", Makelb);
}
#else
#include "lanczos.h"
#endif

#define MAXLANCZ	100
double a[MAXLANCZ], b[MAXLANCZ], evec[MAXLANCZ];

#define XE(e) (e->neigh[0]->x+e->neigh[1]->x+e->neigh[2]->x)/3
#define YE(e) (e->neigh[0]->y+e->neigh[1]->y+e->neigh[2]->y)/3

int nelmt;
int nnode;					/* Global number of elements */
int steps_per_eigen = 15;
int maxiter = 3;

int *proclist = (int *)0;
int nlist = 0;
int chan = 0;
int combtype = 256;
double mid;

#define GROUP_EDGE(node) \
	(node->b != (BDYPTR)0 && (\
		   (!node->b->oppalias) \
		|| (node->b->oppalias != (ALIASPTR)0 && !inmygroup(node->b->oppalias->proc)) \
	))

#define SUB_COMBINE(variable, func, number) \
	excombine(&variable, func, sizeof(variable), number, \
		nlist, proclist, &combtype)

#define SUB_NODE_COMBINE(member, func, number) \
	{struct user_node dummy; \
		node_combine((char *)(&dummy.member) - (char *)(&dummy), \
		sizeof(dummy.member), number, func, 0, nlist, proclist); \
	}
			 
Params()
{
	printf("steps per whack (currently %d max %d)\n", steps_per_eigen, MAXLANCZ);
	get_int(&steps_per_eigen);

	printf("iterations (currently %d)\n", maxiter);
	get_int(&maxiter);
}

Eigen()
{
	NODEPTR node;
	ELMTPTR elmt;
	double get_eigen(), err, min, max;
	int up, down, iter, flip, newproc, moved[2];
	float *z;

	z = (float *)malloc(steps_per_eigen*steps_per_eigen*sizeof(float));
	if(!z)
		abort(1234);
#ifdef SEQ
	chan = 0;
#else
	for(chan=max_doc; chan >0; chan--)
#endif
	{
		Makelb();
		err = 1.0;
		iter = 0;
		do {
			fmulti(stdout);
			if(err > 0.001)
				err = get_eigen(steps_per_eigen, z);
			if(procnum == proclist[0])
				printf("%6.3f ", err);
			fsingl(stdout);
			printf("\n");
			fflush(stdout);
		} while(++iter < maxiter);
		if(nelmt < 5){
			FORALLELMTS(elmt)
				set_balance(elmt, procnum);
			NEXTELMT(elmt)
		} else {
			max = -1.e12;
			min = 1.e12;
			FORALLNODES(node)
				if(node->user->eigen > max) max = node->user->eigen;
				if(node->user->eigen < min) min = node->user->eigen;
			NEXTNODE(node)

			SUB_COMBINE(min, fmin, 1);
			SUB_COMBINE(max, fmax, 1);

			iter = 0;
			do {
				mid = (max + min)/2;
				up = 0;
				FORALLELMTS(elmt)
					elmt->user->up = 
						((elmt->neigh[0]->user->eigen > mid) +
						 (elmt->neigh[1]->user->eigen > mid) +
						 (elmt->neigh[2]->user->eigen > mid)) > 1;
					up += elmt->user->up;
				NEXTELMT(elmt)
				SUB_COMBINE(up, isum, 1);
				down = nelmt - up;
				if(up < down)
					max = mid;
				else
					min = mid;
			} while(abs(up - down) > 1 && iter++ < 20);

			moved[0] = moved[1] = 0;
			FORALLELMTS(elmt)
				for(flip=0; flip<2; flip++){
					newproc = 
						((elmt->user->up) ^ flip) ? 
							(procnum & (~(1<<(chan-1)))) :
							(procnum & (~(1<<(chan-1))))|(1<<(chan-1));
					moved[flip] += (newproc != procnum);
				}
			NEXTELMT(elmt)

			SUB_COMBINE(moved[0], isum, 2);

			flip = (moved[0] < moved[1]) ? 0:1;

			FORALLELMTS(elmt)
				newproc = 
					((elmt->user->up) ^ flip) ? 
						(procnum & (~(1<<(chan-1)))) :
						(procnum & (~(1<<(chan-1))))|(1<<(chan-1));
				set_balance(elmt, newproc);
			NEXTELMT(elmt)
		}

		balance();
	}
	free(z);
}

double get_eigen(niter, z)
int niter;
float *z;
{
	NODEPTR node, eneigh;
	int lancz, j, k, ksecond;
	double dot(), magr(), second, oldsecond = 0;
	double derr, err;

	if(nelmt < 5){
		FORALLNODES(node)
			node->user->eigen = 0.0;
		NEXTNODE(node)
		return 0.0;
	}

	if(niter > nnode)
		niter = nnode;

	FORALLNODES(node)
		node->user->qold = 0;
		node->user->qnew = node->user->eigen;
	NEXTNODE(node)

	orthog();

	FORALLNODES(node)
		node->user->eigen = node->user->r = node->user->qnew;
	NEXTNODE(node)

	b[0] = sqrt(magr());

	FORALLNODES(node)
		node->user->qnew = node->user->r / b[0];
	NEXTNODE(node)
	orthog();

	for(lancz=0; lancz<niter; lancz++){
		matmult();

		FORALLNODES(node)
			node->user->r -= b[lancz] * node->user->qold;
		NEXTNODE(node)

		a[lancz] = dot();

		FORALLNODES(node)
			node->user->r -= a[lancz] * node->user->qnew;
		NEXTNODE(node)

		b[lancz+1] = sqrt(magr());
		if(fabs(b[lancz+1]) < 1.e-8){
			niter = lancz;
			break;
		}

		FORALLNODES(node)
			node->user->qold = node->user->qnew;
			node->user->qnew = node->user->r / b[lancz+1];
		NEXTNODE(node)
		orthog();
	}

	if(niter == 0)
		return 0.0;
		
	tqli(a, b, niter, z);

	second = 1.e12;
	for(k=0; k<niter; k++)
		if(a[k] < second){
			second = a[k];
			ksecond = k;
		}
	for(k=0; k<niter; k++)
		evec[k] = z[k*niter + ksecond];

	FORALLNODES(node)
		node->user->r = node->user->eigen;
		node->user->qold = node->user->eigen = 0;
	NEXTNODE(node)

	b[0] = sqrt(magr());

	FORALLNODES(node)
		node->user->qnew = node->user->r / b[0];
	NEXTNODE(node)
	orthog();

	for(j=0; j<niter; j++){
		FORALLNODES(node)
			node->user->eigen += evec[j] * node->user->qnew;
		NEXTNODE(node)

		matmult();

		FORALLNODES(node)
			node->user->r -= b[j] * node->user->qold;
		NEXTNODE(node)

		a[j] = dot();

		FORALLNODES(node)
			node->user->r -= a[j] * node->user->qnew;
		NEXTNODE(node)
		orthog();

		b[j+1] = sqrt(magr());

		FORALLNODES(node)
			node->user->qold = node->user->qnew;
			node->user->qnew = node->user->r / b[j+1];
		NEXTNODE(node)
	}

	FORALLNODES(node)
		node->user->qnew = node->user->eigen;
	NEXTNODE(node)

	matmult();

	err = 0;
	FORALLNODES(node)
		derr = node->user->r - node->user->eigen * second;
		err += derr*derr;
	NEXTNODE(node)
	SUB_COMBINE(err, fsum, 1);
	err = sqrt(err);
	return err;
}

matmult()
/* multiplies qnew by Q = D-A and puts it in r */
{
	ELMTPTR elmt;
	NODEPTR node, n0, n1;
	int i;

	FORALLNODES(node)
		node->user->r = node->user->qnew * node->user->nneigh;
	NEXTNODE(node)

	FORALLELMTS(elmt)
		for(i=0; i<3; i++){
			n0 = elmt->neigh[i];
			n1 = elmt->neigh[(i+1)%3];
			n0->user->r -= n1->user->qnew;
			if(GROUP_EDGE(n1) && n1->b->clock == n0)
				n1->user->r -= n0->user->qnew;
		}
	NEXTELMT(elmt)

	SUB_NODE_COMBINE(r, fsum, 1);
}

orthog()
{
	NODEPTR node;
	double tot = 0;

	FORALLNODES(node)
		if(node->user->rep)
			tot += node->user->qnew;
	NEXTNODE(node)

	SUB_COMBINE(tot, fsum, 1);

	tot /= nnode;
	FORALLNODES(node)
		node->user->qnew -= tot;
	NEXTNODE(node)
}

double dot()
/* Evaluates qnew dot r */
{
	NODEPTR node;
	double d = 0;

	FORALLNODES(node)
		if(node->user->rep){
			d += node->user->qnew * node->user->r;
		}
	NEXTNODE(node)
	SUB_COMBINE(d, fsum, 1);
	return d;
}

double magr()
/* Evaluates r dot r */
{
	NODEPTR node;
	double d = 0;

	FORALLNODES(node)
		if(node->user->rep)
			d += node->user->r * node->user->r;
	NEXTNODE(node)
	SUB_COMBINE(d, fsum, 1);
	return d;
}

showeigen()
{
	double max, min, size, cont_func();
	NODEPTR node;
/*
	max = -1.e12;
	min = 1.e12;
	FORALLNODES(node)
		if(node->user->eigen > max) max = node->user->eigen;
		if(node->user->eigen < min) min = node->user->eigen;
	NEXTNODE(node)

	SUB_COMBINE(min, fmin, 1);
	SUB_COMBINE(max, fmax, 1);
	size = (max-mid) > (mid-min) ? max-mid:mid-min;
	slow_contor(cont_func, 16, 2, -size, size);
*/
	slow_contor(cont_func, 16, 2, 0.0, 0.0);
}

double cont_func(node)
NODEPTR node;
{
	return node->user->eigen;
}

showgroups()
{
	double cont_g();
	NODEPTR n0, n1, n2;
	ELMTPTR elmt;
	int i;

	disp_erase();
	drawmesh();
	color(YELLOW);
	contor(cont_g, 4, 0, 0.0, 0.0);

	FORALLELMTS(elmt)
		for(i=0; i<3; i++){
			n0 = elmt->neigh[i];
			n1 = elmt->neigh[(i+1)%3];
			if(GROUP_EDGE(n1) && n1->b->clock == n0){
				n2 = elmt->neigh[(i+2)%3];
				move(0.9*n0->x+0.1*n2->x, 0.9*n0->y+0.1*n2->y);
				cont(0.9*n1->x+0.1*n2->x, 0.9*n1->y+0.1*n2->y);
			}
		}
	NEXTELMT(elmt)
	usendplot();
}

double cont_g(node)
NODEPTR node;
{
	return (double)node->user->rep;
}

#ifdef SEQ
#define RAN		(rand()/2.147e9)
#else
#define RAN		(rand()/32768.)
#endif

Makelb()
{
	ELMTPTR elmt;
	NODEPTR node, n0, n1;
	ALIASPTR a, astart;
	int i;

	nlist = 0;
	if(!proclist)
		proclist = (int *)malloc(nprocs*sizeof(int));
	if(!proclist)
		abort(1235);
		
	for(i=0; i<nprocs; i++)
		if(inmygroup(i))
			proclist[nlist++] = i;

	FORALLNODES(node)
		node->user->nneigh = 0;
		node->user->rep = 1;
		if(node->b && node->b->alias){
			a = astart = node->b->alias;
			do {
				if(inmygroup(a->proc) 
				&& (a->proc > procnum
				    || (a->proc == procnum && a->node > node))){
					node->user->rep = 0;
				}
			} while((a = a->next) != astart);
		}
	NEXTNODE(node)

	nelmt = 0;
	FORALLELMTS(elmt)
		nelmt++;
		for(i=0; i<3; i++){
			n0 = elmt->neigh[i];
			n1 = elmt->neigh[(i+1)%3];
			n0->user->nneigh++;
			if(GROUP_EDGE(n1) && n1->b->clock == n0)
				n1->user->nneigh++;
		}
	NEXTELMT(elmt)

	nnode = 0;
	FORALLNODES(node)
		if(node->user->rep){
			nnode++;
			node->user->eigen = RAN;
		} else
			node->user->eigen = 0;
	NEXTNODE(node)

	SUB_COMBINE(nelmt, isum, 1);
	SUB_COMBINE(nnode, isum, 1);
	fmulti(stdout);
	if(procnum == proclist[0])
		printf("%d nodes %d elmts\n", nnode, nelmt);
	fsingl(stdout);

	SUB_NODE_COMBINE(eigen, fsum, 1);
}

inmygroup(proc)
int proc;
{
	return ((proc>>chan) == (procnum>>chan));
}
