#include "dime.h"

struct user_node {
	double unused;
};
struct user_bdy {
	double unused;
};
struct user_elmt {
	int cross;			/* if nonzero means that a pen-line passes thru */
	double area;		/* area */
	double x, y;		/* postion of circumcenter (for soap)*/
};

/*
Here is the definition of the `pen-lines', which are the lines that
create the font, which is Hershey duplex, from the file font.dat.
Each element may be `crossed' or not, depending on whether a pen-line
passes through or not.
*/
#define		MAXLINE 300		/* Maximum number of pen lines */
double      X0[MAXLINE];	/* Pen line from (X0, Y0) to (X1, Y1) */
double      Y0[MAXLINE];
double      X1[MAXLINE];
double      Y1[MAXLINE];
int         nline = 0;		/* Number of pen lines in use */

/*
Refinement: Each `crossed' element is refined if its area is > maxarea
*/
double maxarea;

user_main(mymenu, argc, argv)
MNUPTR mymenu;
int argc;
char **argv;
{
	int readtext();         /* Convert text string to pen-lines */
	int krefine();			/* Refinement around pen-lines */
	int show_crossed();		/* Show in color elements with pen-lines through */
	int soap();             /* Show Voronoi tesselation */
	int font_balance();		/* Give each processor a vertical stripe */

	void font_refine_func();
	void font_swops_func();

	DIME_PREAMBLE;

	menu_add_function(mymenu, "Read_text", readtext);
	menu_add_function(mymenu, "Refine", krefine);
	menu_add_function(mymenu, "Crossed", show_crossed);
	menu_add_function(mymenu, "Soap", soap);
	menu_add_function(mymenu, "Balance", font_balance);

	refine_func = font_refine_func;
	swops_func = font_swops_func;
}

readtext()
{
/*
This function reads in a short text string and looks up the 
pen-strokes required to draw thhose letters using Hershey Duplex.
The lines get scaled and pushed about to fit into a unit square.
*/
	FILE *fp;
	char name[30], c, *ptr;
	int nlet, xmin, xmax, x, y, totx, moveflag, charnum, thischar;
	int i, cpx, cpy, first = 1;
	ELMTPTR elmt;
	double Xmin = 1.e12, Xmax = -1.e12, area(), temp;
	double area();

	if(nline){
		printf("You already have a text string\n");
		return 1;
	}
	if(!the_elmts){
		printf("Please read in a mesh file first\n");
		return 1;
	}

	fp = fopen("font.dat", "r");
	if(fp == (FILE *)NULL){
		printf("Cannot find font.dat\n");
		fflush(stdout);
		exit(0);
	}

	printf("Your name please? ");
	printf("(2 to 6 Upper or lower case letters, no spaces)\n");
	get_string(name);
	ptr = name;

	while((c = *ptr++) != '\0'){
		charnum = c;
		nlet = 0;
		do {
			while(nlet--)
			fscanf(fp, "%d %d", &x, &y);
			if(fscanf(fp, "%d %d %d %d", &thischar, &xmin, &xmax, &nlet)==EOF){
				fprintf(stderr, "Character not found\n");
				goto continu;
			}
		} while(thischar != charnum);
		printf("%c", c); fflush(stdout);
		if(first){
			totx = -xmin;
			first = 0;
		} else
			totx -= xmin;
		moveflag = 1;
		while(nlet--){
			fscanf(fp, "%d %d", &x, &y);
			if(x == -64 && y == 0)
				moveflag = 1;
			else {
				if(moveflag){
					moveflag = 0;
				} else {
					X0[nline] = cpx;
					Y0[nline] = cpy;
					X1[nline] = x+totx;
					Y1[nline] = y;
					if(X0[nline] > Xmax)	Xmax = X0[nline];
					if(X0[nline] < Xmin)	Xmin = X0[nline];
					if(nline++ >= MAXLINE){
						printf("Name too long\n");
						fflush(stdout);
						exit(1);
					}
				}
				cpx = x + totx;
				cpy = y;
			}
		}
		totx += xmax;
		fseek(fp, (long)0, 0);
continu:;
	}
	printf("\n"); fflush(stdout);
	temp = Xmax - Xmin;
	Xmax += 0.12*temp;
	Xmin -= 0.12*temp;
	for(i=0; i<nline; i++){
		X0[i] = (X0[i] -Xmin)/(Xmax - Xmin);
		X1[i] = (X1[i] -Xmin)/(Xmax - Xmin);
		Y0[i] = Y0[i]/(Xmax-Xmin) + 0.4;
		Y1[i] = Y1[i]/(Xmax-Xmin) + 0.4;
	}
	fclose(fp);

	FORALLELMTS(elmt)
		elmt->user->cross = crossed(elmt);
		elmt->user->area = area(elmt);
	NEXTELMT(elmt)
}

soap()
/*
To draw the Voronoi tesselation of the mesh
This looks a bit funny at processor-processor boundaries
*/
{
	NODEPTR node, A, B, C;
	double dx1, dy1, dx2, dy2, cot;
	ELMTPTR elmt, elmtn;
	NODENEIGHPTR n, nstart;	
	color(1);
	FORALLELMTS(elmt)
		A = elmt->neigh[0];
		B = elmt->neigh[1];
		C = elmt->neigh[2];
		dx1 = B->x - A->x;		dy1 = B->y - A->y;
		dx2 = C->x - A->x;		dy2 = C->y - A->y;
		cot = (dx1*dx2 + dy1*dy2)/(dx1*dy2 - dx2*dy1);
		dx1 = B->x - C->x;		dy1 = B->y - C->y;
		elmt->user->x = 0.5*(B->x + C->x + cot*dy1);
		elmt->user->y = 0.5*(B->y + C->y - cot*dx1);
	NEXTELMT(elmt)

	FORALLNODES(node)
		n = nstart = node->n; 
		do {
			elmt = n->elmt;
			if(B_NODE(node) && n == nstart){
				move(elmt->user->x, elmt->user->y);
				cont(0.5*(node->x + node->b->clock->x),
				     0.5*(node->y + node->b->clock->y));
			} else {
				elmtn = n->next->elmt;
				if(elmt > elmtn)
					continue;
				move(elmt->user->x, elmt->user->y);
				cont(elmtn->user->x, elmtn->user->y);
			}
		} while((n = n->next) != nstart);
	NEXTNODE(node)
	usendplot();
}

show_crossed()
/*
To draw all the elements which have a pen line through it 
*/
{
	int i;
	double x, y;
	ELMTPTR elmt;
	disp_erase();
	color(1);
	FORALLELMTS(elmt)
		if(elmt->user->cross)
			initpanel(6, 1);
		else
			initpanel(0, 1);

		for(i=0; i<3; i++){
			DISTORT(elmt->neigh[i], x, y);
			panelpoint(x, y);
		}
		endpanel();
	NEXTELMT(elmt)
	usendplot();
}

krefine()
/*
Keeps on refineing until all elements which have a pen line through them
have area less than maxarea
*/
{
	ELMTPTR elmt;
	int ref, tot;
	printf("Area ? ");fflush(stdout);
	get_double(&maxarea);
	do {
		tot = 0;
		FORALLELMTS(elmt)
			ref = (elmt->user->cross && elmt->user->area > maxarea);
			set_refine(elmt, ref);
			tot += ref;
		NEXTELMT(elmt)
		refine();
		swops();
		GLOBAL_COMBINE(tot, isum, 1);
	} while(tot);
}

crossed(elmt)
ELMTPTR elmt;
{
	int i;
	for(i=0; i<3; i++)
		if(lincrossed(elmt->neigh[i], elmt->neigh[(i+1)%3]))
			return(1);
}

lincrossed(n, m)
NODEPTR n, m;
/*
Checks if any pen line intersests the line from node n to node m
*/
{
	int i, j;
	for(j=0; j<nline; j++)
		if(linsect(n->x, n->y, m->x, m->y, X0[j], Y0[j], X1[j], Y1[j]))
			return(1);
	return(0);
}	

linsect(ax, ay, bx, by, cx, cy, dx, dy)
double  ax, ay, bx, by, cx, cy, dx, dy;
/*
	Checks for intersection of lines a-b and c-d, and returns 1 if they do
*/
{
#define MAX(p, q)		((p>q)?p:q)
#define MIN(p, q)		((p<q)?p:q)
	double acx, acy, abx, aby, cdx, cdy, abt, cdt, det;
	if(MAX(ax, bx) < MIN(cx, dx))	return(0);
	if(MIN(ax, bx) > MAX(cx, dx))	return(0);
	if(MAX(ay, by) < MIN(cy, dy))	return(0);
	if(MIN(ay, by) > MAX(cy, dy))	return(0);
	acx = cx-ax; acy = cy-ay;
	abx = bx-ax; aby = by-ay;
	cdx = dx-cx; cdy = dy-cy;
	det = abx*cdy - aby*cdx;
	abt = acx*cdy - acy*cdx;
	cdt = acx*aby - acy*abx;
	if(det < 0.0){
		det = -det;
		abt = -abt;
		cdt = -cdt;
	}
	return(abt > 0.0 && cdt > 0.0 && abt < det && cdt < det);
}

void
font_refine_func(elmt, node, node1, node2, new, new1, new2)
ELMTPTR elmt, new1, new2;
NODEPTR node, node1, node2, new;
{
	double area();
	new->x = 0.5*(node1->x + node2->x);
	new->y = 0.5*(node1->y + node2->y);
	new1->user->area = new2->user->area = 0.5*elmt->user->area;
	if(lincrossed(node, new))
		new1->user->cross = new2->user->cross = 1;
	else {
		new1->user->cross = (lincrossed(node1, node) || lincrossed(new, node1));
		new2->user->cross = (lincrossed(node2, node) || lincrossed(new, node2));
	}
}

void
font_swops_func(A, B, C, D, ABC, ACD, ABD, BCD)
NODEPTR A, B, C, D;
ELMTPTR ABC, ACD, ABD, BCD;
{
	double area();
	BCD->user->area = area(BCD);
	ABD->user->area = area(ABD);
	BCD->user->cross = crossed(BCD);
	ABD->user->cross = crossed(ABD);
	ABD->user->area = area(ABD);
	BCD->user->area = area(BCD);
}

font_balance()
/*
Each processor gets a vertical stripe of elements.
Note use of global variable `nprocs'.
*/
{
	ELMTPTR elmt;
	double x;
	double xmin, ymin, xmax, ymax;

	get_extents(&xmin, &ymin, &xmax, &ymax);

	FORALLELMTS(elmt)
		x = (elmt->neigh[0]->x+elmt->neigh[1]->x+elmt->neigh[2]->x)/3;
		set_balance(elmt, (int)(nprocs*(xmax - x)/(xmax - xmin)));
	NEXTELMT(elmt)

	balance();
}
