#include "dime.h"
#define NFIELD		2
#define STIFFSIZE	(3*NFIELD*(3*NFIELD+1))/2
#define S			work1
#define W			work2

#define TOLERANCE	0.001
#define REFFRAC		0.25

#define DRICHLT			1	/* Fixed displacement */
#define NEUMANN			2	/* Traction */

double nu = 0.25, nu1, nuh;
double factor, strainfac = 0.01;

struct user_node {
	double phi[NFIELD];
	double rhs[NFIELD];
	double work1[NFIELD], work2[NFIELD];
	int ncopy;
};

struct user_bdy	{
	int bc;
	double bdyslope[NFIELD];
	double rhs[NFIELD];
};

struct user_elmt {
	double stiff[STIFFSIZE];
	double expand;
	double shear;
	double rot;
};

int eigendone = 0;
int setup_called = 0;

user_main(user_menu, argc, argv)
MNUPTR user_menu;
int argc;
char **argv;
{
	int solve(), refine_expand(), refine_shear(), strain();
	int change_bdy(), setup(), poisson_rat(), strain_factor();
	int disp_shear(), disp_expand(), disp_xstrain(), disp_ystrain();
	int stress_read(), stress_write();
	void stress_refine_func();

	DIME_PREAMBLE;

	user_read = stress_read;
	user_write = stress_write;
	refine_func = stress_refine_func;

	menu_add_function(user_menu, "Setup",			setup);
	menu_add_function(user_menu, "Solve",			solve);
	menu_add_function(user_menu, "Change",			change_bdy);
	menu_add_function(user_menu, "Refine_expand",	refine_expand);
	menu_add_function(user_menu, "Refine_shear",	refine_shear);
	menu_add_function(user_menu, "Strain",			strain);
	menu_add_function(user_menu, "Strainfactor",	strain_factor);

	menu_add_function(user_menu, "Expansion",		disp_expand);
	menu_add_function(user_menu, "Shear",			disp_shear);
	menu_add_function(user_menu, "Xstrain",			disp_xstrain);
	menu_add_function(user_menu, "Ystrain",			disp_ystrain);
	menu_add_function(user_menu, "Poisson_Ratio",	poisson_rat);
}

poisson_rat()
{
	double nutemp;
again:
	printf("Poisson Ratio is %f, new value?\n", nu);
	get_double(&nutemp);
	if(nu < 0.0 || nu >= 0.5){
		printf("Poisson ratio must be >= 0 and < 0.5\n");
		goto again;
	}
	nu = nutemp;
	nu1 = 1-nu;
	nuh = 0.5-nu;
}

stress_read(s, type, ptr)
char *s;
int type;
char *ptr;
{
	BDYPTR b;
	int nread;
	double xval, yval;
	if(type == BDY){
		b = (BDYPTR)ptr;
		nread = sscanf(s, "%d %lf %lf", &b->user->bc, &xval, &yval);
		if(nread == 3 && b->user->bc){
			b->user->bdyslope[0] = xval;
			b->user->bdyslope[1] = yval;
		} else {
			b->user->bc = 2;
			b->user->bdyslope[0] = b->user->bdyslope[1] = 0;
		}
		return 1;
	} else
		return 0;
}

stress_write(s, type, ptr)
char *s;
int type;
char *ptr;
{
	BDYPTR b;
	if(type == BDY){
		b = (BDYPTR)ptr;
		if(b->user->bc)
			sprintf(s, "%d %lf %lf", b->user->bc, 
				b->user->bdyslope[0], b->user->bdyslope[1]);
			else
				sprintf(s, "%d", b->user->bc);
			return 1;
	} else
		return 0;
}

setup()
{
	NODEPTR node;

	FORALLNODES(node)
		node->user->phi[0] = node->user->phi[1] = 0.0;
	NEXTNODE(node)
	nu1 = 1-nu;
	nuh = 0.5-nu;
	setup_called = 1;
}

solve()
{
	if(!setup_called)
		setup();
	stiffmat();
	menu_monitor(-3, 1);
	poisson(TOLERANCE);
	eigendone = 0;
}

stiffmat()
{
#define FIXED(node)	(B_NODE(node) && (node->b->user->bc == DRICHLT \
					|| (B_EDGE(node->b->antik) \
					&& node->b->antik->b->user->bc == DRICHLT)))
	NODEPTR node, node1, node2;
	ELMTPTR elmt;
	double a, area(), bi, ci, bj, cj, coeff, dx, dy, stiff;
	int i, j, k;

	FORALLELMTS(elmt)
		a = area(elmt);
		for(i=0; i<3; i++){
			bi = elmt->neigh[(i+1)%3]->y - elmt->neigh[(i+2)%3]->y;
			ci = elmt->neigh[(i+2)%3]->x - elmt->neigh[(i+1)%3]->x;
			for(j=0; j<=i; j++){
				bj = elmt->neigh[(j+1)%3]->y - elmt->neigh[(j+2)%3]->y;
				cj = elmt->neigh[(j+2)%3]->x - elmt->neigh[(j+1)%3]->x;
				elmt->user->stiff[index(i, 0, j, 0)] = (nu1*bi*bj + nuh*ci*cj)/a;
				elmt->user->stiff[index(i, 0, j, 1)] = (nu*bi*cj + nuh*ci*bj)/a;
				elmt->user->stiff[index(i, 1, j, 0)] = (nu*ci*bj + nuh*bi*cj)/a;
				elmt->user->stiff[index(i, 1, j, 1)] = (nu1*ci*cj + nuh*bi*bj)/a;
			}
		}
	NEXTELMT(elmt)

	FORALLNODES(node)
		node->user->ncopy = 1;
		if(!FIXED(node)){
			node->user->S[0] = node->user->S[1] = 0.0;
			continue;
		}
		if(node->b->user->bc == DRICHLT)
			for(k=0; k<NFIELD; k++)
				node->user->S[k] = node->user->phi[k] = 
					node->b->user->bdyslope[k];
		if(B_EDGE(node->b->antik) && node->b->antik->b->user->bc == DRICHLT)
			for(k=0; k<NFIELD; k++)
				node->user->S[k] = node->user->phi[k] = 
					node->b->antik->b->user->bdyslope[k];
	NEXTNODE(node)
	NODE_COMBINE(ncopy, isum, 1)

	matrix_mult();

	FORALLNODES(node)
		for(k=0; k<NFIELD; k++){
			node->user->rhs[k] = -node->user->W[k];
			if(B_NODE(node))
				node->b->user->rhs[k] = 0.0;
		}
	NEXTNODE(node)

	FORALLNODES(node)
		if(B_EDGE(node) && node->b->user->bc == NEUMANN){
			node1 = node->b->clock;
			dx = node->x - node->b->clock->x;
			dy = node->y - node->b->clock->y;
			stiff = 0.5 * sqrt(dx*dx + dy*dy);
			for(k=0; k<NFIELD; k++){
				node ->b->user->rhs[k] += stiff * node->b->user->bdyslope[k];
				node1->b->user->rhs[k] += stiff * node->b->user->bdyslope[k];
			}
		}
	NEXTNODE(node)

	BDY_COMBINE(rhs[0], fsum, NFIELD);

	FORALLNODES(node)
		if(B_NODE(node))
			for(k=0; k<NFIELD; k++)
				node->user->rhs[k] += node->b->user->rhs[k];
	NEXTNODE(node)
}

poisson(tol)
double tol;
{
	NODEPTR node;
	ELMTPTR elmt;
	double alpha, beta, rho = 0.0, rhoold, dot, rhostart = 0.0, rhs;
	int first = 1, count = 0, k;
	fsingl(stdout);

	FORALLNODES(node)
		if(FIXED(node))
			continue;
		for(k=0; k<NFIELD; k++){
			rhs = node->user->rhs[k];
			rhostart += rhs * rhs / node->user->ncopy;
		}
	NEXTNODE(node)
	GLOBAL_COMBINE(rhostart, fsum, 1);

	FORALLNODES(node)
		if(!FIXED(node))
			for(k=0; k<NFIELD; k++)
					node->user->S[k] = node->user->phi[k];
		else
			for(k=0; k<NFIELD; k++)
				node->user->S[k] = 0;
	NEXTNODE(node)

	matrix_mult();

	FORALLNODES(node)
		for(k=0; k<NFIELD; k++)
			node->user->rhs[k] -= node->user->W[k];
	NEXTNODE(node)

	do {
		rho = 0.0;
		FORALLNODES(node)
			if(FIXED(node))	continue;
			for(k=0; k<NFIELD; k++){
				rhs = node->user->rhs[k];
				rho += rhs * rhs / node->user->ncopy;
			}
		NEXTNODE(node)
		GLOBAL_COMBINE(rho, fsum, 1);

		if(first){
			first = 0;
			FORALLNODES(node)
				for(k=0; k<NFIELD; k++)
					node->user->S[k] = node->user->rhs[k];
				if(FIXED(node))
					node->user->S[0] = node->user->S[1] = 0.0;
			NEXTNODE(node)
		} else {
			beta = rho/rhoold;
			FORALLNODES(node)
				if(FIXED(node))	continue;
				for(k=0; k<NFIELD; k++)
					node->user->S[k] = 
						node->user->rhs[k] + beta * node->user->S[k];
			NEXTNODE(node)
		}

		matrix_mult();

		dot = 0.0;
		FORALLNODES(node)
			if(FIXED(node))	continue;
			for(k=0; k<NFIELD; k++)
				dot += node->user->S[k] * node->user->W[k] / node->user->ncopy;
		NEXTNODE(node)
		GLOBAL_COMBINE(dot, fsum, 1);

		alpha = rho/dot;
		FORALLNODES(node)
			if(FIXED(node))	continue;
			for(k=0; k<NFIELD; k++){
				node->user->phi[k] += alpha * node->user->S[k];
				node->user->rhs[k] -= alpha * node->user->W[k];
			}
		NEXTNODE(node)

		rhoold = rho;
		put_monitor(log10(sqrt(rho/rhostart)));
		if(count++ > 500){
			printf("Solve Failed\n");
			return 0;
		}
	} while(rho > rhostart*tol*tol);
}

matrix_mult()
{
	int i, j, k, l;
	NODEPTR node, node1, node2;
	ELMTPTR elmt;

	FORALLNODES(node)
		for(k=0; k<NFIELD; k++)
			node->user->W[k] = 0.0;
	NEXTNODE(node)

	FORALLELMTS(elmt)
		for(i=0; i<3; i++)
			for(j=0; j<3; j++)
				for(k=0; k<NFIELD; k++)
					for(l=0; l<NFIELD; l++)
						elmt->neigh[i]->user->W[k] += 
								elmt->user->stiff[index(i, k, j, l)] *
								elmt->neigh[j]->user->S[l];
	NEXTELMT(elmt)

	NODE_COMBINE(W[0], fsum, NFIELD);
}

index(i, a, j, b)
int i, a, j, b;
{
	int I = i*NFIELD + a, J = j*NFIELD + b;
	if(I <= J)
		return((J*(J+1))/2 + I);
	else
		return((I*(I+1))/2 + J);
}

change_bdy()
{
	NODEPTR node;
	int bc;
	char answer[20];
	double x, y, xbdy, ybdy, s;
	struct curve *curve;

	while(1){
		draw_bdy();
		printf("Do you want to change anything?\n");
		get_string(answer);
		if(answer[0] != 'y' && answer[0] != 'Y')
			return 1;
		printf("Click on a boundary curve\n");
		get_gin(&x, &y);
		closest_to_curve(&curve, &s, x, y);
		printf("Curve '%s': Is this fixed (1) or forced (2)?\n", curve->name);
		get_int(&bc);
		if(bc == DRICHLT)
			printf("What is the Displacement (x y) ?\n");
		else
			printf("What is the Traction (x y) ?\n");
		get_double(&xbdy);
		get_double(&ybdy);

		FORALLNODES(node)
			if(B_EDGE_CURVE(node) != curve)
				continue;
			node->b->user->bc = bc;
			node->b->user->bdyslope[0] = xbdy;
			node->b->user->bdyslope[1] = ybdy;
		NEXTNODE(node)
	}
}

draw_bdy()
{
	NODEPTR node;

	disp_erase();

	FORALLNODES(node)
		if(!B_EDGE(node))
			continue;
		if(FIXED(node) && FIXED(node->b->clock))
			color(FOREGD);
		else 
			color(RED);
		draw(node, node->b->clock);
		arrow((node->x+node->b->clock->x)/2, 
			(node->y+node->b->clock->y)/2, 
			node->b->user->bdyslope[0],
			node->b->user->bdyslope[1]);
	NEXTNODE(node)
}

strain()
{
	void usr_distort();
	double objsize, maxphi = 0;
	double xmin, ymin, xmax, ymax;
	NODEPTR node;

	color(FOREGD);
	FORALLNODES(node)
		if(!B_EDGE(node))
			continue;
		draw(node, node->b->clock);
		if(fabs(node->user->phi[0]) > maxphi) maxphi = fabs(node->user->phi[0]);
		if(fabs(node->user->phi[1]) > maxphi) maxphi = fabs(node->user->phi[1]);
	NEXTNODE(node)

	GLOBAL_COMBINE(maxphi, fmax, 1);

	factor = strainfac/maxphi;

	distort = usr_distort;
	FORALLNODES(node)
		if(!B_EDGE(node))
			continue;
		if(FIXED(node) && FIXED(node->b->clock))
			color(FOREGD);
		else 
			color(RED);
		draw(node, node->b->clock);
	NEXTNODE(node)
	distort = NULL;
	usendplot();
}

void
usr_distort(node, ax, ay)
NODEPTR node;
double *ax, *ay;
{
	*ax = node->x + factor*node->user->phi[0];
	*ay = node->y + factor*node->user->phi[1];
}

refine_shear()
{
	double shear_func();
	refine_fraction(shear_func, REFFRAC);
	swops();
}

refine_expand()
{
	double expand_func();
	refine_fraction(expand_func, REFFRAC);
	swops();
}
	
void
stress_refine_func(elmt, node, node1, node2, new, new1, new2)
ELMTPTR elmt, new1, new2;
NODEPTR node, node1, node2, new;
{
	int k;
	for(k=0; k<NFIELD; k++)
		new->user->phi[k] = 0.5*(node1->user->phi[k] + node2->user->phi[k]);
	if(B_EDGE(node2) && node2->b->clock == new){
		new->b->user->bc = node2->b->user->bc;
		for(k=0; k<NFIELD; k++)
			new->b->user->bdyslope[k] = node2->b->user->bdyslope[k];
	}
}

geteigen(elmt)
ELMTPTR elmt;
{
	NODEPTR n0, n1, n2;
	double x1, y1, x2, y2, u1, v1, u2, v2, a11, a12, a21, a22;
	double det, disc, b, c;

	n0 = elmt->neigh[0];
	n1 = elmt->neigh[1];
	n2 = elmt->neigh[2];

	x1 = n1->x - n0->x;		
	y1 = n1->y - n0->y;
	x2 = n2->x - n0->x;		
	y2 = n2->y - n0->y;
	det = x1*y2 - x2*y1;

	u1 = n1->user->phi[0] - n0->user->phi[0];
	v1 = n1->user->phi[1] - n0->user->phi[1];
	u2 = n2->user->phi[0] - n0->user->phi[0];	
	v2 = n2->user->phi[1] - n0->user->phi[1];

	a11 = ( u1*y2 - u2*y1)/det;
	a12 = (-u1*x2 + u2*x1)/det;
	a21 = ( v1*y2 - v2*y1)/det;
	a22 = (-v1*x2 + v2*x1)/det;

	elmt->user->expand = 0.5*(a11 + a22);
	elmt->user->shear = a11 - a22;
	elmt->user->rot = a12 - a21;
}

disp_shear()
{
	double shear_func();
	void usr_distort();
	ELMTPTR elmt;
	if(!eigendone){
		FORALLELMTS(elmt)
			geteigen(elmt);
		NEXTELMT(elmt)
		eigendone = 1;
	}
	distort = usr_distort;
	elmt_contor(shear_func, BLACK-WHITE, 0.0, 0.0);
	distort = NULL;
}

double shear_func(elmt)
ELMTPTR elmt;
{
	return(elmt->user->shear);
}

disp_expand()
{
	double expand_func();
	void usr_distort();

	ELMTPTR elmt;
	if(!eigendone){
		FORALLELMTS(elmt)
			geteigen(elmt);
		NEXTELMT(elmt)
		eigendone = 1;
	}

	distort = usr_distort;
	elmt_contor(expand_func, BLACK-WHITE, 0.0, 0.0);
	distort = NULL;
}

double  expand_func(elmt)
ELMTPTR elmt;
{
	return(elmt->user->expand);
}

disp_xstrain()
{
	double xstrain();
	void usr_distort();
	color(RED);
	distort = usr_distort;
	contor(xstrain, 32, 0, 0.0, 0.0);
	distort = NULL;
}

double xstrain(node)
NODEPTR node;
{
	return(node->user->phi[0]);
}

disp_ystrain()
{
	double ystrain();
	void usr_distort();
	color(BLUE);
	distort = usr_distort;
	contor(ystrain, 32, 0, 0.0, 0.0);
	distort = NULL;
}

double ystrain(node)
NODEPTR node;
{
	return(node->user->phi[1]);
}

strain_factor()
{
	printf("Current strain factor %f: new?\n", strainfac);
	get_double(&strainfac);
}
