#include "bdy.h"
#include <math.h>

get_extents(xmin, ymin, xmax, ymax)
double *xmin, *ymin, *xmax, *ymax;
{
	struct curve *curve;
	double x, y, s;

	*xmin = *ymin = 1.e12;
	*xmax = *ymax = -1.e-12;

	FORALLCURVES(curve)
		for(s = 0.0; s < curve->length; s += curve->length*0.02){
			get_point(curve, s, &x, &y);
			if(x < *xmin) *xmin = x;
			if(x > *xmax) *xmax = x;
			if(y < *ymin) *ymin = y;
			if(y > *ymax) *ymax = y;
		}
	NEXTCURVE(curve)
}

drawcurves()
{
	struct curve *curve;
	struct segment *seg;
	struct point *point;

	color(6);
	FORALLPOINTS(point)
		draw_point(point);
	NEXTPOINT(point)

	FORALLCURVES(curve)
		draw_point(curve->first);
		draw_point(curve->last);
		FORALLSEGMENTS(curve, seg)
			draw_seg(seg);
		NEXTSEGMENT(curve, seg)
	NEXTCURVE(curve)
	sendplot();
}

draw_seg(seg)
struct segment *seg;
{
	switch(seg->type){
		case LINE  : draw_line  (seg->segdata.line  ); return 1;
		case ARC   : draw_arc   (seg->segdata.arc   ); return 1;
		case CURVETO: draw_curveto(seg->segdata.curveto); return 1;
	}
}

draw_point(point)
struct point *point;
{
	draw_marker(point->x, point->y);
}

draw_marker(x, y)
double x, y;
{
	double screensize();
	marker(4, x, y, 0.01*screensize());
}

draw_line(line)
struct line *line;
{
	move(line->x0, line->y0);
	cont(line->x1, line->y1);
}

draw_arc(arc)
struct arc *arc;
{
	int i, n;
	double th, dth, x, y;

	n = fabs(arc->th1 - arc->th0)*12;
	dth = (arc->th1 - arc->th0)/n;
	for(i=0; i<=n; i++){
		th = arc->th0 + i*dth;
		x = arc->x + arc->r*cos(th);
		y = arc->y + arc->r*sin(th);
		if(i)
			cont(x, y);
		else
			move(x, y);
	}
}

draw_curveto(cu)
struct curveto *cu;
{
    int i;
	double t, x, y;

    for(i=0; i<=20; i++){
        t = (double)i/20;
        x = cu->dx + t*(cu->cx + t*(cu->bx + t*cu->ax));
        y = cu->dy + t*(cu->cy + t*(cu->by + t*cu->ay));
        if(i)
            cont(x, y);
        else
            move(x, y);
    }
}
