/****************************************************/
/*    DIME: contor.c                                */
/*    Turbo Contouring for Dime                     */
/****************************************************/
/*        Roy Williams Copyright Caltech 1988       */
/****************************************************/
#define DIMECODE
#include "dime.h"
/*
	The value of the EMPTY flag is zero if the node is not to be
	colored and 1 if it is to be colored but is not
	part of the current patch, else 2
*/
#define EMPTY		work_space.ni[0]
#define CONT_BDY	work_space.ni[1]
/*
	The value of STATUS is OUTOFBOUNDS if the elmt is not in the
	current strip or has had a contour line thru it already at 
	this level. If in the current strip it is STRIP, if in the
	current patch, which is contained in the strip, it is PATCH,
	and if in the current patch bu already contoured it is
	FINISHED.
*/
#define STATUS		work_space.ei[0]

int __panels, __shade;
int __pointcount;
double (*__func)();
double __level;

contor(funcpass, nlevel, __panelspass, passMin, passMax)
/*
	Draw a contour plot of the function __func which has the form
		double __func(node)
		NODEPTR node;
	with nlevel levels, using colored __panels if __panelspass is 1
	else with lines. Min and Max are the lowest and highest contour
	levels, and if Min >= Max these are found by contor().
*/
double (*funcpass)();
int nlevel, __panelspass;
double passMin, passMax;
{
/*
	If __panelspass, The space is split into vertical strips so that 
	each panel has the topology of a disk, but there can be at 
	most MAXSTRIP of these.
*/
#define MAXSTRIP		128
#define LARGE			1.e12

#define OUTOFBOUNDS		0	/* Elmt outside strip or finished with */
#define STRIP			1	/* Elmt inside current strip */
#define PATCH			2	/* ELMT in current patch which is in strip */
#define FINISHED		3	/* Elmt in patch but already contoured */

#define FILLED			0	/* Node is above current contour level */
#define HOLEEMPTY		1	/* Node is below and not connected to boundary */
#define BDYEMPTY		2	/* Node is below, and node-connected to boundary */

#define INTERIOR		0
#define BDYNOTDONE		1
#define BDYDONE			2
#define OUTBDY			3

	NODEPTR bnode, node;
	double temp, x, y, Max = -LARGE, Min = LARGE;
	int i;

	if(nlevel < 1)
		return 0;
	__panels = __panelspass;
	__func = funcpass;
/*
	If necessary, get global Min and Max
*/
	if(the_nodes != (NODEPTR)NULL)
		Max = Min = (*__func)(the_nodes);
	FORALLNODES(node)
		temp = (*__func)(node);
		if(temp > Max)	Max = temp;
		if(temp < Min)	Min = temp;
	NEXTNODE(node);
	GLOBAL_COMBINE(Max, fmax, 1);
	GLOBAL_COMBINE(Min, fmin, 1);

	printf("Min %f Max %f\n", Min, Max);fflush(stdout);

	if(passMin < passMax){
		Min = passMin;
		Max = passMax;
	}
/*	if(nlevel >= 2)
		Max -= (Max-Min)/nlevel;*/
	if(fabs(Max - Min) < 1.e-12){
		printf("Contor: flat function\n");fflush(stdout);
		return(1);
	}
	fmulti(stdout);
/*
	Establish outer boundary
*/
    temp = LARGE; 
	bnode = NULL;
    FORALLNODES(node)
		node->CONT_BDY = (node->b != (BDYPTR)NULL);
        if(node->CONT_BDY && node->x < temp){
            bnode = node; 
            temp = node->x; 
        }    
    NEXTNODE(node)
     
    if((node = bnode) != (NODEPTR)NULL){
		do {
			node->CONT_BDY = OUTBDY;
		} while((node = node->b->clock) != bnode);
	}

	FORALLNODES(node)
		if(node->CONT_BDY != BDYNOTDONE)
			continue;
		bnode = node;
		do {
			node->CONT_BDY = BDYDONE;
		} while((node = node->b->clock) != bnode);
		node->CONT_BDY = BDYNOTDONE;
	NEXTNODE(node);
/*
	Do the contouring
*/
	for(i=0; i<=nlevel; i++){
		__level = Min + i*(Max-Min)/nlevel;
		if(__panels == 1)
			__shade = RAINSTART + 0.75*(RAINEND - RAINSTART)*(double)i/nlevel;
		else if(__panels == 2)
			__shade = WHITE + (BLACK - WHITE)*(double)i/nlevel;
		contor_level();
		if(mesh_is_distributed)
			usendplot();
		else
			sendplot();
	}
/*
	Draw the border around the boundary
	color(1);
*/
	FORALLNODES(node)
		if(node->b != (BDYPTR)NULL && node->b->oppalias == (ALIASPTR)NULL)
			draw(node, node->b->clock);
	NEXTNODE(node)
	if(mesh_is_distributed)
		usendplot();
	else
		sendplot();
	fsingl(stdout);
}

contor_level()
/*
	To do one contour level, with the globals __level for its
	value and in color __shade if __panels.
*/
{
	double temp, xn, yn, x[3], y[3];
	NODEPTR bnode, node, noden;
	NODENEIGHPTR n, nn, nstart;
	ELMTPTR elmt;
	double cut[MAXSTRIP];
	int i, nstrip = 1, wrong, istrip, cut_here;
/*
	decide which nodes are to be empty as opposed to filled
*/
	FORALLNODES(node)
		node->EMPTY = ((*__func)(node) < __level) 
			|| (node->CONT_BDY == BDYNOTDONE);
	NEXTNODE(node)
/*
	If we're doing lines can forget about topology and the patch
	is just the whole region
*/
	if(!__panels){
		FORALLNODES(node)
			if(node->EMPTY)
				node->EMPTY = BDYEMPTY;
		NEXTNODE(node)

		FORALLELMTS(elmt)
			elmt->STATUS = PATCH;
		NEXTELMT(elmt)
		contor_patch(-LARGE, LARGE);
		return 1;
	}
/*
	Find the empty region which is adjacent to the outer boundary.
	Everthing else is either filled or a hole in a filled region
*/
	FORALLNODES(node)
	if(node->CONT_BDY == OUTBDY && node->EMPTY == HOLEEMPTY){
		node->EMPTY = BDYEMPTY;
		seedfill(node);
	}
	NEXTNODE(node)
/*
	Make a vertical cut through each hole
*/
	cut[0] = -LARGE;
	FORALLNODES(node)
		if(node->EMPTY == HOLEEMPTY){
			DISTORT(node, cut[nstrip++], temp);
			if(nstrip >= MAXSTRIP){
				fprintf(stderr, "Buffer Overflow: Contor will fail\n");
				goto sort;
			}
			FORALLELMTS(elmt)
				for(i=0; i<3; i++)
					DISTORT(elmt->neigh[i], x[i], y[i]);
				DISTORT(node, xn, yn);
				elmt->STATUS = ((x[0] + x[1] + x[2])/3 < xn);
			NEXTELMT(elmt)
/*
	For each cut, eliminate all the holes it has removed
*/
			FORALLNODES(noden)
				if(!(noden->EMPTY))
					continue;
				n = nstart = noden->n;
				do {
					nn = n->next;
					if(n->elmt->STATUS ^ nn->elmt->STATUS)
						noden->EMPTY = BDYEMPTY;
				} while((n = nn) != nstart);
			NEXTNODE(noden)
			seedfill(node);
		}
	NEXTNODE(node);

	cut[nstrip] = LARGE;
/*
	Sort the cuts
*/
sort:
	do {
		wrong = 0;
		for(istrip=1; istrip < nstrip; istrip++)
			if(cut[istrip] < cut[istrip-1]){
				wrong = 1;
				temp = cut[istrip];
				cut[istrip] = cut[istrip-1];
				cut[istrip-1] = temp;
			}
	} while(wrong);
/*
	Contour each strip
*/
	for(istrip=0; istrip < nstrip; istrip++)
		contor_strip(cut[istrip], cut[istrip+1]);
	return 1;
}

contor_strip(xmin, xmax)
/*
	To contour the part of the mesh whose elmts have centers
	between xmin and xmax, using global value __level.
*/
double xmin, xmax;
{
	ELMTPTR elmt;
	double x, xx[3], yy[3];
	int i;

	FORALLELMTS(elmt)
		for(i=0; i<3; i++)
			DISTORT(elmt->neigh[i], xx[i], yy[i]);
		x = (xx[0] + xx[1] + xx[2])/3;
		elmt->STATUS = (x > xmin && x <= xmax);
	NEXTELMT(elmt)
/*
	Do each elmt-connected patch within the strip
*/
	FORALLELMTS(elmt)
		if(elmt->STATUS == STRIP){
			elmtseedfill(elmt);
			contor_patch(xmin, xmax);
		}
	NEXTELMT(elmt)
}

contor_patch(xmin, xmax)
/*
	The elmts have STATUS as follows:
		OUTOFBOUNDS:		Outside the strip (xmin, xmax]
		STRIP:				Inside the strip but not in this patch
		PATCH:				Inside the patch -- connected set
		FINISHED:			None at the moment, but elmts used
					in the contouring will acqire this status.
*/
double xmin, xmax;
{
/*
	These are the states of the contouring process:
*/
#define	INTERIOR		0	/* Are about to cross an elmt: state is defined
							by the elmt to be crossed and the int i which
							is the vertex opposite the edge we came in */

#define BOUNDARY		1	/* Are about to go along the boundary one jump:
							state is defined by the node we are at */

#define	PATCHEDGE		2	/* Are about to go along an edge which is the
							border of a patch and is not a boundary:
							state is define by the node we are at */

	double xx, yy, minx = LARGE;
	int i, state, allup = 1;
	NODEPTR node, node1, node2, nodemin;
	ELMTPTR elmt, start;
	NODENEIGHPTR n, nn, nstart, indexnn();

/*
	Look for an edge which has a node above and a node below.
	We also check here for 'allup' which is all the nodes of
	the patch being above the contouring level. In this case
	we need to fill the patch, so need a seed node which can be
	the leftmost one of the patch
*/
next_panel:
	FORALLELMTS(elmt)
		if(elmt->STATUS != PATCH)
			continue;
		for(i=0; i<3; i++){
			node1 = elmt->neigh[(i+1)%3];
			node2 = elmt->neigh[(i+2)%3];
			DISTORT(node1, xx, yy);
			if(xx < minx){
				minx = xx;
				nodemin = node1;
			}
			if(node1->EMPTY)
				allup = 0;
			if(node1->EMPTY && !(node2->EMPTY)){
				get_interp(node1, node2, &xx, &yy);
				state = INTERIOR;
				start = elmt;
				goto found_panel;
			}
		}
	NEXTELMT(elmt)
/*
	No more __panels to be done. If doing lines we are thru,
	otherwise make sure the allup case is handled.
*/
	if(!__panels)
		return 1;
	if(allup){
		node = nodemin;
		if(node->b == (BDYPTR)NULL || node->n->elmt->STATUS < PATCH)
			state = PATCHEDGE;
		else
			state = BOUNDARY;
		DISTORT(node, xx, yy);
		goto found_panel;
	} else
		goto finish;
/*
	About to fill a panel. We have set up the state properly
	and now let her rip.
*/
found_panel:
	if(__panels){
		initpanel(__shade, 0); 
		__pointcount = 0;
		pnlpoint(xx, yy);
	} else{
/*		color(__shade);*/
		move(xx, yy);
	}

	while(1){
		switch(state){
			case(INTERIOR):
				elmt->STATUS = FINISHED;
				node = elmt->neigh[i];
				node1 = elmt->neigh[(i+1)%3];
				node2 = elmt->neigh[(i+2)%3];
/*
	If a contour line comes into a triangle, there are only
	two ways it can go out depending on whether the opposite
	node to the incoming edge is above or below
*/
				if(!(node->EMPTY)){
					get_interp(node, node1, &xx, &yy);

					if(__panels) { if (pnlpoint(xx, yy)) return 1;}
					else cont(xx, yy);

					n = indexnn(node1, elmt);
					if(node1->b != (BDYPTR)NULL && node1->n == n){
						state = BOUNDARY;
						node = node1;
					} else if(n->next->elmt->STATUS < PATCH){
						state = PATCHEDGE;
						node = node1;
					} else {
						elmt = n->next->elmt;
						i = (indexc(elmt, node1)+2)%3;
					}
						
				} else {
					get_interp(node, node2, &xx, &yy);

					if(__panels) {if(pnlpoint(xx, yy)) return 1;}
					else cont(xx, yy);

					n = indexnn(node, elmt);
					if(node->b != (BDYPTR)NULL && node->n == n)
						state = BOUNDARY;
					else if(n->next->elmt->STATUS < PATCH){
						state = PATCHEDGE;
					} else {
						elmt = n->next->elmt;
						i = (indexc(elmt, node)+2)%3;
					}
				}
				break;

			case(BOUNDARY):
				node1 = node->b->clock;
				if(!(node1->EMPTY)){
					node = node1;

					DISTORT(node, xx, yy);
					if(__panels){if(pnlpoint(xx, yy)) return 1;}

					if(node->n->elmt->STATUS < PATCH)
						state = PATCHEDGE;
				} else {
					get_interp(node, node1, &xx, &yy);

					if(__panels){if(pnlpoint(xx, yy)) return 1;}
					else move(xx, yy);

					state = INTERIOR;
					elmt = node->n->elmt;
					i = (indexc(elmt, node)+1)%3;
				}
				break;

			case(PATCHEDGE):
				n = nstart = node->n;
				do {
					nn = n->next;
					if(n->elmt->STATUS >= PATCH && 
							nn->elmt->STATUS < PATCH)
						break;
				} while((n = nn) != nstart);
				elmt = n->elmt;
				i = (indexc(elmt, node)+1)%3;
				node1 = elmt->neigh[(i+1)%3];
				if(!(node1->EMPTY)){
					node = node1;

					DISTORT(node, xx, yy);
					if(__panels){if(pnlpoint(xx, yy)) return 1;}

					if(node->b != (BDYPTR)NULL && node->n->elmt->STATUS >= PATCH)
						state = BOUNDARY;
				}
				else{
					get_interp(node, node1, &xx, &yy);

					if(__panels){if(pnlpoint(xx, yy)) return 1;}
					else move(xx, yy);

					state = INTERIOR;
				}
				break;
		}
/*
	Here are the criteria for finishing this panel. If allup we
	are filling the entire panel, else we demand getting back to
	the starting elmt again
*/
		if(allup && node == nodemin){
			if(__panels) endpanel();
			goto finish;
		}
		if(!allup && elmt == start && state == INTERIOR){
			if(__panels) endpanel();
			goto next_panel;
		}
	}

finish:
/*
	Cleanup: make sure nobody comes back to this patch again
	since we just did it
*/
	FORALLELMTS(elmt)
		if(elmt->STATUS >= PATCH)
			elmt->STATUS = OUTOFBOUNDS;
	NEXTELMT(elmt)
	return 1;
}

get_interp(node1, node2, xx, yy)
/*
	Linear inverse interpolation for the place between two 
	adjacent nodes where the function value is '__level'
*/
NODEPTR node1, node2;
double *xx, *yy;
{
	double p, x1, y1, x2, y2, z1, z2;

	z1 = (*__func)(node1);
	z2 = (*__func)(node2);
	if(node1->CONT_BDY == BDYNOTDONE && z1 > __level)
		z1 = __level-1.e-12;
	if(node2->CONT_BDY == BDYNOTDONE && z2 > __level)
		z2 = __level-1.e-12;

	if(fabs(z1-z2) > 1.e-12)
		p = (z1 - __level)/(z1 - z2);
	else
		p = 0.5;
	DISTORT(node1, x1, y1);
	DISTORT(node2, x2, y2);
	*xx = p*x2 + (1-p)*x1;
	*yy = p*y2 + (1-p)*y1;
}

seedfill(node)
/*
	Finds largest node-connected set of nodes which has 
	EMPTY = HOLEEMPTY and includes the seed node, and sets the value
	of EMPTY to BDYEMPTY
*/

NODEPTR node;
{
	ELMTPTR elmt;
	NODEPTR noden;
	if(!(node->EMPTY))
		return 1;
	node->EMPTY = BDYEMPTY;
	FORALLNEIGH(node, elmt)
		noden = elmt->neigh[(indexc(elmt, node)+1)%3];
		if(noden->EMPTY == HOLEEMPTY)
			seedfill(noden);	
	NEXTNEIGH(node, elmt)
	if(node->b != (BDYPTR)NULL && noden->EMPTY == HOLEEMPTY)
		seedfill(node->b->clock);
}

elmtseedfill(elmt)
/*
	Finds largest elmt-connected set of elmt which has 
	STATUS set to STRIP and includes the seed elmt, and sets the value
	of STATUS to PATCH
*/
ELMTPTR elmt;
{
	NODENEIGHPTR n, indexnn();
	NODEPTR node;
	int i;
	if(elmt->STATUS != STRIP)
		return 1;
	elmt->STATUS = PATCH;
	for(i=0; i<3; i++){
		node = elmt->neigh[i];
		n = indexnn(node, elmt);
		if(node->b != (BDYPTR)NULL && node->n == n)
			continue;
		elmtseedfill(n->next->elmt);
	}
}

pnlpoint(x, y)
/*
	Same as panelpoint in Plotix, except with error checking
	in case contor gets into an infinite loop.
*/
double x, y;
#define MAXPOINT	1000
{
	panelpoint(x, y);
	if(__pointcount++ > MAXPOINT){
		fprintf(stderr, "Too many points: Contor quits\n");
		endpanel();
		return(1);
	}
	return(0);
}

slow_contor(func, nlevel, __panelspass, passMin, passMax)
/*
    Using color '__shade', do a pressure contour plot with 'nlevel'
    contour levels, spaced equally from the minimum to maximum pressure.
*/
double (*func)();
int nlevel, __panelspass;
double passMin, passMax;
{
    ELMTPTR elmt;
    NODEPTR node;
    double f[3], minlocal, maxlocal, Min, Max, temp;
    int i, imin, imax;
	__panels = __panelspass;
	if(nlevel < 1)
		return 1;
    Max = -1.e20;
    Min = 1.e20;
	FORALLNODES(node)
		temp = (*func)(node);
		if(temp > Max)  Max = temp;
		if(temp < Min)  Min = temp;
	NEXTNODE(node)
    GLOBAL_COMBINE(Max, fmax, 1);
    GLOBAL_COMBINE(Min, fmin, 1);
	printf("Min %f Max %f\n", Min, Max); fflush(stdout);
	if(passMin < passMax){
        Min = passMin;
        Max = passMax;
    }
    if(fabs(Max - Min) < 1.e-12){
        printf("Contor: flat function\n");fflush(stdout);
        return(1);
    }

	Min -= 0.001*(Max - Min);

	FORALLELMTS(elmt)
		minlocal = maxlocal = f[0] = (*func)(elmt->neigh[0]);
		for(i=1; i<3; i++){
			f[i] = (*func)(elmt->neigh[i]);
			if(f[i] < minlocal) minlocal = f[i]; 
			if(f[i] > maxlocal) maxlocal = f[i];
		}
		if(minlocal < Min) minlocal = Min;
		if(maxlocal > Max) maxlocal = Max;
		imin = nlevel*(minlocal - Min)/(Max - Min);
		imax = nlevel*(maxlocal - Min)/(Max - Min);
		for(i=imin; i<= imax; i++){
			if(__panels == 1)
				__shade = RAINSTART + 0.75*(RAINEND - RAINSTART)*(double)i/nlevel;
			else if(__panels == 2)
				__shade = WHITE + (BLACK - WHITE)*(double)i/nlevel;
			contor_an_elmt(elmt, Min + i*(Max-Min)/nlevel, f);
		}
		if(!mesh_is_distributed)
			sendplot();
	NEXTELMT(elmt)
	if(mesh_is_distributed)
		usendplot();
	else
		sendplot();
}

contor_an_elmt(elmt, level, f)
/*
    As above, but only for a single elmt. 'f' are the pressure values
    at the 3 neighboring nodes of the elmt.
*/
struct elmt *elmt;
double level, f[3];
{
    int ou, nu, i, i1, first = 1;
    double zz, xx, yy;
    nu = (f[0] > level);
    for(i=0; i<3; i++){
        ou = nu;
        i1 = (i+1)%3;
        nu = (f[i1] > level);
        if(ou ^ nu){
            zz = (level - f[i])/(f[i1] - f[i]);
            xx = zz*elmt->neigh[i1]->x + (1-zz)*elmt->neigh[i]->x;
            yy = zz*elmt->neigh[i1]->y + (1-zz)*elmt->neigh[i]->y;
            if(first){
                first = 0;
                initpanel(__shade, 0);
            }
            panelpoint(xx, yy);
        }
        if(nu){
            if(first){
                first = 0;
                initpanel(__shade, 0);
            }
            panelpoint(elmt->neigh[i1]->x, elmt->neigh[i1]->y);
        }
    }
    if(!first)
		endpanel();
}
