/****************************************************/
/*    DIME: dime_util.c                             */
/*    These are functions used by many              */
/*    different DIME utilities.                     */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/
#define DIMECODE
#include "dime.h"
#ifdef CUBIX
#ifdef DEBUG
#define STDSYS_SIZE			32768
#else
#define STDSYS_SIZE			128
#endif
#endif
int procnum, curr_doc, max_doc, debug, nprocs, *allprocs;
char *buffer;
char scriptname[50];
int mesh_is_distributed = TRUE;
void (*distort)();
double (*sqdist)();
int (*shall_we_swop)();
void (*refine_func)();
void (*swops_func)();
int (*machine_metric)();

NODEPTR the_nodes;
ELMTPTR the_elmts;

int uns, ues, ubs;

MNUPTR root;

#ifdef IBM
sillysilly(argc, argv)
#else
main(argc, argv)
#endif
int argc;
char **argv;
{
	int num;
	int drawmesh(), logicalmesh(), debug_toggle(), balance_orb(), swops(), pause();
	int buffers(), intact_refine(), dime_exit(), postscript(), disp_erase();
	int dimen(), readmesh(), writemesh();
	int zoom_extents(), zoom_window(), zoom_half(), zoom_exp();
	int relax(), menu_include(), menu_record(), menu_end_record();
	int default_shall_we_swop();
	int uniform_metric(), showbalance();
	int draw_hist();
	double default_sqdist();
	MNUPTR dime_menu, script_menu, zoom_menu, user_menu;

#ifdef CUBIX
	char stdsys_buf[STDSYS_SIZE];

	struct nodenv cube;
	exparam(&cube);
	procnum = cube.procnum;
	nprocs = cube.nprocs;
	for(max_doc = 0; (1<<max_doc) < nprocs; max_doc++);

	setvbuf(stdsys, stdsys_buf, _IOCBF, STDSYS_SIZE);
#else
	procnum = 0;
	max_doc = 0;
	nprocs = 1;
#endif
	curr_doc = max_doc;

	swops_func = NULL;
	refine_func = NULL;
	distort = NULL;
	shall_we_swop = default_shall_we_swop;
	sqdist = default_sqdist;
	user_read = NULL;
	post_read = NULL;
	user_write = NULL;
	machine_metric = uniform_metric;

	the_nodes = (NODEPTR)NULL;
	the_elmts = (ELMTPTR)NULL;

	parse(&argc, argv);

	root = menu_make(ROOT);

	user_menu = menu_make(BRANCH);

#ifdef IBM
	dime_user_main(user_menu, argc, argv);
#else
	user_main(user_menu, argc, argv);
#endif

	add_item(root, "USER", BRANCH, user_menu);

	dime_menu = menu_make(BRANCH);
		add_item(dime_menu, "Readmesh", LEAF, readmesh);
		add_item(dime_menu, "Writemesh", LEAF, writemesh);
		add_item(dime_menu, "Drawmesh", LEAF, drawmesh);
		add_item(dime_menu, "Logicalmesh", LEAF, logicalmesh);
		add_item(dime_menu, "Show_balance", LEAF, showbalance);
		add_item(dime_menu, "Postscript", LEAF, postscript);
		add_item(dime_menu, "Rectrefine", LEAF, intact_refine);
		add_item(dime_menu, "Relax", LEAF, relax);
		add_item(dime_menu, "Toporelax", LEAF, swops);

		if(mesh_is_distributed){
		add_item(dime_menu, "Balance_Orb", LEAF, balance_orb);
		add_item(dime_menu, "Histogram", LEAF, draw_hist);
/*		add_item(dime_menu, "Dimension", LEAF, dimen);*/
		}
		add_item(dime_menu, "Memory", LEAF, buffers);
	add_item(root, "DIME", BRANCH, dime_menu);

	script_menu = menu_make(BRANCH);
		add_item(script_menu, "Use_script", LEAF, menu_include);
		add_item(script_menu, "Make_script", LEAF, menu_record);
		add_item(script_menu, "End_script", LEAF, menu_end_record);
	add_item(root, "SCRIPT", BRANCH, script_menu);

	zoom_menu = menu_make(BRANCH);
		add_item(zoom_menu, "Extents", LEAF, zoom_extents);
		add_item(zoom_menu, "Window", LEAF, zoom_window);
		add_item(zoom_menu, "Half", LEAF, zoom_half);
		add_item(zoom_menu, "Explicit", LEAF, zoom_exp);
	add_item(root, "ZOOM", BRANCH, zoom_menu);

	add_item(root, "Terse", LEAF, debug_toggle);
	add_item(root, "Pause", LEAF, pause);
	add_item(root, "Erase", LEAF, disp_erase);
	add_item(root, "Quit", LEAF, dime_exit);

	fsingl(stdout);
	fsingl(stdin);
	debug = 1;

	buffer = malloc(COM_BUF);

	allprocs = (int *)malloc((1<<max_doc)*sizeof(int));
	for(num=0; num<(1<<max_doc); num++) allprocs[num] = num;

	printf("**************************************************\n");
	printf("*           This is DIME version 1.0             *\n");
	printf("**************************************************\n");
	if(max_doc == 0)
	printf("*           Running on 1 processor               *\n");
	else
	printf("*           Running on %4d processors           *\n", 1<<max_doc);
	printf("**************************************************\n");
	printf("Node size %d bytes + %d user bytes\n", sizeof(_NODE), uns);
	printf("Elmt size %d bytes + %d user bytes\n", sizeof(_ELMT), ues);
	printf("Bdy  size %d bytes + %d user bytes\n", sizeof(_BDY), ubs);
	fflush(stdout);

/*
	sprintf(buffer, "DIME [%d]", 1<<max_doc);
	setplxopt("winlab", buffer);
*/
	init_grafix(0.2, graphics, PLOTIX_BUF);
	if(graphics){
		greyscale(WHITE, BLACK);
		rainbow(RAINSTART, RAINEND);
	}
						 
	if(scriptname[0])
		menu_includefile(scriptname);

	menu_input();
}

pause()
{
	printf("Pausing...hit enter to continue...\n");
	getc(stdin);
}

dime_exit()
{
	printf("DIME - Normal Exit\n");
	exit(0);
}

dimen()
{
	printf("New Cube Dimension?\n"); fflush(stdout);
	get_int(&curr_doc);
	if(curr_doc > max_doc)
		curr_doc = max_doc;
}
    
parse(argc, argv)
     int *argc;
     char *argv[];
{
	int iarg;
	char *arg;
    graphics = 2;
	scriptname[0] = '\0';
    
	iarg = 1;
	while(iarg < *argc){
		arg = argv[iarg];
		if(arg[0] == '-' && arg[1] == 'g'){
			graphics = atoi(arg+2);
			delete_argv(iarg, argc, argv);
		}
		else if(arg[0] == '-' && arg[1] == 'f'){
			strcpy(scriptname, argv[iarg+1]);
			delete_argv(iarg, argc, argv);
			delete_argv(iarg, argc, argv);
		} else
			iarg++;
	}
		
	if(graphics != 2 && !scriptname[0]){
		fprintf(stderr, "Script file please? ");
		fflush(stderr);
		scanf("%s", scriptname);
	}
    return 1;
}

delete_argv(iarg, argc, argv)
int iarg, *argc;
char **argv;
{
	int jarg;
	*argc = *argc - 1;
	for(jarg=iarg; jarg<*argc; jarg++)
		argv[jarg] = argv[jarg+1];
}
