/****************************************************/
/*    DIME: meshio.c                                */
/*    To read and write meshes & their data to disk */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/
#define DIMECODE
#include "dime.h"

int (*user_read)();
void (*post_read)();
int (*user_write)();

#define ELMTNO		work_space.ei[0]	/* Element number from input file */
#define NODENO		work_space.ni[0]	/* Node number from input file */
#define BDYNO		work_space.bi[0]

#define MAXSTRUCT		10000
#define MYSTUFF			(!mesh_is_distributed || num[0]/MAXSTRUCT == procnum)

FILE *in;

readmesh()
{
	char string[256];
	FILE *in;

	printf("Boundary Filename (without suffix)?\n");
	get_string(string);
	strcat(string, ".bdy");
	if((in = fopen(string, "r")) == (FILE *)NULL){
		printf("Cannot open %s\n", string);
		return 0;
	}
	readbdy(in);

	printf("Mesh Filename (without suffix)?\n");
	get_string(string);
	strcat(string, ".mesh");
	if((in = fopen(string, "r")) == (FILE *)NULL){
		printf("Cannot open %s\n", string);
		return 0;
	}
	if(readct(in)){
		sortelmts();
#ifdef DEBUG
printf("sortelmts done\n"); fflush(stdout);
#endif
		zoom_extents();
		drawmesh();
		if(post_read)
			(*post_read)();
	} else
		freemesh();
}

readct(in)
FILE *in;
/*
	To read in a definition of a coarse mesh to feed to dime. The tool
	coarse_tool will produce such a file with the extension .mesh.
*/
{
	ELMTPTR elmt, elmt_alloc();
	NODEPTR node, node_alloc(), find_node(), othernode;
	BDYPTR b_alloc();
	ALIASPTR ali, astart, alias_alloc();
	int i, num[3], currproc, entity;
	int dest, source, nbytes, type, othernum;
	char *ptr, *mailbag, line[1024];
	double x, y;
	struct curve *curve;
	struct point *point;
/* 
	The input file should have a .mesh extension
*/
	freemesh();

	fsingl(in);
	if(mesh_is_distributed)
		printf("Mesh is MULTI\n");
	else
		printf("Mesh is SINGL\n");
#ifdef DEBUG
fmulti(stdout);
#endif
	while(getline(in, line)){
		if(line[0] == ' ' || line[0] == '\t')
			break;

		if(line[0] == '%'){
			(*user_read)(line+1, MISC, NULL);
			break;
		}

		switch(line[0]){
		case 'N':
			sscanf(line+1, "%d %d %d %lf %lf", num, &type, &entity, &x, &y);
#ifdef DEBUG
printf("N %d %d %d %lf %lf\n", num[0], type, entity, x, y);
#endif
			if(!MYSTUFF) break;
			node = node_alloc();
			node->NODENO = num[0];
			switch(type){
			case POINT:
				node->b = b_alloc();
				node->b->type = POINT;
				FORALLPOINTS(point)
					if(point->ptnum == entity){
						node->b->bdy.point = point;
						break;
					}
				NEXTPOINT(point)
				node->x = node->b->bdy.point->x;
				node->y = node->b->bdy.point->y;
				break;
			case CURVE:
				node->b = b_alloc();
				node->b->type = CURVE;
				FORALLCURVES(curve)
					if(curve->curvenum == entity){
						node->b->bdy.curve = curve;
						break;
					}
				NEXTCURVE(curve)
				node->x = x;
				node->y = y;
				node->b->s = x;
				get_point(node->b->bdy.curve, node->b->s, &node->x, &node->y);
				break;
			case INSIDE:
				node->x = x;
				node->y = y;
				break;
			}
			if(user_read){
				ptr = line;
				while(*ptr != '\0')
					if(*ptr++ == '%') {
						(*user_read)(ptr, NODE, node);
						break;
					}
			}
			break;
		case 'B':
			sscanf(line+1, "%d %d", num, num+1);
#ifdef DEBUG
printf("B %d %d\n", num[0], num[1]);
#endif
			if(!MYSTUFF) break;
			node = find_node(num[0]);
			if(node->b == (BDYPTR)NULL)
				node->b = b_alloc();
			node->b->clock = find_node(num[1]);
			if(user_read){
				ptr = line;
				while(*ptr != '\0')
					if(*ptr++ == '%') {
						(*user_read)(ptr, BDY, node->b);
						break;
					}
			}
			break;
		case 'E':
			sscanf(line+1, "%d %d %d", num, num+1, num+2);
#ifdef DEBUG
printf("E %d %d %d\n", num[0], num[1], num[2]);
#endif
			if(!MYSTUFF) break;
			elmt = elmt_alloc();
			for(i=0; i<3; i++)
				elmt->neigh[i] = find_node(num[i]);
			if(user_read){
				ptr = line;
				while(*ptr != '\0')
					if(*ptr++ == '%') {
						(*user_read)(ptr, ELMT, elmt);
						break;
					}
			}
			break;
		case 'A':
			sscanf(line+1, "%d %d", num, num+1);
#ifdef DEBUG
printf("A %d %d\n", num[0], num[1]);
#endif
			if(!MYSTUFF) break;
			node = find_node(num[0]);
			if(node->b == (BDYPTR)NULL)
				node->b = b_alloc();
			add_alias(node, num[1], procnum | HIBITS);
			break;
		case 'a':
			sscanf(line+1, "%d %d", num, num+1);
#ifdef DEBUG
printf("a %d %d\n", num[0], num[1]);
#endif
			if(!MYSTUFF) break;
			node = find_node(num[0]);
			if(node->b == (BDYPTR)NULL)
				node->b = b_alloc();
			add_alias(node, num[1], num[1]/MAXSTRUCT);
			break;
		case 'O':
			sscanf(line+1, "%d %d", num, num+1);
#ifdef DEBUG
printf("O %d %d\n", num[0], num[1]);
#endif
			if(!MYSTUFF) break;
			node = find_node(num[0]);
			if(node->b == (BDYPTR)NULL)
				node->b = b_alloc();
			node->b->oppalias = alias_alloc();
			node->b->oppalias->node = (struct node *)(num[1]);
			node->b->oppalias->proc = procnum|HIBITS;
			break;
		case 'o':
			sscanf(line+1, "%d %d", num, num+1);
#ifdef DEBUG
printf("o %d %d\n", num[0], num[1]);
#endif
			if(!MYSTUFF) break;
			node = find_node(num[0]);
			if(node->b == (BDYPTR)NULL)
				node->b = b_alloc();
			node->b->oppalias = alias_alloc();
			node->b->oppalias->node = (struct node *)(num[1]);
			node->b->oppalias->proc = num[1]/MAXSTRUCT;
			break;
		}
#ifdef DEBUG
fflush(stdout);
#endif
	}

	fclose(in);
#ifdef DEBUG
fsingl(stdout);
#endif

	FORALLELMTS(elmt)
		for(i=0; i<3; i++)
			add_neigh(elmt, elmt->neigh[i]);
	NEXTELMT(elmt)

#define OPPMESS		0
#define ALIMESS		1

	FORALLNODES(node)
		if(node->b == (BDYPTR)NULL)
			continue;

		if(node->b->oppalias != (ALIASPTR)NULL){
			mailbag = buffer;
			xpacket(mailbag, int) = OPPMESS;
			xpacket(mailbag, int) = (int)node->b->oppalias->node;
			xpacket(mailbag, int) = node->NODENO;
			xpacket(mailbag, NODEPTR) = node;
			dest = (node->b->oppalias->proc) & LOWBITS;
			if(Add_to_out_buf(buffer, (int)(mailbag-buffer), 1, &dest, 1) == -1)
					abort(903);                     /* AOB failed in meshio */
		}

		if(node->b->alias != (ALIASPTR)NULL){
			ali = astart = node->b->alias;
			do {
				mailbag = buffer;
				xpacket(mailbag, int) = ALIMESS;
				xpacket(mailbag, int) = (int)ali->node;
				xpacket(mailbag, int) = node->NODENO;
				xpacket(mailbag, NODEPTR) = node;
				dest = (ali->proc) & LOWBITS;
				if(Add_to_out_buf(buffer, (int)(mailbag-buffer), 1, &dest, 1) == -1)
						abort(903);                     /* AOB failed in meshio */
			} while((ali = ali->next) != astart);
		}

	NEXTNODE(node)
#ifdef DEBUG
printf("Into Crystal_Router\n"); fflush(stdout);
#endif
	if(Crystal_router() == -1)
			abort(904);
#ifdef DEBUG
printf("Out of Router\n"); fflush(stdout);
#endif
	while(Get_message(buffer, &source, &nbytes, 1) == 0){
		mailbag = buffer;
		type = xpacket(mailbag, int);
		node = find_node(xpacket(mailbag, int));
		othernum = xpacket(mailbag, int);
		othernode = xpacket(mailbag, NODEPTR);
		if(type == OPPMESS){
			if(node->b && node->b->oppalias
					&& (int)node->b->oppalias->node == othernum)
				node->b->oppalias->node = othernode;
			else
				abort(905);			/* inconsistens oppalias */
		} else {
			if(node->b && node->b->alias){
				ali = astart = node->b->alias;
				do {
					if((int)ali->node == othernum){
						ali->node = othernode;
						goto done;
					}
				} while((ali = ali->next) != astart);
				abort(906);			/* inconsistent alias */
			}
		}
done:;
	}
#ifdef DEBUG
printf("Got messages\n"); fflush(stdout);
#endif

	FORALLNODES(node)
		if(node->b != (BDYPTR)NULL){
			if(node->b->clock == (NODEPTR)NULL){
				printf("Improper Boundaries\n");
				return 0;
			}
			if(node->b != (BDYPTR)NULL && node->b->clock->b != (BDYPTR)NULL)
				node->b->clock->b->antik = node;
		}
		sortout(node);
	NEXTNODE(node)

#ifdef DEBUG
printf("Done Boundaries\n"); fflush(stdout);
#endif
	FORALLNODES(node)
		if(node->b != (BDYPTR)NULL){
			if(B_EDGE(node) || B_EDGE(node->b->antik))
				node->b->is_on_bdy |= 1;
			if(node->b->alias != (ALIASPTR)NULL)
				node->b->is_on_bdy |= 2;
		}
	NEXTNODE(node)

#ifdef DEBUG
printf("Mesh Read\n"); fflush(stdout);
#endif
	fflush(stdout);
	return 1;
}

getline(in, line)
FILE *in;
char *line;
{
	int c;
	char *ptr = line;
	while(1){
		c = getc(in);
		if(c == EOF){
			*ptr++ = '\0';
			return 0;
		}
		if(c == '\n'){
			*ptr++ = '\0';
			return 1;
		}
		*ptr++ = c;
	}
}

NODEPTR find_node(inode)
int inode;
/*
	Returns the address of the node with number inode
*/
{
	NODEPTR node;
	FORALLNODES(node)
		if(node->NODENO == inode)
			return(node);
	NEXTNODE(node)
	printf("Cannot find node %d\n", inode);
	abort(401);							/* find_node failed in readct */
}

writemesh()
{
	char string[30];
	FILE *out;
	printf("Filename (without suffix)?\n");
	get_string(string);
	strcat(string, ".mesh");
	if((out = fopen(string, "w")) == (FILE *)NULL){
		printf("Cannot open %s\n", string);
		fflush(stdout);
	} else
		writect(out);
}

writect(out)
FILE *out;
{
	NODEPTR node, othernode;
	ELMTPTR elmt;
	ALIASPTR ali, astart;
	int i, ibdy, dest, num, source, nbytes, messtype;
	char *mailbag;
	char string[1024];

	if(user_write && (*user_write)(string, MISC, NULL)){
		fprintf(out, "%% %s\n", string);
		fflush(out);
	}

	if(mesh_is_distributed)
		fmulti(out);

	i = MAXSTRUCT*procnum;
	FORALLELMTS(elmt)
		elmt->ELMTNO = i++;
	NEXTELMT(elmt)

	i = MAXSTRUCT*procnum;
	FORALLNODES(node)
		node->NODENO = i++;

		if(!node->b)
			continue;

		if(node->b->oppalias != (ALIASPTR)NULL){
			mailbag = buffer;
			xpacket(mailbag, int) = OPPMESS;
			xpacket(mailbag, int) = node->NODENO;
			xpacket(mailbag, NODEPTR) = node;
			xpacket(mailbag, NODEPTR) = node->b->oppalias->node;
			dest = (node->b->oppalias->proc) & LOWBITS;
			if(Add_to_out_buf(buffer, (int)(mailbag-buffer), 1, &dest, 1) == -1)
					abort(903);                     /* AOB failed in meshio */
		}

		if(node->b->alias != (ALIASPTR)NULL){
			ali = astart = node->b->alias;
			do {
				mailbag = buffer;
				xpacket(mailbag, int) = ALIMESS;
				xpacket(mailbag, int) = node->NODENO;
				xpacket(mailbag, NODEPTR) = node;
				xpacket(mailbag, NODEPTR) = ali->node;
				dest = (ali->proc) & LOWBITS;
				if(Add_to_out_buf(buffer, (int)(mailbag-buffer), 1, &dest, 1) == -1)
						abort(903);                     /* AOB failed in meshio */
			} while((ali = ali->next) != astart);
		}
	NEXTNODE(node)

	if(Crystal_router() == -1)
			abort(904);

	while(Get_message(buffer, &source, &nbytes, 1) == 0){
		mailbag = buffer;
		messtype = xpacket(mailbag, int);
		num = xpacket(mailbag, int);
		othernode = xpacket(mailbag, NODEPTR);
		node = xpacket(mailbag, NODEPTR);
		if(messtype == OPPMESS){
			node->b->oppalias->node = (NODEPTR)num;
		} else {
			ali = astart = node->b->alias;
			do {
				if(ali->node == othernode && ali->proc == source)
					ali->node = (NODEPTR)num;
			} while((ali = ali->next) != astart);
		}
	}


	FORALLNODES(node)
		fprintf(out, "N %d ", node->NODENO);

		if(!B_NODE(node)){
			fprintf(out, "2 0 %f %f %%", node->x, node->y);
		} else {
			if(node->b->type == CURVE)
				fprintf(out, "1 %d %f %%", node->b->bdy.curve->curvenum,
					node->b->s);
			else
				fprintf(out, "0 %d %%", node->b->bdy.point->ptnum);
		}

		if(user_write && (*user_write)(string, NODE, node))
			fprintf(out, "%s \n", string);
		else
			fprintf(out, "\n");
	NEXTNODE(node)

	FORALLNODES(node)
		if(node->b == (BDYPTR)NULL)
			continue;

		if(node->b->oppalias != (ALIASPTR)NULL){
			num = (int)(node->b->oppalias->node);
			if((node->b->oppalias->proc) & HIBITS)
				fprintf(out, "O %d %d\n", node->NODENO, num);
			else
				fprintf(out, "o %d %d\n", node->NODENO, num);
		}

		if(node->b->alias != (ALIASPTR)NULL){
			ali = astart = node->b->alias;
			do {
				num = (int)(ali->node);
				if((ali->proc) & HIBITS)
					fprintf(out, "A %d %d\n", node->NODENO, num);
				else
					fprintf(out, "a %d %d\n", node->NODENO, num);
			} while((ali = ali->next) != astart);
		}

		fprintf(out, "B %d %d %%", node->NODENO, node->b->clock->NODENO);

		if(user_write && (*user_write)(string, BDY, (char *)node->b))
			fprintf(out, " %s\n", string);
		else
			fprintf(out, "\n");
	NEXTNODE(node)

	FORALLELMTS(elmt)
		fprintf(out, "E ");
		for(i=0; i<3; i++)
			fprintf(out, "%d ", elmt->neigh[i]->NODENO);
		fprintf(out, " %% ");

		if(user_write && (*user_write)(string, ELMT, elmt))
			fprintf(out, "%s\n", string);
		 else
			fprintf(out, "\n");
	NEXTELMT(elmt)
	fflush(out);

	fclose(out);
}
