/****************************************************/
/*    DIME: dime.h                                  */
/*    Header file containing structure definitions  */
/*    externs and macros.                           */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/
#include <stdio.h>
#include <math.h>
#include "gr_prim.h"
#include "menu.h"
#include "bdy.h"
#include "colors.h"
#ifdef CUBIX
#include "express.h"
#else
#define ALLNODES	0
#define DONTCARE	0
#endif

typedef struct node { 
	double x, y;					/* Node position */
	struct bdy *b;				/* Boundary structure or NULL */
	struct nodeneigh *n;		/* Pointer to list of neighbors */
	struct node *last, *next;	/* Double list of all nodes */
	struct node *bdynext;
#ifndef DIMECODE
	struct user_node *user;
#else
	char *user;					/* Pointer to user data */
#endif

#define			NODE_WORK_SPACE		4
	union {
		double		 nf[NODE_WORK_SPACE];
		int			 ni[NODE_WORK_SPACE];
		struct node	*na[NODE_WORK_SPACE];
	} work_space;				/* 8 bytes for work space */
} _NODE, *NODEPTR;

typedef struct bdy {
	int is_on_bdy;
	struct node *clock, *antik;	/* Next along boundary (downstream) */
	struct alias *alias;		/* If processor boundary, here are aliases */
	struct alias *oppalias;		/* Pointer to opposite elmt */
	int type;
	union {
		struct curve *curve;
		struct point *point;
	} bdy;
	double s;
#ifndef DIMECODE
	struct user_bdy *user;
#else
	char *user;					/* Pointer to user data */
#endif

#define			BDY_WORK_SPACE		2
	union {
		double       bf[BDY_WORK_SPACE];
		int			 bi[BDY_WORK_SPACE];
		struct node *ba[BDY_WORK_SPACE];
	} work_space;				/* 8 bytes of working space */
} _BDY, *BDYPTR;

typedef struct alias {
	int proc;					/* Procnum of sibling node */
	struct node *node;			/* Address of sibling in other processor */
	struct alias *next;			/* Next alias in linked list */
} ALIAS, *ALIASPTR;

typedef struct nodeneigh {
	struct elmt *elmt;			/* Pointer to this elmt */
	struct nodeneigh *next;		/* Next in neighbor list */
} NODENEIGH, *NODENEIGHPTR;

typedef struct elmt {
	struct node *neigh[3];		/* 3 neighbor nodes of this elmt */
	struct elmt *last, *next;	/* Double list of all elmts */
#ifndef DIMECODE
	struct user_elmt *user;
#else
	char *user;					/* Pointer to user data */
#endif

#define			ELMT_WORK_SPACE		1
	union {
		double	ef[ELMT_WORK_SPACE];
		int		ei[ELMT_WORK_SPACE];
	} work_space;				/* 8 bytes of work space */
} _ELMT, *ELMTPTR;

extern NODEPTR	the_nodes;
extern ELMTPTR	the_elmts;

extern int procnum, curr_doc, max_doc, debug, nprocs;
extern int *allprocs;
extern char *buffer;
extern int uns, ues, ubs;
/*extern char *malloc();*/
extern char *smalloc();
extern char psbuf[];
extern int or(), fsum(), isum(), fmax(), imax(), fmin(), imin();
extern int nntype, atype;		/* used for sfree'ing nodeneigh's */
extern int mesh_is_distributed;
extern MNUPTR root;
extern int combine_is_ok;

extern int (*user_read)();
extern void (*post_read)();
extern int (*user_write)();
extern void (*refine_func)();
extern void (*swops_func)();
extern int (*shall_we_swop)();
extern double (*sqdist)();
extern void (*distort)();
extern int (*machine_metric)();
#define DISTORT(node, p, q)		{if(!distort){p = node->x; q = node->y;}\
									else (*distort)(node, &p, &q);}

#ifdef TINYMEM
#define PLOTIX_BUF			32768
#else
#define PLOTIX_BUF			32767
#endif


#define COM_BUF				32767
#define COA_BUF				2048

#define xpacket(p, mode)  ((mode *)(p += sizeof(mode)))[-1]
#define ANY					999
#define LOWBITS				0xffff
#define HIBITS				0x10000

#define ELMT				1
#define NODE				2
#define BDY					3
#define MISC				4

#define TRUE				1
#define FALSE				0

#define POINT				0
#define CURVE				1
#define INSIDE				2

#define remove		Remove			/*The name 'remove' conflicts with cubix */

#ifdef NOABORT
#define abort		Abort
#endif

#define DIME_PREAMBLE	\
   uns = sizeof(struct user_node);  \
   ues = sizeof(struct user_elmt);  \
   ubs = sizeof(struct user_bdy)

#define FORALLNODES(node) \
	if((node = the_nodes) != (NODEPTR)NULL){ do {

#define NEXTNODE(node) \
	} while((node = node->next) != the_nodes); }

#define FORALLELMTS(elmt) \
	if((elmt = the_elmts) != (ELMTPTR)NULL){ do {

#define NEXTELMT(elmt) \
	} while((elmt = elmt->next) != the_elmts); }

#define GLOBAL_COMBINE(variable, func, number) \
	{int type = 3007; \
	excombine(&variable, func, sizeof(variable), number, \
		ALLNODES, NULL, &type);}

#define NODE_COMBINE(member, func, number) \
	{struct user_node dummy; \
	node_combine((char *)(&dummy.member) - (char *)(&dummy), \
	sizeof(dummy.member), number, func, 0, 0, (int *)0);}

#define BDY_COMBINE(member, func, number) \
	{struct user_bdy dummy; \
	bdy_combine((char *)(&dummy.member) - (char *)(&dummy), \
	sizeof(dummy.member), number, func, 0);}

#define ELMT_EXCHANGE(from, to, number) \
	{struct user_elmt dummy; \
	elmt_exchange((char *)(dummy.from) - (char *)(&dummy), \
	(char *)(dummy.to  ) - (char *)(&dummy), \
	sizeof(dummy.from[0])*(number));}

#define B_EDGE(node) \
	(node->b != (BDYPTR)NULL && node->b->oppalias == (ALIASPTR)NULL)

#define B_NODE(node) \
	(node->b != (BDYPTR)NULL && (node->b->is_on_bdy&1))

#define B_OPP(node) \
	(node->b != (BDYPTR)NULL && (node->b->is_on_bdy&2))

#define FORALLNEIGH(node, Elmt) \
	{NODENEIGHPTR n, nstart;	\
	n = nstart = node->n; do {	\
	Elmt = n->elmt;

#define NEXTNEIGH(node, Elmt)	\
	} while((n = n->next) != nstart);}

#define ELMT_DATA(c, member, type) \
	{struct user_elmt dummy; \
	elmt_data(c, (char *)(&dummy.member) - (char *)(&dummy), type);}

#define NODE_DATA(c, member, type) \
	{struct user_node dummy; \
	node_data(c, (char *)(&dummy.member) - (char *)(&dummy), type);}

#define BDY_DATA(c, member, type) \
	{struct user_bdy dummy; \
	bdy_data(c, (char *)(&dummy.member) - (char *)(&dummy), type);}

#define B_CURVE(node) \
	(B_NODE(node) && (node)->b->type == CURVE ? \
	(node)->b->bdy.curve : (CURVEPTR)0)

#define B_POINT(node) \
	(B_NODE(node) && (node)->b->type == POINT ? \
	(node)->b->bdy.point : (POINTPTR)0)

#define B_EDGE_CURVE(node) \
	(B_EDGE(node) ? (\
		(node)->b->type == CURVE ? (node)->b->bdy.curve : \
			((node)->b->clock->b->type == CURVE ?  \
				(node)->b->clock->b->bdy.curve : (CURVEPTR)0 \
			) \
		) : (CURVEPTR)0 \
	)
