/*
 * bgidriv.c
 *
 * Driver for PLOTIX using the BGI libraries.
 */
#include "plotix.h"
#include "sys/plotix.h"
#include <stdio.h>
#include "bgi.h"
#include "dos.h"
/*
 * This next include file was supplied by Borland and contains code ....
 */
#include "mouse.inc"

/*
 * The _plx_bgi_showmouse, _plx_bgi_hidemouse pair are used here to
 *  protect the screen from being overwritten by the mouse.  They could
 *  be made macros for enhanced speed.
 *
 * Note however!  termbgi in express/common uses these before writing to the
 *  screen.
 */
void _plx_bgi_showmouse() {
    if(DEP(mouse)) mShow();
}

void _plx_bgi_hidemouse() {
    if(DEP(mouse)) mHide();
}

static void _plx_bgi_cont(x,y)
     Coord x,y;
{
    int newx, newy;

    newx = USER_to_X(x);
    newy = USER_to_Y(y);

    _plx_bgi_hidemouse();
    lineto((int)newx, (int)newy);
    _plx_bgi_showmouse();
    return;     
}

static void _plx_bgi_move(x,y)
     Coord x,y;
{
    int newx, newy;

    newx = USER_to_X(x);
    newy = USER_to_Y(y);

    moveto((int)newx, (int)newy);
    return;
}

/*
 * Can share the shading/color definitions between various
 * instances of this driver.
 */
static unsigned char greys[]={
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,     /* Grey 0 */
    0x88, 0x00, 0x22, 0x00, 0x88, 0x00, 0x22, 0x00,     /* Grey 1 */
    0xaa, 0x00, 0xaa, 0x00, 0xaa, 0x00, 0xaa, 0x00,     /* Grey 2 */
    0xaa, 0x44, 0xaa, 0x11, 0xaa, 0x44, 0xaa, 0x11,     /* Grey 3 */
    0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55,     /* Grey 4 */
    0xee, 0x55, 0xbb, 0x55, 0xee, 0x55, 0xbb, 0x55,     /* Grey 5 */
    0xff, 0x55, 0xff, 0x55, 0xff, 0x55, 0xff, 0x55,     /* Grey 6 */
    0xff, 0xdd, 0xff, 0xff, 0xff, 0xdd, 0xff, 0xff,     /* Grey 7 */
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,     /* Grey 8 */
};
static unsigned char not_greys[sizeof(greys)];

static unsigned char hatch[]={
    0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,     /* SUN hatch 1 */
    0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0,     /* SUN hatch 2 */
    0x81, 0x03, 0x06, 0x0c, 0x18, 0x30, 0x60, 0xc0,     /* SUN hatch 3 */
    0x81, 0xc0, 0x60, 0x30, 0x18, 0x0c, 0x06, 0x03,     /* SUN hatch 4 */
    0x18, 0x30, 0x60, 0xc0, 0x81, 0xc0, 0x60, 0x30,     /* SUN hatch 5 */
    0x08, 0x00, 0x00, 0x80, 0xc1, 0x63, 0x36, 0x1c,     /* SUN hatch 6 */
    0xf0, 0xf0, 0xf0, 0xf0, 0x0f, 0x0f, 0x0f, 0x0f,     /* SUN hatch 7 */
    0x01, 0x01, 0x01, 0xff, 0x10, 0x10, 0x10, 0xff,     /* SUN hatch 8 */
    0x3c, 0x66, 0xc3, 0x81, 0x81, 0xc3, 0x66, 0x3c,     /* SUN hatch 9 */
    0x88, 0x88, 0x88, 0xff, 0x88, 0x88, 0x88, 0xff,     /* SUN hatch 10 */
};
#define NGREYS  (sizeof(greys)/8)
#define NHATCH  (sizeof(hatch)/8)

static void initpolys()
{
    int i;
    for(i=0; i<sizeof(greys); i++) not_greys[i] = ~greys[i];
    return;
}

static void _plx_bgi_initpanel(shade,edge)
     int shade,edge;
{
    DEP(sv_shade) = shade;
    DEP(sv_edge)  = edge;
    DEP(polys) = DEP(polyverts);
    DEP(Npts) = 0;
    return;
}

static void _plx_bgi_panelpoint(x, y)
     Coord x, y;
{
    int x_curr, y_curr;

    x_curr = USER_to_X(x);
    y_curr = USER_to_Y(y);

    if(++DEP(Npts) > MAXVERTS) {
        char text[128];
        sprintf(text,"Too many polygon vertices, max = %d\n", MAXVERTS);
        _plx_bgi_hidemouse();
	_ex_print(text);	/* _ex_info ? */
        _plx_bgi_showmouse();
        DEP(Npts) = 0;
	DEP(polys) = DEP(polyverts);
        return;
    }
    *DEP(polys++) = x_curr;
    *DEP(polys++) = y_curr;

    return;
}

static void _plx_bgi_endpanel()
{
    int i, shade, index;
    char *hatch_pat;

/* First we have to close the polygon, if necessary */

    if(*(DEP(polys)-2) != DEP(polyverts[0]) || 
       *(DEP(polys)-1) != DEP(polyverts[1])) {
        if(++DEP(Npts) > MAXVERTS) {
            char text[128];

            sprintf(text,"Too many polygon vertices, max = %d\n", MAXVERTS);
            _plx_bgi_hidemouse();
            _ex_print(text);
            _plx_bgi_showmouse();
            DEP(Npts) = 0;
	    DEP(polys) = DEP(polyverts);
            return;
        }
        *DEP(polys++) = DEP(polyverts[0]);        /* Close polygon for filler */
        *DEP(polys++) = DEP(polyverts[1]);
    }

/* Now select the fill pattern and color */

    if(DEP(sv_shade) < 0) {
        index = (-DEP(sv_shade))+1;
        hatch_pat = hatch+8*(index % NHATCH);
        shade = ((index / NHATCH) % 7) + 1;
        setfillpattern(hatch_pat, DEP(plxmap[shade]));
    }
    else if(IND(grey_lo) <= DEP(sv_shade) && DEP(sv_shade) <= IND(grey_hi)) {
        i = (NGREYS*(DEP(sv_shade)-IND(grey_lo)))/
			(IND(grey_hi)-IND(grey_lo)+1);
        hatch_pat = greys+8*i;
        setfillpattern(hatch_pat, WHITE);
    }
    else if(IND(rain_lo) <= DEP(sv_shade) && DEP(sv_shade) <= IND(rain_hi)) {
        i = (24.0*(DEP(sv_shade)-IND(rain_lo)))/(IND(rain_hi)-IND(rain_lo)+1);
        shade =  2 + i/4;
        hatch_pat = greys + 8*(2 + (i % 4));
        setfillpattern(hatch_pat, DEP(plxmap[shade]));
    }
    else {
        setfillstyle(SOLID_FILL, DEP(plxmap[(DEP(sv_shade) % 8)]));
    }

    _plx_bgi_hidemouse();
    fillpoly(DEP(Npts), DEP(polyverts));
    if(!DEP(sv_edge)) {
        setcolor(BLACK);        /* No borders */
        drawpoly(DEP(Npts), DEP(polyverts));
        setcolor(DEP(plxmap[DEP(fgcolor)]));
    }
    _plx_bgi_showmouse();
    DEP(Npts) = 0;
    DEP(polys) = DEP(polyverts);
}

static void _plx_bgi_point(x,y)
     Coord x,y;
{
    int newx, newy;
    newx = USER_to_X(x);
    newy = USER_to_Y(y);

    _plx_bgi_hidemouse();
    putpixel(newx, newy, DEP(plxmap[DEP(fgcolor)]));
    _plx_bgi_showmouse();

    return;
}

static void _plx_bgi_aspect(xsize,ysize)		/* Inches */
     Coord *xsize,*ysize;
{
    *xsize = 11.0;
    *ysize = 8.0;
    return;
}

static void _plx_bgi_raspect(xsize,ysize)		/* Pixels */
     Coord *xsize,*ysize;
{
    *xsize = IND(devx);
    *ysize = IND(devy);
    return;
}

static void _plx_bgi_closepl()
{
    sendplot();
    _plx_bgi_hidemouse();
    if(DEP(mouse) != 0) mReset();       /* Reset mouse */
    _ex_deinstall_bgi();
    closegraph();
    return;
}

static void _plx_bgi_color(shade)
     int shade;
{
    DEP(fgcolor) = shade % 8;
    setcolor(DEP(plxmap[DEP(fgcolor)]));
    return;
}

static void _plx_bgi_dotext(string, x, y, angle, ljust, hjust)
     char *string;
     Coord x, y;
     int ljust, hjust, angle;
{
    int x_curr,y_curr;
    enum text_just up_down, left_right;

    x_curr = USER_to_X(x);
    y_curr = USER_to_Y(y);

    if(ljust < 0)       left_right = LEFT_TEXT;
    else if(ljust == 0) left_right = CENTER_TEXT;
    else                left_right = RIGHT_TEXT;

    if(hjust < 0)       up_down = BOTTOM_TEXT;
    else if(hjust == 0) up_down = CENTER_TEXT;
    else                up_down = TOP_TEXT;

    settextjustify(left_right, up_down);
    _plx_bgi_hidemouse();
    outtextxy(x_curr, y_curr, string);
    _plx_bgi_showmouse();
    return;
}

static void _plx_bgi_sztext(str, pwd, pht)
     char *str;
     Coord *pwd, *pht;
{
    *pwd = (double)textwidth(str)/IND(mul[0]);
    *pht = (double)textheight(str)/IND(mul[1]);
    return;
}

static void _plx_bgi_intext(prompt, x, y, term, input, maxlen)
     char *prompt;
     Coord x, y;
     char *term, *input;
     int maxlen;
{
    Coord in_x;
    int at_x, at_y;
    char c, *at_ptr, *strchr();
    int got = 0;
    int msize;

    memset(input, 0, maxlen);
    dotext(prompt, x, y, 0, -1, -1);

    in_x = x + ((Coord) textwidth(prompt)/IND(mul[0]));
    at_x = USER_to_X(in_x);
    at_y = USER_to_Y(y);

    at_ptr = input;
    while (kbhit()) getch();	/*  Flush buffer  */
    do {
	c = getch();
	if(c == '\r') c = '\n';

	if((c == 8 || c == 127) && got > 0) {	/* CTRL-H or DEL */
	    int ocol = getcolor();
	    setcolor(BLACK);
            _plx_bgi_hidemouse();
	    bar(at_x,at_y,at_x+textwidth(input),at_y-textheight(input));
            _plx_bgi_showmouse();
	    setcolor(ocol);
	    *--at_ptr = '\0';
	    got--;
	} else if (c >= 32) {		/* Only ASCII, please! */
	    *at_ptr++ = c;
	    got++;
	}
	dotext(input, in_x, y, 0, -1, -1);
    } while(got < maxlen && strchr(term, c) == (char *)0);
    *at_ptr = '\0';
    return;
}

static void _plx_bgi_erase()
{
    _plx_bgi_hidemouse();
    cleardevice();
    _plx_bgi_showmouse();
    return;
}

static int _plx_bgi_gin(key,x,y)
     int *key;
     Coord *x,*y;
{
    int c, done, motion;
    int x0, y0, del_x, del_y;

    if(DEP(mouse)) {      /* Real mouse exists, use MicroSoft interface */
        mouse_gin(key, &DEP(mouse_x), &DEP(mouse_y));
    } else {
        done = 0;
        while(!done) {
            kmShow();
            c = getch();
            if(c == 0x0) {      /* Begins "ESCAPE" sequence */
                c = getch();    /* .. so get other half */
                motion = 1;
            }
            else if (c < 32 && ( c != '\n' && c != '\r') ) motion = 1;
            else motion = 0;
            del_x = del_y = 0;
            if(motion) {
                switch(c) {
                case 0x48:              /* Up arrow */
                case 0x15:
                    del_y = ((c < 32) ? -10*DEP(rate) : -DEP(rate));
                    break;
                case 0x50:              /* Down arrow */
                case 0xe:
                    del_y = ((c < 32) ? 10*DEP(rate) : DEP(rate));
                    break;
                case 0x4d:              /* Right arrow */
                case 0x6:
                    del_x = ((c < 32) ? 10*DEP(rate) : DEP(rate));
                    break;
                case 0x4b:              /* Left arrow */
                case 0x2:
                    del_x = ((c < 32) ? -10*DEP(rate) : -DEP(rate));
                    break;
                case 0x49:              /* Page up */
                    del_y = -DEP(mouse_y);
                    break;
                case 0x51:              /* Page down */
                    del_y = IND(devy) - DEP(mouse_y) - 1;
                    break;
                case 0x47:              /* Home key */
                    if(DEP(rate) == 1) DEP(rate) = 10;
                    else             DEP(rate) = 1;
                    break;
                case 0x3:               /* CTRL-C */
                    closepl();
                    exit(0);
                }
            }
            else {                      /* Non-motion char, terminate */
                *key = c;
                done = 1;
            }

/* "hide" the cursor from its old position before incrementing for next sweep.
*/
            kmHide();
            DEP(mouse_x) += del_x;
            DEP(mouse_y) += del_y;
        }
    }

/* Normalise coordinates and return them */

    y0 = (int) IND(devy) - DEP(mouse_y) - 1;                /* invert y-axis */
    x0 = DEP(mouse_x);

/* Decide whether the coordinates lie within this processor */

    *x = (((Coord) x0) - IND(off[0])) / IND(mul[0]);
    *y = (((Coord) y0) - IND(off[1])) / IND(mul[1]);

    return 0;
}

static int _plx_bgi_gtrack_on()
{
    if(DEP(mouse)) return 0;
    else return -1;
/* NOTREACHED */
}

int _plx_bgi_loopcheck() {
    kbhit();
    return 0;
}

static int _plx_bgi_gtrack(pkey, px, py)
     int *pkey;
     Coord *px, *py;
{
    int x0, y0;
    locRec *m, *mPos();

    if(DEP(mouse)) {      /* Real mouse exists, use MicroSoft interface */

	m = mPos();
	*pkey = m->buttonStatus & 0x7;

/* Normalise coordinates and return them */

	y0 = (int) IND(devy) - m->row - 1;     /* invert y-axis */
	x0 = m->column;

/* Decide whether the coordinates lie within this processor */

        *px = (((Coord) x0) - IND(off[0])) / IND(mul[0]);
        *py = (((Coord) y0) - IND(off[1])) / IND(mul[1]);
    }
    return 0;
}

static void _plx_bgi_gtrack_off()
{
    return;
}

#define EVENTMASK       (0x54)  /* Any mouse button released */

_init_mouse()
{
    resetRec *theMouse;

    theMouse = mReset();
    if(theMouse->exists) {
/*
        mInstTask(EVENTMASK, FP_SEG(handler), FP_OFF(handler));
        theEvents = MK_FP(_psp, 0x00c0);
*/
	_plx_bgi_setcurs(PLX_PTR_CURSOR);
        return theMouse->nButtons;
    }
    else {
        DEP(mouse_x) = IND(devx)/2;
        DEP(mouse_y) = IND(devy)/2;
    }
    return 0;
}

/* Put the cursor up on the screen. Make sure the coordinates are legal. */

static int kmShow()
{
    int x, y, x_lo, x_hi, y_lo, y_hi, *p;

    if(DEP(mouse_x) < 0)        DEP(mouse_x) = 0;
    if(DEP(mouse_x) >= IND(devx)) DEP(mouse_x) = IND(devx)-1;
    if(DEP(mouse_y) < 0)        DEP(mouse_y) = 0;
    if(DEP(mouse_y) >= IND(devy)) DEP(mouse_y) = IND(devy)-1;

/* Now draw the thing on the screen, saving the pixels below in the
   static arrays.
*/
    x_lo = DEP(mouse_x) - KEY_MOUSE_WID;
    x_hi = DEP(mouse_x) + KEY_MOUSE_WID;
    if(x_lo < 0) x_lo = 0;
    if(x_hi >= IND(devx)) x_hi = IND(devx)-1;

    y_lo = DEP(mouse_y) - KEY_MOUSE_WID;
    y_hi = DEP(mouse_y) + KEY_MOUSE_WID;
    if(y_lo < 0) y_lo = 0;
    if(y_hi >= IND(devy)) y_hi = IND(devy)-1;

/* Now draw the cross-hair */

    p = DEP(xpixels);
    y = DEP(mouse_y);
    for(x=x_lo; x<= x_hi; x++) {
        *p++ = getpixel(x, y);
        putpixel(x, y, WHITE);
    }
    p = DEP(ypixels);
    x = DEP(mouse_x);
    for(y=y_lo; y<= y_hi; y++) {
        if(y == DEP(mouse_y)) continue;
        *p++ = getpixel(x, y);
        putpixel(x, y, WHITE);
    }
    return;
}

/* This erases the cursor by putting back whatever used to be in the
   relevant pixels
*/
static int kmHide()
{
    int x, y, x_lo, x_hi, y_lo, y_hi;
    int *p;

    x_lo = DEP(mouse_x) - KEY_MOUSE_WID;
    x_hi = DEP(mouse_x) + KEY_MOUSE_WID;
    if(x_lo < 0) x_lo = 0;
    if(x_hi >= IND(devx)) x_hi = IND(devx)-1;

    y_lo = DEP(mouse_y) - KEY_MOUSE_WID;
    y_hi = DEP(mouse_y) + KEY_MOUSE_WID;
    if(y_lo < 0) y_lo = 0;
    if(y_hi >= IND(devy)) y_hi = IND(devy)-1;

    p = DEP(xpixels);
    y = DEP(mouse_y);
    for(x=x_lo; x<= x_hi; x++) putpixel(x, y, *p++);
    p = DEP(ypixels);
    x = DEP(mouse_x);
    for(y=y_lo; y<= y_hi; y++) {
        if(y == DEP(mouse_y)) continue;
        putpixel(x, y, *p++);
    }

    return;
}

static int mouse_gin(key, xi, yi)
     int *key;
     int *xi, *yi;
{
        int done, stat;
        locRec *m, *mPos();

/* Strip any pending characters */

	while(kbhit()) getch();

/* Phase I. Wait for a key to go down, allowing for keystrokes. */

        do {
                stat = kbhit();
                m = mPos();
        }
        while((m->buttonStatus & 0x7) == 0 && stat == 0);

/* O.K.... now we pressed a key. Wait for it to come up */

	if(stat) *key = getch();
	else *key = m->buttonStatus & 0x7;

        do {
                kbhit();
                m = mPos();
        }
        while((m->buttonStatus != 0) && stat == 0);

        *xi = m->column;
        *yi = m->row;

        return 0;
}

static void _plx_bgi_linemod(style)
     int style;
{
    switch(style%NLINMOD) {
    case 0:     /* Solid */
        setlinestyle(SOLID_LINE, 0, 1);
        break;
    case 1:     /* Longdashed */
        setlinestyle(USERBIT_LINE, 0x0f0f, 1);
        break;
    case 2:     /* Dotted */
        setlinestyle(DOTTED_LINE, 0, 1);
        break;
    case 3: /* dotdashed */
        setlinestyle(USERBIT_LINE, 0xe4e4, 1);
        break;
    }
    return;
}

/* 
 * Various devices supported by Turbo C.
 */
static struct {
    char *mnemon;
    int devcode;
} match_devs[] = {
	{"CGA", CGA},
	{"MCGA", MCGA},
	{"EGA", EGA},
	{"EGA64", EGA64},
	{"EGAMONO", EGAMONO},
	{"IBM8514", IBM8514},
	{"HERCMONO", HERCMONO},
	{"ATT400", ATT400},
	{"VGA", VGA},
	{"PC3270", PC3270},
};


static int _plx_bgi_openpl(bufsiz, grafp)
     int bufsiz;
     FILE *grafp;
{
    int i, Ndevs, driver, mode;
    char *_ex_getcust(), *p, *strchr(), *strupr();
    char BGIpath[128];
/*
 * Attempt to find out if the user has over-ridden the
 * autodetect facility.
 *
 * We allow strings of the form "VGA:1" to indicate using a VGA
 * in mode 1. The mode values are merely those given in the
 * TURBO documentation.
 *
 * If not, or can't be determined, use auto-detect.
 */
    strcpy(BGIpath, _ex_getcust("MONITOR"));
    if((p=strchr(BGIpath, ':')) != (char *)0) *p = '\0';

    Ndevs = sizeof(match_devs)/sizeof(match_devs[0]);
    for(i=0; i<Ndevs; i++) {
	if(!strcmp(BGIpath, strupr(match_devs[i].mnemon))) break;
    }
    if(i < Ndevs) {
	driver = match_devs[i].devcode;
	if(p != (char *)0) mode = atoi(p+1);
	else mode = 0;
    }
    else {
	driver = DETECT;
	mode = 0;
    }
/*
 * Now find out where the drivers are kept using the BGIpath
 * variable.
 */
    strcpy(BGIpath, _ex_getcust("BGI_DRIVERS"));

    initgraph(&driver, &mode, BGIpath);

    IND(devx) = getmaxx();
    IND(devy) = getmaxy() - (DEP(numlines) * (1+textheight("H")));
    if (DEP(numlines) > 0) 
	IND(devy) -= DEP(numlines)*2;  /* room to breathe */

    initpolys();
/* 
 * This line of magic code allows the MS Mouse driver to work properly
 * on a Hercules monitor. 
 */
    if (!stricmp(getdrivername(),"HERC")) poke(0x40,0x49,6);
/* 
 * Now look for a mouse on the system. If there is one remember the
 * number of buttons in mouse.
 */
    DEP(mouse) = _init_mouse();
    if (DEP(numlines) > 0)
        _ex_install_bgi(DEP(numlines));
    _plx_bgi_showmouse();
    return 1;
}

/* ARGSUSED */
static int _plx_bgi_kopenp(bufsiz, user_ptr)
     int bufsiz;
     char *user_ptr;
{
    return _plx_bgi_openpl(0, (FILE *)0);
}

static int _plx_bgi_setval(option, val)
     char *option;
     long val;
{
    if (!strcmp(option, "numlines")) {
	DEP(numlines) = (int) val;
    }
    return -1;
}

static int _plx_bgi_getval(option, pval)
     char *option;
     long *pval;
{
    if(!strcmp(option, "nlcolors")) {
	*pval = 8;
	return 0;
    }
    else if(!strcmp(option, "nlstyles")) {
	*pval = NLINMOD;
	return 0;
    }
    else if(!strcmp(option, "nbuttons")) {
	*pval = DEP(mouse);
	return 0;
    }
    return -1;
}

static int _plx_bgi_setstr(option, str)
     char *option, *str;
{
    return -1;
}

static int _plx_bgi_getstr(option, str, maxlen)
     char *option, *str;
     int maxlen;
{
    return -1;
}

static void _plx_bgi_greyscale(from, to)
     int from, to;
{
    return;
}

static void _plx_bgi_rainbow(from, to)
     int from, to;
{
    return;
}

static void _plx_bgi_sendplot()
{
    return;
}

static void _plx_bgi_usendplot()
{
    return;
}

static void _plx_bgi_asendplot()
{
    return;
}

static int _plx_bgi_plothwm()
{
    return 0;
}

/*
 * Assign the function pointers that describe a particular driver.
 */
static int set_fn_ptrs(ptr)
     struct _plx_ind *ptr;
{
    if(ptr == (struct _plx_ind *)0) return -1;

    ptr->setclip    = 0;
    ptr->endclip    = 0;
    ptr->closepl    = _plx_bgi_closepl;
    ptr->color      = _plx_bgi_color;
    ptr->cont       = _plx_bgi_cont;
    ptr->dotext     = _plx_bgi_dotext;
    ptr->intext     = _plx_bgi_intext;
    ptr->sztext     = _plx_bgi_sztext;
    ptr->erase      = _plx_bgi_erase;
    ptr->gin        = _plx_bgi_gin;
    ptr->gtrack_on  = _plx_bgi_gtrack_on;
    ptr->gtrack     = _plx_bgi_gtrack;
    ptr->gtrack_off = _plx_bgi_gtrack_off;
    ptr->aspect     = _plx_bgi_aspect;
    ptr->raspect    = _plx_bgi_raspect;
    ptr->sendplot   = _plx_bgi_sendplot;
    ptr->asendplot  = _plx_bgi_asendplot;
    ptr->usendplot  = _plx_bgi_usendplot;
    ptr->plothwm    = _plx_bgi_plothwm;
    ptr->point      = _plx_bgi_point;
    ptr->linemod    = _plx_bgi_linemod;
    ptr->move       = _plx_bgi_move;
    ptr->openpl     = _plx_bgi_openpl;
    ptr->kopenp     = _plx_bgi_kopenp;
    ptr->rainbow    = _plx_bgi_rainbow;
    ptr->greyscale  = _plx_bgi_greyscale;
    ptr->initpanel  = _plx_bgi_initpanel;
    ptr->panelpoint = _plx_bgi_panelpoint;
    ptr->endpanel   = _plx_bgi_endpanel;
    ptr->polgn      = 0;
    ptr->box        = 0;
    ptr->setval     = _plx_bgi_setval;
    ptr->getval     = _plx_bgi_getval;
    ptr->setstr     = _plx_bgi_setstr;
    ptr->getstr     = _plx_bgi_getstr;
    ptr->setcurs    = _plx_bgi_setcurs;
    return 0;
}

/*
 * This routine serves a VERY IMPORTANT purpose. It must allocate
 * space for the newly created device's "dependent" and "independent"
 * structures (or allocate them statically) and fill in any entries
 * that must have sensible defaults.
 */

static struct _plx_ind bgidev;
static struct _plx_dep bgidep;

int _plx_bgi_ctor(pdev)
     struct _plx_ind **pdev;
{
    *pdev = &bgidev;
    _plx_curdev = &bgidev;
    IND(devdep) = &bgidep;

    set_fn_ptrs(&bgidev);
/*
 * Set hardware dependent defaults.
 */
    DEP(plxmap[0]) = BLACK;
    DEP(plxmap[1]) = WHITE;
    DEP(plxmap[2]) = LIGHTRED;
    DEP(plxmap[3]) = LIGHTGREEN;
    DEP(plxmap[4]) = LIGHTBLUE;
    DEP(plxmap[5]) = LIGHTCYAN;
    DEP(plxmap[6]) = LIGHTMAGENTA;
    DEP(plxmap[7]) = YELLOW;
    DEP(fgcolor) = 1;
    DEP(mouse) = 0;
    DEP(rate) = 10;
    DEP(numlines) = 0;	/* i.e. use FULL screen by default */
/*
 * ... and hardware independent defaults.
 */
    IND(devx) = DEV_x;
    IND(devy) = DEV_y;
    return 0;
}

static int _plx_bgi_cursors[4][16] = {
    {0x3fff,		/* 0011111111111111   Screen Mask for  */
     0x1fff,		/* 0001111111111111    standard cursor */
     0x0fff,            /* 0000111111111111   Hot Spot: -1,-1  */
     0x07ff,		/* 0000011111111111 */
     0x03ff,		/* 0000001111111111 */
     0x01ff,		/* 0000000111111111 */
     0x00ff,		/* 0000000011111111 */
     0x007f,		/* 0000000001111111 */
     0x003f,		/* 0000000000111111 */
     0x001f,		/* 0000000000011111 */
     0x01ff,		/* 0000000111111111 */
     0x10ff,		/* 0001000011111111 */
     0x30ff,		/* 0011000011111111 */
     0xf87f,		/* 1111100001111111 */
     0xf87f,		/* 1111100001111111 */
     0xfc3f},		/* 1111110000111111 */

    {0x0000,		/* 0000000000000000   Cursor Mask for  */
     0x4000,		/* 0100000000000000    standard cursor */
     0x6000,            /* 0110000000000000 */
     0x7000,		/* 0111000000000000 */
     0x7800,		/* 0111100000000000 */
     0x7c00,		/* 0111110000000000 */
     0x7e00,		/* 0111111000000000 */
     0x7f00,		/* 0111111100000000 */
     0x7f80,		/* 0111111110000000 */
     0x7fc0,		/* 0111111111000000 */
     0x7c00,		/* 0111110000000000 */
     0x4600,		/* 0100011000000000 */
     0x0600,		/* 0000011000000000 */
     0x0300,		/* 0000001100000000 */
     0x0300,		/* 0000001100000000 */
     0x0000},		/* 0000000000000000 */

    {0x0000,		/* 0000000000000000 */
     0x0000,		/* 0000000000000000   Screen Mask for  */
     0x0000,		/* 0000000000000000    hourglass       */
     0x0000,            /* 0000000000000000   Hot Spot: 7,7    */
     0x8001,		/* 1000000000000001 */
     0xc003,		/* 1100000000000011 */
     0xe007,		/* 1110000000000111 */
     0xf00f,		/* 1111000000001111 */
     0xe007,		/* 1110000000000111 */
     0xc003,		/* 1100000000000011 */
     0x8001,		/* 1000000000000001 */
     0x0000,		/* 0000000000000000 */
     0x0000,		/* 0000000000000000 */
     0x0000,		/* 0000000000000000 */
     0x0000,		/* 0000000000000000 */
     0xffff},		/* 1111111111111111 */

    {0x0000,		/* 0000000000000000   Cursor Mask for  */
     0x7ffe,		/* 0111111111111110    hourglass       */
     0x6006,            /* 0110000000000110 */
     0x300c,		/* 0011000000001100 */
     0x1818,		/* 0001100000011000 */
     0x0c30,		/* 0000110000110000 */
     0x0660,		/* 0000011001100000 */
     0x03c0,		/* 0000001111000000 */
     0x0660,		/* 0000011001100000 */
     0x0c30,		/* 0000110000110000 */
     0x1998,		/* 0001100110011000 */
     0x33cc,		/* 0011001111001100 */
     0x67e6,		/* 0110011111100110 */
     0x7ffe,		/* 0111111111111110 */
     0x0000,		/* 0000000000000000 */
     0x0000}		/* 0000000000000000 */
};


/* Sets graphic cursor shape */
static void _plx_bgi_setgraphcur(int hotx, int hoty, int *pmasks)
{
    struct SREGS  seg;

    inreg.x.ax = 9;                                    /* function 9 */
    inreg.x.bx = hotx;                /* cursor hot spot: horizontal */
    inreg.x.cx = hoty;                  /* cursor hot spot: vertical */
    inreg.x.dx = FP_OFF(pmasks);
    seg.es = FP_SEG(pmasks);
    int86x (0x33, &inreg, &outreg, &seg);
}

static int _plx_bgi_setcurs(which)
    int which;
{
    _plx_bgi_hidemouse();
    switch(which) {
    case PLX_HRG_CURSOR:
	_plx_bgi_setgraphcur(7,7,_plx_bgi_cursors[2]);
	break;
    case PLX_PTR_CURSOR:
	_plx_bgi_setgraphcur(-1,-1,_plx_bgi_cursors[0]);
	break;
    default:
	return -1;
    }
    _plx_bgi_showmouse();
    return 1;
}
