/*****************************************************/
/*         Contour Front-end: see contour_low.c      */
/*         Roy Williams  May 1986                    */
/*         All material copyright Caltech            */
/*****************************************************/
#include <stdio.h>
#include "plotix.h"
#ifdef isMPW
#define __SEG__ PLOTIX
#include <:sys:config.h>
#include <:sys:plotix.h>
#else
#include "sys/config.h"
#include "sys/plotix.h"
#endif

#if __STDC__                    /* Be nice to ourselves, use prototypes */
static int shallwecut(char *left, char *right);
static int findcut(void);
static int corner(void);
static void coords(Coord *xf, Coord *yf);
static int nextpoint(void);
static double p(double z0, double z1);
#else                           /* Can't do prototypes, but declare anyway */
static int shallwecut();
static int findcut();
static int corner();
static void coords();
static int nextpoint();
static double p();
#endif                          /* __STDC__ */

static int colors[] = {1, 2, 3, 4, 5, 6, 7, 0};
static int ncolor = (sizeof(colors) / sizeof(colors[0]));

void contour(zpass, nx, ny, levelmin, levelmax, nlevel, panels)
double (*zpass)();          /* Function to be contoured */
int nx, ny;                /* Grid Size */
double levelmin, levelmax;  /* Minimum, Maximum contour levels */
int nlevel;                /* Number of levels */
int panels;                /* =0: Lines, =1: Color Panels */
{
        int ilevel, colorlabel, endflag, status, ix, iy;
        double level, zmin, zmax, z;
        Coord xf, yf;
        colorlabel = 0;      /* First color */
        if(nlevel < 0 || levelmin == levelmax){       
			/* nlevel -ve means find your own contour
			   Minimum & Maximum */
                if(nlevel < 0) nlevel = -nlevel;
                zmin = (*zpass)(0, 0);
                zmax = zmin;
                for(ix = 0; ix<nx; ix++)
                        for(iy = 0; iy<ny; iy++){
                                if((z = (*zpass )(ix, iy)) < zmin) zmin = z;
                                if(z > zmax) zmax = z;
                        }
                levelmin = zmin + (zmax-zmin)/(nlevel+1);
                levelmax = zmax - (zmax-zmin)/(nlevel+1);
        }
        for(ilevel=0; ilevel<nlevel; ilevel++){
                if(nlevel == 1)     /* If nlevel=1 use levelmin, ignore levelmax */
                        level = levelmin;
                else
                        level = levelmin + ilevel*(levelmax - levelmin)/(nlevel - 1);
                                                                /* initialise this level */
#ifdef DEBUG
fprintf(debugfp,
     "initlevel: nx=%d, ny=%d, level=%f, panels=%d\n", nx, ny, level, panels);
fflush(debugfp);
#endif
                initlevel(zpass, nx, ny, level, panels);
                endflag = 1;
                while(status = getpoint(&xf, &yf)){
                        if(status == 2){
                                if(panels)
                                        endpanel();
                                endflag = 1;
                        }else{
                                if(panels){
                                        if(endflag)
                                                initpanel(colors[colorlabel], 0);
                                        panelpoint(xf, yf);
                                }else
                                        if(endflag)
                                                move(xf, yf);
                                        else
                                                cont(xf, yf);
                                endflag = 0;
                        }
                }
                if(++colorlabel >= ncolor)
                        colorlabel = 0;
                usendplot();
        }
}

/********************************************************/
/*                                                      */
/*    Contouring by Roy Williams March 1986             */
/*            Copyright Caltech 1986                    */
/*                                                      */
/********************************************************/
#define TABLE(x,y) table[(x)*ny+y]
#define ONBIT(x,y,bit) table[(x)*ny+y]&mask[bit]
#define KILL(x,y,bit) table[(x)*ny+y]^=mask[bit]

extern char *_ex_malloc();  extern void _ex_free();

/*      These are common variables for contour  */
#ifdef __STDC__
static double (*z)(int x, int y);       /* Pointer to the data function */
#else
static double (*z)();                           /* Pointer to the data function */
#endif
static double level;                            /* Contouring level */
static int nx, ny;                                      /* Numbers of points in x, y directions */
static int nx1, ny1;                            /* nx1 = nx-1 and ny1 = ny-1 */
static int xleft, xright;                       /* Vertical cut positions */
static int wantedge;                            /* Edge style required (input) */
static char mask[8];                            /* mask[i] = 1<<i */
static char *table;                                     /* Lookup table: nx*ny bytes malloced */
static char *hihere, *hileft;           /* 2*ny bytes malloced */
static int looksofar;                           /* When looking for the next polygon,
                                                                                don't look where we've been already */
static int x, y, bit;                           /* Current position and direction */
static int xstart, ystart, bitstart;/* Starting position and direction */
static int endflag;                                     /* If 1, send ENDPOLY next time around */
static int allupflag;                           /* If 1, all points above level, so
                                                                                draw a box */

void initlevel(zpass, nxpass, nypass, levelpass, passwantedge)
/*
Contouring a function consists of calling initlevel with the array, sizes,
contouring level, and edge style required and then (repeatedly) getpoint() 
returns 
        0 for no more contours at this level, 
        1 for a point found on the present contour, or 
        2 for the end of a contour polygon. 
The contour line direction is such that the z>level part is on the left.
Wantedge (or its incarnation passwantedge, same thing) talks about
how to deal with the bounding rectangle: 
        0 means only internal segments and possibly a single edge point 
                at an end of the contour line; useful for vector plotters.
        1 means the edge points are connected, and corners included, so that each
                contour line is part of a closed polygon, and 
        2 means that all border points are included, in case the whole picture 
                is to be warped by something down the line.
If wantedge is 1 or 2, all polygons have the topology of a disk, so that
a hardware fill can be used to make colors or shading.
*/
#ifdef __STDC__
double (*zpass)(int x, int y);
#else
double (*zpass)();
#endif
double levelpass;
int nxpass, nypass, passwantedge;
{
        int i;
        level = levelpass;
#ifdef DEBUG
        fprintf(stderr, "INITLEVEL: %f\n", level);
#endif
        nx = nxpass;    ny = nypass;    z = zpass;      wantedge = passwantedge;
        nx1 = nx - 1;
        ny1 = ny - 1;
        mask[0] = 1;
        for(i=1; i<8; i++)
                mask[i] = mask[i-1] << 1;
        if((table = _ex_malloc(nx*ny)) == 0
                || (hihere = _ex_malloc(ny)) == 0
                || (hileft = _ex_malloc(ny)) == 0){
#ifndef DEVIND
                        fprintf(stderr, "malloc failed\n"); 
#endif
                        abort(-101);
        }
        xright = 0;
        allupflag = 1;
        findcut();
}

static int findcut()
{
        int hinow, hithere;
/*
This routine moves a vertical line rightward until shallwecut() decides
to cut the picture. In this way complex topologies such as an annulus
are avoided, because the hardware fill cannot deal with them.
The next section tells the lookup table whats above and whats below the 
contouring level.  An UP point for example has z>level at (x, y) and 
z<=level in the +y direction.
*/
#define UP    mask[0]   /*         ^ 0          */
#define LEFT  mask[1]   /*         |            */
#define RIGHT mask[2]   /*    1 <--*--> 2       */
#define DOWN  mask[3]   /*         |            */
                        /*         V 3          */
        if(xright == nx1)
                return(0);
        xleft = xright;
        for(x=xleft; x<nx; x++){
                if(x == xleft)
                        for(y=0; y<ny; y++){
                                TABLE(xleft, y) = 0;
                                hihere[y] = ((*z)(xleft, y) > level);
                        }
                for(y=0; y<ny; y++){
                        hinow = hihere[y];
                        if(!hinow)
                                allupflag = 0;
                        if(y > 0){
                                hithere = hihere[y-1];
                                if( hinow && !hithere)
                                        TABLE(x, y) |= DOWN;
                                if( !hinow && hithere)
                                        TABLE(x, y-1) |= UP;
                        }
                        if(x > xleft){
                                hithere = hileft[y];
                                if( hinow && !hithere)
                                        TABLE(x, y) |= LEFT;
                                if( !hinow && hithere)
                                        TABLE(x-1, y) |= RIGHT;
                        }
                }
                if(x == nx1)
                        break;
                for(y=0; y<ny; y++){
                        TABLE(x+1, y) = 0;
                        hileft[y] = hihere[y];
                        hihere[y] = ((*z)(x+1, y) > level);
                }
                if(x > xleft && x < nx1){
                        if(shallwecut(hileft, hihere)){
                                xright = x;
                                goto foundcut;
                        }
                }
        }
        xright = nx1;
/*
Tell the lookup table where the edges are, where edge includes the fake
edges decided on by shallwecut().
*/
#define UPEDGE     mask[6]
#define LEFTEDGE   mask[4]
#define RIGHTEDGE  mask[7]
#define DOWNEDGE   mask[5]
        foundcut:
        for(x=xleft; x<=xright; x++){
                TABLE(x, 0) |= DOWNEDGE;
                TABLE(x, ny1) |= UPEDGE;
        }
        for(y=0; y<ny; y++){
                TABLE(xleft, y) |= LEFTEDGE;
                TABLE(xright, y) |= RIGHTEDGE;
        }
#ifdef DEBUG
        fprintf(stderr, "FINDCUT: %d %d\n", xleft, xright);
        for(x=xleft; x<=xright; x++)
	    for(y=0; y<ny; y++)
		    fprintf(stderr, "%d %d %f %x\n", x, y, (*z)(x, y) - level,
                                TABLE(x, y));
#endif
        looksofar = xleft*ny;
        bit = 999;
        endflag = 0;
        return(1);
}
static int shallwecut(left, right)
/*
Decide whether to cut (i.e. introduce a righthand edge) here.
If two z>level regions become one, then a cut is made. Example:
an annulus has four possible places, being enter penumbra, enter umbra, 
leave umbra, and leave penumbra. We want xright to be at leave umbra.
*/
char *left, *right;
{
        int i;
        if(wantedge == 0)
                return(0);
        i=0;
        again:
                while( !left[i]){
                        i++; 
                        if(i >= ny1) 
                                return(0);
                }
                while( left[i]){
                        i++; 
                        if(i >= ny1) 
                                return(0);
                }
                if( right[i-1] || right[i] ){
                        if( right[i-1]) 
                                i++;
                        while( right[i]){
                                i++; 
                                if(i >= ny1) 
                                        return(0);
                                if( left[i] || left[i+1])
                                        return(1);
                        }
                }
        goto again;
}
int getpoint(xf, yf)
/*
This routine is called repeatedly and it puts new coordinates into the
addresses xf and yf, scaled to fit in the unit square. When it returns
ENDPOLY or NOPOLYS, these addresses contain garbage. A polygon is 
completed in the sense that the first and last points are the same,
so you don't need to close the jaws yourself. When the last point has 
been sent, the next call to getpoint returns ENDPOLY, with garbage in xf, yf.
All the real work is done by nextpoint, and getpoint just does the 
housekeeping of the wantedge business.
*/
Coord *xf, *yf;
{
#define NOPOLYS                 0
#define FOUNDPOINT              1
#define ENDPOLY                 2
#define FOUNDEDGEPOINT  3
#define FOUNDCORNER             4
        int status;
        static int endsent;             /* Used when wantedge is 0, to stop huge numbers
                                                                of consecutive ENDPOLY's being sent.    */
newpoint:
        status = nextpoint();
#ifdef DEBUG
        fprintf(stderr, "(%d %d %d) %d (%d %d %d) %x\n", x, y, bit, status, 
                xstart, ystart, bitstart, TABLE(x, y));
#endif
        if(status == NOPOLYS)
                if(findcut())
                        goto newpoint;
                else{
                        _ex_free(table); _ex_free(hileft); _ex_free(hihere);
                        return(NOPOLYS);
                }
        coords(xf, yf);
        if(status <= ENDPOLY || wantedge == 2){
                if(status != ENDPOLY)
                        endsent=0;
                return(status);
        }
        if(wantedge == 0 && !endsent){
                endsent = 1;
                return(ENDPOLY);
        }
        if(wantedge == 1 && status == FOUNDCORNER)
                return(FOUNDPOINT);
        goto newpoint;
}
static void coords(xf, yf)
Coord *xf, *yf;
{
        double p();
#define ON_EDGE 100 
        if(bit >= ON_EDGE){
                *xf = x;
                *yf = y;
        }
        else{
                switch(bit){                    /* This switch defines th meaning of bit, 
                                                                        and is consistent with that in the #defines
                                                                        for UP, LEFT etc. above. */
                        case 0: *xf = x;
                                        *yf = y + p((*z)(x, y), (*z)(x, y+1));
                                        break;
                        case 1: *xf = x - p((*z)(x, y), (*z)(x-1, y));
                                        *yf = y;
                                        break;
                        case 2: *xf = x + p((*z)(x, y), (*z)(x+1, y));
                                        *yf = y;
                                        break;
                        case 3: *xf = x;
                                        *yf = y - p((*z)(x, y), (*z)(x, y-1));
                                        break;
                }
        }
/*
Lets return coordinates in the unit square.
*/
        *xf /= (double)nx1;
        *yf /= (double)ny1;
}
static double p(z0, z1)
double z0, z1;
{
        return((z0-level)/(z0-z1));
}
static int nextpoint()
{
        int test, xtry, ytry, istart, tester;
/*
When we stand on the brink of a new square, these arrays tell which edges 
to examine.
*/
        static int xinc[4][4] = {{-1, 0, -1, 0}, {-1, 0, 0, 0}, 
                                                        {0, 0, 0, 1},   {0, 1, 0, 1}};
        static int yinc[4][4] = {{0, 0, 1, 0},   {-1, -1, 0, 0}, 
                                                        {0, 0, 1, 1},   {0, -1, 0, 0}}; 
        static int testbit[4][3] = {{1, 0, 2}, {3, 1, 0}, {0, 2, 3}, {2, 3, 1}};
/*
Having sent the last point to close a polygon send an ENDPOLY 
*/
        if(endflag){    
                endflag = 0;
                bit = 999;
                if(allupflag == 2)
                        allupflag = 3;
                return(ENDPOLY);
        }
        if(allupflag == 3)
                return(NOPOLYS);
/*
allupflag deals with the case of all z being >level, so that the box
(0, 0), (0, 1), (1, 1), (1, 0), (0, 0), ENDPOLY, is sent. This has to be done
artificially because no contours will have been found.
*/
        if(allupflag == 2){
                switch(bit){
                        case 100: y-=1; break;
                        case 101: x+=1; break;
                        case 102: x-=1; break;
                        case 103: y+=1; break;
                }
                if(x == 0 && y == 0)
                        endflag = 1;
                return(corner());
        }
/*
When bit is 999 we have to search around for the start of a new polygon.
When it's found, return the starting point, or if no polygon, return
NOPOLYS. looksofar makes sure we don't go over the same ground twice.
*/
        if(bit == 999){
                for(istart=looksofar; istart < (xright+1)*ny; istart++){
                        if(table[istart] & 0xf){
                                x = istart/ny;
                                y = istart%ny;
                                looksofar = istart;
                                for (test=0; test<4; test++) {
                                        if(ONBIT(x, y, test)){
                                                bit = test;
                                                xstart = x;
                                                ystart = y;
                                                bitstart = bit;
                                                KILL(x, y, bit);
                                                return(FOUNDPOINT);
                                        }
                                }
                        }
                }
                if(allupflag == 1 && wantedge != 0){
                        x = 0;          y = 0;          bit = 101;
                        allupflag = 2;
                        return(FOUNDCORNER);
                }else
                        return(NOPOLYS);
        }
/* 
We have just dealt with a point on an edge between two squares, 
and are standing on that edge, of type bit, with a high point (x, y) 
at one end of the edge. Check first to see if our nose is in the 
border fence, and if so turn left, because the contours are meant to 
have the high part on the left.
Otherwise look at the three possible edges for the next point.  
*/ 
#define MAGIC_NUMBER 3
        if(bit >= ON_EDGE){
                test = MAGIC_NUMBER - (bit - ON_EDGE);
                if(x == xstart && y== ystart && test == bitstart){
                        bit = test;
                        if(allupflag == 2)
                                bit =test + ON_EDGE;
                        endflag = 1;
                        return(FOUNDPOINT);
                }
                if(ONBIT(x, y, test)){          /* Shall we venture off the edge? */
                        KILL(x, y, test);
                        bit = test;
                        return(FOUNDPOINT);
                }
                switch(bit){                    /* No, just go anticlockwise on it */
                        case 100:                               /* Left edge */
                                y -= 1; break;
                        case 101:                               /* Bottom edge */
                                x += 1; break;
                        case 102:                               /* Upper edge */
                                x -= 1; break;
                        case 103:                               /* Right edge */
                                y += 1; break;
                }
                return(corner());
        }
/*
Is our nose in the border fence?
*/
        if(ONBIT(x, y, bit + 4)){
                bit += ON_EDGE;
                return(corner());
        }
/* Forget edges, we're in the wide world. Check 3 possible edges */
        for(tester=0; tester<3; tester++){
                test = testbit[bit][tester];
                xtry = x + xinc[bit][test];
                ytry = y + yinc[bit][test];
                if(xtry == xstart && ytry == ystart && test == bitstart){
                        x = xtry;
                        y = ytry;
                        bit = test;
                        endflag = 1;
                        return(FOUNDPOINT);
                }
                if(ONBIT(xtry, ytry, test)){
                        x = xtry;
                        y = ytry;
                        bit = test;
                        KILL(x, y, bit);
                        return(FOUNDPOINT);
                }
        }
/*
We should never get here.
*/
                bit = 999;
                return(ENDPOLY);
}
static int corner()
/*
Deal with the corners. For use when tooling around the edge.
*/
{
        switch(bit){
                case 100:                               /* Left edge */
                        if(y == 0){
                                bit = 101;
                                return(FOUNDCORNER);
                        }
                        else
                                return(FOUNDEDGEPOINT);
                case 101:                               /* Bottom edge */
                        if(x == xright){
                                bit = 103;
                                return(FOUNDCORNER);
                        }
                        else
                                return(FOUNDEDGEPOINT);
                case 102:                               /* Upper edge */
                        if(x == xleft){
                                bit = 100;
                                return(FOUNDCORNER);
                        }
                        else
                                return(FOUNDEDGEPOINT);
                case 103:                               /* Right edge */
                        if(y == ny1){
                                bit = 102;
                                return(FOUNDCORNER);
                        }
                        else
                                return(FOUNDEDGEPOINT);
        }
}

/*
 * From here down is basically the various FORTRAN interfaces to
 * this stuff.
 *
 * This requires more caution than usual because of the problem
 * with C compilers that convert float to double automatically.
 * In cases such as this we have to play funny games with pointers
 * as demonstrated in the Sun code, below.
 */
#ifdef _3L
static float (*f_func)();
double fake_dbl();

double c_func(i,j)
    int i, j;
{
    return (double)((*f_func)(&i, &j));
}

double fake_dbl(pval)
     void *pval;
{
    return *(float *)pval;
}

kcntor(func, nx, ny, levelmin, levelmax, nlevs, panels)
    float (*func)();
    int *nx, *ny;
    float *levelmin, *levelmax;
    int *nlevs, *panels;
{
    f_func = func;
    contour(c_func, *nx, *ny, *levelmin, *levelmax, *nlevs, *panels);
}

kgetpt(xf, yf)
    float *xf, *yf;
{
    Coord p, q;
    int stat;

    stat = getpoint(&p ,&q);
    *xf = p;
    *yf = q;
    return stat;
}

kinitl(func, nx, ny, level, edge)
    float (*func)();
    int *nx, *ny;
    float *level;
    int *edge;
{
    f_func = func;
    initlevel(c_func, *nx, *ny, *level, *edge);
}
#endif		/* _3L */

#if COMPILER == SUNF77
static int (*f_func)();
static double fake_dbl();

static double c_func(i,j)
    int i, j;
{
    int temp;
    temp = ((*f_func)(&i, &j));
    return fake_dbl((void *)&temp);
}

static double fake_dbl(pval)
     void *pval;
{
    return *(float *)pval;
}

kcntor_(func, nx, ny, levelmin, levelmax, nlevs, panels)
    int (*func)();
    int *nx, *ny;
    float *levelmin, *levelmax;
    int *nlevs, *panels;
{
    f_func = func;
    contour(c_func, *nx, *ny, (double)*levelmin, (double)*levelmax, 
				*nlevs, *panels);
}

kgetpt_(xf, yf)
    float *xf, *yf;
{
    Coord p, q;
    int stat;

    stat = getpoint(&p ,&q);
    *xf = p;
    *yf = q;
    return stat;
}

kinitl_(func, nx, ny, level, edge)
    int (*func)();
    int *nx, *ny;
    float *level;
    int *edge;
{
    f_func = func;
    initlevel(c_func, *nx, *ny, *level, *edge);
}
#endif		/* SUNF77 */
