/**********
* devices *	Copyright ParaSoft Corporation, 1990.
***********

Code which manipulates the various plotting devices, creating instances
etc......
*/
#include "plotix.h"
#ifdef isMPW
#define __SEG__ PLOTIX
#include <:sys:config.h>
#include <:sys:plotix.h>
#include <:sys:meta.h>
#else
#include "sys/config.h"
#include "sys/plotix.h"
#include "sys/meta.h"
#endif

/*
 * Maintain an array of "in use" devices. A Linked list would
 * be better but that requires malloc which we don't want to
 * assume to be available.
 */
#define NDEVS	(5)		/* Maximum number of simultaneous devs. */
static struct _plx_ind *used_device[NDEVS];

#ifdef DEVIND
static int used_indx[NDEVS];
#endif

#ifdef DEVIND
extern int _plx_remote_mkdisp();
extern int _plx_remote_swdisp();
extern int _plx_remote_lsdisp();
#endif

/*
 * Can't call the library function because it might not be defined
 * in FORTRAN.
 */
static int mystreq();
/*
 * Create an instance of the named device, building internal
 * data structures and merging it into the list of devices.
 *
 * Return an integer that allows the user to make use of the 
 * device.
 */
int make_display(choice)
     char *choice;
{
    int stat, slot, host_indx;
    struct _plx_ind *devstruc;
    PLOTIX_DEVS *p;
/*
 * Ms-Windows compiler is too stupid to link this stuff properly unless we
 * actually call functions in all appropriate files.
 */
#ifdef MSWINDOWS
     _plx_dummy();
#endif
/*
 * Search the available list to find out if the indicated device is
 * supported on this system. Note that the local list may be different
 * from that supported on the real host.
 */
#ifdef DEVIND
    p = plx_devlist;	/* It's always the first */
#else
    p = plx_devlist;
    while(choice != (char *)0 && p->abbrev != (char *)0) {
	if(mystreq(choice, p->abbrev)) break;
	p++;
    }
    if(p->abbrev == (char *)0) return -2;
#endif
/*
 * If a device of this class has been initialized return the same 
 * pointer unless we're allowed multiple instances.
 */
    if(p->mul_allowed == 0 && p->last_inst > 0) {
	_plx_curdev = used_device[p->last_inst-1];
	return p->last_inst;
    }
/*
 * Find an empty slot in the "used" list and fail if there isn't one.
 */
    for(slot=0; slot<NDEVS; slot++) {
	if(used_device[slot] == (struct _plx_ind *)0) break;
    }
    if(slot >= NDEVS) return -1;
/*
 * OK, we've found it so build the appropriate data structures by invoking
 * the "ctor" function.
 */
#ifdef OLD_FN_CALLS
    if((*(p->devctor))(&devstruc) < 0) return -3;
#else
    if((p->devctor)(&devstruc) < 0) return -3;
#endif
/*
 * Insert the device handle into the used list.
 */
    used_device[slot] = devstruc;
#ifdef DEVIND
    if((host_indx = _plx_remote_mkdisp(choice)) < 0) return stat;
    used_indx[slot] = host_indx; 
#endif
    if(devstruc != (struct _plx_ind *)0) _plx_curdev = devstruc;
    p->last_inst = slot+1;
    return slot+1;
}

/*
 * Swap the active device to the one indicated ... if it exists.
 * Return the currently active index.
 */
int swap_display(devno)
     int devno;
{
    struct _plx_ind *devstruc;
    int i, ret;

    ret = 0;
    if(_plx_curdev != (struct _plx_ind *)0) {
        for(i=0; i<NDEVS; i++) {
	    if(used_device[i] == _plx_curdev) {
		ret = i+1;
		break;
	    }
        }
    }

    if(devno < 1 || devno >= NDEVS) return -1;
    devstruc = used_device[devno-1];

#ifdef DEVIND
    _plx_remote_swdisp(used_indx[devno-1]);
#endif

    if(devstruc != (struct _plx_ind *)0) {
	_plx_curdev = devstruc;
	return ret;
    }
    else return -1;

/* NOTREACHED */
}

/*
 * Finally one that gets rid of stuff.
 */
void kill_display(devno)
     int devno;
{
    devno--;
    if(devno >= 0 && devno < NDEVS) {
	used_device[devno] = (struct _plx_ind *)0;
#ifdef DEVIND
	used_indx[devno] = 0;
#endif
    }
    return;
}

void list_display(dev_abbr_list)
    char *dev_abbr_list;
{
    PLOTIX_DEVS *p;
    char *q;

#ifdef DEVIND
    _plx_remote_lsdisp(dev_abbr_list);
#else
    *dev_abbr_list = '\0';
    p = plx_devlist;
    while(p->abbrev != (char *)0) {
	*dev_abbr_list++ = ' ';
	q = p->abbrev;
	while(*q) *dev_abbr_list++ = *q++;
	p++;
    }
    *dev_abbr_list = '\0';
#endif
    return;
}

/*
 * Return non-zero if strings match.
 */
static int mystreq(p1, p2)
     char *p1, *p2;
{
    int i;

    i = strlen(p1);
    if(i != strlen(p2)) return 0;

    while(i--) {
	if(*p1 != *p2) return 0;
	p1++;
	p2++;
    }
    return 1;
}

#ifdef _3L
kmkdsp(name, nmlen)
     char *name;
     int nmlen;
{
    char *p, tmp[80];

    if(nmlen > sizeof(tmp)) return -1;

    p = tmp;
    while(nmlen--) *p++ = *name++;
    *p++ = '\0';

    if(mystreq(tmp, "default"))
        return make_display((char *)0);
    else
	return make_display(tmp);
/* NOTREACHED */
}

int kswdsp(pto)
     int *pto;
{
    return swap_display(*pto);
}

klsdsp(buffer, blen)
     char *buffer;
     int blen;
{
    list_display(buffer);
    return;
}
#endif

#if COMPILER == SUNF77
kmkdsp_(name, nmlen)
     char *name;
     int nmlen;
{
    char *p, tmp[80];

    if(nmlen > sizeof(tmp)) return -1;

    p = tmp;
    while(*name) *p++ = *name++;
    *p++ = '\0';

    if(mystreq(tmp, "default"))
        return make_display((char *)0);
    else
	return make_display(tmp);
/* NOTREACHED */
}

int kswdsp_(pto)
     int *pto;
{
    return swap_display(*pto);
}

void klsdsp_(buffer, blen)
     char *buffer;
     int blen;
{
    list_display(buffer);
    return;
}
#endif		/* SUNF77 */
