/***********
* plotix.h *
************

Copyright ParaSoft Corporation, 1990.
*/

/* 
 * Standard header file to be used with PLOTIX programs.
 */

#ifndef _PLOTIXdotH
#define _PLOTIXdotH

/*
 * This structure is used to build an instance of a particular
 * device, and to list those available in a given system.
 */
typedef struct plx_devs {
    char *abbrev;	/* Character abbrev for device */
    int (*devctor)();	/* Function which creates an instance */
    int mul_allowed;	/* Boolean: Are multiple instances allowed? */
    int last_inst;	/* Device index of last instance */
} PLOTIX_DEVS;
extern PLOTIX_DEVS plx_devlist[];

/*
 * The following functions each create an instance of a particular
 * device.
 */
extern int  	_plx_ps_ctor(), _plx_x_ctor(), _plx_sun_ctor(),
		_plx_bgi_ctor(), _plx_mac_ctor(), _plx_msw_ctor(),
		_plx_rt_ctor(), _plx_nec_ctor(), _plx_nod_ctor();
/*
 * The following is an exhaustive list of supported devices
 * and the "magic" parameters used to create instances of them.
 */
#define PLX_PS	{"ps", 	_plx_ps_ctor, 1, 0}	/* PostScript */
#define PLX_X	{"X", 	_plx_x_ctor, 1, 0}	/* X-Windows */
#define PLX_SUN	{"sun",	_plx_sun_ctor, 1, 0}	/* SunView */
#define PLX_BGI	{"bgi",	_plx_bgi_ctor, 0, 0}	/* Turbo C, under DOS */
#define PLX_MAC	{"mac",	_plx_mac_ctor, 0, 0}	/* Macintosh */
#define PLX_MSW	{"msw",	_plx_msw_ctor, 0, 0}	/* MicroSoft Windows */
#define PLX_RT	{"rt",	_plx_rt_ctor, 0, 0}	/* NCUBE graphics system */
#define PLX_NEC {"nec", _plx_nec_ctor, 0, 0}  	/* NEC-98 IBM PC clone */
#define PLX_NOD {"nod", _plx_nod_ctor, 1, 0}  	/* Device indep. Node program */

#define PLX_NUL {(char *)0, 0, 0, 0}

/*
 * The following definitions provide support for the MENU system.
 */
typedef struct {
    char *icon;			/* String to place in box */
    int (*func)();		/* Function to invoke when selected */
    int code;			/* How to fill the box in the menu */
} PLXMENU, *PLXMENUPTR;

/*
 * This macro can be passed to the menu_process function to indicate
 * that no "miss" processing should be performed.
 */
#define NO_MISS_PROC	((int (*)())0)

/*
 * The following macros denote the "styles" supported by the
 * menu drawing package.
 */
#define MENU_SIMPLE	(0)	/* Simple "boxes" */
#define MENU_3D		(1)	/* Boxes with "3d" shadowing */
#define MENU_SHADED	(2)	/* Greyscale shading effects */

/*
 * The following cursor types are supported.
 */
#define PLX_PTR_CURSOR	(0)	/* Default, pointer */
#define PLX_HRG_CURSOR	(1)	/* Hourglass */

#ifdef __STDC__
#include <stdio.h>
#endif

#ifdef MSWINDOWS
extern int _g_spawnwin;
#endif

/* 
 * Decide whether to use Floating point or integer coordinates.
 * Default to FLTCORD if neither is defined.
 */
#ifndef FLTCORD
#ifndef INTCORD
#define FLTCORD
#endif
#endif

#ifdef FLTCORD
typedef double Coord;
#endif

#ifdef INTCORD
typedef unsigned int Coord;
#endif

#ifdef __STDC__                 /* Function protypes */
void aclosepl(void);
void aerase(void);
int agin(int *key, Coord *x, Coord *y);
int aopenpl(int bufsiz, FILE *fp);
void asendplot(void);
void aspect(Coord *xsize, Coord *ysize);
void box(Coord,Coord,Coord,Coord,int,int);
void closepl(void);
void color(int shade);
void cont(Coord x, Coord y);
void contour(double (*zpass)(), int nx, int ny,
        double levelmin, double levelmax, int nlevel, int panels);
void dotext(char *s, Coord x, Coord y, int angle,
        int hjust, int vjust);
void endclip(void);
void endpanel(void);
#ifndef isMPW
void erase(void);
#endif
int getplxopt(char *, ...);
int getpoint(Coord *xf, Coord *yf);
int gin(int *key,Coord *x,Coord *y);
void greyscale(int lo, int hi);
void initlevel(double (*zpass)(int x, int y), int nxpass, int nypass,
        double levelpass, int passwantedge);
void initpanel(int color, int edge);
void label(char *s, Coord x, Coord y);
void linemod(int s);
void marker(int mkr, Coord x, Coord y, Coord m_size);
void move(Coord x, Coord y);
int openpl(int bufsiz, FILE *fp);
void ortho_space(Coord x0,Coord y0,Coord x1,Coord y1,int position);
void panelpoint(Coord x, Coord y);
int plothwm(void);
void polgn(int num, Coord *x, Coord *y, int color, int edge);
void rainbow(int lo, int hi);
void sendplot(void);
void setclip(Coord x0, Coord y0, Coord x1, Coord y1);
int setplxopt(char *, ...);
int setvport(int);
void space(Coord x0, Coord y0, Coord x1, Coord y1);
void usendplot(void);
int vport(Coord x0, Coord y0, Coord x1, Coord y1);
#else           /* No function protypes, but try to be nice anyway */
void aclosepl();
void aerase();
int agin();
int aopenpl();
void asendplot();
void aspect();
void box();
void closepl();
void color();
void cont();
void contour();
void dotext();
void endclip();
void endpanel();
#ifndef isMPW
void erase();
#endif
int getplxopt();
int getpoint();
int gin();
void greyscale();
void initlevel();
void initpanel();
void label();
void linemod();
void marker();
void move();
int openpl();
void ortho_space();
void panelpoint();
int plothwm();
void polgn();
void rainbow();
void sendplot();
void setclip();
int setplxopt();
int setvport();
void space();
void usendplot();
int vport();
#endif          /* __STDC__ */
#endif		/* PLOTIXdotH */
