/*********
* plxopt *              J.W.Flower, ParaSoft Corporation.
**********

This file contains machine-specific options supported by a particular 
device. Because of the vagaries of "varargs" functions and the like
we adopt a rather simple interface - a string is used to indicate
what operation is required and a single 32-bit parameter is allowed.

NOTE we are implicitly assuming that everything interesting can
fit into 32-bits, OR is a string
*/
#include <stdio.h>
#include "plotix.h"
#ifdef isMPW
#define __SEG__ PLOTIX
#include <:sys:config.h>
#include <:sys:plotix.h>
#else
#include "sys/config.h"
#include "sys/plotix.h"
#endif

int setplxopt(option, val)
     char *option;
     long val;
{
    if(_plx_curdev == (struct _plx_ind *)0) return -1;
    return INVOKE(setval)(option, val);
}

int getplxopt(option, pval)
     char *option;
     long *pval;
{
    if(_plx_curdev == (struct _plx_ind *)0) return -1;
    return INVOKE(getval)(option, pval);
}

/* 
 * Serves same purpose but with characters string valued parameters.
 */
int setplxstr(option, str)
     char *option, *str;
{
    if(_plx_curdev == (struct _plx_ind *)0) return -1;
    return INVOKE(setstr)(option, str);
}

int getplxstr(option, str, maxlen)
     char *option, *str;
     int maxlen;
{
    if(_plx_curdev == (struct _plx_ind *)0) return -1;
    return INVOKE(getstr)(option, str, maxlen);
}

#ifdef _3L
int kpxgop(pstr, pval, len)
     char *pstr;
     long *pval;
     long len;
{
    char *p, tmp[80];
    if(len >= sizeof(tmp)) len = sizeof(tmp) - 1;
    p = tmp;
    while(len--) *p++ = *pstr++;
    *p++ = '\0';
		 
    return getplxopt(tmp, pval);
}

int kpxsop(pstr, pval, len)
     char *pstr;
     long *pval;
     long len;
{
    char *p, tmp[80];
    if(len >= sizeof(tmp)) len = sizeof(tmp) - 1;
    p = tmp;
    while(len--) *p++ = *pstr++;
    *p++ = '\0';
		 
    return setplxopt(tmp, (long*)pval);
}

int kpxsst(popt, pstr, lenopt, lenstr)
     char *popt, *pstr;
     int lenopt, lenstr;
{
    char *p, tmpopt[80], tmpstr[80];

    if(lenopt > sizeof(tmpopt)) return -1;
    p = tmpopt;
    while(lenopt--) *p++ = *popt++;
    *p = '\0';

    if(lenstr > sizeof(tmpstr)) return -1;
    p = tmpstr;
    while(lenstr--) *p++ = *pstr++;
    *p = '\0';
    return setplxstr(tmpopt, tmpstr);
}

int kpxgst(popt, pstr, pmaxlen, lenopt, lenstr)
     char *popt, *pstr;
     int *pmaxlen;
     int lenopt, lenstr;
{
    char *p, tmpopt[80];

    if(lenopt > sizeof(tmpopt)) return -1;
    p = tmpopt;
    while(lenopt--) *p++ = *popt++;
    *p = '\0';

    return getplxstr(tmpopt, pstr, *pmaxlen);
}
#endif

#if COMPILER == SUNF77
int kpxgop_(pstr, pval, len)
     char *pstr;
     long *pval;
     long len;
{
    char *p, tmp[80];
    if(len >= sizeof(tmp)) len = sizeof(tmp) - 1;
    p = tmp;
    while(len--) *p++ = *pstr++;
    *p++ = '\0';
		 
    return getplxopt(tmp, pval);
}

int kpxsop_(pstr, pval, len)
     char *pstr;
     long *pval;
     long len;
{
    char *p, tmp[80];
    if(len >= sizeof(tmp)) len = sizeof(tmp) - 1;
    p = tmp;
    while(len--) *p++ = *pstr++;
    *p++ = '\0';
		 
    return setplxopt(tmp, (long*)pval);
}

int kpxsst_(popt, pstr, lenopt, lenstr)
     char *popt, *pstr;
     int lenopt, lenstr;
{
    char *p, tmpopt[80], tmpstr[80];

    if(lenopt > sizeof(tmpopt)) return -1;
    p = tmpopt;
    while(lenopt--) *p++ = *popt++;
    *p = '\0';

    if(lenstr > sizeof(tmpstr)) return -1;
    p = tmpstr;
    while(lenstr--) *p++ = *pstr++;
    *p = '\0';
    return setplxstr(tmpopt, tmpstr);
}

int kpxgst_(popt, pstr, pmaxlen, lenopt, lenstr)
     char *popt, *pstr;
     int *pmaxlen;
     int lenopt, lenstr;
{
    char *p, tmpopt[80];

    if(lenopt > sizeof(tmpopt)) return -1;
    p = tmpopt;
    while(lenopt--) *p++ = *popt++;
    *p = '\0';

    return getplxstr(tmpopt, pstr, *pmaxlen);
}
#endif		/* SUNF77 */
