#include <stdio.h>
#include "plotix.h"
#if 0
PLOTIX_DEVS plx_devlist[] = {PLX_PS, PLX_NUL};
#endif

char msg[1024];

extern int std_test(), intext_test(), sketch_test(), boxit_test(),
	clip_test(), menu_test(), logo_test();

struct {
    char *text;
    int (*func)();
} tests[] = {
    {"Quit", (int (*)())0},		/* Must always be first */
    {"\"Standard test\"", std_test},
    {"intext", intext_test},
    {"Freehand drawing", sketch_test},
    {"Boxing text", boxit_test},
    {"Simple clipping", clip_test},
    {"Menu test", menu_test},
    {"Logo", logo_test},
};

main(argc, argv)
     int argc;
     char *argv[];
{
    int i, stat, Ntests, main_dev, test;
    char tmp[80];
    long option, _ex_getnum();

    main_dev = make_display(NULL);
    printf("Status from make_display = %d\n", main_dev);
    if(main_dev < 0) exit(0);

/*
    stat = make_display("ps");
    printf("Status from make_display(ps) = %d\n", stat);
    if(stat < 0) exit(0);
*/
    setplxstr("winlab", "Plotix TEST");
    setplxstr("psfile", "jwf.ps");
    setplxopt("landscape", 1);

    swap_display(main_dev);

/*
 * Offer the selection of test, pick one and execute it....
 */
    option = 1;
    while(option) {
        strcpy(msg, "\nSelect test option:\n");
	Ntests = sizeof(tests)/sizeof(tests[0]);
	for(i=0; i<Ntests; i++) {
	    sprintf(tmp, "   %3d. %s\n", i, tests[i].text);
	    strcat(msg, tmp);
	}
	strcat(msg, "Enter option number: ");
	_ex_print(msg);
	memset(tmp, 0, sizeof(tmp));
	gets(tmp);
	if(sscanf(tmp, "%ld", &option) != 1 || option < 0 || option > Ntests)
	    continue;
	if(option > 0) {
	    safe_openpl();
	    tests[option].func();
	    _ex_print("Hit enter to return to main menu\n");
	    gets(tmp);
	    closepl();
	}
    }
    _ex_info("Quitting\n");
    exit(0);
}

#define X (double)ix
#define Y (double)iy

std_test()
{
    int x, y, ix, iy, colour, vleft, vright;
    int stat, key, left, right;
    double xx = 0.2,yy = 0.2;
    double x0, y0, x1, y1;
    int position;

    greyscale(32, 47);
    rainbow(64, 79);
    vleft = vport(0.0, 0.0, 0.5, 1.0);
    vright = vport(0.5, 0.0, 1.0, 1.0);
    ortho_space(-1., -1., 18., 18., 0);
    box(0.,0.,17.,17.,4,1);

    colour = 0;
    for(ix = 0; ix<16; ix++) {
        for(iy = 0; iy<16; iy++)
            myrect(ix, iy, colour++);
       sendplot();
    }
    sendplot();
    for(ix = 1; ix<=8 ; ix++)
        myrect(17, ix-1, -ix);
    sendplot();

    setvport(vleft);
    ortho_space(0.0, 0.0, 1.0, 1.0, 0);
    box(0.0, 0.0, 1.0, 1.0, 5, 1);
    setclip(0.0, 0.0, 0.7, 0.7);
    box(0.1, 0.1, 0.6, 0.6, 0, 1);
    sendplot();

    colour = 1;
    while(1){
        printf("Hit the Mouse wiv all your might !!!\n");
        gin(&key, &xx, &yy);
        color(colour);
        linemod(colour);
        marker(colour, xx, yy, 0.2);
        sendplot();
        if(colour++ > 6){
            dotext("The", .35,.35, 0, 0, -1);
            dotext("End", .35,.35, 0, 0, 1);
	    return;
        }
    }
}

myrect(ix, iy, colour)
int ix, iy, colour;
{
    char number[128];
    initpanel(colour, 1);
    panelpoint(X, Y);
    panelpoint(X + 1, Y);
    panelpoint(X + 1, Y + 1);
    panelpoint(X, Y + 1);
    panelpoint(X, Y);
    endpanel();
    sprintf(number, "%4d", colour);
    label(number, X + 0.2, Y + 0.5);
}

logo_test()
{
    ortho_space(0., 0., 1., 1.,0);
    box(0.2, 0.2, 0.8, 0.8, 0, 1);
    exlogo(0.2, 0.2, 0.8, 0.8);
    sendplot();
    return;
}

intext_test()
{
    char mybuf[16];
    int i = 10;
    while(i--) {
        intext("This is a test: ", .5, .5, "012d3\n", mybuf, sizeof(mybuf));
        dotext(mybuf, .5, .6, 0, 0, 0);
	sendplot();
	sleep(1);
	erase();
    }
    return;
}

sketch_test()
{
    draw_freehand();
    return;
}

draw_freehand()
{
    long nbuts, allmask;
    int stat;
    int key;
    double xx, yy;

    gtrack_on();
    if(getplxopt("nbuttons", &nbuts) < 0 || nbuts < 1) exit(0);
    allmask = (1 << nbuts) - 1;
    do {
	stat = gtrack(&key, &xx, &yy);
	if(key > 0) color(key);
	marker(0, xx, yy, 0.1);
	sendplot();
    }
    while(key != allmask);
}

boxit_test()
{
    boxit(0.5, 0.5, "This is a test");
    sendplot();
    printf("Hit enter to exit\n");
    return;
}

boxit(atx, aty, str)
     double atx, aty;
     char *str;
{
    double x, y;

    sztext(str, &x, &y);
    box(atx-0.5*x, aty-0.5*y, atx+0.5*x, aty+0.5*y, 0, 1);
    dotext(str, atx, aty, 0, 0, 0);
    sendplot();
}

safe_openpl()
{
    int stat;
    stat = openpl(8192, (FILE *)0);
    if(stat < 0) {
        _ex_fatal("Failed to initialise graphics system.\n");
        exit(1);
    }
}

clip_test()
{
    setclip(0.5, 0.5, 1.0, 1.0);
    move(1.0, 1.0);
    cont(0.0, 0.0);
    endclip();

    setclip(0.25, 0.25, 0.75, 0.75);
    box(0.1, 0.1, 0.9, 0.9, -3, 1);
    endclip();
    sendplot();
    return;
}

/*
 * Used in menu processing.
 */
int give_index(index)
    int index;
{
    return index;
}

PLXMENU test_menu[] = {
	{"Done", give_index, 0},
	{"Sel #1.", give_index, 0},
	{"Sel #2.", give_index, 0},
};

int my_miss_proc(key, x, y)
     int key;
     double x, y;
{
    printf("Missed menu: key=%d @ (%f,%f)\n", key, x, y);
    return 1;
}

int menu_test()
{
    PLXMENU *pmen;
    int which, style;
    double l, r, b, t;

    space(0., 0., 1., 1.);

    l = .7; r = 1.0;
    b = 0.; t = 1.0;
    style = MENU_SHADED;
    init_menu(style, l, b, r, t);

    push_menu(test_menu, sizeof(test_menu)/sizeof(test_menu[0]));
    while((which=process_menu(my_miss_proc)) != 0) {
	pmen = test_menu + which;
	if(pmen->code == 0) {
	    check_menu(which, pmen);
	    pmen->code = 1;
	}
	else {
	    uncheck_menu(which, pmen);
	    pmen->code = 0;
	}
    }
    pop_menu(0);
}

char *_ex_malloc (n)
int n;
{
	return malloc(n);
}
_ex_free (ptr)
char *ptr;
{
	free(ptr);
}
_ex_print (s) char *s; { printf("%s", s); }
_ex_error (s) char *s; { fprintf(stderr, "%s", s); }
_ex_info (s) char *s; { fprintf(stderr, "%s", s); }
_ex_fatal (s) char *s; { fprintf(stderr, "%s", s); }
exidle (){}
exregidle (){}
