/*
*	space.c
*/
#include <stdio.h>
#include "plotix.h"
#ifdef isMPW
#define __SEG__ PLOTIX
#include <:sys:config.h>
#include <:sys:plotix.h>
#else
#include "sys/config.h"
#include "sys/plotix.h"
#endif

static void _vpinit()
{
    struct _vptab *thisvp;
    int i;

    for(i=0; i<NVPORT; i++) {
        thisvp = &IND(vptab[i]);
        thisvp->V0[0] = (Coord)0;
        thisvp->V0[1] = (Coord)0;
#ifdef FLTCORD
        thisvp->V1[0] = (Coord)1;
        thisvp->V1[1] = (Coord)1;
#else
        thisvp->V1[0] = (Coord)IND(devx);
        thisvp->V1[1] = (Coord)IND(devy);
#endif
        thisvp->W0[0] = (Coord)0;
        thisvp->W0[1] = (Coord)0;
#ifdef FLTCORD
        thisvp->W1[0] = (Coord)1;
        thisvp->W1[1] = (Coord)1;
#else
        thisvp->W1[0] = (Coord)IND(devx);
        thisvp->W1[1] = (Coord)IND(devy);
#endif
        thisvp->clip = 0;
    }
    IND(vpinited) = 1;
    IND(cvport) = 0;
    IND(nvport) = 0;
    return;
}

/**************
* ortho_space * Roy Williams
***************
Given a user rectangle, finds a larger rectangle which has no
circle-to-ellipse distortion and centers the user's rectangle
within it.
*/
void ortho_space(x0, y0, x1, y1, position)
Coord x0, y0, x1, y1;
int position;
{
        struct _vptab *thisvp;
        Coord xp0, yp0, xp1, yp1;
        double portasp, size, surfasp, xtemp, ytemp;

        if(!IND(vpinited)) _vpinit();
        aspect(&xtemp, &ytemp);
        surfasp = xtemp/ytemp;

        thisvp = &IND(vptab[IND(cvport)]);
        portasp=(surfasp*(thisvp->V1[0]-thisvp->V0[0]))/
                                (thisvp->V1[1]-thisvp->V0[1]);

        if(portasp*(y1-y0) > (x1-x0)){
                size = (y1-y0)*portasp;
                if(position == 0){
                        xp0 = 0.5*(x0+x1) - 0.5*size;
                        xp1 = 0.5*(x0+x1) + 0.5*size;
                } else if(position < 0){
                        xp0 = x0;
                        xp1 = xp0 + size;
                } else {
                        xp1 = x1;
                        xp0 = xp1 - size;
                }
                yp0 = y0;
                yp1 = y1;
        } else {
                xp0 = x0;
                xp1 = x1;
                size = (x1-x0)/portasp;
                if(position == 0){
                        yp0 = 0.5*(y0+y1) - 0.5*size;
                        yp1 = 0.5*(y0+y1) + 0.5*size;
                } else if(position < 0){
                        yp0 = y0;
                        yp1 = yp0 + size;
                } else {
                        yp1 = y1;
                        yp0 = yp1 - size;
                }
        }
        space(xp0, yp0, xp1, yp1);
}

/********
* space *               J.W.Flower   
*********

Sets the world coordinates defined by the entire viewport
*/
void space(x0,y0,x1,y1)
Coord x0,y0,x1,y1;
{
    struct _vptab *thisvp;

    if(!IND(vpinited)) _vpinit();

    thisvp = &IND(vptab[IND(cvport)]);

    thisvp->W0[0] = x0; thisvp->W1[0] = x1;
    thisvp->W0[1] = y0; thisvp->W1[1] = y1;

    IND(mul[0])=IND(devx)*(thisvp->V1[0]-thisvp->V0[0])/
                            (thisvp->W1[0]-thisvp->W0[0]);
    IND(mul[1])=IND(devy)*(thisvp->V1[1]-thisvp->V0[1])/
                            (thisvp->W1[1]-thisvp->W0[1]);
    
    IND(off[0]) = 
        IND(devx)* (thisvp->W1[0]*thisvp->V0[0] - thisvp->W0[0]*thisvp->V1[0])/
                         (thisvp->W1[0]-thisvp->W0[0]);
    IND(off[1]) = 
        IND(devy)* (thisvp->W1[1]*thisvp->V0[1] - thisvp->W0[1]*thisvp->V1[1])/
                         (thisvp->W1[1]-thisvp->W0[1]);
    return;
}

/********
* vport *               J.W.Flower
*********

Sets the viewport in coords relative to the whole screen (0 to 1).
*/

int vport(x0,y0,x1,y1)
Coord x0,x1,y0,y1;
{
    struct _vptab *thisvp;

    if(!IND(vpinited)) _vpinit();

    if(IND(nvport) >= NVPORT) return -1;
    
    IND(cvport) = IND(nvport++);

    thisvp = &IND(vptab[IND(cvport)]);
    thisvp->V0[0] = x0; thisvp->V1[0] = x1;
    thisvp->V0[1] = y0; thisvp->V1[1] = y1;
    
    thisvp->clip = 0;
    
    setvport(IND(cvport));              /* Make new viewport current */
    return IND(cvport);
}

/***********
* setvport *
************

Switch to a new viewport. Index numbers are returned by vport.
*/
int setvport(vpnum)
     int vpnum;
{
    struct _vptab *thisvp;

    if(!IND(vpinited)) _vpinit();

    if(vpnum < 0 || vpnum >= IND(nvport)) return -1;

/* Save the relevant information about the current vport */

    thisvp = &IND(vptab[IND(cvport)]);

    thisvp->clip = IND(clip);   /* Save old clipping state */
    if(IND(clip)) 
        _getclip(thisvp->clipwin);

/* Now switch to the new vport */

    IND(cvport) = vpnum;
    thisvp = &IND(vptab[vpnum]);

    IND(mul[0]) = IND(devx)*(thisvp->V1[0]-thisvp->V0[0])/
                        (thisvp->W1[0]-thisvp->W0[0]);
    IND(mul[1]) = IND(devy)*(thisvp->V1[1]-thisvp->V0[1])/
                        (thisvp->W1[1]-thisvp->W0[1]);
    
    IND(off[0]) = 
        IND(devx)*(thisvp->W1[0]*thisvp->V0[0]-thisvp->W0[0]*thisvp->V1[0])/
                         (thisvp->W1[0]-thisvp->W0[0]);

    IND(off[1]) = 
        IND(devy)*(thisvp->W1[1]*thisvp->V0[1]-thisvp->W0[1]*thisvp->V1[1])/
                         (thisvp->W1[1]-thisvp->W0[1]);

    if(thisvp->clip) {
        setclip(thisvp->clipwin[0], thisvp->clipwin[1],
                thisvp->clipwin[2], thisvp->clipwin[3]);
    }
    else endclip();
    return 0;
}

int _inwindow(x, y)
         Coord x, y;
{
    struct _vptab *thisvp;

    if(!IND(vpinited)) _vpinit();

    thisvp = &IND(vptab[IND(cvport)]);

    if(x < thisvp->W0[0]) return 0;
    if(x > thisvp->W1[0]) return 0;
    if(y < thisvp->W0[1]) return 0;
    if(y > thisvp->W1[1]) return 0;
    return 1;
}

/* Set Clipping rectangle to be current viewport limits */
/* February 1990 */
int clipvp()
{
    struct _vptab *thisvp;

    if(!IND(vpinited)) _vpinit();

    thisvp = &IND(vptab[IND(cvport)]);
    setclip( thisvp->W0[0], thisvp->W0[1], thisvp->W1[0], thisvp->W1[1]);
    return 1;
}

#ifdef _3L
kortho(x0, y0, x1, y1, position)
     float *x0, *y0, *x1, *y1;
     int *position;
{
    ortho_space((Coord)*x0, (Coord)*y0, (Coord)*x1, (Coord)*y1, *position);
}

kspace(x0,y0,x1,y1)
     float *x0, *y0, *x1, *y1;
{
    space((Coord)*x0, (Coord)*y0, (Coord)*x1, (Coord)*y1);
}

kvport(x0,y0,x1,y1)
     float *x0, *y0, *x1, *y1;
{
    vport((Coord)*x0, (Coord)*y0, (Coord)*x1, (Coord)*y1);
}

kclpvp()
{
    clipvp();
}

ksetvp(vpnum)
     int *vpnum;
{
    setvport(*vpnum);
}
#endif                  /* _3L */

#if COMPILER == SUNF77
kortho_(x0, y0, x1, y1, position)
     float *x0, *y0, *x1, *y1;
     int *position;
{
    ortho_space((Coord)*x0, (Coord)*y0, (Coord)*x1, (Coord)*y1, *position);
}

kspace_(x0,y0,x1,y1)
     float *x0, *y0, *x1, *y1;
{
    space((Coord)*x0, (Coord)*y0, (Coord)*x1, (Coord)*y1);
}

kvport_(x0,y0,x1,y1)
     float *x0, *y0, *x1, *y1;
{
    vport((Coord)*x0, (Coord)*y0, (Coord)*x1, (Coord)*y1);
}

kclpvp_()
{
    clipvp();
}

ksetvp_(vpnum)
     int *vpnum;
{
    setvport(*vpnum);
}
#endif                  /* SUNF77 */
