/*********
* config *		J.W.Flower, ParaSoft Corporation, April 1988
**********

This file contains master configuration information. In order to configure
EXPRESS for a new system one has to basically specify three things: the host
operating system under which everything will run, the node operating
system and the target CPU node processor for which we are aiming. 

After this it should all just work out fine ????
*/
#ifndef _CONFIGdotH
#define _CONFIGdotH

/* CHOICES 1/2 ------------- HOST/NODE OPERATING SYSTEM ------------------ */

/* The following operating systems currently support CUBIX */

#define XENIX		(1)		/* MicroSoft XENIX */
#define AXIS		(2)		/* NCUBE's AXIS */
#define DOS		(3)		/* MS-DOS */
#define BSD		(4)		/* Berkeley UNIX 4.x */
#define EXPRESS		(5)		/* ParaSoft EXPRESS */
#define VERTEX		(6)		/* NCUBE Vertex */
#define VMS             (7)             /* DEC-VMS version ??? */
#define MPW		(8)		/* Macintosh */
#define UNIX5		(9)		/* AT&T Unix */

/* Set the following variable to the host operating system you intend
   to use.
*/

#ifndef H_OPSYS
#define H_OPSYS		XENIX
#endif

/* And also set the node operating system. This should be EXPRESS in all
   cases but on NCUBE devices
*/
#ifndef N_OPSYS
#define N_OPSYS		EXPRESS
#endif

/* ------------------------- HOST/NODE OPERATING SYSTEM ------------------ */

/* CHOICE 3. -------------------- TARGET NODE PROCESSOR ------------------ */

/* There are currently two supported node processors; the NCUBE chip
   and the various sorts of Transputers. Define the one that you have.
*/

#define NCUBE		(101)		/* NCUBE custom chip */
#define TRANSPUTER	(102)		/* T414 and T800 Transputers */
#define SYMULT		(103)		/* Ametek 2010 series */
#define NCUBE2		(104)		/* Second generation NCUBE */
#define I860		(105)		/* Intel i860 */

#ifndef N_CPU
#define N_CPU		TRANSPUTER
#endif

/* ------------------------------ TARGET NODE PROCESSOR ------------------ */


/* Now we get to the crux of the matter. Having given an operating system
   we have to set a few other OS dependent parameters....
*/
#if H_OPSYS == XENIX
#define sysV			/* Use system V terminal driver */
#endif

#if H_OPSYS == AXIS || H_OPSYS == BSD
#define sys7			/* Use system 7 terminal driver */
#endif

#if H_OPSYS == DOS
#undef sys7				/* DOS has neither terminal driver !! */
#undef sysV
#endif

/* The following sections are used to define various compiler combinations
   for linking FORTRAN and C together. Aain this section exists mainly
   because the #if mechanism is still up in the air. 
*/
#define GENERIC	(0)		/* Irrelevent ... */
#define CFG	(1)		/* CFG's NCUBE compiler */
#define SUNF77	(2)		/* SUN Fortran, F77 */
#define NDP	(3)		/* NDP, MicroWay 80386 */
#define MSC	(4)		/* Microsoft */
#define ZTC	(5)		/* ZORTECH */

#ifndef COMPILER
#define COMPILER	SUNF77
#endif

/* Any particular features of a compiler should be placed here. */

#if COMPILER == NDP
#define SILENT		(1)
#define HOSTCODE	(1)
#endif

/* The following section is a list of possible values for the __PS_machine
    variable.  Any changes to this list should be accompanied with similar
    changes to _ex_custrd()
*/
#define M_UNKNOWN	(0)
#define M_QUN		(1)
#define M_TOP		(2)
#define M_SUN_B011	(3)
#define M_SUN_KOBE	(4)
#define M_SUN_TOPO	(5)
#define M_LEVCO		(6)
#define M_MEIKO		(7)
#define M_MEIKO_200	(8)
#define M_SUN_B014	(9)
#define M_SUN_TOPD	(10)
#define M_SUN_NET	(11)
#define M_STD		(12)
#define M_SUN_386	(13)
#define M_SUN_NSC	(14)

/* the following describes ibterfaces between hosts and boards for dma.
We use dma on suns to indicate that we are using memory maped instead
of kernel */

#define DM_NO		(0)
#define DM_B004		(1)
#define DM_DEF		(2)
#define DM_MEMMAP	(3)
#define DM_ALTA		(4)
#endif		/* CONFIGdotH */
