/********
* xdriv *
*********

Driver code for the X-windows implementation of PLOTIX.
*/
#include "plotix.h"
#ifdef isMPW
#include <:sys:plotix.h>
#define __SEG__ PLOTIX
#else
#include "sys/plotix.h"
#endif
#include "xplot.h"
#include <stdio.h>

static char text[80];   /* For errors, etc. */
int _plx_x_loopcheck();
static void processevent();

static void _plx_x_cont(x,y)
     Coord x,y;
{
    int oldx, oldy, newx, newy;

    oldx = (int) USER_to_X(IND(xold));
    oldy = (int) USER_to_Y(IND(yold));
    newx = (int) USER_to_X(x);
    newy = (int) USER_to_Y(y);
    XDrawLine(DEP(display), DEP(win), DEP(gc), oldx, oldy ,newx, newy);
    move(x,y);
}

static void _plx_x_move(x, y)
     Coord x, y;
{
    return;
}

/*
 * This is safely static since it can be shared between
 * different incarnations of the window.
 */
static unsigned char hatch[]={
    0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* SUN hatch 1 */
    0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0,  /* SUN hatch 2 */
    0x81, 0x03, 0x06, 0x0c, 0x18, 0x30, 0x60, 0xc0,
    0x81, 0xc0, 0x60, 0x30, 0x18, 0x0c, 0x06, 0x03,
    0x18, 0x30, 0x60, 0xc0, 0x81, 0xc0, 0x60, 0x30,
    0x08, 0x00, 0x00, 0x80, 0xc1, 0x62, 0x36, 0x1c,
    0xf0, 0xf0, 0xf0, 0xf0, 0x0f, 0x0f, 0x0f, 0x0f,
    0x01, 0x01, 0x01, 0xff, 0x10, 0x10, 0x10, 0xff,
    0x3c, 0x66, 0xc3, 0x81, 0x81, 0xc3, 0x66, 0x3c,
    0x88, 0x88, 0x88, 0xff, 0x88, 0x88, 0x88, 0xff   /* SUN hatch 10 */
};
#define NHATCH          (sizeof(hatch)/8)

static void _plx_x_initpanel(shade,edge)
     int shade,edge;
{
    DEP(sv_shade) = shade;
    DEP(sv_edge)  = edge;
    DEP(polys) = DEP(polyverts);
    DEP(Npts) = 0;
    return;
}

static void _plx_x_panelpoint(x, y)
     Coord x, y;
{
    int x_curr, y_curr;

    x_curr = (int) USER_to_X(x);
    y_curr = (int) USER_to_Y(y);

    if(++DEP(Npts) > MAXVERTS) {
        sprintf(text,"Too many polygon vertices, max = %d", MAXVERTS);
        if (_ex_error(text))
            exit(1);
        DEP(Npts) = 0;
        return;
    }
    *DEP(polys++) = x_curr;
    *DEP(polys++) = y_curr;

    return;
}

static void _plx_x_endpanel()
{
    int i, shade, index;
    Pixmap stipple=NULL;

    /* First we have to close the polygon, if necessary */

    if(*(DEP(polys)-2) != DEP(polyverts[0]) || 
       *(DEP(polys)-1) != DEP(polyverts[1])) {
        if(++DEP(Npts) > MAXVERTS) {
            sprintf(text,"Too many polygon vertices, max = %d", MAXVERTS);
            if (_ex_error(text)) exit(1);
            DEP(Npts) = 0;
            return;
        }
        *DEP(polys++) = DEP(polyverts[0]);        /* Close polygon for filler */
        *DEP(polys++) = DEP(polyverts[1]);
    }

    /* Now select the proper fill pattern and color */
    if (DEP(mono)) DEP(sv_shade) *= -1;
    if(DEP(sv_shade) < 0) {
        index = -DEP(sv_shade)-1;
        shade = (index - (index % NHATCH))/NHATCH;
        shade %= 7;
        shade++;
        while (shade > 7) shade -= 7;
        stipple = XCreateBitmapFromData(DEP(display),
                RootWindow(DEP(display),DEP(screen)), 
                &(hatch[8*(index%NHATCH)]),8,8);
        XSetStipple(DEP(display),DEP(gc),stipple);
        XSetForeground(DEP(display),DEP(gc),DEP(cmap[shade]));
        XSetFillStyle(DEP(display),DEP(gc),FillOpaqueStippled);
    }
    else {
        if ( IND(grey_lo) <= DEP(sv_shade) && DEP(sv_shade) <= IND(grey_hi)) 
            XSetForeground(DEP(display),DEP(gc), DEP(cmap[DEP(sv_shade)]));
        else if (IND(rain_lo) <= DEP(sv_shade) && DEP(sv_shade) <= IND(rain_hi)) 
            XSetForeground(DEP(display),DEP(gc),DEP(cmap[DEP(sv_shade)]));
        else      /* Undefined color --> map into original 8 */
            XSetForeground(DEP(display),DEP(gc),DEP(cmap[DEP(sv_shade) % 8])); 
    }

    /* convert formats */
    for (i = 0; i < DEP(Npts); i++) {
        DEP(polyarray[i]).x = DEP(polyverts[i << 1]);
        DEP(polyarray[i]).y = DEP(polyverts[(i << 1) + 1]);
    }
    XFillPolygon(DEP(display),DEP(win),DEP(gc),DEP(polyarray),DEP(Npts),Complex,CoordModeOrigin);
    if (stipple != NULL) {
        XFreePixmap(DEP(display),stipple);
        XSetFillStyle(DEP(display),DEP(gc),FillSolid);
    }
    XSetForeground(DEP(display),DEP(gc),DEP(cmap[DEP(fgcolor)]));
    if(DEP(sv_edge)) 
        XDrawLines(DEP(display),DEP(win),DEP(gc),DEP(polyarray),DEP(Npts),CoordModeOrigin);
    DEP(Npts) = 0;
}

static void _plx_x_point(x,y)
     Coord x,y;
{
    int newx, newy;

    newx = USER_to_X(x);
    newy = USER_to_Y(y);
    XDrawPoint(DEP(display), DEP(win), DEP(gc), newx, newy);
}

static void _plx_x_aspect(xsize,ysize) 
     Coord *xsize,*ysize;
{
    *xsize = IND(devx);
    *ysize = IND(devy);
    return;
}

static void _plx_x_raspect(xsize,ysize)
     Coord *xsize,*ysize;
{
    *xsize = IND(devx);
    *ysize = IND(devy);
    return;
}

static void _plx_x_closepl()
{
    if (IND(open)) {
        sendplot();
        IND(open) = 0;  
#if 0
        XUnloadFont(DEP(display),DEP(font)->fid);
        XFreeCursor(DEP(display),DEP(ptrcursor));
        XFreeCursor(DEP(display),DEP(hrgcursor));
        XFreeGC(DEP(gc));
        XCloseDisplay(DEP(display));
#endif
    }
}

static void _plx_x_color(shade)
     int shade;
{
    DEP(fgcolor) = shade % 256;
    if (DEP(fgcolor) > 1 && DEP(mono)) DEP(fgcolor) = 1;
    XSetForeground(DEP(display),DEP(gc),DEP(cmap[DEP(fgcolor)]));
}

static void _plx_x_dotext(string, x, y, angle, ljust, hjust)
     char *string;
     Coord x, y;
     int ljust, hjust, angle;
{
    int x_curr,y_curr,high,wide,length,dummy;
    XCharStruct temp;

    XSetForeground(DEP(display),DEP(gc),DEP(cmap[1]));   /* Black */
    x_curr = (int) USER_to_X(x);
    y_curr = (int) USER_to_Y(y);
    length = strlen(string);
    XTextExtents(DEP(font),string,length,&dummy,&dummy,&dummy,&temp);
    high = temp.ascent+temp.descent;
    wide = temp.width;

    if (ljust == 0)     x_curr -= (wide >> 1);
    else if (ljust > 0) x_curr -= wide;

    if (hjust == 0)     y_curr += (high >> 1);
    else if (hjust > 0) y_curr += high;

    XDrawString(DEP(display),DEP(win),DEP(gc),x_curr,y_curr,string,length);
    XSetForeground(DEP(display),DEP(gc),DEP(cmap[DEP(fgcolor)]));
}

static void _plx_x_sztext(str, pwd, pht)
     char *str;
     Coord *pwd, *pht;
{
    int dummy;
    XCharStruct temp;

    XTextExtents(DEP(font),str,strlen(str),&dummy,&dummy,&dummy,&temp);
    *pwd = (double)(temp.width)/IND(mul[0]);
    *pht = (double)(temp.ascent+temp.descent)/IND(mul[1]);
    return;
}

#ifdef DBG_INTEXT
static char give_str[] = "ab\177cdefghijklmnopqrstuvwxyz";
static char *give_ptr = give_str;
#endif

static int poll_getchar()
{
    int c;

#ifndef DBG_INTEXT
    DEP(loopmode) = QueuedAfterFlush;
    do {
        DEP(fallout) = 0;
        while(!DEP(fallout)) _plx_x_loopcheck();
    } while(DEP(button) < 32 && DEP(button) > 0177);
    DEP(loopmode) = MODEDEFAULT;
    return DEP(button);
#else
    return *give_ptr++;
#endif
/* NOTREACHED */
}

static void _plx_x_intext(prompt, x, y, term, input, maxlen)
     char *prompt;
     Coord x, y;
     char *term, *input;
     int maxlen;
{
    Coord in_x;
    int dummy, at_x, at_y;
    char c, *at_ptr, *strchr();
    int got = 0;
    XCharStruct msize;

#ifdef DBG_INTEXT
    give_ptr = give_str;
#endif
    memset(input, 0, maxlen);
    dotext(prompt, x, y, 0, -1, -1);

    XTextExtents(DEP(font), prompt, strlen(prompt), 
	&dummy, &dummy, &dummy, &msize);

    in_x = x + ((double)msize.width)/IND(mul[0]);
    at_x = USER_to_X(in_x);
    at_y = USER_to_Y(y);

    at_ptr = input;
    do {
	c = poll_getchar();
	if(c == '\r') c = '\n';

	if((c == 8 || c == 127) && got > 0) {	/* CTRL-H or DEL */
	    XTextExtents(DEP(font), input, got, 
		&dummy, &dummy, &dummy, &msize);
	    XClearArea(DEP(display), DEP(win), 
		at_x, at_y-msize.ascent, 	/* ascent > 0 */
		msize.width, msize.ascent+msize.descent, False);
	    *--at_ptr = '\0';
	    got--;
	}
	else if(c >= 32) {		/* Only ASCII, please! */
	    *at_ptr++ = c;
	    got++;
	}
	dotext(input, in_x, y, 0, -1, -1);
	sendplot();

    } while(got < maxlen && strchr(term, c) == (char *)0);
    *at_ptr = '\0';
    return;
}

static void _plx_x_erase()
{
    XClearWindow(DEP(display),DEP(win));
    return;
}

static int _plx_x_gin(pbutton, px, py)
    int *pbutton;
    Coord *px,*py;
{
    DEP(loopmode) = QueuedAfterFlush;
    DEP(fallout) = 0;
    while (!DEP(fallout)) _plx_x_loopcheck();
    DEP(loopmode) = MODEDEFAULT;
    DEP(mouse_x) = X_to_USER(DEP(mx));
    DEP(mouse_y) = Y_to_USER(DEP(my));
    *pbutton = DEP(button);
    *px = DEP(mouse_x);
    *py = DEP(mouse_y);

    return 0;
}

static int _plx_x_gtrack_on()
{
    return 0;
}

static int _plx_x_gtrack(pkey, px, py)
     int *pkey;
     Coord *px, *py;
{
    Bool stat;
    int root_x, root_y, win_x, win_y;
    Window root, child;
    unsigned int keys;

    do {
	_plx_x_loopcheck();
        stat = XQueryPointer(DEP(display), DEP(win), &root, &child, 
				&root_x, &root_y,
				&win_x, &win_y, &keys);
	if(stat == True) {
	    if(win_x > IND(devx)) stat = False;
	    if(win_y > IND(devy)) stat = False;
	}
    } while(stat == False);
/*
 * Now decode what it's told us.
 */
    *pkey = 0;
    if(keys & Button1Mask) *pkey |= 0x01;
    if(keys & Button2Mask) *pkey |= 0x02;
    if(keys & Button3Mask) *pkey |= 0x04;
    if(keys & Button4Mask) *pkey |= 0x08;
    if(keys & Button5Mask) *pkey |= 0x10;
    *px = X_to_USER(win_x);
    *py = Y_to_USER(win_y);

    return 0;
}

static void _plx_x_gtrack_off()
{
    return;
}

/*
 * The different line types can also be safely shared.
 */
static int dash_list_length[]= {2,4,2,2,3};
static unsigned char dotted[2]={3,1};
static unsigned char dot_dashed[4]={3,4,3,1};
static unsigned char short_dashed[2]={4,4};
static unsigned char long_dashed[2]={4,7};
static unsigned char odd_dashed[3]={1,2,3};
static unsigned char *dash_list[]={
        dotted,
        dot_dashed,
        short_dashed,
        long_dashed,
        odd_dashed
};

static void _plx_x_linemod(style)
     int style;
{
    XGCValues gcv;

    DEP(linestyle) = style % NUMLIN;
    if (DEP(linestyle) == 0) {
        XSetLineAttributes(DEP(display),DEP(gc),1,LineSolid,CapButt,JoinMiter);
    } else {
        XSetDashes(DEP(display),DEP(gc),0,dash_list[DEP(linestyle)-1],
                dash_list_length[DEP(linestyle)-1]);
        XSetLineAttributes(DEP(display),DEP(gc),1,LineOnOffDash,CapButt,JoinMiter);
    }
    return;
}

static int _plx_x_setval(option, val)
     char *option;
     long val;
{
    if(!strcmp(option, "redraw"))         DEP(redraw) = val;
    else if(!strcmp(option, "width"))     DEP(wishx) = val;
    else if(!strcmp(option, "height"))    DEP(wishy) = val;
    else if(!strcmp(option, "winx"))      DEP(atx) = (int)val;
    else if(!strcmp(option, "winy"))      DEP(aty) = (int)val;
    else return -1;

    return 0;
}

static int _plx_x_getval(option, pval)
     char *option;
     long *pval;
{
    unsigned char maxbuts[8];

    if(!strcmp(option, "nlcolors")) {
        if(DEP(mono)) *pval = 2;
        else *pval = 256;
        return 0;
    }
    else if(!strcmp(option, "nlstyles")) {
        *pval = NUMLIN;
        return 0;
    }
    else if(!strcmp(option, "nbuttons")) {
	if(DEP(display) != (Display *)0)
	    *pval = XGetPointerMapping(DEP(display), maxbuts, sizeof(maxbuts));
	else
	    *pval = 0;
	return 0;
    }
    return -1;
}

static int _plx_x_setstr(option, str)
     char *option, *str;
{
    if(!strcmp(option, "winlab")){
        DEP(winlab)[NCHARLAB-1] = '\0';
        strncpy(DEP(winlab), str, sizeof(DEP(winlab)));
    }
    else
        return -1;
    return 0;
}

static int _plx_x_getstr(option, str, maxlen)
     char *option, *str;
     int maxlen;
{
    return -1;
}

static void _plx_x_sendplot()
{
    DEP(loopmode) = QueuedAfterFlush;
    while(exidle());
    DEP(loopmode) = MODEDEFAULT;
}

static void _plx_x_usendplot()
{
    _plx_x_sendplot();
}

static void _plx_x_asendplot()
{
    _plx_x_sendplot();
}

static int _plx_x_plothwm()
{
    return 0;
}

static void setcmap(r,g,b,pos)
    unsigned int r,g,b;
    int pos;
{
    XColor colorcelldef;

    colorcelldef.red = r;
    colorcelldef.blue = b;
    colorcelldef.green = g;
    colorcelldef.flags = DoRed | DoGreen | DoBlue;
    if (XAllocColor(DEP(display),DEP(xcmap),&colorcelldef) == 0) 
        DEP(cmap[pos]) = pos;
    else
        DEP(cmap[pos]) = colorcelldef.pixel;
    return;
}

static int _plx_x_openpl(bufsiz, grafp)
     int bufsiz;
     FILE *grafp;
{
    char *display_name = NULL;
    int display_width,display_height;
    XSizeHints size_hints;
    XGCValues values;
    unsigned char map[3];
    int depth;

    if (!IND(open)) {
        if ( (DEP(display) = XOpenDisplay(display_name)) == NULL ) {
            return -1;
        }
        DEP(screen) = DefaultScreen(DEP(display));
        display_width  = DisplayWidth(DEP(display),DEP(screen));
        display_height = DisplayHeight(DEP(display),DEP(screen));
        if(DEP(wishx) == 0) DEP(wishx) = 2*display_width/3;
        if(DEP(wishy) == 0) DEP(wishy) = display_height/2;
        DEP(win) = XCreateSimpleWindow(DEP(display),RootWindow(DEP(display),DEP(screen)),
                0,0,DEP(wishx),DEP(wishy),1,
                BlackPixel(DEP(display),DEP(screen)),
                WhitePixel(DEP(display),DEP(screen)) ); 
        size_hints.flags = PPosition | PSize | PMinSize;
/* Center on display */
        if(DEP(atx) == 0) size_hints.x = DEP(wishx)/6;
        else              size_hints.x = DEP(atx);
        if(DEP(aty) == 0) size_hints.y = DEP(wishy)/4;
        else              size_hints.y = DEP(aty);

        size_hints.width = DEP(wishx);
        size_hints.height = DEP(wishy);
        size_hints.min_width = 350;
        size_hints.min_height = 250;
        if(DEP(winlab)[0] == '\0')
            strcpy(DEP(winlab), "Plotix");
        XSetStandardProperties(DEP(display),DEP(win),DEP(winlab),DEP(winlab),
                None, NULL, 0, &size_hints);
        XSelectInput(DEP(display),DEP(win),ExposureMask | KeyPressMask |
                ButtonPressMask | StructureNotifyMask );

        DEP(font) = XLoadQueryFont(DEP(display),"8x13");
        if (DEP(font) == NULL) return -1;

        DEP(gc) = XCreateGC(DEP(display),DEP(win),0L,&values);
        XSetFont(DEP(display),DEP(gc),DEP(font)->fid);
        XSetForeground(DEP(display),DEP(gc),BlackPixel(DEP(display),DEP(screen)));
        XSetBackground(DEP(display),DEP(gc),WhitePixel(DEP(display),DEP(screen)));
        XMapWindow(DEP(display),DEP(win));
        DEP(ptrcursor) = XCreateFontCursor(DEP(display),XC_top_left_arrow); /* 132 */
	DEP(hrgcursor) = XCreateFontCursor(DEP(display),XC_watch /* 150 */ );
        XDefineCursor(DEP(display),DEP(win),DEP(ptrcursor));
        DEP(mouse) = XGetPointerMapping(DEP(display),map,3);  /* Get # of buttons */    
        depth = DisplayPlanes(DEP(display),DEP(screen));
        if (depth == 1) DEP(mono) = 1;
        DEP(xcmap) = DefaultColormap(DEP(display),DEP(screen));

        setcmap(MAXCOL,MAXCOL,MAXCOL,0);/* White (background) */
        setcmap(  0,  0,  0,1);         /* Black (foreground) */
        setcmap(MAXCOL,  0,  0,2);      /* Red */
        setcmap(  0,MAXCOL,  0,3);      /* Green */
        setcmap(  0,  0,MAXCOL,4);      /* Blue */
        setcmap(  0,MAXCOL,MAXCOL,5);   /* Cyan */
        setcmap(MAXCOL,  0,MAXCOL,6);   /* Magenta */
        setcmap(MAXCOL,MAXCOL,  0,7);   /* Yellow */

        IND(open) = 1;
	exregidle(_plx_x_loopcheck);
        while(IND(devx) == 0 || IND(devy) == 0)
            _plx_x_loopcheck();
    }
    return 1;
}

/* ARGSUSED */
static int _plx_x_kopenp(bufsiz, user_ptr)
     int bufsiz;
     char *user_ptr;
{
    return _plx_x_openpl(0, (FILE *)0);
}

static int _plx_x_setcurs(which)
    int which;
{
    switch(which) {
    case PLX_HRG_CURSOR:
        XDefineCursor(DEP(display),DEP(win),DEP(hrgcursor));
	break;
    case PLX_PTR_CURSOR:
        XDefineCursor(DEP(display),DEP(win),DEP(ptrcursor));
	break;
    default:
	return -1;
    }
    return 1;
}

static void _plx_x_greyscale(from, to)
     int from, to;
{
    int i;
    unsigned int rgb;

    if (to < from) {
        i = from; from = to; to = i;
    }
    if (to > 255) to = 255;
    IND(grey_lo) = from;
    IND(grey_hi) = to;
    for (i = from; i < to; i++) {
        rgb = (unsigned int) ((double) (MAXCOL * (i - IND(grey_lo)))/ 
                        (double)(IND(grey_hi) - IND(grey_lo)));
        setcmap(rgb,rgb,rgb,i);
    }
    return;
}

static void _plx_x_rainbow(from, to)
    int from, to;
{
    int i,j;
    unsigned int r,g,b;
    double h,f,n;

    if (to < from) {
        i = from; from = to; to = i;
    }
    if (to > 255) to = 255;
    IND(rain_lo) = from;
    IND(rain_hi) = to;

    for (j = from; j < to; j++) {
        h = (double) (IND(rain_hi) - j) / (IND(rain_hi) - IND(rain_lo));
        if (h > .999) h = 0.;
        h *= 6.;
        i = (int) h;
        f = h - (double) i;
        n = 1. - f;
        switch(i) {
        case 0: r = MAXCOL; g=(unsigned int) (f * MAXCOL); b = 0; break;
        case 1: r = (unsigned int) (n * MAXCOL); g = MAXCOL; b = 0; break;
        case 2: r = 0; g = MAXCOL; b = (unsigned int) (f * MAXCOL); break;
        case 3: r = 0; g = (unsigned int) (n * MAXCOL); b = MAXCOL; break;
        case 4: r = (unsigned int) (f * MAXCOL); g = 0; b = MAXCOL; break;
        case 5: r = MAXCOL; g = 0; b=(unsigned int) (n * MAXCOL); break;
        }
        setcmap(r,g,b,j);
    }
    return;
}

static void _plx_x_setclip(x0, y0, x1, y1)
     Coord x0, y0, x1, y1;
{
    int l,b,r,t; 
    XRectangle rect[1];

    l = (int) USER_to_X(x0);
    b = (int) USER_to_Y(y0);
    r = (int) USER_to_X(x1);
    t = (int) USER_to_Y(y1);

    rect[0].x = 0; 
    rect[0].y = 0;
    rect[0].width = (r - l + 1);
    rect[0].height = (b - t + 1);
    XSetClipRectangles(DEP(display),DEP(gc),l,t,rect,1,YXSorted);
    IND(clip) = 1;
}

static void _plx_x_endclip()
{
    int l,b,r,t; 
    XRectangle rect[1];

    l = (int) (IND(vptab[IND(cvport)].V0[0]) * IND(devx));
    b = (int) ( ((Coord) 1 - IND(vptab[IND(cvport)].V0[1])) * IND(devy));
    r = (int) (IND(vptab[IND(cvport)].V1[0]) * IND(devx));
    t = (int) ( ((Coord) 1 - IND(vptab[IND(cvport)].V1[1])) * IND(devy));

    rect[0].x = 0; 
    rect[0].y = 0;
    rect[0].width = (r - l + 1);
    rect[0].height = (b - t + 1);

    XSetClipRectangles(DEP(display),DEP(gc),l,t,rect,1,YXSorted);

    IND(clip) = 0;
    return;
}

int _plx_x_loopcheck()
{
    int numevents;

    if (!IND(open)) return 0;
    numevents = XEventsQueued(DEP(display),DEP(loopmode));
    if (!numevents) return 0;
    processevent();
    return (numevents - 1);
}


/* 
 * This procedure assumes there is an event waiting... if not, it will
 * wait for one. 
 */
static void processevent() {
    XEvent report;
    int count;
    char buffer;
    KeySym key;

    XNextEvent(DEP(display),&report);
    switch(report.type) {
    case Expose:
        while (XCheckTypedEvent(DEP(display),Expose,&report));
        if(DEP(redraw) != NULL) DEP(redraw)();
        break;
    case ConfigureNotify:
        IND(devx) = (Coord) report.xconfigure.width;
        IND(devy) = (Coord) report.xconfigure.height;
        setvport(IND(cvport));
        break;

    case ButtonPress:
        DEP(mx) = report.xbutton.x; 
        DEP(my) = report.xbutton.y;
        DEP(fallout) = 1;
        DEP(button) = report.xbutton.button;
        break;

    case KeyPress:
        count = XLookupString(&report,&buffer,1,&key,NULL);
        if (count > 0) {
            DEP(fallout) = 1;
            DEP(button) = (int) buffer;
        }
        /* This is done so we can process more keysyms later if necessary by
        adding code something like this:
        switch(key) {
        case XK_return:
            DEP(fallout) = 1; DEP(button) = '\n'; break;
        }
        and so on...  These are keys without ASCII strings */

        if (DEP(fallout)) {
            DEP(mx) = report.xkey.x;
            DEP(my) = report.xkey.y;
        }
        break;
    default:
        break;  
    }
    return;
}
/*
 * Assign the function pointers that describe a particular driver.
 */
static int set_fn_ptrs(ptr)
     struct _plx_ind *ptr;
{
    if(ptr == (struct _plx_ind *)0) return -1;

    ptr->setclip    = _plx_x_setclip;
    ptr->endclip    = _plx_x_endclip;
    ptr->closepl    = _plx_x_closepl;
    ptr->color      = _plx_x_color;
    ptr->cont       = _plx_x_cont;
    ptr->dotext     = _plx_x_dotext;
    ptr->sztext     = _plx_x_sztext;
    ptr->intext     = _plx_x_intext;
    ptr->erase      = _plx_x_erase;
    ptr->gin        = _plx_x_gin;
    ptr->gtrack_on  = _plx_x_gtrack_on;
    ptr->gtrack     = _plx_x_gtrack;
    ptr->gtrack_off = _plx_x_gtrack_off;
    ptr->aspect     = _plx_x_aspect;
    ptr->raspect    = _plx_x_raspect;
    ptr->sendplot   = _plx_x_sendplot;
    ptr->asendplot  = _plx_x_asendplot;
    ptr->usendplot  = _plx_x_usendplot;
    ptr->plothwm    = _plx_x_plothwm;
    ptr->point      = _plx_x_point;
    ptr->linemod    = _plx_x_linemod;
    ptr->move       = _plx_x_move;
    ptr->openpl     = _plx_x_openpl;
    ptr->kopenp     = _plx_x_kopenp;
    ptr->rainbow    = _plx_x_rainbow;
    ptr->greyscale  = _plx_x_greyscale;
    ptr->initpanel  = _plx_x_initpanel;
    ptr->panelpoint = _plx_x_panelpoint;
    ptr->endpanel   = _plx_x_endpanel;
    ptr->polgn      = 0;
    ptr->box        = 0;
    ptr->setval     = _plx_x_setval;
    ptr->getval     = _plx_x_getval;
    ptr->setstr     = _plx_x_setstr;
    ptr->getstr     = _plx_x_getstr;
    ptr->setcurs    = _plx_x_setcurs;
    return 0;
}
 
/*
 * This routine serves a VERY IMPORTANT purpose. It must allocate
 * sapce for the nexly created deivces "dependent" and "independent"
 * structures (or allocate them statically) and fill in any entries
 * that must have sensible defaults.
 */
int _plx_x_ctor(pdev)
     struct _plx_ind **pdev;
{
    extern char *calloc();
    struct _plx_ind *xdev;
    struct _plx_dep *xdep;

    xdev = (struct _plx_ind *)calloc(1, sizeof(*xdev));
    xdep = (struct _plx_dep *)calloc(1, sizeof(*xdep));
    if(xdep == (struct _plx_dep *)0 || xdev == (struct _plx_ind *)0)
        return -1;

    *pdev = xdev;
    _plx_curdev = xdev;
    IND(devdep) = xdep;

    set_fn_ptrs(xdev);
/*
 * Set defaults for hardware dependent parameters.
 */
    DEP(fgcolor) = 1;             /* Foreground drawing color */
    DEP(mouse) = 0;               /* Number of mouse buttons */

    DEP(redraw) = NULL;           /* Pointer to repaint procedure */

    DEP(wishx) = 0;
    DEP(wishy) = 0;               /* Requested screen size */
    DEP(atx) = 0;
    DEP(aty) = 0;

    DEP(loopmode) = MODEDEFAULT;
    DEP(fallout) = 0;
    DEP(button) = 0;
    DEP(mx) = DEP(my) = 0;

    memset(DEP(winlab), 0, sizeof(DEP(winlab)));

    IND(devx) = DEV_x;
    IND(devy) = DEV_y;
    return 0;
}
