/*
Some routines to do parallel random number generation.  The idea is
to take a standard linear congruential algorithm, and have every
processor compute the Pth iterate of that algorithm (where P is the
number of processors.)  If the seeds are set up properly (staggered)
the processors leapfrog over one another, and it is just as good as
having used the basic algorithm on a sequential machine. See Chap 12
of the Book.
*/

#include "strings.h"
#include <fcntl.h>

#define MULT	1103515245
#define ADD	12345
#define MASK	( 0x7fffffff )

static double _twoto31 = 2147483648.0 ;
static long AAA,BBB;
long randx;

pranset(seed)
/*
Unlike the sequential algorithm, you MUST call pranset, even if
you want a seed of 1.  Otherwise all processors will produce the
same sequence.
*/
long seed;
{
	int proc;

	AAA= 1;
	BBB = 0;
	for (proc=0;proc<nproc;++proc) {
	   AAA = (MULT * AAA) & MASK;
	   BBB = (MULT * BBB + ADD) & MASK;
	   if (proc == procnum) randx = (AAA*seed + BBB) & MASK;
	}
}

double ranf()
/*
	Return a random double in [0, 1.0)
*/
{
	double retvalue;

	retvalue = randx / _twoto31;
	randx = (AAA*randx+BBB)&MASK;
	return(retvalue);
}

iranf()
/*
	Return a random int in [0, 2^31)
*/
{

	randx = (AAA*randx+BBB)&MASK;
	return(randx);
}


/**********
* setseed *    J.W.Flower & the infamous Dr. O  March 1985
***********

Used to set the seed for the parallel random number system -
MUST be called or all processors will execute same sequence! - see
Chapter 12 of *THE BOOK*.  Two options:
  i) Input seed from terminal                        
  ii) Read in seed from standard file "Seed".  
ALSO: at the end of production runs, don't forget to store current
seed in "Seed" by running option "d". 
N.B. You don't need to do this any more, since the seed is dumped at 
the end of the backup file. 
*/

setseed()
{
    FILE *fp ;
    int mode ;
    long seed ;

    printf("Select mode of seed setting : (0=input,1=file) : ") ;
    fflush(stdout);
    scanf("%d",&mode) ;
    printf("%d\n",mode) ;
    fflush(stdout);

    if (mode == 0) {
       printf("Input seed : "); fflush(stdout);
       scanf("%ld",&seed);
       pranset(seed);
       printf("Seed set to %ld\n",seed); fflush(stdout);
    }  
    else if (mode != 1) {
       perror("setseeds parameter must be 0 or 1\n") ;
       exit(-42);
    }

    /* Set up the names for the seeds and pointers files */
    if(lambdaE == 0.0) {
	sprintf(seedsname, "N%d/seedsA%4.3f_%d", nnode, lambdaA, dim);
	sprintf(ptsname, "N%d/ptsA%4.3f_%d", nnode, lambdaA, dim);
    }
    else {
	sprintf(seedsname, "N%d/seedsE%4.3f_%d", nnode, lambdaE, dim);
	sprintf(ptsname, "N%d/ptsE%4.3f_%d", nnode, lambdaE, dim);
    }

    /* Initialize the TWC random number generator 
     * (uses the Prand rng to set up the seeds). */
    if (mode == 0) {
      init_seeds();
    }  
    else {
      read_seeds(seedsname,ptsname);
    }
    printf("Initialized random number generator\n"); fflush(stdout);

    return(0) ;
}

