
/* This program converts the raw data for the strings runs to 
   averages and errors, and works out the autocorrelations.
   The input can be multiple files of data.

   N.B. Data should be analyzed on the same number of processors 
   as produced it.
*/

#include <express.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#define NOBS  15       /* number of observables */
#define NAUTO 6        /* number of observables to print out 
			  autocorrelations for */
#define MAXNA 1000     /* max distance of autocorr measurements */
#define M     NOBS

double lastNEA[NOBS][MAXNA], EA[NOBS][MAXNA], EA2[NOBS][MAXNA]; 
double errEA[NOBS][MAXNA];

float energy[NOBS];
double spheat[NOBS];
double errener[NOBS], errspheat[NOBS];
float maxenergy[NOBS], minenergy[NOBS];
double avenergy[NOBS], avenersq[NOBS];
double avspheat[NOBS], avspheatsq[NOBS], avbinenersq[NOBS];
double partenergy[NOBS], partenersq[NOBS];
double binmeansq;

float lambda, mass;
int i,j,m,n, idata, binsz, numbins, sites;
int cntenergy, NA, sweeps, ndata;

FILE *infp, *fp, *fopen(), *fpax;

/* Express variables */
int procnum, nprocs;

main()
{
char *tmp_ptr, fileerror[128],filename[128];
char outname[128],dataname[128];
short next;

   /* Set up the Express parameters: 
    *    nprocs  = number of processors
    *    procnum = processor number 
    */
   struct nodenv nodedata;
   exparam(&nodedata);
   procnum = nodedata.procnum;
   nprocs  = nodedata.nprocs;

   scanf("%d",&sites);    /* size of nodes in the mesh */
   scanf("%d",&binsz);    /* size of bins for error estimates */
   scanf("%d",&NA);       /* number of autocorrs measured */
   scanf("%d",&sweeps);   /* sweeps per measurement */
   scanf("%s",outname);   /* name of output file extension */

   printf("nodes = %d\n",sites);
   printf("bin size = %d\n",binsz);
   printf("NA = %d\n",NA);
   printf("sweeps per measurement = %d\n",sweeps);
   fflush(stdout);

    /* initialize everything */

    cntenergy=0; 

for (m=0;m<NOBS;m++) {

    maxenergy[m] = -100000.0; 
    minenergy[m] =  100000.0;

    avenergy[m]=0.; avenersq[m]=0.; 
    avspheat[m]=0.; avspheatsq[m]=0.; 
    partenergy[m]=0.; partenersq[m]=0.; 
    avbinenersq[m]=0; 
    
    for(i=0; i<MAXNA; i++) {
       lastNEA[m][i] = 0.0;
       EA[m][i] = 0.0; 
       EA2[m][i] = 0.0;
    }
}


while (scanf("%s",filename) != EOF) {

    scanf("%d", &ndata);
    printf("Reading %d measurements from file %s ..... \n", ndata, filename);
    fflush(stdout);

    if ((fp = fopen(filename,"r")) == NULL) {
       sprintf(fileerror, "Data file %s not found\n", filename);
       fprintf(stderr,fileerror);
       exit(-1);
    }
    fmulti(fp);

    /* Looking for EOF as in sequential code doesn't work, so need
       to know ndata and loop over it */
    for (idata=0;idata<ndata;idata++) {

       /* This is pretty horrible, but need to read in everything
	  up to the fflush when they were printed out, and this
	  is the only way I know how to do it. 
	  Note that if it was binary data it would be trivial -
          fread(energy, sizeof(float), NOBS, fp);
        */
       fscanf(fp,"%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f", 
	      &energy[0],  &energy[1],  &energy[2],  &energy[3],  &energy[4], 
	      &energy[5],  &energy[6],  &energy[7],  &energy[8],  &energy[9], 
	      &energy[10], &energy[11], &energy[12], &energy[13], &energy[14]);
       cntenergy++;
       analyze();
     }

     printf("Finished reading data from %s\n", filename);
     fflush(stdout);

} /* end of loop over input files */

     numbins = cntenergy/binsz;
     for (m=0;m<M;m++) {
        avenergy[m] /= cntenergy;
        avenersq[m] /= cntenergy;
        avbinenersq[m] /= numbins;
        spheat[m] = avbinenersq[m] - avenergy[m]*avenergy[m];
        errener[m] = sqrt( spheat[m] / numbins );
        avspheat[m] /= numbins;
        avspheatsq[m] /= numbins;
        spheat[m] = avspheatsq[m] - avspheat[m]*avspheat[m];
        errspheat[m] = sqrt( spheat[m] / numbins );
        spheat[m] = avenersq[m] - avenergy[m]*avenergy[m];

        if (cntenergy >= NA) { 	 /* finish energy autocorr */
           finish_autocorr(lastNEA,EA,avenergy,cntenergy,m);
	}
     }
     /* Specific heat for the gaussian action is multiplied
      * by the number of nodes - 1, so have to scale it.
      */
     spheat[0]   *= (double)(sites-1) / (double)sites;
     avspheat[0] *= (double)(sites-1) / (double)sites;

     printf("Finished analyzing data\n");
     printf("Number of measurements = %d\n", cntenergy);
     printf("Number of bins = %d\n", numbins);
     printf("\n");
     fflush(stdout);

   /* Print out results for each processor */

   for (n=0; n<nprocs; n++) {

      /* Open output file */
      sprintf(dataname, "anal.%s.P%d", outname,n+1); 
      fpax = fopen(dataname,"w");
      fmulti(fpax);

      if (procnum == n) {
         print_data(n);
      } 
      fflush(fpax);

      /* close output file */
      fclose(fpax);

   } /* end of loop over processors */

   /* Average data over processors */
   proc_average();

   /* Print out averages */

   /* Open output file */
   sprintf(dataname, "anal.%s", outname); 
   fpax = fopen(dataname,"w");
   fsingl(fpax);

   print_data(nprocs);
   fflush(fpax);

   /* close output file */
   fclose(fpax);


   exit(0);

} /* end of main */



/* FUNCTIONS */


analyze()
{
   for (m=0;m<M;m++) {
      partenergy[m] += energy[m];
      partenersq[m] += energy[m] * energy[m];
      if (energy[m] < minenergy[m]) minenergy[m] = energy[m];
      if (energy[m] > maxenergy[m]) maxenergy[m] = energy[m];
      if (cntenergy % binsz == 0) {
	 binmeansq = (partenergy[m]*partenergy[m])/(binsz*binsz);
	 spheat[m] = partenersq[m]/binsz - binmeansq;
	 avspheat[m] += spheat[m];
	 avspheatsq[m] += spheat[m] * spheat[m];
         avenergy[m] += partenergy[m];
         avenersq[m] += partenersq[m];
	 avbinenersq[m] += binmeansq;
         partenergy[m] = 0.;
         partenersq[m] = 0.;
      }
      /* Energy auto-correlations */
      autocorrs(lastNEA,EA,energy,m);
   }
}


autocorrs(lastNA,A,value,m)
double lastNA[M][MAXNA], A[M][MAXNA];
float value[M];
int m;
{
   for (i=1;i<NA;i++) lastNA[m][i-1] = lastNA[m][i];
   lastNA[m][NA-1] = value[m];
   for (i=0;i<NA;i++) {
      A[m][i] += lastNA[m][0] * lastNA[m][i];
   }
}


finish_autocorr(lastNA,A,average,count,m)
double lastNA[M][MAXNA], A[M][MAXNA];
double average[M];
int count, m;
{
   for (j=0;j<NA;j++) {
      for (i=1;i<NA;i++)
	 lastNA[m][i-1] = lastNA[m][i];
      lastNA[m][NA-1] = 0;
      for (i=0;i<NA;i++)
	 A[m][i] += lastNA[m][0] * lastNA[m][i];
   }
   for (i=0;i<NA;i++) {
      A[m][i] /= (count-i);
      A[m][i] -= (average[m] * average[m]);
   }
}


print_data(n)
int n;
{

   if (n == nprocs) {
      fprintf(fpax,"average over %d processors\n", nprocs);
      cntenergy *= nprocs;
   }
   else {
      fprintf(fpax,"processor %d of %d\n", n+1,nprocs);
   }
   fprintf(fpax,"nodes = %d\n",sites);
   fprintf(fpax,"bin size = %d\n",binsz);
   fprintf(fpax,"NA = %d\n",NA);
   fprintf(fpax,"sweeps per measurement = %d\n",sweeps);

   fprintf(fpax, "Number of measurements = %d\n", cntenergy);
   fprintf(fpax, "Number of bins = %d\n", numbins);
   fprintf(fpax, "\n");

   /* Average of energies */
   for (m=0;m<M;m++) {
      fprintf(fpax,"ave (%2d) %12.5f  +/- %9.5f\n", m,avenergy[m],errener[m]);
   }
   fprintf(fpax,"\n");

   /* Specific heats */
   for (m=0;m<M;m++) {
      fprintf(fpax,"var (%2d) %12.5f  :  %12.5f  +/- %9.5f\n", 
      	      m,sites*spheat[m],sites*avspheat[m],sites*errspheat[m]);
   }
   fprintf(fpax,"\n");

   /* Maxima and minima of energies */
   for (m=0;m<M;m++) {
      fprintf(fpax,"%3d   min =%12.6f    max =%12.6f\n", 
	           m, minenergy[m],maxenergy[m]);
   }
   fprintf(fpax,"\n");

   /* Autocorrelations in the energies */
   if (cntenergy >= NA) {
      /* Print out autocorrelation functions */
      print_autocorrs(0,NAUTO);
      print_autocorrs(NAUTO,2*NAUTO);
      /* Print out autocorrelation lengths */
      print_autolen(0,NAUTO);
      print_autolen(NAUTO,2*NAUTO);
   }

}


print_autocorrs(first,last)
int first, last;
{
  fprintf(fpax,"Auto correlations\n");
  fprintf(fpax,"\n");
  for (i=0;i<NA;i++) {
     fprintf(fpax,"%3d  ", i); 
     for (m=first;m<last;m++) {
	if (EA[m][0] != 0.0)
           fprintf(fpax,"%9.5f  ", EA[m][i]/EA[m][0]);
	else
           fprintf(fpax,"  0.00000  ");
     }
     fprintf(fpax,"\n");
  }
  fprintf(fpax,"\n");
}


print_autolen(first,last)
int first, last;
{

  /* Autocorrelation lengths */
  fprintf(fpax,"Auto correlation length\n");
  fprintf(fpax,"\n");
  for (i=1;i<NA;i++) {
     fprintf(fpax,"%3d  ", i);
     for (m=first;m<last;m++) {
	if (EA[m][i] != 0.0)
           mass = EA[m][i-1]/EA[m][i];
	else
           mass = 0.0;
        if ((mass > 0) && (mass != 1.0)) 
	   mass = 1./log(mass);
        else 
	   mass = 0.;
	if (mass < 0.0) mass = 0.0;
        fprintf(fpax,"%9.5f  ", mass);
     }
     fprintf(fpax,"\n");
  }
  fprintf(fpax,"\n");
}


proc_average()
{
/* Work out averages and variances (specific heats) and their errors 
   by averaging over data on different processors, which should be 
   completely uncorrelated.
*/
   double err_ave, err_var, sigmasq;
   /* These are just a dummies */
   int type1=37, type2=38, type3=39, type4=40;  
   int type5=41, type6=42, type7=43, type8=44;  
   int fsum(), fmin(), fmax();

   if (nprocs > 1) {
        printf("Averaging over processors.... "); fflush(stdout);
        
	for(i=0; i<NOBS; i++) {
	   avspheat[i] = spheat[i];
	   avenersq[i]   = avenergy[i] * avenergy[i];
	   avspheatsq[i] = avspheat[i] * avspheat[i];
	}
	for(i=0; i<NOBS; i++) {
	for(n=0; n<NA; n++) {
	   EA2[i][n] = EA[i][n] * EA[i][n];
	}}

	/* Combine the data over processors */
	excombine(avenergy,   fsum, sizeof(double), NOBS, 
		  ALLNODES, NULLPTR, &type1);
	excombine(avenersq,   fsum, sizeof(double), NOBS,
		  ALLNODES, NULLPTR, &type2);
	excombine(avspheat,   fsum, sizeof(double), NOBS,
		  ALLNODES, NULLPTR, &type3);
	excombine(avspheatsq, fsum, sizeof(double), NOBS,
		  ALLNODES, NULLPTR, &type4);
	excombine(minenergy,  fmin, sizeof(float), NOBS,
		  ALLNODES, NULLPTR, &type5);
	excombine(maxenergy,  fmax, sizeof(float), NOBS,
		  ALLNODES, NULLPTR, &type6);
	excombine(EA,  fsum, sizeof(double), MAXNA*NOBS, 
		  ALLNODES, NULLPTR, &type7);
	excombine(EA2, fsum, sizeof(double), MAXNA*NOBS, 
		  ALLNODES, NULLPTR, &type8);

        /* Work out averages and errors */
	for(i=0; i<NOBS; i++) {
	   avenergy[i]   /= (double)nprocs;
	   avenersq[i]   /= (double)nprocs;
	   avspheat[i]   /= (double)nprocs;
	   avspheatsq[i] /= (double)nprocs;
	   sigmasq = avenersq[i] - (avenergy[i] * avenergy[i]);
	   spheat[i] = sigmasq;
	   /* Remember that gaussian must be scaled by (nodes-1) */
	   if (i == 0) spheat[i] *= (double)(sites-1)/(double)sites;
	   /* Correction because sum of squares != square of sum */
	   spheat[i] = avspheat[i] + spheat[i];
	   errener[i] = sqrt(sigmasq/(double)nprocs);
	   sigmasq = avspheatsq[i] - (avspheat[i] * avspheat[i]);
	   errspheat[i] = sqrt(sigmasq/(double)nprocs);
	}
        /* Work out averages and errors of autocorrs */
	for(i=0; i<NOBS; i++) {
	for(n=0; n<NA; n++) {
	   EA[i][n]  /= (double)nprocs;
	   EA2[i][n] /= (double)nprocs;
	   sigmasq = EA2[i][n] - (EA[i][n] * EA[i][n]);
	   errEA[i][n] = sqrt(sigmasq/(double)nprocs);
	}}
        printf("Done.\n"); fflush(stdout);
   }
   else {
      printf("ERROR! nprocs = %d!\n", nprocs);
      fflush(stdout);
   }
}


fsum(a1, a2, size)
double *a1, *a2; int size; { *a1 = *a2 = *a1 + *a2; return(0); }

fmax(a1, a2, size)
float *a1, *a2; int size; { if(*a1 < *a2) *a1 = *a2; return(0); }

fmin(a1, a2, size)
float *a1, *a2; int size; { if(*a1 > *a2) *a1 = *a2; return(0); }


