#include "dime.h"
int istage, totclus;

struct user_bdy{ double unused; };
struct user_elmt {
	int newproc;			/* My putative new processor */
	ELMTPTR eneigh[3];
	double choose_prob;		/* Relative chance of picking me as a seed */
	int in_cluster;			/* Flag for the seed fill algorithm */
	int yoursub[3];			/* Subscript in my neighbors LBE for me */
};
struct user_node { double unused; };

user_main(user_menu, argc, argv)
MNUPTR user_menu;
int argc;
char **argv;
{
	int Anneal(), Lbsetup();
	DIME_PREAMBLE;
	menu_add_function(user_menu, "Balance_Anneal", Anneal);
	menu_add_function(user_menu, "Anneal_Params", Lbsetup);
}

/*
	RAN should be a random number in [0.0, 1.0)
*/
#define RAN	(rand()/2.148e9)

#define LARGE	1.e12
#define NOTHING	-1

/*
	Center of an element for graphics
*/
#define XE(e) (e->neigh[0]->x+e->neigh[1]->x+e->neigh[2]->x)/3
#define YE(e) (e->neigh[0]->y+e->neigh[1]->y+e->neigh[2]->y)/3

/*
	Use temp space in element to point to my LBE
*/

#define MAXPROC		256
int nelmts;					/* Global number of elements */
int globproc[MAXPROC];		/* This processors view of the balance situation */
int cluster_bdy[MAXPROC];	/* Histogram of processor numbers at the boundary
								of a cluster */

double bdycoeff;			/* Energy = loadcoeff*sum of squares */
double loadcoeff;			/*	+ Mu * bdycoeff * sum of metric */
double elmts_per_proc;		/* Number of elements / Number of processors */
double Mu = 0.15;			/* Importance of communication */
double Temperature = 0.25;	/* Starting temp for Anneal */
double Cluster_prob = 0.58;	/* Determines cluster sizes */
int Stages = 20;			/* Number of temp steps / communication steps */
int Success_per_stage = 1;	/* How many accepted moves before communicating */
int NNprocs = -1;			/* Number of processors */
int maxprocs;
ELMTPTR seed;

Lbsetup()
{
	ELMTPTR elmt;
	double min = 20.;
	FORALLELMTS(elmt)
		if(fabs(elmt->neigh[0]->x-.5) + fabs(elmt->neigh[0]->y-.5) < min){
			min = fabs(elmt->neigh[0]->x-.5) + fabs(elmt->neigh[0]->y-.5);
			seed = elmt;
		}
	NEXTELMT(elmt)

	printf("Cluster Prob (currently %f)\n", Cluster_prob);
	get_double(&Cluster_prob);

	printf("Stages (currently %d)\n", Stages);
	get_int(&Stages);
}

Anneal()
{
	int new, isweep, success, tries, must_do;
	int bdy, min, max;
	ELMTPTR elmt;
	double temp, de, E, running_energy, Delta_cluster(), Energy();

	NNprocs = 1;
	Makelb();
	FORALLELMTS(elmt)
		elmt->user->in_cluster = -1;
	NEXTELMT(elmt)
	/*for(Cluster_prob = 0.; Cluster_prob < 0.8; Cluster_prob += 0.02){*/
	Cluster_prob = 0.655;
	totclus = 0;
	for(istage=0; istage<Stages; istage++){
		Delta_cluster(&new, &must_do);
		if(istage % 20 == 19)
		printf("%f %f\n", Cluster_prob, (double)totclus/(istage+1));
	}
}

double Delta_cluster(to_proc, must_do)
/*
	Make a cluster, with a suggested destination processor. Some
	moves evade the Metropolis by having "must_do" set. If a processor
	owns no elements there is an energy barrier to overcome to give it some
*/
int *to_proc, *must_do;
{
	ELMTPTR elmt, seed, eneigh;
	int from_proc, iproc, i, ip;
	int n_cluster, at_bdy, bdy, deltabdy, deltaload;
	double sum_choose_prob = 0;

	Cluster_seedfill(seed);

	FORALLELMTS(elmt)
		if(elmt->user->in_cluster == istage)
			totclus++;
	NEXTELMT(elmt)
}

metric(p, q)
int p, q;
{
	return (p == q ? 0:1);
}

Cluster_seedfill(elmt)
ELMTPTR elmt;
{
	ELMTPTR eneigh;
	int i;
	if(elmt->user->in_cluster == istage)
		return;
	elmt->user->in_cluster = istage;
	for(i=0; i<3; i++){
		if(!(eneigh = elmt->user->eneigh[i]))
			continue;
		if(RAN < Cluster_prob)
			Cluster_seedfill(eneigh);
	}
}

Makelb()
{
	ELMTPTR elmt, elmtn;
	NODEPTR node;
	NODENEIGHPTR n, nstart;
	int i, in;

	nelmts = 0;
	FORALLELMTS(elmt)
		nelmts++;
		elmt->user->newproc = 0;
		for(i=0; i<3; i++)
			elmt->user->eneigh[i] = (ELMTPTR)0;
	NEXTELMT(elmt)
/*
	Run round all the edges, setting the appropriate neighbor reference
	if its local, or else send a message
*/
	FORALLNODES(node)
		n = nstart = node->n;
		do {
			elmt = n->elmt;
			i = (indexc(elmt, node) + 1)%3;

			if(!(node->b && n == nstart)){
				elmtn = n->next->elmt;
				in = (indexc(elmtn, node) + 2)%3;
				elmt->user->eneigh[i] = elmtn;
				elmt->user->yoursub[i] = in;
			}
		} while((n = n->next) != nstart);
	NEXTNODE(node)

/*
	Get useful coefficients
*/
	elmts_per_proc = (double)nelmts/NNprocs;
	loadcoeff = 1./(elmts_per_proc*sqrt(elmts_per_proc));
	bdycoeff = 1;

	if(debug){
		disp_erase();
		FORALLELMTS(elmt)
			Showlbe(elmt);
		NEXTELMT(elmt);
	}
}

double Energy(bdy, min, max)
int *bdy, *min, *max;
{
	ELMTPTR elmt, eneigh;
	int i, ip, iproc, load = 0;

	*bdy = 0;
	for(iproc=0; iproc<maxprocs; iproc++)
		globproc[iproc] = 0;
/*
	Get global histogram and boundary. Note each edge is counted twice
*/
	FORALLELMTS(elmt)
		globproc[elmt->user->newproc]++;
		for(i=0; i<3; i++){
			if(eneigh = elmt->user->eneigh[i]){
				ip = eneigh->user->newproc;
				if(ip != elmt->user->newproc )
					*bdy += metric(elmt->user->newproc, ip);
			}
		}
	NEXTELMT(elmt)

/*
	Min and Max occupancy and sum of squares
*/
	*max = 0;
	*min = 1<<30;
	for(iproc=0; iproc<NNprocs; iproc++){
		if(globproc[iproc] > *max)
			*max = globproc[iproc];
		if(globproc[iproc] < *min)
			*min = globproc[iproc];
		load += globproc[iproc]*globproc[iproc];
	}

	return(load*loadcoeff + *bdy*Mu*bdycoeff);
}

Changeproc(elmt, new)
/*
	Change the color of this element and inform the neighbors
*/
ELMTPTR elmt;
int new;
{
	int i, old;
	double dbias;

	if(elmt->user->newproc == new)
		return;

	old = elmt->user->newproc;
/*
	Update (my view of) the global histogram
*/
	globproc[old]--;
	elmt->user->newproc = new;
	globproc[new]++;

	if(debug)
		Showlbe(elmt);
}

Showlbe(elmt)
ELMTPTR elmt;
{
	int i, p;
	double xe, ye, x[3], y[3];
	xe = XE(elmt);
	ye = YE(elmt);
	for(i=0; i<3; i++){
		x[i] = elmt->neigh[i]->x;
		y[i] = elmt->neigh[i]->y;
	}

	initpanel(PROC_COLOR(elmt->user->newproc), 0);
	for(i=0; i<3; i++)
		panelpoint((xe+x[i])/2, (ye+y[i])/2);
	endpanel();
}
