#include "bdy.h"
#include <math.h>

struct segment *segment_alloc();
struct segment *currseg = (struct segment *)NULL;
struct segment *startseg = (struct segment *)NULL;

struct curve *curve_alloc(), *findcurvename();
struct curve *currcurve = (struct curve *)NULL;
struct curve *the_bdy = (struct curve *)NULL;

struct point *point_alloc(), *findptname();
struct point *currpt = (struct point *)NULL;
struct point *the_pts = (struct point *)NULL;

double currx, curry;

make_moveto(name)
char *name;
{
	struct point *point = findptname(name);
	currx = point->x;
	curry = point->y;
}

moveto(point)
struct point *point;
{
	currx = point->x;
	curry = point->y;
}

struct point *
make_point(name, x, y)
char *name;
double x, y;
{
	struct point *new = point_alloc();

	strcpy(new->name, name);
	new->x = currx = x;	
	new->y = curry = y;	
	return(new);
}

make_alias(name1, name2)
char *name1, *name2;
{
	struct curve *c1 = findcurvename(name1);
	struct curve *c2 = findcurvename(name2);
	c1->alias = c2;
	c2->alias = c1;
}

struct segment *
make_line(x1, y1)
double x1, y1;
{
	struct segment *new = segment_alloc();
	double dist();

	new->type = LINE;
	new->segdata.line.x0 = currx;	
	new->segdata.line.y0 = curry;
	new->segdata.line.x1 = x1;	
	new->segdata.line.y1 = y1;
	new->length = dist(currx, curry, x1, y1);
	currx = x1;
	curry = y1;
	return(new);
}

double dist(x0, y0, x1, y1)
double x0, y0, x1, y1;
{
	double dx = x1-x0, dy = y1-y0;
	return(sqrt(dx*dx + dy*dy));
}

struct segment *
make_arc(x, y, r, th0, th1)
double x, y, r, th0, th1;
{
	struct segment *new = segment_alloc();

	new->type = ARC;
	new->segdata.arc.x = x;	
	new->segdata.arc.y = y;	
	new->segdata.arc.r = r;	
	while(th0 >= 360.0) th0 -= 360;
	while(th0 <    0.0) th0 += 360;
	while(th1 >= th0+360)th1 -= 360;
	while(th1 < th0)     th1 += 360;
	th0 *= PI/180;
	th1 *= PI/180;
	new->segdata.arc.th0 = th0;
	new->segdata.arc.th1 = th1;	
	new->length = r*fabs(th1 - th0);
	currx = x + r*cos(th1);
	curry = y + r*sin(th1);
	return(new);
}

struct segment *
make_arcn(x, y, r, th0, th1)
double x, y, r, th0, th1;
{
	struct segment *new = segment_alloc();

	new->type = ARC;
	new->segdata.arc.x = x;	
	new->segdata.arc.y = y;	
	new->segdata.arc.r = r;	
	while(th0 >= 360.0) th0 -= 360;
	while(th0 <    0.0) th0 += 360;
	while(th1 >= th0) 	th1 -= 360;
	while(th1 < th0-360)th1 += 360;
	th0 *= PI/180;
	th1 *= PI/180;
	new->segdata.arc.th0 = th0;
	new->segdata.arc.th1 = th1;	
	new->length = r*fabs(th1 - th0);
	currx = x + r*cos(th1);
	curry = y + r*sin(th1);
	return(new);
}

struct segment *
make_curveto(x1, y1, x2, y2, x3, y3)
double x1, y1, x2, y2, x3, y3;
{
    struct segment *new = segment_alloc();
    double ax, ay, bx, by, cx, cy, dx, dy;
	double t, x, y, xold, yold, dist();
	int i;
	struct curveto *cu;

    new->type = CURVETO;
    dx = currx;             dy = curry;
    cx = 3*(x1-dx);         cy = 3*(y1-dy);
    bx = 3*(x2-x1)-cx;      by = 3*(y2-y1)-cy;
    ax = x3-dx-cx-bx;       ay = y3-dy-cy-by;

	cu = &(new->segdata.curveto);
    cu->ax = ax;
    cu->bx = bx;
    cu->cx = cx;
    cu->dx = dx;
    cu->ay = ay;
    cu->by = by;
    cu->cy = cy;
    cu->dy = dy;

	new->length = 0;
    for(i=0; i<=20; i++){
        t = (double)i/20;
        x = cu->dx + t*(cu->cx + t*(cu->bx + t*cu->ax));
        y = cu->dy + t*(cu->cy + t*(cu->by + t*cu->ay));
        if(i)
			new->length += dist(xold, yold, x, y);
		xold = x;
		yold = y;
    }

    currx = x3;
    curry = y3;
	return(new);
}

struct point *point_alloc()
{
	struct point *new;
	static int npts=0;

	new = (struct point *)malloc(sizeof(struct point));
	if(!new){
		printf("point_alloc: Too many points\n");
		exit(101);
	}
	new->next = (struct point *)NULL;
	new->ptnum = npts++;
	if(currpt)
		currpt->next = new;
	else
		the_pts = currpt = new;

	currpt = new;
	return(new);
}

struct segment *segment_alloc()
{
	struct segment *new;

	new = (struct segment *)malloc(sizeof(struct segment));
	if(!new){
		printf("segment_alloc: Too many segments\n");
		exit(101);
	}
	new->next = (struct segment *)NULL;
	if(currseg){
		new->sstart = currseg->sstart + currseg->length;
		currseg->next = new;
	} else {
		startseg = new;
		new->sstart = 0.0;
	}

	currseg = new;
	return(new);
}

make_curve(curvename, firstptname, lastptname)
char *curvename, *firstptname, *lastptname;
{
	struct curve *new;
	struct point *first = findptname(firstptname);
	struct point *last  = findptname(lastptname);
	static int ncurve = 0;

	if(!startseg)
		make_line(first->x, first->y, last->x, last->y);

	new = (struct curve *)malloc(sizeof(struct curve));
	if(!new){
		printf("closecurve: Too many curves\n");
		exit(102);
	}

	strcpy(new->name, curvename);
	new->curvenum = ncurve++;
	new->first = first;
	new->last = last;
	new->alias = (struct curve *)NULL;
	new->start = startseg;
	new->length = currseg->sstart + currseg->length;
	currseg = startseg = (struct segment *)NULL;
	new->next = (struct curve *)NULL;

	if(!the_bdy)
		the_bdy = currcurve = new;
	else
		currcurve->next = new;

	currcurve = new;
}

bdy_inventory()
{
	struct curve *curve;
	struct segment *segment;
	struct point *point;

	FORALLPOINTS(point)
		printf("Point %s at (%0.3f, %0.3f)\n", 
			point->name, point->x, point->y);
	NEXTPOINT(point)

	printf("Curve Inventory:\n");
	FORALLCURVES(curve)
		printf("Curve %s: From %s to %s, Length %0.3f\n", 
			curve->name, curve->first->name, curve->last->name, 
			curve->length);
		if(curve->alias)
			printf("Aliased to %s\n", curve->alias->name);
		FORALLSEGMENTS(curve, segment)
			printf("\t%s (%0.3f to %0.3f)\n",
				(segment->type == POINT  ? "POINT  " :
				(segment->type == LINE   ? "LINE   " :
				(segment->type == ARC    ? "ARC    " :
				                           "CURVETO"))),
				segment->sstart, segment->sstart + segment->length);
		NEXTSEGMENT(curve, segment)
	NEXTCURVE(curve)

	printf("Done\n");
}

struct curve *findcurvename(name)
char *name;
{
	struct curve *curve;
	FORALLCURVES(curve)
		if(!strcmp(curve->name, name))
			return(curve);
	NEXTCURVE(curve)
	printf("findcurvename: %s not found\n", name);
}

struct point *findptname(name)
char *name;
{
	struct point *point;
	FORALLPOINTS(point)
		if(!strcmp(point->name, name))
			return(point);
	NEXTPOINT(point)
	printf("findpointname: %s not found\n", name);
}
