#include "bdy.h"
#include <math.h>

struct segment *
get_point(curve, s, x, y)
struct curve *curve;
double s, *x, *y;
{
	struct segment *seg;
	struct arc *arc;
	struct curveto *cu;
	double ds, p, t;

	while(s < 0.0)           s += curve->length;
	while(s > curve->length) s -= curve->length;

	FORALLSEGMENTS(curve, seg)
		if(s >= seg->sstart && s <= seg->sstart + seg->length)
			break;
	NEXTSEGMENT(curve, seg)

	ds = s - seg->sstart;
	switch(seg->type){
		case LINE:
			p = ds/seg->length;
			*x = seg->segdata.line.x0*(1-p) + seg->segdata.line.x1*p;
			*y = seg->segdata.line.y0*(1-p) + seg->segdata.line.y1*p;
			return seg;
		case ARC:
			arc = &(seg->segdata.arc);
			ds /= arc->r;
			if(arc->th0 > arc->th1) ds = -ds;
			*x = arc->x + arc->r*cos(arc->th0 + ds);
			*y = arc->y + arc->r*sin(arc->th0 + ds);
			return seg;
		case CURVETO:
			cu = &(seg->segdata.curveto);
			t = ds/seg->length;
			*x = cu->dx + t*(cu->cx + t*(cu->bx + t*cu->ax));
			*y = cu->dy + t*(cu->cy + t*(cu->by + t*cu->ay));
			return seg;
	}
}

tangent(curve, s, x, y)
struct curve *curve;
double s, *x, *y;
{
	struct segment *seg;
	struct arc *arc;
	struct curveto *cu;
	double ds, p, t, norm;

	if(!curve){
		*x = *y = 0;
		return 0;
	}

	while(s < 0.0)           s += curve->length;
	while(s > curve->length) s -= curve->length;

	FORALLSEGMENTS(curve, seg)
		if(s >= seg->sstart && s <= seg->sstart + seg->length)
			break;
	NEXTSEGMENT(curve, seg)

	ds = s - seg->sstart;
	switch(seg->type){
		case LINE:
			p = ds/seg->length;
			*x = -seg->segdata.line.x0 + seg->segdata.line.x1;
			*y = -seg->segdata.line.y0 + seg->segdata.line.y1;
			break;
		case ARC:
			arc = &(seg->segdata.arc);
			ds /= arc->r;
			if(arc->th0 > arc->th1) ds = -ds;
			*x = -sin(arc->th0 + ds);
			*y =  cos(arc->th0 + ds);
			break;
		case CURVETO:
			cu = &(seg->segdata.curveto);
			t = ds/seg->length;
			*x = cu->cx + t*(2*cu->bx + 3*t*cu->ax);
			*y = cu->cy + t*(2*cu->by + 3*t*cu->ay);
			break;
	}
	norm = sqrt((*x)*(*x) + (*y)*(*y));
	*x /= norm;
	*y /= norm;
	return 1;
}

double closest_to_point(pointmin, x, y)
struct point **pointmin;
double x, y;
{
	struct point *point;
	double d, dmin = 1.e12;
	double dist();

	FORALLPOINTS(point)
		d = dist (x, y, point->x, point->y);
		if(d < dmin){
			dmin = d;
			*pointmin = point;
		}
	NEXTPOINT(point)
	return(dmin);
}

double closest_to_curve(curvemin, smin, x, y)
struct curve **curvemin;
double *smin, x, y;
{
	struct curve *curve;
	struct segment *seg;
	double s, d, dmin = 1.e12;
	double d_to_point(), d_to_line(), d_to_arc(), d_to_curveto();

	FORALLCURVES(curve)
		FORALLSEGMENTS(curve, seg)
			switch(seg->type){
			case LINE  : 
			d = d_to_line  (x, y, &s, &(seg->segdata)); break;
			case ARC   : 
			d = d_to_arc   (x, y, &s, &(seg->segdata)); break;
			case CURVETO: 
			d = d_to_curveto(x, y, &s, &(seg->segdata)); s *= seg->length; break;
			}
			if(d < dmin){
				*smin = s + seg->sstart;
				dmin = d;
				*curvemin = curve;
			}
		NEXTSEGMENT(curve, seg)
	NEXTCURVE(curve)
	return(dmin);
}

double d_to_line(x, y, s, line)
double x, y, *s;
struct line *line;
{
    double px, py, qx, qy, cos, lsq;
	double dist();

    qx = line->x1 - line->x0;
    qy = line->y1 - line->y0;
    px = x - line->x0;
    py = y - line->y0;
    lsq = qx*qx + qy*qy;
    cos = (px*qx + py*qy)/lsq;
    if(cos < 0.0){
		*s = 0.0;
        return(dist(x, y, line->x0, line->y0));
    }
    if(cos > 1.0){
		*s = sqrt(lsq);
        return(dist(x, y, line->x1, line->y1));
    }
	*s = cos*sqrt(lsq);
	return(dist(x, y, line->x0 + qx*cos, line->y0 + qy*cos));
}

double d_to_arc(x, y, s, arc)
double x, y, *s;
struct arc *arc;
{
	double th, d0, d1, dist();
	int inarc;

	th = atan2(y - arc->y, x - arc->x) - 4*PI;
	if(arc->th1 > arc->th0){
		while(th < arc->th0) th += 2*PI;
		inarc = (th < arc->th1);
	} else {
		while(th < arc->th1) th += 2*PI;
		inarc = (th < arc->th0);
	}
	if(inarc){
		*s = arc->r*fabs(arc->th0 - th);
		return(fabs(dist(x, y, arc->x, arc->y) - arc->r));
	} else {
		d0 = dist(x, y,
		arc->x + arc->r*cos(arc->th0), arc->y + arc->r*sin(arc->th0));
		d1 = dist(x, y,
		arc->x + arc->r*cos(arc->th1), arc->y + arc->r*sin(arc->th1));
		if(d0 < d1){
			*s = 0.0;
			return(d0);
		} else {
			*s = arc->r*fabs(arc->th0 - arc->th1);
			return(d1);
		}
	}
}

double d_to_curveto(x, y, s, cu)
double x, y, *s;
struct curveto *cu;
{
#define NSEG	50
    int i;
    double dist(), dmin = 1.e12, t, d, xc, yc, tmin;
 
    for(i=0; i<=NSEG; i++){
        t = (double)i/NSEG;
        xc = cu->dx + t*(cu->cx + t*(cu->bx + t*cu->ax));
        yc = cu->dy + t*(cu->cy + t*(cu->by + t*cu->ay));
        d = dist(xc, yc, x, y);
        if(d < dmin){
			dmin = d;
			tmin = t;
		}
    }
	*s = tmin; 
    return(dmin);
}

double arcdist(curve, sa, sc)
struct curve *curve;
double sa, sc;
{
	double s, x, y, xold, yold, length = 0;
	double dist();
	int i;

    for(i=0; i<=20; i++){
        s = sa + i*(sc - sa)/20;
		get_point(curve, s, &x, &y);
        if(i)
            length += dist(xold, yold, x, y);
        xold = x;
        yold = y;
    }
	return length;
}

double curvature(curve, s)
struct curve *curve;
double s;
{
	struct segment *seg;
	struct arc *arc;
	struct curveto *cu;
	double ds, p, x, y, t, xt, yt, xtt, ytt, rt;

	while(s < 0.0)           s += curve->length;
	while(s > curve->length) s -= curve->length;

	FORALLSEGMENTS(curve, seg)
		if(s >= seg->sstart && s <= seg->sstart + seg->length)
			break;
	NEXTSEGMENT(curve, seg)

	ds = s - seg->sstart;
	switch(seg->type){
		case LINE:
			return(0.0);
		case ARC:
			return(1.0/seg->segdata.arc.r);
		case CURVETO:
			cu = &(seg->segdata.curveto);
			t = ds/seg->length;
			x = cu->dx + t*(cu->cx + t*(cu->bx + t*cu->ax));
			y = cu->dy + t*(cu->cy + t*(cu->by + t*cu->ay));
			xt = cu->cx + t*(2*cu->bx + 3*t*cu->ax); 
			yt = cu->cy + t*(2*cu->by + 3*t*cu->ay); 
			xtt = 2*cu->bx + 6*t*cu->ax;
			ytt = 2*cu->by + 6*t*cu->ay;
			rt = xt*xt + yt*yt;
			return((xt*ytt - yt*xtt) / (rt*sqrt(rt)));
	}
}
