#include <stdio.h>
#include "menu.h"
#include "curvetool.h"

MNUPTR root;
extern int graphics;
FILE *ps_file;
int plotting = 0;
extern int verbose;
int explicit = 0;

int grid = 0;
double gridspace = 0.1;

int snap = 0;
double snapspace = 0.1;

extern int menuport, pixport;
extern double menu_width;

main(argc, argv)
int argc;
char **argv;
{
	MNUPTR zoom_menu, script_menu, extras_menu;
	int zoom_extents(), zoom_window(), zoom_half(), zoom_exp();
	int addpoints(), disp_erase(), draw_segments(), _connect();
	int exit(), adjust(), drawclean(), split(), postscript();
	int toggle_explicit(), name_point(), name_curve(), write_bdy();
	int straighten(), grid_toggle(), snap_toggle();
	int menu_include(), menu_record(), menu_end_record();

	root = menu_make(ROOT);
	zoom_menu = menu_make(BRANCH);
		add_item(zoom_menu, "Extents", LEAF, zoom_extents);
		add_item(zoom_menu, "Window", LEAF, zoom_window);
		add_item(zoom_menu, "Half", LEAF, zoom_half);
		add_item(zoom_menu, "Explicit", LEAF, zoom_exp);
	add_item(root, "ZOOM", BRANCH, zoom_menu);
 
    script_menu = menu_make(BRANCH);
        add_item(script_menu, "Use_script", LEAF, menu_include);
        add_item(script_menu, "Make_script", LEAF, menu_record);
        add_item(script_menu, "End_script", LEAF, menu_end_record);
    add_item(root, "SCRIPT", BRANCH, script_menu);

	extras_menu = menu_make(BRANCH);
		add_item(extras_menu, "Grid", LEAF, grid_toggle);
		add_item(extras_menu, "Snap", LEAF, snap_toggle);
		add_item(extras_menu, "Name_curve", LEAF, name_curve);
		add_item(extras_menu, "Name_point", LEAF, name_point);
		add_item(extras_menu, "Explicit", LEAF, toggle_explicit);
    add_item(root, "EXTRAS", BRANCH, extras_menu);

	add_item(root, "Redraw", LEAF, draw_segments);
	add_item(root, "Addpoint", LEAF, addpoints);
	add_item(root, "Connect", LEAF, _connect);
	add_item(root, "Adjust", LEAF, adjust);
	add_item(root, "Split", LEAF, split);
	add_item(root, "Straighten", LEAF, straighten);
	add_item(root, "Drawcurves", LEAF, drawclean);
	add_item(root, "Postscript", LEAF, postscript);
	add_item(root, "Write_bdy", LEAF, write_bdy);
	add_item(root, "Erase", LEAF, disp_erase);
	add_item(root, "Quit", LEAF, exit);


	printf("**************************************************\n");
    printf("*           This is CURVETOOL version 1.0        *\n");
    printf("**************************************************\n");
    printf("*            Copyright Caltech 1990              *\n");
    printf("**************************************************\n");

	setplxopt("winlab", "Curvetool");
	init_grafix(0.2, 2, 32768);
	graphics = 2;

	if(argc > 1)
		menu_includefile(argv[1]);

	menu_input();
}

drawclean()
{
	verbose = 0;
	disp_erase();
	draw_segments();
	verbose = 1;
}

addpoints()
{
	double x, y;
	struct point *new, *point_alloc();

	get_a_point(&x, &y);
	new = point_alloc();
	new->x = x;
	new->y = y;
	draw_point(new);
}

_connect()
{
	struct point *p1, *p2, *pick_point();
	double x, y;

	printf("Pick start Point\n");
	get_gin(&x, &y);
	p1 = pick_point(x, y);

	get_gin(&x, &y);
	printf("Pick end Point\n");
	p2 = pick_point(x, y);
	connect_points(p1, p2);
}

connect_points(p1, p2)
struct point *p1, *p2;
{
	struct segment *seg, *segment_alloc();
	struct endseg *e1, *e2, *add_endseg();
	struct curve *curve, *curve_alloc();

	if(p1 == p2){
		printf("Cannot connect a Point to itself\n");
		return 0;
	}
	e1 = add_endseg(p1);
	e2 = add_endseg(p2);

	curve = curve_alloc();
	curve->start = e1;
	curve->pstart = p1;
	curve->pend = p2;

	seg = segment_alloc();
	seg->e1 = e1;		e1->segment = seg;
	seg->e2 = e2;		e2->segment = seg;
	straighten_segment(seg);
	color(FOREGD);
	draw_segment(seg);
	draw_point(p1);
	draw_point(p2);
}

#define NONE	0
#define LEFT	1
#define RIGHT	4

#define FIRST	1
#define MID		0
#define LAST	2

adjust()
{
	double x, y;
	int key, type, oldkey;
	char *ptr, *pick_pointish();
	double xmin, ymin, xmax, ymax;

	sleep(1);
	get_screen(&xmin, &ymin, &xmax, &ymax);

	while(1){
		setvport(menuport);
/*		box(0.0, 0.0, menu_width, 1.0, explicit?GREEN:BLUE, 1);*/
		setvport(pixport);

		oldkey = NONE;
		while(1){
			gtrack(&key, &x, &y);
			if(snap){
				x = snapspace*nint(x/snapspace);
				y = snapspace*nint(y/snapspace);
			}
			if(!key){
				if(oldkey == RIGHT)
					process_right(LAST, x, y);
				oldkey = NONE;
				continue;
			}
			if(x < xmin){
				printf("End of Adjust\n");
				disp_erase();
				draw_segments();
			/*	gin(&key, &x, &y);*/
				return 1;
			}

			if(key == LEFT)
				process_left(oldkey == NONE ? FIRST:MID, x, y);
			else if(key == RIGHT)
				process_right(oldkey == NONE ? FIRST:MID, x, y);

			oldkey = key;
		}
	}
}

struct point *oldpoint = NULL;

process_right(state, x, y)
int state;
double x, y;
{
	static struct point *p1, *p2;
	struct point *point, *point_alloc();
	static double oldx, oldy;
	double sqdist(), screensize();

	point = pick_point(x, y);
	if(point && sqrt(sqdist(x, y, point->x, point->y)) > 0.02*screensize())
		point = NULL;

	if(oldpoint && point != oldpoint){
		color(FOREGD);
		draw_point(oldpoint);
		oldpoint = NULL;
	} 

	if(point){
		color(GREEN);
		draw_point(point);
		oldpoint = point;
	} 

	switch(state){
		case FIRST:
			if(point)
				p1 = point;
			else {
				p1 = point_alloc();
				p1->x = oldx = x;
				p1->y = oldy = y;
			}
			break;
		case MID:
			color(BACKGD);
			move(p1->x, p1->y);
			cont(oldx, oldy);
			color(FOREGD);
			move(p1->x, p1->y);
			cont(oldx = x, oldy = y);
			break;
		case LAST:
			if(point)
				p2 = point;
			else {
				p2 = point_alloc();
				p2->x = oldx;
				p2->y = oldy;
			}
			connect_points(p1, p2);
			break;
	}
}

process_left(first, x, y)
int first;
double x, y;
{
	struct point *point;
	struct endseg *endseg;
	static int type;
	static char *ptr;

	if(first)
		ptr = pick_pointish(&type, x, y);

	if(!ptr)
		return;

	color(BACKGD);
	switch(type){
		case POINT:
			point = (struct point *)ptr;
			draw_point(point);
			adjust_point(point, x, y);
			break;
		case ENDSEG:
			endseg = (struct endseg *)ptr;
			draw_endseg(endseg);
			adjust_endseg(endseg, x, y);
			if(endseg->interior)
				adjust_endseg(endseg->loop, x, y);
			break;
		case KNOT:
			endseg = (struct endseg *)ptr; 
			draw_knot(endseg);
			adjust_knot(endseg, x, y);
			break;
		default:
			break;
	}
}

straighten()
{
	struct segment *seg, *pick_segment();
	double x, y, x1, y1, x2, y2;

	do {
		printf("Pick a segment\n");
		get_gin(&x, &y);
	} while(!(seg = pick_segment(x, y)));

	color(BACKGD);
	draw_segment(seg);

	x1 = seg->e1->x;		y1 = seg->e1->y;
	x2 = seg->e2->x;		y2 = seg->e2->y;

	adjust_knot(seg->e1, (2*x1+x2)/3, (2*y1+y2)/3);
	adjust_knot(seg->e2, (2*x2+x1)/3, (2*y2+y1)/3);

	color(FOREGD);
	draw_segment(seg);
}

split()
{
	struct segment *seg, *pick_segment();
	double x, y;

	do {
		printf("Pick a segment\n");
		get_gin(&x, &y);
	} while(!(seg = pick_segment(x, y)));
	splitseg(x, y, seg);
}

splitseg(x, y, seg)
double x, y;
struct segment *seg;
{
	struct segment *seg1, *seg2, *segment_alloc(), *pick_segment();
	struct endseg *endseg1, *endseg2, *endseg_alloc();
	double dx, dy;

	color(BACKGD);
	draw_segment(seg);

	endseg1 = endseg_alloc();
	endseg2 = endseg_alloc();

	endseg1->interior = endseg2->interior = 1;
	get_middle(seg, &x, &y);
	endseg1->x = endseg2->x = x;
	endseg1->y = endseg2->y = y;

	dx = (seg->e2->x - seg->e1->x)/6;
	dy = (seg->e2->y - seg->e1->y)/6;
	endseg1->kx = x - dx;
	endseg1->ky = y - dy;
	endseg2->kx = x + dx;
	endseg2->ky = y + dy;

	seg2 = segment_alloc();
	seg->e2->segment = seg2;
	seg2->e2 = seg->e2;
	seg1 = seg;

	endseg1->segment = seg1;
	endseg2->segment = seg2;

	endseg1->loop = endseg2;
	endseg2->loop = endseg1;

	seg1->e2 = endseg1;
	seg2->e1 = endseg2;

	setup_segment(seg1);
	setup_segment(seg2);

	color(FOREGD);
	draw_segment(seg1);
	draw_segment(seg2);
}

get_a_point(x, y)
double *x, *y;
{
	if(explicit){
		printf("x, y please\n");
		get_double(x);
		get_double(y);
	} else {
		get_gin(x, y);
		if(!snap)
			return 1;
		*x = snapspace*nint(*x/snapspace);
		*y = snapspace*nint(*y/snapspace);
	}
}

toggle_explicit()
{
	explicit = 1-explicit;
	if(explicit)
		printf("Explicit Mode\n");
	else
		printf("Mouse mode\n");
}

snap_toggle()
{
	snap = 1-snap;
	if(snap){
		printf("Snap spacing?\n");
		get_double(&snapspace);
		if(snapspace < SMALL)
			snap = 0;
	}
	if(!snap)
		printf("Snap Off\n");
}

grid_toggle()
{
	grid = 1-grid;
	if(grid){
		printf("Grid spacing?\n");
		get_double(&gridspace);
		if(gridspace < SMALL)
			grid = 0;
	}
	if(!grid)
		printf("Grid Off\n");
}
