#include "curvetool.h"
#include <math.h>

int nres = 20;
int verbose = 1;

extern int grid;
extern double gridspace;
extern int snap;
extern double snapspace;

get_extents(xmin, ymin, xmax, ymax)
double *xmin, *ymin, *xmax, *ymax;
{
	struct point *point;
	struct segment *segment;
	struct endseg *endseg;
	double x, y, w;

	*xmin = *ymin = LARGE;
	*xmax = *ymax = -LARGE;

	FORALLPOINTS(point)
		x = point->x;
		y = point->y;
		if(x < *xmin) *xmin = x;
		if(x > *xmax) *xmax = x;
		if(y < *ymin) *ymin = y;
		if(y > *ymax) *ymax = y;
	NEXTPOINT(point)

	FORALLSEGMENTS(segment)
		if(segment->xmin < *xmin) *xmin = segment->xmin;
		if(segment->xmax > *xmax) *xmax = segment->xmax;
		if(segment->ymin < *ymin) *ymin = segment->ymin;
		if(segment->ymax > *ymax) *ymax = segment->ymax;
	NEXTSEGMENT(segment)

	FORALLENDSEGS(endseg)
		if(endseg->kx < *xmin) *xmin = endseg->kx;
		if(endseg->kx > *xmax) *xmax = endseg->kx;
		if(endseg->ky < *ymin) *ymin = endseg->ky;
		if(endseg->ky > *ymax) *ymax = endseg->ky;
	NEXTENDSEG(endseg)

	w = 0.03*(*xmax - *xmin);
	*xmax += w;
	*xmin -= w;
	w = 0.03*(*ymax - *ymin);
	*ymax += w;
	*ymin -= w;

	return 1;
}

draw_segments()
{
	struct segment *seg;
	struct point *point;
	struct endseg *endseg;

	if(grid)
		make_grid();
	color(FOREGD);

	FORALLSEGMENTS(seg)
		draw_segment(seg);
	NEXTSEGMENT(seg)

	if(verbose){
		FORALLPOINTS(point)
			draw_point(point);
		NEXTPOINT(point)
	}
}

draw_segment(seg)
struct segment *seg;
{
    int i;
	double t, x, y;

	if(verbose){
		draw_endseg(seg->e1);
		draw_endseg(seg->e2);
	}

    for(i=0; i<=nres; i++){
        t = (double)i/nres;
        x = seg->dx + t*(seg->cx + t*(seg->bx + t*seg->ax));
        y = seg->dy + t*(seg->cy + t*(seg->by + t*seg->ay));
        if(i)
            cont(x, y);
        else
            move(x, y);
    }
}

draw_points()
{
	struct point *point;
	FORALLPOINTS(point)
		draw_point(point);
	NEXTPOINT(point)
}

draw_point(point)
struct point *point;
{
	double screensize();
	marker(POINT_MARK, point->x, point->y, 0.02*screensize());
}

draw_endseg(endseg)
struct endseg *endseg;
{
	double screensize();
	marker(ENDSEG_MARK, endseg->x, endseg->y, 0.01*screensize());
	draw_knot(endseg);
	if(endseg->interior)
		draw_knot(endseg->loop);
}

draw_knot(endseg)
struct endseg *endseg;
{
	move(endseg->x, endseg->y);
	cont(endseg->kx, endseg->ky);
}

make_grid()
{
	double xmin, ymin, xmax, ymax;
	int i, ixmin, iymin, ixmax, iymax;

	get_screen(&xmin, &ymin, &xmax, &ymax);
	ixmin = xmin/gridspace -1;
	ixmax = xmax/gridspace +1;
	iymin = ymin/gridspace -1;
	iymax = ymax/gridspace +1;

	color(YELLOW);
	for(i=ixmin; i<= ixmax; i++){
		move(i*gridspace, ymin);
		cont(i*gridspace, ymax);
	}
	for(i=iymin; i<= iymax; i++){
		move(xmin, i*gridspace);
		cont(xmax, i*gridspace);
	}
}
