/****************************************************/
/*    DIME: dime_util.c                             */
/*    These are functions used by many              */
/*    different DIME utilities.                     */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/
#define DIMECODE
#include "dime.h"
#ifndef CUBIX
#define stdsys stdout
#endif

NODEPTR bdystart;
int combine_is_ok = 0;
double lasttime, commtime, worktime;

relax()
{
#define XX			work_space.nf[1]
#define YY			work_space.nf[2]
#define NN			work_space.nf[3]
#define SCLOCK 		work_space.bf[0]
#define SANTIK		work_space.bf[1]
    NODEPTR node, onode, nodec;
    ELMTPTR elmt;
    int i, nodeoffset, bdyoffset;
	struct curve *curve;
	double s, sc;

	{NODEPTR dummy;
	nodeoffset = (char *)(&dummy->XX) - (char *)(dummy);
	}
	{BDYPTR dummy;
	bdyoffset = (char *)(&dummy->SCLOCK) - (char *)(dummy);
	}

    FORALLNODES(node)
        node->NN = node->XX = node->YY = 0.0;
        if(B_NODE(node))
            node->b->SCLOCK = node->b->SANTIK = 0.0;
    NEXTNODE(node)
 
    FORALLNODES(node)
        if(B_EDGE(node) && node->b->type != INSIDE){
			nodec = node->b->clock;
			curve = NULL;
			if(node->b->type == CURVE){
				curve = node->b->bdy.curve;
				s = node->b->s;
			} 
			if(nodec->b->type == CURVE){
				curve = nodec->b->bdy.curve;
				sc = nodec->b->s;
			}
			if(!curve){
				marker(5, node->x, node->y, 0.05);
				sendplot();
				printf("Relax: Internal Error\n");
				sleep(300);
				abort(907);
			}
			if(node->b->type == POINT){
				if(curve->first == node->b->bdy.point)
					s = 0;
				if(curve->last == node->b->bdy.point)
					s = curve->length;
			}
			if(nodec->b->type == POINT){
				if(curve->first == nodec->b->bdy.point)
					sc = 0;
				if(curve->last == nodec->b->bdy.point)
					sc = curve->length;
			}
			nodec->b->SANTIK = s;
            node->b->SCLOCK = sc;
        }
    
        FORALLNEIGH(node, elmt)
            for(i=0; i<3; i++){
                onode = elmt->neigh[i];
                if(onode == node)
                    continue;
                node->XX += onode->x;
                node->YY += onode->y;
                node->NN += 1.0;
            }
        NEXTNEIGH(node, elmt)
    NEXTNODE(node)
 
	node_combine(nodeoffset, sizeof(double), 3, fsum, 1, 0, (int *)0);
	bdy_combine(bdyoffset, sizeof(double), 2, fsum, 1);
 
    FORALLNODES(node)
        if(B_NODE(node)){
			if(node->b->type == CURVE){
				node->b->s = 
				0.25*(node->b->SCLOCK + 2*node->b->s + node->b->SANTIK);
				get_point(node->b->bdy.curve, node->b->s, 
					&node->x, &node->y);
			}  
		} else {
            node->x = node->XX/node->NN;
            node->y = node->YY/node->NN;
        }
    NEXTNODE(node)
}

int stattype;

setup_combine()
{
	NODEPTR node, noden;
	ALIASPTR a;
	int dest, olddest, count;

	bdystart = NULL;
	FORALLNODES(node)
		node->bdynext = NULL;
	NEXTNODE(node)

	FORALLNODES(node)
		if(node->b != (BDYPTR)NULL && node->bdynext == (NODEPTR)NULL){
			noden = node;
			do {
				noden->bdynext = noden->b->clock;
			} while((noden = noden->bdynext) != node);

			if(bdystart != (NODEPTR)NULL){
				noden = node->bdynext;
				node->bdynext = bdystart->bdynext;
				bdystart->bdynext = noden;
			} else
				bdystart = node;
		}
	NEXTNODE(node)

	combine_is_ok = 1;
	stattype = 1;
}

#define FORALLALIAS(node, a) \
	{ALIASPTR astart;					\
	if((node = bdystart) != (NODEPTR)NULL){ do {	\
	if(node->b->alias != (ALIASPTR)NULL){			\
	a = astart = node->b->alias; do { 

#define NEXTALIAS(node, a) \
	} while((a = a->next) != astart);}				\
	} while((node = node->bdynext) != bdystart); }}

#ifdef TIMING
double floptime()
{
#define NFLOP	2000
	double a = 1., b = 1., c= 1.;
	double time;
	int i;

	time = extime();
	for(i=0; i<NFLOP; i++)
		a = b*c;
	time = extime() - time;
	GLOBAL_COMBINE(time, fsum, 1);
	return(time/(nprocs*NFLOP));
}

start_timing()
{
	commtime = worktime = 0.0;
	lasttime = extime();
}

end_timing(comm, work)
double *comm, *work;
{
	double nowtime;

	nowtime = extime();
	worktime += nowtime - lasttime;
	lasttime = nowtime;

	GLOBAL_COMBINE(commtime, fsum, 1);
	GLOBAL_COMBINE(worktime, fsum, 1);
	commtime /= nprocs;
	worktime /= nprocs;
	*comm = commtime;
	*work = worktime;
}
#endif

make_hist(hist, nhist, histincrem)
int *hist, *histincrem, nhist;
{
	NODEPTR node;
	ALIASPTR a, astart;
	int maxcount, count, hi;
	int i, proc, nextproc, dest;

	for(i=0; i<nhist; i++)
		hist[i] = 0;

	maxcount = 0;

	nextproc = 32767;
	do {
		count = 0;
		proc = nextproc;
		nextproc = -1;
		FORALLALIAS(node, a)
			dest = (a->proc) & LOWBITS;
			if(dest == proc)
				count++;
			if(dest < proc && dest > nextproc)
				nextproc = dest;
		NEXTALIAS(node, a)

		if(count > maxcount)
			maxcount = count;
	} while(nextproc != -1);

	GLOBAL_COMBINE(maxcount, imax, 1);

	hi = maxcount/(nhist-1) + 1;
	if(hi == 0)
		hi = 1;
	*histincrem = hi;

	nextproc = 32767;
	do {
		count = 0;
		proc = nextproc;
		nextproc = -1;
		FORALLALIAS(node, a)
			dest = (a->proc) & LOWBITS;
			if(dest == proc)
				count++;
			if(dest < proc && dest > nextproc)
				nextproc = dest;
		NEXTALIAS(node, a)

		if(count)
			hist[count/hi]++;
	} while(nextproc != -1);

	GLOBAL_COMBINE(hist[0], isum, nhist);
}

node_combine(offset, size, number, func, for_dime, Nnodes, nodelist)
int offset, size, number, for_dime;
int (*func)();
int Nnodes, *nodelist;
{
	NODEPTR node;
	ALIASPTR a;
	char *mailbag;
	int nbytes, i, offoff, totbytes = size*number, dest;
	int proc, nextproc, olddest;
	int messsize = totbytes + sizeof(NODEPTR);
	int nmess = 0, src, type, messnow, imess;
	double nowtime;

	if(!combine_is_ok)
		setup_combine();
#ifdef TIMING
	nowtime = extime();
	worktime += nowtime - lasttime;
	lasttime = nowtime;
#endif

	type = stattype;

	nextproc = 32767;
	do {
		proc = nextproc;
		nextproc = -1;
		mailbag = buffer;
		FORALLALIAS(node, a)
			dest = (a->proc) & LOWBITS;
			if(dest == proc){
				nmess++;
				xpacket(mailbag, NODEPTR) = a->node;
				if(for_dime)
					copybuf((char *)node + offset, mailbag, totbytes);
				else
					copybuf(node->user + offset, mailbag, totbytes);
				mailbag += totbytes;
			}
			if(dest < proc && dest > nextproc && inlist(Nnodes, nodelist, dest))
				nextproc = dest;
		NEXTALIAS(node, a)

		if(mailbag > buffer)
			exwrite(buffer, (int)(mailbag-buffer), &proc, &type);

	} while(nextproc != -1);

	while(nmess){
		src = DONTCARE;
		type = stattype;
		nbytes = exread(buffer, COM_BUF, &src, &type);
		messnow = nbytes / messsize;
		for(imess=0; imess<messnow; imess++){
			mailbag = buffer + imess*messsize;
			node = xpacket(mailbag, NODEPTR);
			for(i=0; i<number; i++){
				offoff = i*size;
				if(for_dime)
				(*func)((char *)node + offset + offoff, mailbag + offoff, 1);
				else
				(*func)(node->user + offset + offoff, mailbag + offoff, 1);
			}
			nmess--;
		}
	}
	stattype = (stattype+1)%256;
#ifdef TIMING
	nowtime = extime();
	commtime += nowtime - lasttime;
	lasttime = nowtime;
#endif
}

inlist(Nnodes, nodelist, dest)
int Nnodes, *nodelist, dest;
{
	int i;
	if(nodelist == (int *)0)
		return 1;
	for(i=0; i<Nnodes; i++)
		if(*(nodelist+i) == dest)
			return 1;
	return 0;
}

bdy_combine(offset, size, number, func, for_dime)
int offset, size, number, for_dime;
int (*func)();
{
	NODEPTR node;
	ALIASPTR a, astart;
	char *mailbag;
	int source, nbytes, i, offoff, totbytes = size*number, dest;
	int type, nmess = 0, src;
	double nowtime;
#ifdef TIMING
	nowtime = extime();
	worktime += nowtime - lasttime;
	lasttime = nowtime;
#endif
	FORALLNODES(node)
		if(B_NODE(node) && (node->b->alias != (ALIASPTR)NULL)){ 
			a = astart = node->b->alias;
			do { 
				mailbag = buffer;
				xpacket(mailbag, NODEPTR) = a->node;
				if(for_dime)
					copybuf((char *)node->b + offset, mailbag, totbytes);
				else
					copybuf(node->b->user + offset, mailbag, totbytes);
				mailbag += totbytes;
				dest = (a->proc) & LOWBITS;
				type = stattype;
				exwrite(buffer, (int)(mailbag-buffer), &dest, &type);
				nmess++;
			} while((a = a->next) != astart);
		}
	NEXTNODE(node)

	while(nmess--){
		src = DONTCARE;
		type = stattype;
		exread(buffer, COM_BUF, &src, &type);
		mailbag = buffer;
		node = xpacket(mailbag, NODEPTR);
		for(i=0; i<number; i++){
			offoff = i*size;
			if(for_dime)
			(*func)((char *)node->b + offset + offoff, mailbag + offoff, 1);
			else
			(*func)(node->b->user + offset + offoff, mailbag + offoff, 1);
		}
	}
	stattype = (stattype+1)%256;
#ifdef TIMING
	nowtime = extime();
	commtime += nowtime - lasttime;
	lasttime = nowtime;
#endif
}


NODENEIGHPTR indexnn(node, elmt)
/*
	Finds the nodeneigh around 'node' which points to 'elmt'
*/
ELMTPTR elmt;
NODEPTR node;
{
	NODENEIGHPTR n, nstart;
	n = nstart = node->n;
	do {
		if(n->elmt == elmt)
			return(n);
	} while((n = n->next) != nstart);
	abort(206);					/* indexnn failed */
}

buffers()
{
/*
	Prints out buffer usage, structures assigned, and memory usage.
*/
	int sums[6], maxs[6];
	NODEPTR node;
	ELMTPTR elmt;
#define nelmt	sums[3]
#define nnode	sums[4]
#define nbdy	sums[5]
	nnode = nelmt = nbdy = 0;
	FORALLNODES(node)
		nnode++;
		if(node->b != (BDYPTR)NULL)
			nbdy++;
	NEXTNODE(node)
	FORALLELMTS(elmt)
		nelmt++;
	NEXTELMT(elmt)
	maxs[0] = sums[0] = plothwm();
	maxs[1] = sums[1] = crystal_hwm();
	maxs[2] = sums[2] = s_inventory(0);
	maxs[3] = sums[3];

	GLOBAL_COMBINE(maxs[0], imax, 4);
	GLOBAL_COMBINE(sums[0], isum, 6);

	printf("Buffers:    Messages %d; Plot %d\n", maxs[1], maxs[0]);
	printf("Memory:     Total %d bytes; Maximum %d bytes\n", sums[2], maxs[2]);
	printf("Structures: Total Nodes %d; Total Elmts %d (Max %d); Total Bdys %d\n", 
						nnode, nelmt, maxs[3], nbdy);
#ifdef CUBIX
	printf("Available Memory:\n");
	fflush(stdout);
	fmulti(stdout);
	printf("Processor %d has %ld bytes\n", procnum, malloc_avail());
#endif
	fflush(stdout);
	fsingl(stdout);
}

/*
	Here are the standard combining functions declared in dime.h 
*/
fmax(a1, a2, size)
double *a1, *a2; int size; { if(*a1 < *a2) *a1 = *a2; return(0); }

imax(a1, a2, size)
int *a1, *a2, size; { if(*a1 < *a2) *a1 = *a2; return(0); }

fmin(a1, a2, size)
double *a1, *a2; int size; { if(*a1 > *a2) *a1 = *a2; return(0); }

imin(a1, a2, size)
int *a1, *a2, size; { if(*a1 > *a2) *a1 = *a2; return(0); }

fsum(a1, a2, size)
double *a1, *a2; int size; { *a1 = *a2 = *a1 + *a2; return(0); }

isum(a1, a2, size)
int *a1, *a2, size; { *a1 = *a2 = *a1 + *a2; return(0); }

or(a1, a2, size)
int *a1, *a2, size; { *a1 = *a2 = (*a1) | (*a2); return(0); }

add_neigh(elmt, this)
ELMTPTR elmt;
NODEPTR this;
/*
	Puts a nodeneigh into node pointing to elmt. The place is arbitrary
	so the node will need sorting eventually.
*/
{
	NODENEIGHPTR nodeneigh_alloc(), n;
	n = nodeneigh_alloc();
	n->elmt = elmt;
	if(this->n == (NODENEIGHPTR)NULL){
		n->next = n;
		this->n = n;
	} else {
		n->next = this->n->next;
		this->n->next = n;
	}
}

NODENEIGHPTR before(nn)
NODENEIGHPTR nn;
/*
	Returns the nodeneigh immediately BEFORE the argument.
*/
{
	NODENEIGHPTR n = nn;
	do {
		if(n->next == nn)
			return(n);
	} while((n = n->next) != nn);
}

indexc(elmt, node)
/*
	Returns the subscript of 'node' from 'elmt' -- that is
	we solve for 'in' the equation "elmt->neigh[in] = node".
*/
ELMTPTR elmt;
NODEPTR node;
{
	int in;
	NODEPTR n;
	for(in=0; in<3; in++)
		if(elmt->neigh[in] == node)
			return(in);
	return(-1);
}

double area(elmt)
ELMTPTR elmt;
/*
	Returns the area of the elmt.
*/
{
	NODEPTR n0, n1, n2;
	double ax, ay, bx, by;
	n0 = elmt->neigh[0];
	n1 = elmt->neigh[1];
	n2 = elmt->neigh[2];
	ax = n0->x - n1->x;			ay = n0->y - n1->y;
	bx = n1->x - n2->x;			by = n1->y - n2->y;
	return(0.5*(ax*by - ay*bx));
}

nodeinfo(node)
NODEPTR node;
/*
	For debugging to print stuff into stdsys before crash & burn.
*/
{
    int bb = (node->b != (BDYPTR)NULL);
    fprintf(stdsys, "%d::(%f %f) %c%c\n", procnum, node->x, node->y, 
						(bb && node->b->oppalias != (ALIASPTR)NULL)?'O' : 0,
                        (bb && node->b->alias != (ALIASPTR)NULL)?'P':' ');
}

insert(node, elmt, newneigh, newelmt)
/*
    In node 'node', insert the nodeneigh 'new' in position after
    the nodeneigh which points to the elmt 'elmt', and make it point
    to the elmt 'newelmt'.
*/
NODEPTR node;
ELMTPTR elmt, newelmt;
NODENEIGHPTR newneigh;
{
    NODENEIGHPTR nstart, n;
    n = nstart = node->n;
    do {
		if(n->elmt == elmt){
            newneigh->next = n->next;
            n->next = newneigh;
            newneigh->elmt = newelmt;
            return 1;
        }
    } while((n = n->next) != nstart);
    abort(207);						/* insert failed */
}

insert_before(node, elmt, newneigh, newelmt)
/*
    In node 'node', insert the nodeneigh 'new' in position before
    the nodeneigh which points to the elmt 'elmt', and make it point
    to the elmt 'newelmt'.
*/
NODEPTR node;
ELMTPTR elmt, newelmt;
NODENEIGHPTR newneigh;
{
    NODENEIGHPTR nstart, n;
    n = nstart = node->n;
    do {
		if(n->next->elmt == elmt){
            newneigh->next = n->next;
            n->next = newneigh;
            newneigh->elmt = newelmt;
            return 1;
        }
    } while((n = n->next) != nstart);
    abort(207);						/* insert failed */
}

NODENEIGHPTR remove(node, victim_elmt)
/*
    In node 'node', removes the nodeneigh which points to 'elmt', and
    returns a pointer to it for recycling or free()ing.
*/
NODEPTR node;
ELMTPTR victim_elmt;
{
    NODENEIGHPTR nstart, n, return_n;
    n = nstart = node->n;
    do {
        if(n->next->elmt == victim_elmt){
            if((return_n = n->next) == n)
				node->n = (NODENEIGHPTR)NULL;
			else {
				n->next = n->next->next;
				node->n = n;  	/* Mustn't remove the one the node points to */
			}
			return(return_n);
        }
    } while((n = n->next) != nstart);
    abort(208);							/* remove failed */
}

get_bdy_nn(node)
NODEPTR node;
/* 
	To make sure that if node is on the boundary, then node->n points
	to the elmt just before the actual boundary.
*/
{
	NODENEIGHPTR n, nstart;
	if(node->b == (BDYPTR)NULL)
		return 1;
	n = nstart = node->n;
	do {
		if(!antik(n, n->next, node)){
			node->n = n;
			return 1;
		}
	} while((n = n->next) != nstart);
	abort(209);			/* On boundary but not on boundary */
}

copybuf(from, to, nbytes)
char *from, *to;
int nbytes;
{
	if(to == (char *)NULL || from == (char *)NULL)
		return 1;
	while(nbytes--)
		*to++ = *from++;
}

add_alias(node, newnode, receiver)
NODEPTR node, newnode;
int receiver;
/*
	Add an alias to this node which points to receiver::newnode, unless
	we already have such an alias.
*/
{
	ALIASPTR newa, alias_alloc(), ali, astart;
	BDYPTR b_alloc();
	if(newnode == node && receiver == procnum)
		return 1;
	if(node->b == (BDYPTR)NULL)
		node->b = b_alloc();
	else if(node->b->alias != (ALIASPTR)NULL){
		ali = astart = node->b->alias;
		do {
			if(ali->node == newnode && ali->proc == receiver)
				return 1;
		} while((ali = ali->next) != astart);
	}
	newa = alias_alloc();
	newa->node = newnode;
	newa->proc = receiver;
	if(!(node->b->alias != (ALIASPTR)NULL)){
		node->b->alias = newa;
		newa->next = newa;
	} else {
		newa->next = node->b->alias->next;
		node->b->alias->next = newa;
	}
}

double default_sqdist(n1, n2)
NODEPTR n1, n2;
{
	double dx, dy;
	dx = n1->x - n2->x;
	dy = n1->y - n2->y;
	return(dx*dx + dy*dy);
}

showbalance()
/*
    Show procnums of all the elements, with lines between processors
    width-coded by metric
*/
{
    NODEPTR node;
    ELMTPTR elmt;
    char num[10];
    double x1, y1, x2, y2, ux, uy, d;
    int m;
/*
    sprintf(num, "%d", procnum);
    FORALLELMTS(elmt)
        dotext(num, XE(elmt), YE(elmt), 0.0, 0, 0);
    NEXTELMT(elmt)
    usendplot();
*/
 
    FORALLNODES(node)
        if(!node->b || !node->b->oppalias)
            continue;
        x1 = node->x;
        y1 = node->y;
        x2 = node->b->clock->x;
        y2 = node->b->clock->y;
        m = (*machine_metric)(procnum, node->b->oppalias->proc);
        linemod(m);
        move(x1, y1);
        cont(x2, y2);
    NEXTNODE(node)
 
    usendplot();
}

uniform_metric(p, q)
int p, q;
{
    return(p!=q);
}
