/****************************************************/
/*    DIME: grafix.c                                */
/*    Various graphics functions for dime.          */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/
#define DIMECODE
#include "dime.h"
#define xold		work_space.nf[0]		/* For picture() */
#define yold		work_space.nf[1]

FILE *ps_file;
int plotting = 0;

drawmesh()
/*
	Draw the edges of each elmt, but only do each edge once.
*/
{
	ELMTPTR this;
	NODEPTR n0, n1, node;
	double x, y, x1, y1;
	int in, bedge;
	fmulti(stdout);
	color(4);
/*
	Do internal edges
*/
	FORALLELMTS(this)
		for(in=0; in<3; in++){
			n0 = this->neigh[in];
			n1 = this->neigh[(in+1)%3];
			bedge = (n0->b != (BDYPTR)NULL) && (n1->b != (BDYPTR)NULL) && 
					((n0->b->clock == n1) || (n1->b->clock == n0));
			if(!bedge && (long)n0 < (long)n1){
				DISTORT(n0, x, y);
				move(x, y);
				DISTORT(n1, x, y);
				cont(x, y);
			}
		}
	NEXTELMT(this)
/*
	Do boundaries, with real boundaries and processor boundaries different
*/
	FORALLNODES(node)
		if(node->b != (BDYPTR)NULL){
			DISTORT(node, x, y);
			if(node->b->oppalias == (ALIASPTR)NULL)
				color(3);
			else if(node->b->oppalias->proc & HIBITS)
				color(5);
			else
				color(2);
			move(x, y);
			DISTORT(node->b->clock, x1, y1);
			cont(x1, y1);
		}
	NEXTNODE(node)
	if(mesh_is_distributed)
		usendplot();
	else
		sendplot();
	fsingl(stdout);
}

drawelmt(this, shade)
/*
	Using color 'shade', draw the elmt 'this' as a filled triangle a little
	smaller than the whole elmt.
*/
ELMTPTR this;
int shade;
{
	int i;
	double x[3], y[3], xc, yc;
	for(i=0; i<3; i++)
		DISTORT(this->neigh[i], x[i], y[i]);
	xc = (x[0] + x[1] + x[2])/3;
	yc = (y[0] + y[1] + y[2])/3;
	initpanel(shade, 0);
	for(i=0; i<3; i++){
		panelpoint(0.2*xc + 0.8*x[i], 0.2*yc + 0.8*y[i]);
	}
	endpanel();
}

void
distortold(node, axold, ayold)
NODEPTR node;
double *axold, *ayold;
{
	*axold = node->xold;
	*ayold = node->yold;
}

drawoldelmt(this, shade)
ELMTPTR this;
int shade;
{
	int i;
	void distortold(), (*keep)();
	double x[3], y[3], xc, yc;
	keep = distort;
	distort = distortold;
	for(i=0; i<3; i++)
		DISTORT(this->neigh[i], x[i], y[i]);
	xc = (x[0] + x[1] + x[2])/3;
	yc = (y[0] + y[1] + y[2])/3;
	initpanel(shade, 0);
	for(i=0; i<3; i++){
		panelpoint(0.5*xc + 0.5*x[i], 0.5*yc + 0.5*y[i]);
	}
	endpanel();
	distort = keep;
}

drawsmallelmt(this, shade)
ELMTPTR this;
int shade;
{
	int i;
	double x[3], y[3], xc, yc;
	for(i=0; i<3; i++)
		DISTORT(this->neigh[i], x[i], y[i]);
	xc = (x[0] + x[1] + x[2])/3;
	yc = (y[0] + y[1] + y[2])/3;
	initpanel(shade, 0);
	for(i=0; i<3; i++){
		panelpoint(0.75*xc + 0.25*x[i], 0.75*yc + 0.25*y[i]);
	}
	endpanel();
}

draw(A, B)
NODEPTR A, B;
/*
	Draw a line from node A to node B with current color
*/
{
	double x, y;
	DISTORT(A, x, y);
	move(x, y);
	DISTORT(B, x, y);
	cont(x, y);
}

debug_toggle()
{
	if(debug)
		printf("Debug Off\n");
	else
		printf("Debug On\n");
	debug = 1-debug;
}

postscript()
{
#define HDSIZ	256
	char string[256], hdbuf[HDSIZ];
	int n;
	FILE *pshead;
	double xmin, ymin, xmax, ymax;

	if(plotting){
		fflush(ps_file);
		fsingl(ps_file);
		fprintf(ps_file, "\nshowpage\n");
		fclose(ps_file);
		printf("Plot finished\n"); fflush(stdout);
	} else {
		strcpy(string, getenv("DIME"));
		strcat(string, "/src/dime/ps.head");
		if((pshead = fopen(string, "r")) == (FILE *)NULL){
			printf("Cannot find Postscript header %s\n", string);
			fflush(stdout);
		}

		do {
			printf("Postscript file?\n"); fflush(stdout);
			fsingl(stdin);
			get_string(string);
		} while((ps_file = fopen(string, "w")) == (FILE *)NULL);

		if(pshead != (FILE *)NULL){
			while((n = fread(hdbuf, 1, HDSIZ, pshead)) != 0)
					  fwrite(hdbuf, 1, n, ps_file);
			fclose(pshead);
		}
		get_screen(&xmin, &ymin, &xmax, &ymax);
		fprintf(ps_file, "%f %f %f %f z\n", xmin, ymin, xmax, ymax);
		fflush(ps_file);
		if(mesh_is_distributed)
			fmulti(ps_file);
	}
	plotting = !plotting;
	return(1);
}

#define NHIST 20
draw_hist()
{
	int hist[NHIST], hi;
	int i, maxhist, maxelmt, minelmt, sumelmt, nelmt = 0;
	int nmess, totmess;
	ELMTPTR elmt;
	char s[100];

	FORALLELMTS(elmt)
		nelmt++;
	NEXTELMT(elmt)

	sumelmt = minelmt = maxelmt = nelmt;
	GLOBAL_COMBINE(maxelmt, imax, 1);
	GLOBAL_COMBINE(minelmt, imin, 1);
	GLOBAL_COMBINE(sumelmt, isum, 1);
	printf("Avg elements %6.1f, min %d max %d\n", 
		(double)sumelmt/nprocs, minelmt, maxelmt);

	setup_combine();
	make_hist(hist, NHIST, &hi);

	zoom_explicit(-0.2, -0.2, 1.0, 1.0);

	color(FOREGD);
	box(1.0, (double)procnum/nprocs, 
		1.0 - nelmt*0.5/maxelmt, (double)(procnum+1)/nprocs,
		YELLOW, 1);
	sprintf(s, "%d in %d", nelmt, procnum);
	dotext(s, 0.9, (procnum+0.5)/nprocs, 0.0, 1, 0);
	usendplot();

	fmulti(stdout);
	if(procnum == 0){
		maxhist = nmess = totmess = 0;
		for(i=0; i<NHIST; i++){
			nmess += hist[i];
			totmess += (i+0.5)*hi*hist[i];
			if(hist[i] > maxhist)
				maxhist = hist[i];
		}
		for(i=0; i<NHIST; i++){
			if(maxhist > 0)
				box(0.0, (double)i/NHIST,
					hist[i]*0.5/maxhist, (double)(i+1)/NHIST,
					CYAN, 1);
			if(hist[i] > 0){
				sprintf(s, "%d in [%d,%d]", hist[i], i*hi, (i+1)*hi-1);
				dotext(s, 0.1, (i+0.5)/NHIST, 0.0, -1, 0);
			}
		}
		sprintf(s, "Elmts per Processor");
		dotext(s, 0.9, -0.05, 0.0, 1, 0);
		sprintf(s, "Global Message Histogram");
		dotext(s, 0.0, -0.05, 0.0, -1, 0);
		sprintf(s, "Messages %d, Total size %d", nmess, totmess);
		dotext(s, 0.0, -0.15, 0.0, -1, 0);

		printf("Messages %d, Total size %d\n", nmess, totmess);
	}
	fsingl(stdout);
	usendplot();
}

balance_orb()
{
#define NEWPROC		work_space.ei[0]
#ifdef CUBIX
	int balance(), *nextelmt();
	static int first = 1;
	ELMTPTR elmt;
	if(!mesh_is_distributed)
		return 0;
	FORALLELMTS(elmt)
		set_balance(elmt, procnum);
	NEXTELMT(elmt)
	orb(balance, nextelmt, first);
	first = 0;
#endif
	return(1);
}

int *nextelmt(ax, ay, aflag)
double *ax, *ay;
int *aflag;
{
	static ELMTPTR elmt;
	double x[3], y[3];
	int i;
	if(*aflag){
		elmt = the_elmts;
		if(elmt == (ELMTPTR)NULL)
			return((int *)NULL);
	}
	if(*aflag == 0 && (elmt = elmt->next) == the_elmts)
		return((int *)NULL);
	*aflag = 0;
	for(i=0; i<3; i++)
		DISTORT(elmt->neigh[i], x[i], y[i]);
	*ax = (x[0] + x[1] + x[2])/3;
	*ay = (y[0] + y[1] + y[2])/3;
	return(&(elmt->NEWPROC));
}

intact_refine()
{
	double tol, xmin, ymin, xmax, ymax, dx, dy, xc, yc;
	int yes, i;
	ELMTPTR elmt;
	printf("Max length?");fflush(stdout);
	fsingl(stdin);
	get_double(&tol);
	printf("Click lower left corner\n"); fflush(stdout);
	get_gin(&xmin, &ymin);
	printf("Click upper right corner\n"); fflush(stdout);
	get_gin(&xmax, &ymax);
	rect_refine(tol, xmin, ymin, xmax, ymax);
}

rect_refine(tol, xmin, ymin, xmax, ymax)
double tol, xmin, ymin, xmax, ymax;
{
	double xc, yc, tolsq, d2;
	double x[3], y[3];
	int yes, i, nref, count = 0;
	ELMTPTR elmt;
	tolsq = tol*tol;
	do {
		nref = 0;
		FORALLELMTS(elmt)
			for(i=0; i<3; i++)
				DISTORT(elmt->neigh[i], x[i], y[i]);
			xc = (x[0] + x[1] + x[2])/3;
			yc = (y[0] + y[1] + y[2])/3;
			if(xc < xmin || xc > xmax || yc < ymin || yc > ymax){
				set_refine(elmt, 0);
				continue;
			}
			if(tol < 0.0)
				yes = 1;
			else {
				yes = 0;
				for(i=0; i<3; i++){
					d2 = (*sqdist)(elmt->neigh[i], elmt->neigh[(i+1)%3]);
					yes |= (d2 > tolsq);
				}
			}
			set_refine(elmt, yes);
			nref += yes;
		NEXTELMT(elmt)
		refine();
		GLOBAL_COMBINE(nref, isum, 1);
	} while(nref && tol >= 0.0 && count++ < 6);
	/*swops();*/
}

logicalmesh()
/*
	To draw a picture demonstating the entire structural part of the mesh.
*/
{
#define OPPALIAS		1
	int source, i, nbytes, nneigh, aliastype, aliasproc, dest, shade;
	void distortold(), (*keep)();
	double x[3], y[3];
	double screensize(), markersize = 0.01*screensize();
	NODEPTR node, mailnode;
	ELMTPTR elmt;
	NODENEIGHPTR n, nstart;
	ALIASPTR ali, astart;
	double xc, yc;
	char *mailbag;

	FORALLNODES(node)
		if(node->b != (BDYPTR)NULL && (node->b->alias != (ALIASPTR)NULL
				|| node->b->oppalias != (ALIASPTR)NULL)){
			node->xold = node->yold = 0.0;
			nneigh = 0;
			n = nstart = node->n;
			do {
				elmt = n->elmt;
				for(i=0; i<3; i++)
					DISTORT(elmt->neigh[i], x[i], y[i]);
				xc = (x[0] + x[1] + x[2])/3;
				yc = (y[0] + y[1] + y[2])/3;
				++nneigh;
				node->xold += xc;
				node->yold += yc;
			} while((n = n->next) != nstart);
			DISTORT(node, x[0], y[0]);
			node->xold = 0.5*node->xold/nneigh + 0.5*x[0];
			node->yold = 0.5*node->yold/nneigh + 0.5*y[0];
		} else {
			DISTORT(node, node->xold, node->yold);
		}
	NEXTNODE(node)

	fmulti(stdout);
	linemod(0);

	keep = distort;
	distort = distortold;
	FORALLNODES(node)
		shade = 1;
		n = nstart = node->n;
		do {
			elmt = n->elmt;
			for(i=0; i<3; i++)
				DISTORT(elmt->neigh[i], x[i], y[i]);
			xc = (x[0] + x[1] + x[2])/3;
			yc = (y[0] + y[1] + y[2])/3;

			DISTORT(node, x[0], y[0]);
			color(4);
			move(x[0], y[0]);
			cont((x[0] + xc)/2, (y[0] + yc)/2);
		} while((n = n->next) != nstart);
	NEXTNODE(node)
	linemod(0);
	if(mesh_is_distributed)
		usendplot();
	else
		sendplot();

	FORALLNODES(node)
		if(node->b == (BDYPTR)NULL)
			continue;
		DISTORT(node, xc, yc);
		if(B_NODE(node)){
			color(4+node->b->type);
			marker(1, xc, yc, markersize);
		}
		if(B_OPP(node)){
			color(2);
			marker(3, xc, yc, markersize);
		}
		if(node->b->oppalias == (ALIASPTR)NULL)
			color(3);
		else if(node->b->oppalias->proc & HIBITS)
			color(2);
		else
			color(5);
		move(xc, yc);
		DISTORT(node->b->clock, xc, yc);
		cont(xc, yc);

		DISTORT(node, xc, yc);
		if(node->b->alias != (ALIASPTR)NULL){
			ali = astart = node->b->alias;
			do {
				mailbag = buffer;
				xpacket(mailbag, NODEPTR ) = ali->node;
				xpacket(mailbag, double) = xc;
				xpacket(mailbag, double) = yc;
				xpacket(mailbag, int) = ali->proc;
				xpacket(mailbag, int) = !OPPALIAS;
				dest = (ali->proc) & LOWBITS;
				if(Add_to_out_buf(buffer, (int)(mailbag - buffer), 1, &dest,1) == -1)
					abort(301);			/* AOB alias failed in picture */
			} while((ali = ali->next) != astart);
		}
		if(node->b->oppalias != (ALIASPTR)NULL){
			mailbag = buffer;
			xpacket(mailbag, NODEPTR ) = node->b->oppalias->node;
			xpacket(mailbag, double) = xc;
			xpacket(mailbag, double) = yc;
			xpacket(mailbag, int) = node->b->oppalias->proc;
			xpacket(mailbag, int) = OPPALIAS;
			dest = (node->b->oppalias->proc) & LOWBITS;
			if(Add_to_out_buf(buffer, (int)(mailbag - buffer), 1, &dest, 1) == -1)
				abort(302);			/* AOB oppalias failed in picture */
		}
	NEXTNODE(node)
	if(mesh_is_distributed)
		usendplot();
	else
		sendplot();

	FORALLELMTS(elmt)
		drawoldelmt(elmt, PROC_COLOR(procnum));
	NEXTELMT(elmt)
	distort = distortold;
	if(mesh_is_distributed)
		usendplot();
	else
		sendplot();

	if(Crystal_router() == -1)
		abort(303);						/* CR failed in picture */

	while(Get_message(buffer, &source, &nbytes, 1) == 0){
		mailbag = buffer;
		mailnode = xpacket(mailbag, NODEPTR );
		xc = xpacket(mailbag, double);
		yc = xpacket(mailbag, double);
		aliasproc = xpacket(mailbag, int);
		aliastype = xpacket(mailbag, int);

		if(aliasproc & HIBITS)
			color(2);
		else
			color(5);
		if(aliastype == OPPALIAS)
			color(5);
		else
			color(7);
		DISTORT(mailnode, x[0], y[0]);
		move((x[0] + xc)/2, (y[0] + yc)/2);
		cont(xc, yc);
	}

	if(mesh_is_distributed)
		usendplot();
	else
		sendplot();

	fsingl(stdout);
	distort = keep;
	return(1);
}

elmt_contor(func, nlevel, min, max)
/*
	To put a color number on each element corresponding to 
	the function func.
*/
double (*func)();
int nlevel;
double min, max;
{
	ELMTPTR elmt;
	NODEPTR node;
#define LARGE		1.e12
	int i, shade;
	double temp, x, y, Max = -LARGE, Min = LARGE;
	if(nlevel < 1)
		return 0;
	if(Min >= Max){
		if(the_elmts != (ELMTPTR)NULL)
			Max = Min = (*func)(the_elmts);
		FORALLELMTS(elmt)
			temp = (*func)(elmt);
			if(temp > Max)	Max = temp;
			if(temp < Min)	Min = temp;
		NEXTELMT(elmt);
		GLOBAL_COMBINE(Max, fmax, 1);
		GLOBAL_COMBINE(Min, fmin, 1);
	} else {
		Min = min;
		Max = max;
	}
	printf("Min %f Max %f\n", Min, Max);fflush(stdout);
	if(fabs(Max - Min) < 1.e-12){
		printf("Contor: flat function\n");fflush(stdout);
		return(1);
	}
	fmulti(stdout);
	FORALLELMTS(elmt)
		shade = BLACK + (WHITE-BLACK)*((Max - (*func)(elmt))/(Max - Min));
		initpanel(shade, 0);
		for(i=0; i<3; i++){
			DISTORT(elmt->neigh[i], x, y);
			panelpoint(x, y);
		}
		endpanel();
	NEXTELMT(elmt)

	color(1);
	FORALLNODES(node)
		if(B_EDGE(node))
			draw(node, node->b->clock);
	NEXTNODE(node)
	if(mesh_is_distributed)
		usendplot();
	else
		sendplot();
	fsingl(stdout);
}

arrow(x, y, vx, vy)
double x, y, vx, vy;
{
	double dx, dy;
	move(x, y);
	cont(x = x+vx, y = y+vy);
	vx *= 0.2;
	vy *= 0.2;
	dx = 0.6*vx + 0.4*vy;
	dy =-0.4*vx + 0.6*vy;
	cont(x - dx, y - dy);
	move(x, y);
	dx = 0.6*vx - 0.4*vy;
	dy = 0.4*vx + 0.6*vy;
	cont(x - dx, y - dy);

/* Filled arrow for video
	initpanel(3, 0);
	panelpoint(x + vx,  y + vy);
	vx *= 0.2;
	vy *= 0.2;
	panelpoint(x - vy,  y + vx);
	panelpoint(x + vy,  y - vx);
	endpanel();
*/
}

get_extents(xmin, ymin, xmax, ymax)
double *xmin, *ymin, *xmax, *ymax;
{
	double x, y;
	NODEPTR node;

	if(the_nodes)
		DISTORT(the_nodes, x, y);

	*xmin = LARGE; *xmax = -LARGE;
	*ymin = LARGE; *ymax = -LARGE;
	FORALLNODES(node)
		DISTORT(node, x, y);
		if(x < *xmin) *xmin = x; 
		if(x > *xmax) *xmax = x;
		if(y < *ymin) *ymin = y;
		if(y > *ymax) *ymax = y;
	NEXTNODE(node);
#ifdef DEBUG
fmulti(stdout);
printf("(%f %f) to (%f %f)\n", xmin[0], ymin[0], xmax[0], ymax[0]);
fflush(stdout);
fsingl(stdout);
#endif
	GLOBAL_COMBINE(xmin[0], fmin, 1);
exsync();
	GLOBAL_COMBINE(ymin[0], fmin, 1);
exsync();
	GLOBAL_COMBINE(xmax[0], fmax, 1);
exsync();
	GLOBAL_COMBINE(ymax[0], fmax, 1);
exsync();
#ifdef DEBUG
fmulti(stdout);
printf("(%f %f) to (%f %f)\n", xmin[0], ymin[0], xmax[0], ymax[0]);
fflush(stdout);
fsingl(stdout);
#endif
}
