#ifdef CUBIX
/****************************************************/
/*    DIME: orb.c                                   */
/*    Orthogonal Recursive Bisection for            */
/*    Load Balancing.                               */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/
#include <stdio.h>
#include <cros.h>
#define MAXDOC			10
#define MEDIAN_STEPS	12	
#define LARGE			1.e12
#define START			1

extern int curr_doc, procnum;

orb(send, next, first)
int (*send)(), *(*next)();
int first;
/*
	This function accomplishes Orthogonal Recursive Bisection of a set of
	objects in 2D. The function next() takes 2 pointers to double as 
	arguments and returns a pointer to int with the double filled with the 
	(x, y) coordinates of the next object and the int being a place where
	the new procnum for the object is stored. When all done next() returns
	NULL. send() uses the information in the addresses returned by next()
	to actually route the mail to other processors. The rectangle
	(xmin, ymin, xmax, ymax) is the universe of discourse and first is 
	1 for the first call and subsequently 0.

	First we split on the highest channel on x then on the next highest
	on y and so on.
*/
{
	int i, chan, uscnt[2], themcnt[2], mask, step, *newproc, flag, total;
	double x, y, proportion, box[4], min, max, line;
	int splitx, fmin(), fmax(), bit, oddbit[2];
#define minx	box[0]
#define miny	box[1]
#define maxx	box[2]
#define maxy	box[3]
	oddbit[0] = oddbit[1] = 0;

	for(chan=0; chan < curr_doc; chan++){
		printf("Splitting on channel %d ", chan); fflush(stdout);
		mask = 1<<chan;
		minx = miny = LARGE;
		maxx = maxy = -LARGE;
		flag = START;
		while((newproc = (*next)(&x, &y, &flag)) != (int *)NULL){
			if(x < minx)	minx = x;
			if(x > maxx)	maxx = x;
			if(y < miny)	miny = y;
			if(y > maxy)	maxy = y;
		}
		sub_combine(chan, curr_doc, &minx, fmin, sizeof(double), 2);
		sub_combine(chan, curr_doc, &maxx, fmax, sizeof(double), 2);
		
		splitx = (chan+1)%2;
		if(splitx){
			min = minx;
			max = maxx;
		} else {
			min = miny;
			max = maxy;
		}
		line = 0.5*(max - min);
		flag = START;
		while((newproc = (*next)(&x, &y, &flag)) != (int *)NULL){
			if((splitx ? x: y) > line)
				bit = 1; else bit = 0;
			bit = bit ^ oddbit[splitx];
			if(bit)
				*newproc = procnum | mask;
			else
				*newproc = procnum &~ mask;
		}
/*
	Do a binary search for the median split along this channel
*/
		for(step=0; step<MEDIAN_STEPS; step++){
			uscnt[0] = uscnt[1] = 0;
			flag = START;
/*
	Get the number of objects owned by me and destined for me in uscnt[0]
	and the number owned by me and destined for someone else in uscnt[1].
*/
			while((newproc = (*next)(&x, &y, &flag)) != (int *)NULL){
				if((*newproc) & mask)
					uscnt[1]++;
				else
					uscnt[0]++;
			}
/*
	And the same in themcnt for the neighbor on this channel
*/
			for(i=chan; i<curr_doc; i++){
				cshift(themcnt, 1<<i, 2*sizeof(int), 
					   uscnt  , 1<<i, 2*sizeof(int));
				uscnt[0] += themcnt[0];
				uscnt[1] += themcnt[1];
			}
			total = uscnt[0] + uscnt[1];

			proportion = (total > 0) ? (double)(uscnt[0])/total : 0.5;
/*
	Get new approximation to the median by binary search
*/
			if((proportion > 0.5) ^ oddbit[splitx]){
				max = line;
				line = 0.5*(min + line);
			} else {
				min = line;
				line = 0.5*(max + line);
			}

			flag = START;
			while((newproc = (*next)(&x, &y, &flag)) != (int *)NULL){
				if((splitx ? x: y) > line)
					bit = 1; else bit = 0;
				bit = bit ^ oddbit[splitx];
				if(bit)
					*newproc = procnum | mask;
				else
					*newproc = procnum &~ mask;
			}
		}
		if((procnum>>chan)&1)
			oddbit[splitx] = 1-oddbit[splitx];
/*
	Now actually send objects according to *newproc values
*/
		(*send)();
	}
}

sub_combine (minchan, maxchan, buf, func, size, nitems)
int maxchan, minchan;
char *buf;
int (*func)();
int size;
int nitems;
{
#define BIG		256
	int i, j, status, error;
	char temp[BIG];

	error = 0;
	for (i=minchan; i<maxchan; ++i) {
		status = cshift (temp, 1<<i, size*nitems, 
				 buf, 1<<i, size*nitems);
		if (status == -1) error = -1;
		for (j=0; j<nitems*size; j+=size)
			(*func)(buf+j, temp+j, size);
	}
	if (error == -1) return(error);
	else return(status);
}
#endif
