/**********
* macdriv *		Copyright ParaSoft Corporation, 1990
***********

Device driver for the Macintosh implementation of PLOTIX.
*/
#include <stdio.h>
#include "plotix.h"
#include <:sys:plotix.h>
#pragma segment PLOTIX
#include "quickd.h"

/*
 * These variables are used by the "gin" mechanism.
 */
extern int mousechar;
extern Point mousePos;
Boolean mousedown = false;

static void _plx_mac_setclip(l, b, r, t)
     Coord l, b, r, t;
{
    Rect cliprect;
	
    cliprect.left = USER_to_DEV_x(l);
    cliprect.right = USER_to_DEV_x(r);
    cliprect.top = USER_to_DEV_y(t);
    cliprect.bottom = USER_to_DEV_y(b);
    ClipRect(&cliprect);
    IND(clip) = 1;
    return;
}

static void _plx_mac_endclip()
{
    Rect rect;
	
    rect.left = 0;
    rect.right = IND(devx)-1;
    rect.top = 0;
    rect.bottom = IND(devy)-1;
    ClipRect(&rect);

    IND(clip) = 0;
    return;	
}

static void _plx_mac_cont(x,y)
     Coord x,y;
{
    int x_curr,y_curr;
    x_curr = USER_to_DEV_x(x);
    y_curr = USER_to_DEV_y(y);
    LineTo(x_curr, y_curr);
    return;        
}

static void _plx_mac_move(x,y)
     Coord x,y;
{
    int x_curr,y_curr;
    x_curr = USER_to_DEV_x(x);
    y_curr = USER_to_DEV_y(y);
    MoveTo(x_curr, y_curr);

    return;     
}

static long rainb[] = {redColor, magentaColor, blueColor, 
			cyanColor, greenColor, yellowColor, redColor};

/* The following make up the grey scale and hatch patterns used by
   good old PLOTIX.
*/
static Pattern greys[] = {
    {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},   /* Grey 0 - backgrnd */
    {0x88, 0x00, 0x22, 0x00, 0x88, 0x00, 0x22, 0x00},   /* Grey 1 */
    {0xaa, 0x00, 0xaa, 0x00, 0xaa, 0x00, 0xaa, 0x00},   /* Grey 2 */
    {0xaa, 0x44, 0xaa, 0x11, 0xaa, 0x44, 0xaa, 0x11},   /* Grey 3 */
    {0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55},   /* Grey 4 */
    {0xee, 0x55, 0xbb, 0x55, 0xee, 0x55, 0xbb, 0x55},   /* Grey 5 */
    {0xff, 0x55, 0xff, 0x55, 0xff, 0x55, 0xff, 0x55},   /* Grey 6 */
    {0xff, 0xdd, 0xff, 0xff, 0xff, 0xdd, 0xff, 0xff},   /* Grey 7 */
    {0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff},   /* Grey 8 - foregrnd */
};

static Pattern hatch[] = {
    {0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},   /* SUN hatch 1 */
    {0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0},   /* SUN hatch 2 */
    {0x81, 0x03, 0x06, 0x0c, 0x18, 0x30, 0x60, 0xc0},   /* SUN hatch 3 */
    {0x81, 0xc0, 0x60, 0x30, 0x18, 0x0c, 0x06, 0x03},   /* SUN hatch 4 */
    {0x18, 0x30, 0x60, 0xc0, 0x81, 0xc0, 0x60, 0x30},   /* SUN hatch 5 */
    {0x08, 0x00, 0x00, 0x80, 0xc1, 0x63, 0x36, 0x1c},   /* SUN hatch 6 */
    {0xf0, 0xf0, 0xf0, 0xf0, 0x0f, 0x0f, 0x0f, 0x0f},   /* SUN hatch 7 */
    {0x01, 0x01, 0x01, 0xff, 0x10, 0x10, 0x10, 0xff},   /* SUN hatch 8 */
    {0x88, 0x88, 0x88, 0xff, 0x88, 0x88, 0x88, 0xff},	/* SUN hatch 9 */
    {0x3c, 0x66, 0xc3, 0x81, 0x81, 0xc3, 0x66, 0x3c},	/* SUN hatch 10 */
};
#define NGREY		(sizeof(greys)/sizeof(greys[0]))
#define NPATS   	(sizeof(hatch)/sizeof(hatch[0]))
#define MOD(i, n)	((i < 0) ? ((-i)%n) : (i%n))

static void _plx_mac_initpanel(color,edge)
     int color,edge;
{
    DEP(polygon) = OpenPoly();
    DEP(first) = 1;    
    DEP(sv_color) = color;
    DEP(sv_edge) = edge;
    return;
}

static void _plx_mac_panelpoint(x, y)
     Coord x, y;
{
    short x_curr,y_curr;

    x_curr = USER_to_DEV_x(x);
    y_curr = USER_to_DEV_y(y);

    if(DEP(first)) {
	MoveTo(x_curr, y_curr);
	DEP(first) = 0;
	DEP(first_x) = x_curr;
	DEP(first_y) = y_curr;
    }
    else LineTo(x_curr, y_curr);
	
    return;
}

static void _plx_mac_endpanel()
{
    double h, f;
    int i;
	
    LineTo(DEP(first_x), DEP(first_y));	/* Close the polygon */
    ClosePoly();

/* Now paint the region appropriately if we've selected a color */

    if (!DEP(colorFlag)) {
	if(IND(grey_lo) <= DEP(sv_color) && DEP(sv_color) <= IND(grey_hi)) {
	    FillPoly(DEP(polygon),greys[(NGREY*(DEP(sv_color)-IND(grey_lo)))/
					(IND(grey_hi)-IND(grey_lo)+1)]);
	}
	else if(IND(rain_lo) <= DEP(sv_color) && DEP(sv_color)<=IND(rain_hi)) {
	    FillPoly(DEP(polygon),greys[(NGREY*(DEP(sv_color)-IND(rain_lo)))/
					(IND(rain_hi)-IND(rain_lo)+1)]);
	}
	else if(DEP(sv_color) == 1) {
	    FillPoly(DEP(polygon), greys[NGREY-1]);
	}
	else if(DEP(sv_color) == 0) {
	    FillPoly(DEP(polygon), greys[0]);
	}
	else if(DEP(sv_color) != 0) {
	     FillPoly(DEP(polygon), hatch[MOD(DEP(sv_color), NPATS)]);
	}
    } 
    else {
/*
 * In this section we assume that we are beginning with the 
 * Background color set to white. If this condition is ever violated 
 * we have to go and restore things the way they were. We also have 
 * to save the foreground color and reset it so that lines etc. come 
 * out correctly.
 */
	if(IND(grey_lo) <= DEP(sv_color) && DEP(sv_color) <= IND(grey_hi)) {
	    ForeColor(blackColor);
	    FillPoly(DEP(polygon),greys[(NGREY*(DEP(sv_color)-IND(grey_lo)))/
						(IND(grey_hi)-IND(grey_lo)+1)]);
	}
	else if(IND(rain_lo) <= DEP(sv_color) && DEP(sv_color)<=IND(rain_hi)) {
	    h = 6.0*((double)(DEP(sv_color)-IND(rain_lo)))/
				((double)(IND(rain_hi)-IND(rain_lo)+1));
	    i = (int)h;
	    f = h - (double)i;
	    ForeColor(rainb[i+1]);
	    BackColor(rainb[i]);
	    FillPoly(DEP(polygon),greys[(int)((NGREY-1)*f)+1]);
	    BackColor(whiteColor);
	}
	else if(DEP(sv_color) < 0) {
	    ForeColor(blackColor);
	    FillPoly(DEP(polygon), hatch[MOD(DEP(sv_color), NPATS)]);
	}
	else {			/* Solid case of PLOTIX colors */
	    ForeColor(DEP(plxcols[MOD(DEP(sv_color), 8)]));
	    FillPoly(DEP(polygon), greys[NGREY-1]);
	}
	ForeColor(DEP(forecol));
    }
/*
 * Draw the edge, if we asked for one. Do this after painting the 
 * interior so that the edges get "done" properly
 */
	if(DEP(sv_edge)) {
	    FramePoly(DEP(polygon));
	}
	KillPoly(DEP(polygon));
	return;
}

static void _plx_mac_point(x,y)
     Coord x,y;
{
    short x_curr, y_curr;

    x_curr = USER_to_DEV_x(x);
    y_curr = USER_to_DEV_y(y);
    MoveTo(x_curr, y_curr);
    LineTo(x_curr, y_curr);	
/* ... it's worked before !! */
    return;
}

static void _plx_mac_color(shade)
     int shade;
{
    if(!DEP(colorFlag)) {
        if(color == 0) PenPat(qd.white);
        else           PenPat(qd.black);
    } 
    else {
	DEP(forecol) = (shade >= 0) ? (shade % 8) : ((-shade) % 8);
	ForeColor(DEP(plxcols[DEP(forecol)]));
    }
    return;
}

static void _plx_mac_dotext(s,x,y,angle, hjust, vjust)
     char *s;
     Coord x,y;
     int angle;
     int hjust, vjust;
{
    short hpatch, vpatch,x_curr,y_curr;

    x_curr = USER_to_DEV_x(x);
    y_curr = USER_to_DEV_y(y);

    if(hjust < 0) hpatch = 0;
    else if(hjust == 0) hpatch = stringwidth(s)/2;
    else hpatch = stringwidth(s);

    if(vjust < 0) vpatch = 0;
    else if(vjust == 0) vpatch = DEP(fontinfo.ascent)/2;
    else vpatch = DEP(fontinfo.ascent)+DEP(fontinfo.descent);
	
    MoveTo(x_curr-hpatch, y_curr+vpatch);
    ForeColor(blackColor);
    drawstring(s);
    return;
}

static void _plx_mac_sztext(str, pwd, pht)
     char *str;
     Coord *pwd, *pht;
{
    *pwd = (double)(stringwidth(s))/IND(mul[0]);
    *pht = (double)(DEP(fontinfo.ascent) + DEP(fontinfo.descent))/IND(mul[1]);
    return;
}

#ifdef DBG_INTEXT
static char give_str[] = "ab\177cdefghijklmnopqrstuvwxyz";
static char *give_ptr = give_str;
#endif

int _plx_mac_loopcheck() {
    loopcheck(10);	/* 10 is a more or less random number */
    return 0;	/* Do we know if more is waiting? */
}

static int poll_getchar()
{
#ifndef DBG_INTEXT
    int c;
    do {
		mousechar = 0;		/* Makesure clicks are separate */
    	while(!mousechar)loopcheck(100);
    	c = mousechar;
    	mousechar = 0;
	}
	while(c < 32);
	return c;
#else
    return *give_ptr++;
#endif
}

static void _plx_mac_intext(prompt, x, y, term, input, maxlen)
     char *prompt;
     Coord x, y;
     char *term, *input;
     int maxlen;
{
    Coord in_x;
    short at_x, at_y;
    char c, *at_ptr, *strchr();
    int got = 0;
    int msize_x;
    Rect rect;

#ifdef DBG_INTEXT
    give_ptr = give_str;
#endif
    memset(input, 0, maxlen);
    dotext(prompt, x, y, 0, -1, 0);
    
    msize_x = stringwidth(prompt);
    in_x = x + ((double)msize_x)/IND(mul[0]);
    at_x = USER_to_DEV_x(in_x);
    at_y = USER_to_DEV_y(y);
    
    at_ptr = input;
    do {
    	c = poll_getchar();
	if(c == '\r') c = '\n';
	
	if((c == 8 || c == 127) && got > 0) {	/* CTRL-H or DEL */
	    msize_x = stringwidth(input);
   	    rect.left = at_x;
    	    rect.right = at_x + msize_x;
    	    rect.top = at_y - DEP(fontinfo.ascent);
    	    rect.bottom = at_y + DEP(fontinfo.descent);	
    	    EraseRect(&rect);
	    *--at_ptr = '\0';
	    got--;
	}
	else if(c >= 32) {			/* Only ASCII, please! */
	    *at_ptr++ = c;
	    got++;
	}
	dotext(input, in_x, y, 0, -1, 0);
	sendplot();
	
    } while(got < maxlen && strchr(term, c) == (char *)0);
    *at_ptr = '\0';
    return;
}

static void _plx_mac_erase()
{
    Rect rect;
	
    rect.left = 0;
    rect.right = IND(devx)-1;
    rect.top = 0;
    rect.bottom = IND(devy)-1;
	
    EraseRect(&rect);
    return;
}

static int _plx_mac_gin(key,x,y)
     int *key;
     Coord *x,*y;
{
    long i;
    int svkey;

    mousechar = 0;		/* Makesure clicks are separate */
    while(!mousechar)loopcheck(100);
    svkey = mousechar;
    mousechar = 0;
    for(i=0;i<100;i++){
	if(mousechar) break;
	loopcheck(0);
    }
    *key = svkey;
    if(mousechar) *key = 2;

    *x = (((double)mousePos.h) - IND(off[0])) / IND(mul[0]);
    *y = (((double)(IND(devy)-mousePos.v)) - IND(off[1])) / IND(mul[1]);

    return 0;
}

static int _plx_mac_gtrack_on()
{
    return 0;
}

static int _plx_mac_gtrack(pkey, px, py)
     int *pkey;
     Coord *px, *py;
{
    GrafPtr savePort;
/*
 * Switch to graphics window.
 */
    GetPort(&savePort);
    SetPort(DEP(myGraphWindow));
/*
 * Where is the mouse?
 */
    doEvent(0);
    GetMouse(&mousePos);
    *px = (((double)mousePos.h) - IND(off[0])) / IND(mul[0]);
    *py = (((double)(IND(devy)-mousePos.v)) - IND(off[1])) / IND(mul[1]);
/*
 * Is the button depressed?
 */
    if(mousedown) *pkey = 1;
    else *pkey = 0;
/*
 * Restore the previously active viewport.
 */
    SetPort(savePort);
    return 0;
}

static void _plx_mac_gtrack_off()
{
    return;
}

static void _plx_mac_linemod(style)
     int style;
{
    return;
}

static void _plx_mac_closepl()
{
    return;
}


static void _plx_mac_rainbow(from, to)
     int from, to;
{
    return;
}

static void _plx_mac_greyscale(from, to)
     int from, to;
{
    return;
}

static int _plx_mac_openpl(size, fp)
     int size;
     FILE *fp;
{
    if(IND(open)) return 0;

    ShowWindow(DEP(myGraphWindow));
    SelectWindow(DEP(myGraphWindow));
    SetPort(DEP(myGraphWindow));

/* Set font size and style..... this is anybody's choice */

    TextFont(geneva);
    TextSize(10);
    GetFontInfo(&DEP(fontinfo));
/* 
 * Get rid of any events conected with opening the window 
 */
    doEvent(0);	
    return 0;
}

/* ARGSUSED */
static int _plx_mac_kopenp(size, user_ptr)
     int size;
     char *user_ptr;
{
    return _plx_mac_openpl(0, (FILE *)0);
}

static void _plx_mac_aspect(xsize,ysize)
     Coord *xsize,*ysize;
{
    *xsize = IND(devx);
    *ysize = IND(devx);

    return;
}

static void _plx_mac_raspect(xsize,ysize)
     Coord *xsize,*ysize;
{
    *xsize = IND(devx);
    *ysize = IND(devx);

    return;
}

static int _plx_mac_setval(option, val)
    char *option;
    long val;
{
    return -1;
}

static int _plx_mac_getval(option, pval)
    char *option;
    long *pval;
{
    if(!strcmp(option, "nlcolors")) {
	if(DEP(colorFlag)) *pval = 8;
	else *pval = 2;
    }
    else if(!strcmp(option, "nlstyles")) {
	*pval = 1;		/* 'oops!' */
    }
    else if(!strcmp(option, "nbuttons")) {
	*pval = 1;
    }
    else return -1;
	
    return 0;
}

static int _plx_mac_setstr(option, str)
     char *option, *str;
{
    return -1;
}

static int _plx_mac_getstr(option, str, maxlen)
     char *option, *str;
     int maxlen;
{
    return -1;
}

static void _plx_mac_usendplot()                     
{                                       
    return;
}

static void _plx_mac_sendplot()
{   
    return;
}

static void _plx_mac_asendplot()
{
    return;
}

static int _plx_mac_plothwm()
{
    return 0;
}

/*
 * Assign the function pointers that describe a particular driver.
 */
static int set_fn_ptrs(ptr)
     struct _plx_ind *ptr;
{
    if(ptr == (struct _plx_ind *)0) return -1;

    ptr->setclip    = _plx_mac_setclip;
    ptr->endclip    = _plx_mac_endclip;
    ptr->closepl    = _plx_mac_closepl;
    ptr->color      = _plx_mac_color;
    ptr->cont       = _plx_mac_cont;
    ptr->dotext     = _plx_mac_dotext;
    ptr->intext     = _plx_mac_intext;
    ptr->sztext     = _plx_mac_sztext;
    ptr->erase      = _plx_mac_erase;
    ptr->gin        = _plx_mac_gin;
    ptr->gtrack_on  = _plx_mac_gtrack_on;
    ptr->gtrack     = _plx_mac_gtrack;
    ptr->gtrack_off = _plx_mac_gtrack_off;
    ptr->aspect     = _plx_mac_aspect;
    ptr->raspect    = _plx_mac_raspect;
    ptr->sendplot   = _plx_mac_sendplot;
    ptr->asendplot  = _plx_mac_asendplot;
    ptr->usendplot  = _plx_mac_usendplot;
    ptr->plothwm    = _plx_mac_plothwm;
    ptr->point      = _plx_mac_point;
    ptr->linemod    = _plx_mac_linemod;
    ptr->move       = _plx_mac_move;
    ptr->openpl     = _plx_mac_openpl;
    ptr->kopenp     = _plx_mac_kopenp;
    ptr->rainbow    = _plx_mac_rainbow;
    ptr->greyscale  = _plx_mac_greyscale;
    ptr->initpanel  = _plx_mac_initpanel;
    ptr->panelpoint = _plx_mac_panelpoint;
    ptr->endpanel   = _plx_mac_endpanel;
    ptr->polgn      = 0;
    ptr->box        = 0;
    ptr->setval     = _plx_mac_setval;
    ptr->getval     = _plx_mac_getval;
    ptr->setstr     = _plx_mac_setstr;
    ptr->getstr     = _plx_mac_getstr;
    ptr->setcurs    = 0;
    return 0;
}

/*
 * This routine serves a VERY IMPORTANT purpose. It must allocate
 * sapce for the nexly created deivces "dependent" and "independent"
 * structures (or allocate them statically) and fill in any entries
 * that must have sensible defaults.
 */

static struct _plx_ind macdev;
static struct _plx_dep macdep;
/* 
 * Because of the horrible interaction between the Mac PLOTIX
 * and the "shell" that surrounds applications we can only
 * allow this function to be called once.
 */
static int mac_init = 0;

int _plx_mac_ctor(pdev)
     struct _plx_ind **pdev;
{
    if(mac_init) return -1;	
    else mac_init = 1;
	
    *pdev = &macdev;
    _plx_curdev = &macdev;
    IND(devdep) = &macdep;

    set_fn_ptrs(&macdev);
/*
 * Set hardware dependent deafults.
 */
    DEP(forecol) = blackColor;
    DEP(plxcols[0]) = whiteColor;
    DEP(plxcols[1]) = blackColor;
    DEP(plxcols[2]) = redColor;
    DEP(plxcols[3]) = greenColor;
    DEP(plxcols[4]) = blueColor;
    DEP(plxcols[5]) = cyanColor;
    DEP(plxcols[6]) = magentaColor;
    DEP(plxcols[7]) = yellowColor;
/*
 * ... and hardware independent defaults.
 */
    IND(devx) = DEV_x;
    IND(devy) = DEV_y;
    return 0;
}
