/*******
* menu *                
********

Author: J.W.Flower, ParaSoft Corporation, June 1989
        R.Salvador, ParaSoft Corporation, Feb  1990

This file contains the routines that manipulate the "menu" system.
They're rather primitive but still worthwhile. Of
course, on workstations, we can use the canned primitives .......
.................."Dream on !!!!! "
*/
#include <stdio.h>
#include "plotix.h"

static float cur_menht;
static double menu_left, menu_right, menu_top, menu_bottom;
static double sv_left, sv_right, sv_top, sv_bottom;
static int menu_style = MENU_SIMPLE;

/*
 * Need to save the state of a pushed button in order to
 * "unpush" it on the next trip. This requires that we "stack" all
 * the active menus in order to get all the buttons on the
 * way back up the tree.
 */
#define MAXMENDEPTH	(4)
typedef struct _menu_stack {
    PLXMENUPTR menu;
    int num_icons;
    int pushed_slot;
    unsigned long highlights;
} MENUSTACK;

static MENUSTACK menu_stack[MAXMENDEPTH];
static int cur_mendep = -1;
static MENUSTACK *cur_menu = (MENUSTACK *)0; 
		/* Menu on display at the moment */

static void draw_button();
void draw_menu();
static void fill_menu(), menu_clear();

/*
 * This function sets up the area on the display where the menus
 * will be drawn.
 */
int init_menu(style, left, bottom, right, top)
     int style;
     Coord left, bottom, right, top;
{
    sv_left = left;
    sv_bottom = bottom;
    sv_right = right;
    sv_top = top;

    menu_left   = left + 0.02*(right-left);
    menu_bottom = bottom + 0.02*(top-bottom);
    menu_right  = right - 0.02*(right-left);
    menu_top    = top - 0.02*(top-bottom);
    menu_style  = style;
    if(menu_style == MENU_SHADED) greyscale(100,110);

    return 0;
}

void push_menu(menu, num_icons)
     PLXMENUPTR menu;
     int num_icons;
{
    cur_mendep++;
    if(cur_mendep >= MAXMENDEPTH) {
	_ex_fatal("Ran out of menu depth!!!\n");
	exit(0);
    }
    cur_menu = &menu_stack[cur_mendep];
    cur_menu->menu = menu;
    cur_menu->num_icons = num_icons;
    cur_menu->pushed_slot = -1;
    cur_menu->highlights = 0;
    draw_menu();
    return;
}

/*
 * This code returns one level through the menu hierarchy.
 */
void pop_menu(redraw)
     int redraw;
{
    PLXMENUPTR pslot;
    if(cur_mendep < 0) {
	_ex_fatal("Popped off top of menu stack\n");
	exit(0);
    }
    cur_mendep--;
    cur_menu = &menu_stack[cur_mendep];
    if(redraw) draw_menu();
    return;
}

/* 
 * This is the basic menu processing function. It takes a MENU as
 * argument and sits forever processing it.... simple really.
*/
int process_menu(miss_func)
     int (*miss_func)();
{
    double x, y;
    long _ex_getnum();
    int done, key, icon, stat,save;
    PLXMENU *pmen;

    if(cur_menu == (MENUSTACK *)0) return 0;

    if(cur_menu->pushed_slot >= 0) {
	pmen = cur_menu->menu + cur_menu->pushed_slot;
	fill_menu(cur_menu->pushed_slot, pmen, 0);
	sendplot();
	cur_menu->pushed_slot = -1;
    }
/* 
 * We've drawn the boxes, let's go into gin mode and select a function to
 * execute. Don't allow selections that aren't within the menu area...
 * unless the user has supplied a "miss_func" in which case call it
 * with whatever "gin" returns.
 */
    done = 0;
    while(!done) {              /* Loop forever */
#ifdef NO_GIN
	_ex_print("Option ? : ");
	icon = _ex_getnum();
#else
	gin(&key, &x, &y);
	icon = (int) ((menu_top - y)/cur_menht);

	if(x < menu_left || x > menu_right || 
	   icon < 0 || icon >= cur_menu->num_icons) {
	    if(miss_func) {
#ifdef OLD_FN_CALLS
		return (*miss_func)(key, x, y);
#else
		return miss_func(key, x, y);
#endif
	    }
	    else continue;
	}
#endif
/*
 * Draw the button in its "pushed" state after saving its location.
 */
        pmen = cur_menu->menu + icon;
	cur_menu->pushed_slot = icon;
	fill_menu(icon, pmen, 1);
	sendplot();
#ifdef OLD_FN_CALLS
	stat = (*pmen->func)(icon);
#else
	stat = pmen->func(icon);
#endif

	done = 1;
    }
    return stat;
}

void draw_menu()
{
    int i;
#ifdef NO_GIN
    char text[64];
#endif

    menu_clear();

    if(cur_menu == (MENUSTACK *)0) return;

    if(cur_menu->num_icons > 12) 
	cur_menht = .98*(float)(menu_top-menu_bottom)/
		(float)cur_menu->num_icons;
    else cur_menht = .98 * (float) (menu_top-menu_bottom) / 12.0;
/* 
 * Draw boxes with their text. Start at the top and work down 
 */
    for(i=0; i<cur_menu->num_icons; i++) {
#ifdef NO_GIN
	sprintf(text,"%3d. %-20s ", i, cur_menu[i].icon);
	_ex_print(text);
	if(((i+1) % 3) == 0) _ex_print("\n");
#endif
	fill_menu(i, &cur_menu->menu[i], 0);
    }
#ifdef NO_GIN
    if((cur_icons % 3) != 0) _ex_print("\n");
#endif
    sendplot();
    return;
}

/*
 * Somehow indicate that the menu has been clicked and has a new "value".
 * This is "style" dependent.
 */
static void add_highlight(slot, pmen)
     int slot;
     PLXMENUPTR pmen;
{
    double dx, lcheck, rcheck, bcheck, tcheck, shade_x;

    color(1);

    switch(menu_style) {
    case MENU_SIMPLE:
    default:
	lcheck = menu_left + .85*(menu_right-menu_left);
	rcheck = menu_left + .94*(menu_right-menu_left);
	bcheck = menu_top-(slot+1)*cur_menht+.27*cur_menht;
	tcheck = menu_top-slot*cur_menht-.19*cur_menht;
	break;

    case MENU_3D:		/* Place a "check" mark in the box */
	shade_x = 0.02*(menu_right - menu_left);
	lcheck = menu_left + .90*(menu_right-shade_x-menu_left);
	rcheck = menu_left + .97*(menu_right-shade_x-menu_left);
	bcheck = menu_top-(slot+1)*cur_menht+.27*cur_menht;
	tcheck = menu_top-slot*cur_menht-.19*cur_menht;
	break;

    case MENU_SHADED:
	dx = .08*cur_menht;
	lcheck = menu_left + .87*(menu_right-dx-menu_left);
	rcheck = menu_left + .93*(menu_right-dx-menu_left);
	bcheck = menu_top-(slot+1)*cur_menht+3.3*dx;
	tcheck = menu_top-slot*cur_menht-3.3*dx;
	break;
    }
/*
 * Now draw the check mark on the box.
 */
    move(lcheck, bcheck+.2*(tcheck-bcheck));
    cont(lcheck+.2*(rcheck-lcheck), bcheck);
    cont(rcheck, tcheck);
    sendplot();
    return;
}

void check_menu(slot, pmen)
     int slot;
     PLXMENUPTR pmen;
{
    if(slot < 32) cur_menu->highlights |= (((unsigned long)1)<<slot);
    add_highlight(slot, pmen);
    sendplot();
    return;
}

/*
 * Undo the action of "check".
 */
void uncheck_menu(slot, pmen)
     int slot;
     PLXMENUPTR pmen;
{
    if(slot < 32) cur_menu->highlights &= ~(((unsigned long)1) <<slot);
    fill_menu(slot, pmen, 0);
    sendplot();
    return;
}

static void fill_menu(slot, pmen, pushed)
     int slot;                  /* Which slot on the screen */
     PLXMENUPTR pmen;           /* Menu item to fill in */
     int pushed;		/* Is the button pushed or not? */
{
    double shade_x;

    color(1);

    switch(menu_style) {
    case MENU_SIMPLE:
    default:
        if(pmen->icon != (char *) 0) {
            box(1.01*menu_left, menu_top-slot*cur_menht,
    	        .99*menu_right, menu_top-(slot+1)*cur_menht, pushed, 1);
            move(1.01*menu_left, menu_top-slot*cur_menht);
            cont(.99*menu_right, menu_top-slot*cur_menht);
	    dotext(pmen->icon, menu_left+.1*(menu_right-menu_left),
		menu_top-slot*cur_menht-.5*cur_menht,0, -1, 0);
	}
	break;

    case MENU_3D:
        if(pmen->icon != (char *)0) {
            shade_x = 0.02*(menu_right - menu_left);
            if(pushed) {
	        color(0);
	        box(menu_left, menu_top-(slot+1)*cur_menht+.08*cur_menht,
		    menu_right-shade_x, menu_top-slot*cur_menht, 0, 1);
	        color(1);
	        box(menu_left+shade_x, menu_top-(slot+1)*cur_menht,
	            menu_right, menu_top-slot*cur_menht, 0, 1);
            }
            else {
                box(menu_left+shade_x, menu_top-(slot+1)*cur_menht,
		    menu_right, menu_top-slot*cur_menht-.08*cur_menht, 1, 1);
	        box(menu_left, menu_top-(slot+1)*cur_menht+.08*cur_menht,
		    menu_right-shade_x, menu_top-slot*cur_menht, 0, 1);
            }
	    dotext(pmen->icon, 0.5*(menu_left+menu_right),
	        menu_top-slot*cur_menht-0.5*cur_menht, 0, 0, 0);
	}
	break;

    case MENU_SHADED:
        draw_button(menu_left, menu_top-(slot+1)*cur_menht,
              menu_left + .95*(menu_right-menu_left),  menu_top-slot*cur_menht,
             pmen->icon, 0, 0, pushed);
	break;
    }
/*
 * Look for a possible "highlight"
 */
    if(slot < 32 && (cur_menu->highlights & (1<<slot)))
	add_highlight(slot, pmen);
/*
 * Don't "sendplot" here because we want the menu to come out fast.
 */
    return;
}

/* 
 * Make the menu disappear - trickier than it looks because of
 * the finite width of the lines. So...... ? 
 */
static void menu_clear()
{
    box(sv_left,sv_bottom,sv_right,sv_top,0,0);
    sendplot();
    return;
}

/*
 * Draw a single "cute" menu button.
 */
static void draw_button(xi, yi, xf, yf, text, hjust, vjust, on)
     double xi, yi, xf, yf;
     char *text;
     int hjust, vjust, on;
{
    int coff = 103;
    float dx = .08*(yf-yi);
 
    if(text == (char *)0) return;
 
    box(xi+dx, yi+dx, xf-dx, yf-dx, 101, 1);
 
    if (on)
        initpanel(0, 1);
    else   
        initpanel(coff, 1); /* S */
    panelpoint(xi, yi);
    panelpoint(xi+dx, yi+dx);
    panelpoint(xf-dx, yi+dx);
    panelpoint(xf, yi);
    endpanel();
  
    if (on)
        initpanel(coff, 1);
    else   
        initpanel(0, 1); /* W */
    panelpoint(xi, yi);
    panelpoint(xi, yf);
    panelpoint(xi+dx, yf-dx);
    panelpoint(xi+dx, yi+dx);
    endpanel();
 
    if (on)
        initpanel(coff, 1);
    else   
        initpanel(0, 1); /* N */
    panelpoint(xi, yf);
    panelpoint(xf, yf);
    panelpoint(xf-dx, yf-dx);
    panelpoint(xi+dx, yf-dx);
    endpanel();
     
    if (on)
        initpanel(0, 1);
    else   
        initpanel(coff, 1); /* E */
    panelpoint(xf, yf);
    panelpoint(xf, yi);
    panelpoint(xf-dx, yi+dx);
    panelpoint(xf-dx, yf-dx);
    endpanel();
  
    if(text != (char *)0)
        dotext(text, (xi+xf)/2., (yi+yf)/2., 0, hjust, vjust);
 
    return;
}
