/**********
* mswdriv *		Copyright ParaSoft Corporation, 1990
***********

Device driver for the MicroSoft Windows implementation of PLOTIX.
*/
#include <stdio.h>
#include "plotix.h"
#include "sys/plotix.h"
#include "mswin.h"

#ifndef NOSCROLL
extern HWND oldhWnd;    /* defined in OPENPL.C */
#endif

#ifndef NOSHELL
extern char *abb;       		/* defined in MSSHELL */
extern int _g_plotix,_g_spawnwin;	/* defined in MSSHELL */
extern HANDLE ghInstance;		/* defined in MSSHELL */
HWND oldhWnd;			/* Used to remember Parent window */
#else
int _g_spawnwin = 0;	/* no window spawning without MSSHELL */
int _g_plotix = 0;
#endif

#define X_to_USER(x)  (((Coord) (x) - IND(off[0])) / IND(mul[0]))
#define Y_to_USER(y)  ((IND(devy) - IND(off[1]) - (Coord) (y))/IND(mul[1]))

extern int _fallout,_button,_convert;
extern int _g_mx,_g_my, _exitok;
extern Coord _g_devx, _g_devy;

void exit(), _ex_info();
void wait_for_user_event();
int _ex_error();
/* 
 * This routine initializes the hDCCur, and sets the current viewport,
 * line styles, and background brushes, as needed.
 * The calling routine is responsible for finishing with the command
       _closewin()  
 */
static int wopen = 0;

static void _openwin() 
{
    int l,b,r,t; /* Clipping coordinates left,bottom,right,top */

    if (wopen)       /* Already open, do nothing */
        return;

    wopen = 1;
    DEP(hDCCur) = GetDC(hWndCur);
    if (IND(clip)) {
        l = (int) USER_to_X(IND(vptab[IND(cvport)].clipwin[0]));
        b = (int) USER_to_Y(IND(vptab[IND(cvport)].clipwin[1]));
        r = (int) USER_to_X(IND(vptab[IND(cvport)].clipwin[2]));
        t = (int) USER_to_Y(IND(vptab[IND(cvport)].clipwin[3]));
    }
    else {
        l = (int) (IND(vptab[IND(cvport)].V0[0]) * IND(devx)); /* Convert */
        b = (int) ( ((Coord) 1 - IND(vptab[IND(cvport)].V0[1])) * IND(devy));
        r = (int) (IND(vptab[IND(cvport)].V1[0]) * IND(devx));
        t = (int) ( ((Coord) 1 - IND(vptab[IND(cvport)].V1[1])) * IND(devy));
    }

    DEP(hRgn) = CreateRectRgn(l,t,r,b);      /* Do rect clipping */
    SelectClipRgn(DEP(hDCCur),DEP(hRgn));
    DEP(hPen) = CreatePen(DEP(linestyle),0, DEP(plxmap[DEP(fgcolor)]) );
    /* 0: Solid  1:  Longdashed  2:  Dotted  3:  dotdashed 4: dotdashdash */
    if (DEP(hPen) != NULL)
        SelectObject(DEP(hDCCur),DEP(hPen));
    SetBkMode(DEP(hDCCur),TRANSPARENT); /* Good for all but Filling */
}

static void _closewin() 
{
    if (wopen) {
        ReleaseDC(hWndCur,DEP(hDCCur));
        if (DEP(hPen) != NULL)
	    if (DeleteObject(DEP(hPen)) == 0)
		if (_ex_error("Unable to delete Pen"))
		    exit(1);
	if (DeleteObject(DEP(hRgn)) == 0)
	    if (_ex_error("Unable to delete clipping Region"))
		exit(1);
        wopen = 0;
    }
}

/*
 * The clipping is taken care of by the _open/_closewin combination.
 */
static void _plx_msw_setclip(xl, yb, xr, yt)
     Coord xl, yb, xr, yt;
{
    IND(clip) = 1;
    return;
}

static void _plx_msw_endclip()
{
    IND(clip) = 0;
    return;
}

static void _plx_msw_cont(x,y)
     Coord x,y;
{
    int newx,newy;

    _openwin();

    newx = (int) USER_to_X(IND(xold));
    newy = (int) USER_to_Y(IND(yold));
    MoveTo(DEP(hDCCur), newx, newy);

    newx = (int) USER_to_X(x);
    newy = (int) USER_to_Y(y);
    LineTo(DEP(hDCCur),newx,newy);

    _closewin();
}

static void _plx_msw_move(x,y)
     Coord x,y;
{
    return;	/* Have to do this in cont(). */
}

#define NHATCH          (6)
static int short hatch[NHATCH]={
    HS_BDIAGONAL,HS_CROSS,HS_DIAGCROSS,
    HS_FDIAGONAL,HS_HORIZONTAL,HS_VERTICAL };

static void _plx_msw_initpanel(shade,edge)
     int shade,edge;
{
    DEP(sv_shade) = shade;
    DEP(sv_edge)  = edge;
    DEP(polys) = DEP(polyverts);
    DEP(Npts) = 0;
    return;
}

static void _plx_msw_panelpoint(x, y)
     Coord x, y;
{
    int x_curr, y_curr;
    char text[80];

    x_curr = (int) USER_to_X(x);
    y_curr = (int) USER_to_Y(y);

    if(++DEP(Npts) > MAXVERTS) {
        sprintf(text,"Too many polygon vertices, max = %d", MAXVERTS);
        if (_ex_error(text))
            exit(1);
        DEP(Npts) = 0;
	DEP(polys) = DEP(polyverts);
        return;
    }
    *DEP(polys)++ = x_curr;
    *DEP(polys)++ = y_curr;

    return;
}

static POINT polyarray[MAXVERTS];

static void _plx_msw_endpanel()
{
    int i, shade, index, r, g, b;
    double h,f,n;       /* For rainbows! */
    char text[80];
    HBRUSH hBrush;
    HPEN tPen;

    /* First we have to close the polygon, if necessary */

    if(*(DEP(polys)-2) != DEP(polyverts[0]) || 
       *(DEP(polys)-1) != DEP(polyverts[1])) {
        if(++DEP(Npts) > MAXVERTS) {
            sprintf(text,"Too many polygon vertices, max = %d", MAXVERTS);
            if (_ex_error(text)) exit(1);
            DEP(Npts) = 0;
            DEP(polys) = DEP(polyverts);
	    return;
        }
        *DEP(polys)++ = DEP(polyverts[0]);        /* Close polygon for filler */
        *DEP(polys)++ = DEP(polyverts[1]);
    }

    _openwin();
    SetBkMode(DEP(hDCCur),OPAQUE);
    /* Now select a Brush with the fill pattern and color */
    if(DEP(sv_shade) < 0) {
        index = -DEP(sv_shade)-1;
        shade = (index - (index % NHATCH))/NHATCH;
        shade %= 7;
        shade++;
	while (shade > 7) shade -= 7;
        hBrush = CreateHatchBrush(hatch[index % NHATCH], DEP(plxmap[shade]));
    }
    else if(DEP(sv_shade) < 8) {
        hBrush = CreateSolidBrush(DEP(plxmap[DEP(sv_shade)]));
    }
    else if ( IND(grey_lo) <= DEP(sv_shade) && DEP(sv_shade) <= IND(grey_hi)) {
        i = (int) ( (float) ( NGREYS * (DEP(sv_shade) - IND(grey_lo)) ) / 
			(float) (IND(grey_hi) - IND(grey_lo)));
        hBrush = CreateSolidBrush( RGB(i,i,i) );
    }
    else if (IND(rain_lo) <= DEP(sv_shade) && DEP(sv_shade) <= IND(rain_hi)) {
							/* ==> Rainbow */
        h = (double) (IND(rain_hi) - DEP(sv_shade)) / 
			(double) (IND(rain_hi) - IND(rain_lo));
        if (h > .999)  h = 0.;
        h *= 6.;
        i = (int) h;
        f = h - (double) i;
        n = 1. - f;
        switch(i) {
        case 0: r = 255;           g = (int) (f * 255); b = 0;
            break;
        case 1: r = (int) (n*255); g = 255;             b = 0;
            break;
        case 2: r = 0;             g = 255;             b = (int) (f*255);
            break;
        case 3: r = 0;             g = (int) (n * 255); b = 255;
            break;
        case 4: r = (int) (f*255); g = 0;               b = 255;
            break;
        case 5: r = 255;           g = 0;               b = (int) (n*255);
            break;
        }
        hBrush = CreateSolidBrush( RGB(r,g,b) );
    }
    else {      /* Undefined color --> map into original 8 */
        hBrush = CreateSolidBrush(DEP(plxmap[DEP(sv_shade) % 8]));
    }

    SelectObject(DEP(hDCCur), hBrush);
    SetPolyFillMode(DEP(hDCCur),WINDING);

    /* convert formats */
    for (i = 0; i < DEP(Npts); i++) {
        polyarray[i].x = DEP(polyverts[i << 1]);
        polyarray[i].y = DEP(polyverts[(i << 1) + 1]);
    }
    Polygon(DEP(hDCCur),polyarray,DEP(Npts));
    if(!DEP(sv_edge)) {
        tPen = CreatePen(0, 0, DEP(plxmap[0]));
        SelectObject(DEP(hDCCur),tPen);
        if (DEP(hPen) != NULL)
           if (DeleteObject(DEP(hPen)) == 0)
               if (_ex_error("Unable to delete Pen"))
                   exit(1);
        DEP(hPen) = tPen;
        Polyline(DEP(hDCCur),polyarray,DEP(Npts));
    }
/*  If changed later to no _closewin, add SetBkMode(DEP(hDCCur),TRANSPARENT); */
    _closewin();
    if (DeleteObject(hBrush) == 0)
        if (_ex_error("Unable to delete Brush!"))
            exit(1);
    DEP(Npts) = 0;
    DEP(polys) = DEP(polyverts);
}

static void _plx_msw_point(x,y)
     Coord x,y;
{
    int newx, newy;
    _openwin();
    newx = (int) USER_to_X(x);
    newy = (int) USER_to_Y(y);
    SetPixel(DEP(hDCCur),(int) newx, (int) newy, DEP(plxmap[DEP(fgcolor)]));
    _closewin();
}

static void _plx_msw_aspect(xsize,ysize)
     Coord *xsize,*ysize;
{
    *xsize = IND(devx); /** hDC.right - hDC.left **/
    *ysize = IND(devy); /** hDC.bottom - hDC.top **/
    return;
}

/* WORRY -jwf- */
static void _plx_msw_raspect(xsize,ysize)
     Coord *xsize,*ysize;
{
    *xsize = IND(devx); /** hDC.right - hDC.left **/
    *ysize = IND(devy); /** hDC.bottom - hDC.top **/
    return;
}


static void _plx_msw_closepl()
{
    if (_g_plotix) {
        sendplot();
	_closewin();
#ifndef NOSCROLL
	if (_g_plotix != 1) {	/* We've spawned a window */
	    _g_plotix = 0;		/* Needed for below    --\  */
	    DestroyWindow(hWndCur);     /*                  <----/  */
            hWndCur = oldhWnd;
	} else {        /* Redraw scroll bars, etc. */
#else
	if (1) {
#endif
            _g_plotix = 0;      /* need for updatewindow() */
	    /* SIZENORMAL is a blatant lie, but seems to work fine */
            SendMessage(hWndCur,WM_SIZE,SIZENORMAL /* ? poss. prob. ?*/,
                     ((long) IND(devx) & ((long) IND(devy) << 16)) );
            InvalidateRect(hWndCur,NULL,TRUE);
	    UpdateWindow(hWndCur);
        }
    }
}

static void _plx_msw_color(shade)
     int shade;
{
    DEP(fgcolor) = shade % 8;
}

int strlen(char *);

/* #define _TRIAL_  ** Doesn't work yet 
*	it's purpose is to allow special fonts, and more particularly,
*	fonts pointing in the Plotix directions.
*/

static void _plx_msw_dotext(string, x, y, angle, ljust, hjust)
     char *string;
     Coord x, y;
     int ljust, hjust, angle;
{
    int x_curr,y_curr,high,wide;
    int length;
    DWORD extent;
#ifdef _TRIAL_
    LOGFONT logfont;
    HFONT hFont;
#endif

    _openwin();

#ifdef _TRIAL_
/*  logfont.lfEscapement = (short) (angle * 10);
    logfont.lfOrientation = (short) (angle * 10); 
    logfont.lfStrikeOut = 1; */
    hFont = CreateFontIndirect(&logfont);
    SelectObject(DEP(hDCCur),hFont);
#endif

    x_curr = (int) USER_to_X(x);
    y_curr = (int) USER_to_Y(y);
    length = strlen(string);
    extent = GetTextExtent(DEP(hDCCur),string,length);
    high = HIWORD(extent);
    wide = LOWORD(extent);

    if (ljust == 0)     x_curr -= (wide >> 1);
    else if (ljust > 0) x_curr -= wide;

    if (hjust == 0)     y_curr -= (high >> 1);
    else if (hjust < 0) y_curr -= high;

    TextOut(DEP(hDCCur),x_curr,y_curr,string,length);

    _closewin();

#ifdef _TRIAL_
    DeleteObject(hFont);
#endif
}

static void _plx_msw_sztext(str, pwd, pht)
     char *str;
     Coord *pwd, *pht;
{
    DWORD extent;
    int high,wide;
    
    _openwin();
    extent = GetTextExtent(DEP(hDCCur), str, strlen(str));

    high = HIWORD(extent);
    wide = LOWORD(extent);
    *pwd = (double)high/IND(mul[0]);
    *pht = (double)wide/IND(mul[1]);

    _closewin();
    return;
}

#ifdef DBG_INTEXT
static char give_str[] = "ab\177cdefghijklmnopqrstuvwxyz";
static char *give_ptr = give_str;
#endif

static int poll_getchar()
{
    int c;

#ifndef DBG_INTEXT
    _button = 0;
    while(_button < 32  && _button != '\r' && _button != '\n' && _button != 8) 
	wait_for_user_event();
    return _button;
#else
    return *give_ptr++;
#endif
}

static void _plx_msw_intext(prompt, x, y, term, input, maxlen)
     char *prompt;
     Coord x, y;
     char *term, *input;
     int maxlen;
{
    Coord in_x;
    int high, wide, at_x, at_y;
    char c, *at_ptr, *strchr();
    int got = 0;
    DWORD extent;

#ifdef DBG_INTEXT
    give_ptr = give_str;
#endif

    memset(input, 0, maxlen);
    dotext(prompt, x, y, 0, -1, -1);

    _openwin();
    extent = GetTextExtent(DEP(hDCCur),prompt,strlen(prompt));
    high = HIWORD(extent);
    wide = LOWORD(extent);
    _closewin();

    in_x = x + ((double)wide)/IND(mul[0]);
    at_x = USER_to_X(in_x);
    at_y = USER_to_Y(y);

    at_ptr = input;
    do {
	c = (char) poll_getchar();
	if(c == '\r') c = '\n';

	if((c == 8 || c == 127) && got > 0) {	/* CTRL-H or DEL */
    	    RECT temp;
            HBRUSH hBrush;

            _openwin();
	    extent = GetTextExtent(DEP(hDCCur),input,strlen(input));
	    high = HIWORD(extent);
	    wide = LOWORD(extent);
	    temp.top = at_y - high;
	    temp.left = at_x;
	    temp.right = (int) at_x + wide;
	    temp.bottom = (int) at_y;
   	    hBrush = GetStockObject(WHITE_BRUSH);
	    FillRect(DEP(hDCCur), &temp, hBrush);
            _closewin();
	    *--at_ptr = '\0';
	    got--;
	}
	else if(c >= 32) {		/* Only ASCII, please! */
	    *at_ptr++ = c;
	    got++;
	}
	dotext(input, in_x, y, 0, -1, -1);
	sendplot();
    } while(got < maxlen && strchr(term, c) == (char *)0);
    *at_ptr = '\0';
    return;
}

static void _plx_msw_erase()
{
    HBRUSH hBrush;
    RECT temp;

    _openwin();
    hBrush = GetStockObject( WHITE_BRUSH );
    temp.top = 0;
    temp.left = 0;
    temp.right = (int) IND(devx);
    temp.bottom = (int) IND(devy);
    FillRect(DEP(hDCCur), &temp, hBrush);
    _closewin();
    return;
}

static int _plx_msw_gin(pbutton, px, py)
    int *pbutton;
    Coord *px,*py;
{
    wait_for_user_event();
    if (_convert) {
        DEP(mouse_x) = X_to_USER(_g_mx);
        DEP(mouse_y) = Y_to_USER(_g_my);
        _convert = 0;
    }
    *pbutton = _button;
    *px = DEP(mouse_x);
    *py = DEP(mouse_y);

    return 0;
}

static int _plx_msw_gtrack_on()
{
    return 0;
}

/*
 * These variables are set within the main window processing loop,
*	which is found in the shells\mswin directory.
 */
extern int _g_msatx, _g_msaty, _g_msbutmask;

static int _plx_msw_gtrack(pkey, px, py)
     int *pkey;
     Coord *px, *py;
{
    *pkey = _g_msbutmask;
    *px = X_to_USER(_g_msatx);
    *py = Y_to_USER(_g_msaty);
    return 0;
}

static void _plx_msw_gtrack_off()
{
    return;
}

void wait_for_user_event() 
{
    MSG msg;
    int stat;

    _closewin();
    do {
        if (_fallout == 3) {
	    IND(devx) = _g_devx;
	    IND(devy) = _g_devy;
            setvport(IND(cvport));
	}
        _fallout = 0;
        while ( !_fallout && (stat = GetMessage(&msg,NULL,0,0) ) != 0 ) {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    } while (_fallout == 3);
    if (!stat) {                /* ==> Program has been terminated */
	if (!_exitok)
	    _ex_info("User prematurely ended program.");
	exit();
    }
    return;
}

/*
* This routines looks for a waiting event, and processes 1 event if found.
* The return value specifies whether an event was processed.  Thus a
* call of (while(_plx_msw_loopcheck());) will empty a waiting event queue.
* Additionaly, resize messages are processed PROMPTLY and accurately,
* i.e., in the case of a resize message, we immediately process the waiting
* REPAINT message, after changing coordinate systems.  The loopcheck found
* in mswloop skips this step, and MUST NOT BE USED WHILE PLOTIX IS OPENED!!
*/
int _plx_msw_loopcheck() 
{
    MSG msg;
    int stat;

    _closewin();
    do {
        if (_fallout == 3 && _g_plotix != 0) {
	    IND(devx) = _g_devx;
	    IND(devy) = _g_devy;
            setvport(IND(cvport));
	}
        _fallout = 0;
	if ((stat = PeekMessage(&msg,NULL,0,0,PM_REMOVE) ) != 0) {
	    if (msg.message == WM_QUIT) {
	        if (!_exitok)
	            _ex_info("User prematurely ended program.");
	        exit();
	    }
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    } while (_fallout == 3);
    return stat;
}

static void _plx_msw_linemod(style)
     int style;
{
    DEP(linestyle) = style % NUMLIN;   
	/* Setting occurs in function _openwin() */
    return;
}

extern int _g_textx,_g_texty;

static int _plx_msw_openpl(bufsiz, grafp)
     int bufsiz;
     FILE *grafp;
{
/*
fprintf(stdaux,"openpl: plotix=%d, spawn=%d\r\n", _g_plotix, _g_spawnwin);
*/
    if (!_g_plotix) {
    /* Now look for a mouse on the system.  If there is, assume two buttons */
	if (GetSystemMetrics(SM_MOUSEPRESENT))
	    DEP(mouse) = 2;   /* Assume two buttons, as MS Windows does */
	_g_plotix = _g_spawnwin+1;
#ifndef NOSHELL
        if (_g_spawnwin) {
	    char temp[20];

            oldhWnd = hWndCur;
            sprintf(temp,"%sPLOTIX",abb);
            hWndCur = CreateWindow(temp, "PLOTIX",WS_CAPTION |
		WS_THICKFRAME | WS_OVERLAPPED | WS_MAXIMIZEBOX |
		WS_MINIMIZEBOX,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,
		CW_USEDEFAULT,NULL /* oldhWnd */,NULL,ghInstance,NULL);
            if (hWndCur == NULL) {
                if (_ex_error("Unable to open PLOTIX!!"))
                    exit(0);
		hWndCur = oldhWnd;
		_g_plotix = 0;
                return -1;
            }
            ShowWindow(hWndCur, SW_SHOWNORMAL);
            UpdateWindow(hWndCur);
	} 
	else {
#else
	if (1) {
#endif
            SetScrollRange(hWndCur,SB_VERT,0,0,TRUE);   /* Remove scroll bars */
            SetScrollRange(hWndCur,SB_HORZ,0,0,TRUE);
	    ValidateRect(hWndCur,NULL);
            _g_devx = (Coord) _g_textx; _g_devy = (Coord)  _g_texty;
	}
	if(_g_devx == (Coord) 0 || _g_devy == (Coord) 0) {
	    _ex_fatal("Window size zero");
	    exit(0);
	} else {
	    IND(devx) = _g_devx;
	    IND(devy) = _g_devy;
	}
        exregidle(_plx_msw_loopcheck);
    }
    return 1;
}

/* ARGSUSED */
static int _plx_msw_kopenp(bufsiz, user_ptr)
     int bufsiz;
     char *user_ptr;
{
    return _plx_msw_openpl(0, (FILE *)0);
}

extern int _g_msw_cursor;
extern HCURSOR _ex_msw_arrow,_ex_msw_hourglass;

static int _plx_msw_setcurs(which)
    int which;
{
    switch(which) {
    case PLX_HRG_CURSOR:
	_g_msw_cursor = 1;
	SetCursor(_ex_msw_hourglass);
	break;
    case PLX_PTR_CURSOR:
	_g_msw_cursor = 0;
	SetCursor(_ex_msw_arrow);
	break;
    default:
	return -1;
    }
    return 1;
}

static  void _plx_msw_greyscale(from, to)
    int from, to;
{
    return;
}

static void _plx_msw_rainbow(from, to)
     int from, to;
{
    return;
}

static int _plx_msw_setval(option, val)
     char *option;
     long val;
{
    return -1;
}

static int _plx_msw_getval(option, pval)
     char *option;
     long *pval;
{
    int strcmp();

    if(!strcmp(option, "nlcolors")) {
	*pval = 8;
	return 0;
    }
    else if(!strcmp(option, "nlstyles")) {
	*pval = NUMLIN;
	return 0;
    }
    else if(!strcmp(option, "nbuttons")) {
	*pval = 0;		/* WORRY -jwf- */
	return 0;
    }
    return -1;
}

static int _plx_msw_setstr(option, str)
     char *option, *str;
{
    return -1;
}

static int _plx_msw_getstr(option, str, maxlen)
     char *option, *str;
     int maxlen;
{
    return -1;
}

static void _plx_msw_sendplot()
{
    while(_plx_msw_loopcheck());
}

static void _plx_msw_usendplot()
{
    sendplot();
}

static void _plx_msw_asendplot()
{
    sendplot();
}

static int _plx_msw_plothwm()
{
    return 0;
}

/*
 * Assign the function pointers that describe a particular driver.
 */
static int set_fn_ptrs(ptr)
     struct _plx_ind *ptr;
{
    if(ptr == (struct _plx_ind *)0) return -1;

    ptr->setclip    = _plx_msw_setclip;
    ptr->endclip    = _plx_msw_endclip;
    ptr->closepl    = _plx_msw_closepl;
    ptr->color      = _plx_msw_color;
    ptr->cont       = _plx_msw_cont;
    ptr->dotext     = _plx_msw_dotext;
    ptr->intext     = _plx_msw_intext;
    ptr->sztext     = _plx_msw_sztext;
    ptr->erase      = _plx_msw_erase;
    ptr->gin        = _plx_msw_gin;
    ptr->gtrack_on  = _plx_msw_gtrack_on;
    ptr->gtrack     = _plx_msw_gtrack;
    ptr->gtrack_off = _plx_msw_gtrack_off;
    ptr->aspect     = _plx_msw_aspect;
    ptr->raspect    = _plx_msw_raspect;
    ptr->sendplot   = _plx_msw_sendplot;
    ptr->asendplot  = _plx_msw_asendplot;
    ptr->usendplot  = _plx_msw_usendplot;
    ptr->plothwm    = _plx_msw_plothwm;
    ptr->point      = _plx_msw_point;
    ptr->linemod    = _plx_msw_linemod;
    ptr->move       = _plx_msw_move;
    ptr->openpl     = _plx_msw_openpl;
    ptr->kopenp     = _plx_msw_kopenp;
    ptr->rainbow    = _plx_msw_rainbow;
    ptr->greyscale  = _plx_msw_greyscale;
    ptr->initpanel  = _plx_msw_initpanel;
    ptr->panelpoint = _plx_msw_panelpoint;
    ptr->endpanel   = _plx_msw_endpanel;
    ptr->polgn      = 0;
    ptr->box        = 0;
    ptr->setval     = _plx_msw_setval;
    ptr->getval     = _plx_msw_getval;
    ptr->setstr     = _plx_msw_setstr;
    ptr->getstr     = _plx_msw_getstr;
    ptr->setcurs    = _plx_msw_setcurs;
    return 0;
}

/*
 * This routine serves a VERY IMPORTANT purpose. It must allocate
 * space for the newly created device's "dependent" and "independent"
 * structures (or allocate them statically) and fill in any entries
 * that must have sensible defaults.
 */

static struct _plx_ind mswdev;
static struct _plx_dep mswdep;
/* 
 * Because of the horrible interaction between the Mac PLOTIX
 * and the "shell" that surrounds applications we can only
 * allow this function to be called once.
 */

int _plx_msw_ctor(pdev)
     struct _plx_ind **pdev;
{
    static int msw_init = 0;

    if(msw_init) return -1;	
    else msw_init = 1;

    *pdev = &mswdev;
    _plx_curdev = &mswdev;
    IND(devdep) = &mswdep;

    set_fn_ptrs(&mswdev);
/*
 * Set hardware dependent defaults.
 */
    DEP(fgcolor) = 1;
    DEP(mouse) = 0;
    DEP(plxmap[0]) = RGB( 255, 255, 255);          /* WHITE */
    DEP(plxmap[1]) = RGB(   0,   0,   0);          /* BLACK */
    DEP(plxmap[2]) = RGB( 255,   0,   0);          /* LIGHTRED */
    DEP(plxmap[3]) = RGB(   0, 255,   0);          /* LIGHTGREEN */
    DEP(plxmap[4]) = RGB(   0,   0, 255);          /* LIGHTBLUE */
    DEP(plxmap[5]) = RGB(   0, 255, 255);          /* LIGHTCYAN */
    DEP(plxmap[6]) = RGB( 255,   0, 255);          /* LIGHTMAGENTA */
    DEP(plxmap[7]) = RGB( 255, 255,   0);          /* YELLOW */
/*
 * ... and hardware independent defaults. The window size will
 * actually be set in a call-back function later on!!!
 */
#define DEV_x	(0)
#define DEV_y	(0)

    IND(devx) = (Coord) DEV_x;
    IND(devy) = (Coord) DEV_y;
    return 0;
}
