/**********
* necdriv *
***********

Driver code for the NEC implementation of PLOTIX.
*/
#include "plotix.h"
#include "sys/plotix.h"
#include "nec.h"
#include <dos.h>
#include <stdio.h>

static void _plx_nec_cont(x,y)
     Coord x,y;
{
    int newx, newy, oldx, oldy;

    oldx = USER_to_X(IND(xold));
    oldy = USER_to_Y(IND(yold));
    newx = USER_to_X(x);
    newy = USER_to_Y(y);

    ega_line(oldx, oldy, newx, newy);
    return;        
}

static void _plx_nec_move(x, y)
     Coord x, y;
{
    return;
}

static void _plx_nec_initpanel(color,edge)
     int color,edge;
{
    init_ylist();
    ega_state.polyedge = edge;
    DEP(save_col) = ega_state.color;
    DEP(panel_col) = color;
    DEP(first) = 1;
    return;
}

static void _plx_nec_panelpoint(x, y)
     Coord x, y;
{
    int x_curr, y_curr;

    x_curr = USER_to_X(x);
    y_curr = USER_to_Y(y);

    if(DEP(first)) {
	DEP(start_x) = x;
	DEP(start_y) = y;
	init_poly(DEP(panel_col), x_curr, y_curr);
	if(ega_state.polyedge) move(x, y);
	DEP(first) = 0;
    }
    else {
	sys_polyverts(x_curr, y_curr);
	if(ega_state.polyedge) cont(x, y);
    }
    return;
}

static void _plx_nec_endpanel()
{
    int x_curr, y_curr;

    if(ega_state.polyedge) cont(DEP(start_x), DEP(start_y));

    color(DEP(panel_col));
    x_curr = USER_to_X(DEP(start_x));
    y_curr = USER_to_Y(DEP(start_y));
    sys_polyverts(x_curr, y_curr);
    sys_polyfill();
    color(DEP(save_col));
    DEP(first) = 1;
}

static void _plx_nec_point(x,y)
     Coord x,y;
{
    Coord x_curr, y_curr;

    x_curr = USER_to_X(x);
    y_curr = USER_to_Y(y);

    ega_dot(x_curr, y_curr);
    return;
}

static void _plx_nec_aspect(xsize,ysize)
     Coord *xsize,*ysize;
{
    *xsize = 10.5;
    *ysize = 7.5;		/* in inches, of course !! */
    return;
}

static void _plx_nec_raspect(xsize, ysize)
     Coord *xsize, *ysize;
{
    *xsize = 640;
    *ysize = 350;
    return;
}

static void _plx_nec_closepl()
{
    ega_end();
    ega_close();
    return;
}

static void _plx_nec_greyscale(from, to)
     int from, to;
{
    ega_state.grey_lo = from;
    ega_state.grey_hi = to;
    return;
}

static void _plx_nec_rainbow(from, to)
     int from, to;
{
    ega_state.grey_lo = from;
    ega_state.grey_hi = to;
    return;
}

static void _plx_nec_color(shade)
     int shade;
{
    ega_state.color = shade;
    return;
}

static void _plx_nec_dotext(s,x,y, angle, hjust, vjust)
     char *s;
     Coord x,y;
     int angle, hjust, vjust;
{
    int x_curr,y_curr;
    int x_off, y_off, N = strlen(s);

    if(IND(softclip) && !_pointclip(x,y))  return;

    ega_state.cur_x = USER_to_X(x);
    ega_state.cur_y = USER_to_Y(y);

/* Now we have pixels. Cater for the alignment. */

    x_off = 0;					/* Chars are 8 pixels wide */
    if(hjust == 0) x_off = 4*N;
    else if(hjust > 0) x_off = 8*N;

    y_off = 0;
    if(vjust == 0) y_off = 7;	/* ... and 14 pixels high */
    if(vjust > 0)  y_off = 14;

    ega_state.cur_x -= x_off;
    ega_state.cur_y += y_off;	/* Y axis is inverted for EGA */

    writegtext(strlen(s), s);
}

static void _plx_nec_sztext(str, pwd, pht)
     char *str;
     Coord *pwd, *pht;
{
    *pwd = (double)(8*strlen(str))/IND(mul[0]);
    *pht = 14.0/IND(mul[1]);
    return;
}

static void _plx_nec_intext(prompt, x, y, term, input, maxlen)
     char *prompt;
     Coord x, y;
     char *term, *input;
     int maxlen;
{
    Coord in_x;
    int dummy, at_x, at_y;
    char c, *at_ptr, *strchr();
    int got = 0;
    char *str, *base, *_ex_malloc();

#ifdef DBG_INTEXT
    give_ptr = give_str;
#endif

    dotext(prompt, x, y, 0, -1, 0);

    base = (char*) _ex_malloc(sizeof(prompt) + maxlen + 1);
    memset(input, 0, maxlen);
    memset(base, 0, sizeof(prompt) + maxlen + 1);
    str = base + sizeof(prompt);
    at_ptr = input;

    do {
	c = getchar();
	if(c == '\r') c = '\n';

	if((c == 8 || c == 127) && got > 0) {	/* CTRL-H or DEL */
	    *--at_ptr = '\0';
	    *--str = '\0';
	    got--;
	}
	else if(c >= 32) {		/* Only ASCII, please! */
	    *at_ptr++ = c;
		*str++ = c;
	    got++;
	}

	dotext(base, x, y, 0, -1, 0);
	sendplot();

    } while(got < maxlen && strchr(term, c) == (char *)0);
    *at_ptr = '\0';
    _ex_free(base);
    return;
}

static void _plx_nec_erase()
{
    ega_erase();
    return;
}

static int _plx_nec_gin(key,x,y)
     int *key;
     Coord *x,*y;
{
    int x0, y0;

    if(DEP(mouse)) mouse_gin(key, &x0, &y0);
    else           ega_gin(key, &x0, &y0);
    y0 = IND(devy) - y0 - 1;		/* invert y-axis */

/* Decide whether the coordinates lie within this processor */

    *x = (((Coord)x0) - IND(off[0])) / IND(mul[0]);
    *y = (((Coord)y0) - IND(off[1])) / IND(mul[1]);

    return 0;
}

static int _plx_nec_gtrack_on()
{    
    if(DEP(mouse)) return 0;
    else return -1;
/* NOTREACHED */
}

static int _plx_nec_gtrack(pkey, px, py)
     int *pkey;
     Coord *px, *py;
{
    union REGS regs;
    int x0, y0;
        
    if(DEP(mouse)) {

	kbhit();

	erase_gin();
	regs.x.ax = 0x03;

	int86(0x33, &regs, &regs);

	draw_gin(regs.x.cx, regs.x.dx);
 
        *pkey = 0;
        if(regs.x.ax != 0) *pkey |= 1;   /* Case of left button ON */
        if(regs.x.bx != 0) *pkey |= 2;

	x0 = regs.x.cx;
	y0 = IND(devy) - regs.x.dx - 1;

	*px = (((Coord)x0) - IND(off[0])) / IND(mul[0]);
	*py = (((Coord)y0) - IND(off[1])) / IND(mul[1]);
    }
    return 0;
}

static void _plx_nec_gtrack_off()
{
    return;
}

static void _plx_nec_linemod(style)
     int style;
{
    return;
}

static int _plx_nec_openpl(bufsiz, fp)
     int bufsiz;
     FILE *fp;
{
    union REGS regs;
    struct SREGS sregs;

    regs.h.ah = 0x0;
    regs.h.al = 0x10;
    int86(0x10, &regs, &regs);	/* Set video mode 10 */

    regs.x.ax = 0;
    int86(0x33, &regs, &regs);

    if(regs.x.ax != 0) DEP(mouse) = 2;
    else DEP(mouse) = 0;

    ega_start();
    return 0;
}

/* ARGSUSED */
static int _plx_nec_kopenp(bufsiz, user_ptr)
     int bufsiz;
     char *user_ptr;
{
    return _plx_nec_openpl(0, (FILE *)0);
}

static int _plx_nec_setval(option, val)
     char *option;
     long val;
{
    return -1;
}

static int _plx_nec_getval(option, pval)
     char *option;
     long *pval;
{
    if(!strcmp(option, "nlcolors")) {
	*pval = 8;
	return 0;
    }
    else if(!strcmp(option, "nlstyles")) {
	*pval = 1;
	return 0;
    }
    else if(!strcmp(option, "nbuttons")) {
	printf("I think there are %ld buttons\n", (long)DEP(mouse));
	*pval = DEP(mouse);
	return 0;
    }
    return -1;
}

static int _plx_nec_setstr(option, str)
     char *option, *str;
{
    return -1;
}

static int _plx_nec_getstr(option, str, maxlen)
     char *option, *str;
     int maxlen;
{
    return -1;
}

static int _plx_nec_plothwm()
{
    return 0;
}

static void _plx_nec_sendplot()
{
    return;
}

static void _plx_nec_usendplot()
{
    return;
}

static void _plx_nec_asendplot()
{
    return;
}

/*
 * Assign the function pointers that describe a particular driver.
 */
static int set_fn_ptrs(ptr)
     struct _plx_ind *ptr;
{
    if(ptr == (struct _plx_ind *)0) return -1;

    ptr->setclip    = 0;
    ptr->endclip    = 0;
    ptr->closepl    = _plx_nec_closepl;
    ptr->color      = _plx_nec_color;
    ptr->cont       = _plx_nec_cont;
    ptr->dotext     = _plx_nec_dotext;
    ptr->intext     = _plx_nec_intext;
    ptr->sztext     = _plx_nec_sztext;
    ptr->erase      = _plx_nec_erase;
    ptr->gin        = _plx_nec_gin;
    ptr->gtrack_on  = _plx_nec_gtrack_on;
    ptr->gtrack     = _plx_nec_gtrack;
    ptr->gtrack_off = _plx_nec_gtrack_off;
    ptr->aspect     = _plx_nec_aspect;
    ptr->raspect    = _plx_nec_raspect;
    ptr->sendplot   = _plx_nec_sendplot;
    ptr->asendplot  = _plx_nec_asendplot;
    ptr->usendplot  = _plx_nec_usendplot;
    ptr->plothwm    = _plx_nec_plothwm;
    ptr->point      = _plx_nec_point;
    ptr->linemod    = _plx_nec_linemod;
    ptr->move       = _plx_nec_move;
    ptr->openpl     = _plx_nec_openpl;
    ptr->kopenp     = _plx_nec_kopenp;
    ptr->rainbow    = _plx_nec_rainbow;
    ptr->greyscale  = _plx_nec_greyscale;
    ptr->initpanel  = _plx_nec_initpanel;
    ptr->panelpoint = _plx_nec_panelpoint;
    ptr->endpanel   = _plx_nec_endpanel;
    ptr->polgn      = 0;
    ptr->box        = 0;
    ptr->setval     = _plx_nec_setval;
    ptr->getval     = _plx_nec_getval;
    ptr->setstr     = _plx_nec_setstr;
    ptr->getstr     = _plx_nec_getstr;
    ptr->setcurs    = 0;
    return 0;
}
 
/*
 * This routine serves a VERY IMPORTANT purpose. It must allocate
 * sapce for the nexly created deivces "dependent" and "independent"
 * structures (or allocate them statically) and fill in any entries
 * that must have sensible defaults.
 */
/*
 * Make an instance of the NEC-window
 */
static struct _plx_ind necdev;
static struct _plx_dep necdep;

int _plx_nec_ctor(pdev)
     struct _plx_ind **pdev;
{
    *pdev = &necdev;
    _plx_curdev = &necdev;
    IND(devdep) = &necdep;
 
    set_fn_ptrs(&necdev);
/*
 * Set defaults for hardware dependent parameters.
 */
    DEP(mouse) = 0;               /* Number of mouse buttons */

    IND(devx) = DEV_x;
    IND(devy) = DEV_y;
    return 0;
}
