/***********
* nodedriv *
************

PLOTIX driver for the nodes of an Express system.
*/
#include <stdio.h>
#include "sys/config.h"
#include "plotix.h"
#include "sys/plotix.h"
#include "sys/meta.h"
#include "express.h"
#include "sys/typedef.h"
#include "sys/eltix.h"
#include "sys/msgtype.h"
#include "plxnode.h"

/* 
 * This doesn't need to check for overflow because we've already done
 * it in chkpad.
 */
static void gputch(c)
     int c;
{
    DEP(buffr[DEP(bufptr)++]) = c;
}

static void gputint(val)	
		/* Just spit out two bytes, least significant first */
     int val;
{
    gputch(val & 0xff);
    gputch((val >> 8) & 0xff);
}

/* 
 * Reserve space in the PLOTIX buffer for a command containing nchars 
 * bytes and nwords integer parameters. Returns -1 if something didn't
 * work (e.g., no space, label too long) and 0 otherwise. Remember that
 * the buffer size is already a multiple of PLX_QUANTUM.
 */
static int chkpad(nchars, nwords)
     int nchars, nwords;
{
    int left, length;

    if(DEP(bufptr) >= DEP(bufsiz)) return -1;

    length = nchars + 2*nwords;
    left   = PLX_QUANTUM - (DEP(bufptr) % PLX_QUANTUM);

    if(length <= left) return 0;
    else if((DEP(bufsiz)-DEP(bufptr)) < PLX_QUANTUM) return -1;
    else {
	_ex_memset(DEP(buffr)+DEP(bufptr), PL_NOOP, left);
	DEP(bufptr) += left;
	return 0;
    }
/* NOTREACHED */
}
#define static
static void _plx_nod_usendplot()                     
{                                       
    if(DEP(bufptr) > DEP(bufhwm)) DEP(bufhwm) = DEP(bufptr);

    if(DEP(plotfp) != (FILE *)0) {
#ifdef meta_file_support
	if(!ismulti(DEP(plotfp))) fmulti(DEP(plotfp));
	fwrite(DEP(buffr),DEP(bufptr), 1, DEP(plotfp));
#endif
    }
    else {
	gplxfilt(PL_UPLOT, DEP(buffr), DEP(bufptr), (char *)0, 0);
    }
    DEP(bufptr)=0;
}

static void _plx_nod_sendplot()
{       
    if(DEP(bufptr) > DEP(bufhwm)) DEP(bufhwm) = DEP(bufptr);

    if(DEP(plotfp) != (FILE *)0) {
#ifdef meta_file_support
	if(ismulti(DEP(plotfp)) || isasync(DEP(plotfp))) fsingl(DEP(plotfp));
	fwrite(DEP(buffr),DEP(bufptr),1,DEP(plotfp));
#endif
    }
    else {
	gplxfilt(PL_PLOT, DEP(buffr), DEP(bufptr), (char *)0, 0);
    }
    DEP(bufptr)=0;
}

static void _plx_nod_asendplot()
{
    if(DEP(bufptr) > DEP(bufhwm)) DEP(bufhwm) = DEP(bufptr);

    if(DEP(plotfp) != (FILE *)0) {
#ifdef meta_file_support
	if(!isasync(DEP(plotfp))) fasync(DEP(plotfp));
	fwrite(DEP(buffr), DEP(bufptr), 1, DEP(plotfp));
#endif
    }
    else {
	IND(_plx_mode) = PLX_ASYNC;
	gplxfilt(PL_APLOT, DEP(buffr), DEP(bufptr), (char *)0, 0);
    }
    DEP(bufptr) = 0;
}

static int _plx_nod_plothwm()
{
    if(DEP(bufptr) > DEP(bufhwm)) return DEP(bufptr);
    else                          return DEP(bufhwm);
}

static void _plx_nod_cont(x,y)
     Coord x,y;
{
    int newx, newy;

    newx = USER_to_X(x);
    newy = USER_to_Y(y);

    if(chkpad(1, 2) == 0) {
	gputch(PL_CONT);
	gputint(newx);
	gputint(newy);
    }

    return;        
}

static void _plx_nod_move(x,y)
     Coord x,y;
{
    int newx, newy;

    newx = USER_to_X(x);
    newy = USER_to_Y(y);

    if(chkpad(1, 2) == 0) {
	gputch(PL_MOVE);
	gputint(newx);
	gputint(newy);
    }
    return;        
}

static void _plx_nod_initpanel(color,edge)
     int color,edge;
{
    if(chkpad(1, 2) == 0) {
        gputch(PL_INITP);
        gputint(color);
        gputint(edge);
    }
}

static void _plx_nod_panelpoint(x, y)
     Coord x, y;
{
    int newx, newy;

    newx = USER_to_X(x);
    newy = USER_to_Y(y);

    if(chkpad(1, 2) == 0) {
        gputch(PL_PANELP);
        gputint(newx);
        gputint(newy);
    }
}

static void _plx_nod_endpanel()
{
    if(chkpad(1,0) == 0) gputch(PL_ENDP);
}

static void _plx_nod_point(x,y)
     Coord x,y;
{
    int newx, newy;

    newx = USER_to_X(x);
    newy = USER_to_Y(y);

    if(chkpad(1,2) == 0) {
	gputch(PL_POINT);
	gputint(newx);
	gputint(newy);
    }
}

static void _plx_nod_closepl()
{
    gplxfilt(PL_CLOSE, (char *)0, 0, (char *)0, 0);

    if(DEP(bufsiz) != 0 && DEP(buffr) != (char *)0) 
	free(DEP(buffr));

    return;
}

static void _plx_nod_greyscale(from, to)
     int from, to;
{
    if(chkpad(1, 2) == 0) {
	gputch(PL_GREY);
	gputint((int)from);
	gputint((int)to);
    }
    return;
}

static void _plx_nod_rainbow(from, to)
     int from, to;
{
    if(chkpad(1, 2) == 0) {
	gputch(PL_RAIN);
	gputint((int)from);
	gputint((int)to);
    }
    return;
}

static void _plx_nod_color(shade)
     int shade;
{
    if(chkpad(1,1) == 0) {
        gputch(PL_COLOR);
        gputint(shade);
    }
    return;
}

static void _plx_nod_dotext(s,x,y,angle, hjust, vjust)
     char *s;
     Coord x,y;
     int angle;
     int hjust, vjust;
{
    int newx, newy;
    int N;
    char *p;

    newx = USER_to_X(x);
    newy = USER_to_Y(y);

    p = s;		/* Avoid calling strlen becuase of library clashes */
    N = 0;
    while(*p++) N++;

    if(chkpad(1+1+N, 5) == 0) {
	gputch(PL_TEXT);
	gputint(newx);
	gputint(newy);
	gputint((int)angle);
	gputint(hjust);
	gputint(vjust);

	do {
	    gputch(*s);
	} while(*s++ != '\0');
    }
    return;
}

static void _plx_nod_sztext(str, pwd, pht)
     char *str;
     Coord *pwd, *pht;
{
    int stat;
    int vals[2];

    stat = gplxfilt(PL_SZTEXT, (char *)str, strlen(str)+1, (char *)vals, 0);
				/* Nothing to go out, BUT read vals */

    *pwd = (double)vals[0] / IND(mul[0]);
    *pht = (double)vals[1] / IND(mul[1]);
    return;
}

/*
 * This is a "bitch" because the host needs so much info.
 * We try to pack it here but to do so requires space.....
 * Assume that we can use the official buffer - after all
 * we'vew probably just done a "sendplot" --- if not it's
 * now declared to be an error.
 */
static void _plx_nod_intext(prompt, x, y, term, input, maxlen)
     char *prompt;
     Coord x, y;
     char *term, *input;
     int maxlen;
{
    int newx, newy;
    char *p;
/*
 * Force the graphics buffer to be empty and then fill it with an "intext"
 * command.
 */
    DEP(bufptr) = 0;

    gputch(PL_INTEXT);
    newx = USER_to_X(x);
    newy = USER_to_Y(y);
    gputint(newx);
    gputint(newy);
/*
 * Put the strings in the buffer.
 */
    p = prompt;
    while(*p) gputch(*p++);
    gputch('\0');

    p = term;
    while(*p) gputch(*p++);
    gputch('\0');
    gputch(PL_PLOT);
/*
 * Cheat here and put in the last argument something that can
 * be used to get at the buffer coming back.
 */
    gplxfilt(PL_INTEXT, (char *)DEP(buffr), DEP(bufptr), input, maxlen);
    return;
}

static void _plx_nod_erase()
{
    gplxfilt(PL_ERASE, (char *)0, 0, (char *)0, 0);
			/* Call to cubix erase command */
    return;
}

static int _plx_nod_gin(key,x,y)
     int *key;
     Coord *x,*y;
{
    int stat;
    int vals[3];

    stat = gplxfilt(PL_GIN, (char *)0, 0, (char *)vals, 0);
				/* Nothing to go out, BUT read vals */

    if(stat < 0) return 0;		/* Operation failed */

/* Decide whether the coordinates lie within this processor */

    *key = vals[0];
    *x = (((double)vals[1]) - IND(off[0])) / IND(mul[0]);
    *y = (((double)vals[2]) - IND(off[1])) / IND(mul[1]);

    return 0;
}

static int _plx_nod_gtrack_on()
{
    return gplxfilt(PL_TRKON, (char *)0, 0, (char *)0, 0);	
}

static int _plx_nod_gtrack(pkey, px, py)
     int *pkey;
     Coord *px, *py;
{
    int stat;
    int vals[3];

    stat = gplxfilt(PL_TRACK, (char *)0, 0, (char *)vals, 0);
				/* Nothing to go out, BUT read vals */

    if(stat < 0) return 0;		/* Operation failed */

/* Decide whether the coordinates lie within this processor */

    *pkey = vals[0];
    *px = (((double)vals[1]) - IND(off[0])) / IND(mul[0]);
    *py = (((double)vals[2]) - IND(off[1])) / IND(mul[1]);

    return 0;
}

static void _plx_nod_gtrack_off()
{
    gplxfilt(PL_TRKOFF, (char *)0, 0, (char *)0, 0);
    return;
}

static void _plx_nod_linemod(style)
     int style;
{
    if(chkpad(1,1) == 0) {
	gputch(PL_LINEMOD);
	gputint(style);
    }
    return;
}

static int do_open()
{
    int stat, vals[3];

    stat=gplxfilt(PL_OPEN, (char *)0, 0, (char *)vals, 0);
				/* Nothing to go out, BUT read vals */
    if(stat >= 0) {
	DEP(hostdev_x) = vals[0];
	DEP(hostdev_y) = vals[1];
	stat = vals[2];
    }
    return stat;
}

static int _plx_nod_openpl(bufsiz, fp)
     int bufsiz;
     FILE *fp;
{
    char *calloc();
    DEP(plotfp) = fp;
/* 
 * First make sure that the CUBIX version does it's I/O into a large buffer,
 * i.e., our own buffr !! Make the buffer size be a multiple of
 * PLX_QUANTUM for the benefit of chkpad.
 */
    DEP(bufsiz) = (bufsiz > 0) ? bufsiz : 8192;
    DEP(bufsiz) = ((DEP(bufsiz)+PLX_QUANTUM-1)/PLX_QUANTUM)*PLX_QUANTUM;
    DEP(buffr) = (char *)calloc(1, DEP(bufsiz));

    if(DEP(buffr) == (char *)0) return -1; 
					/* Asked for too much ..... fail !! */
    return do_open();
}

static int _plx_nod_kopenp(size, buffer)
     int size;
     char *buffer;
{
    DEP(plotfp) = (FILE *)0;
/* 
 * First make sure that we do I/O into the user buffer.
 * Make the buffer size be a multiple of PLX_QUANTUM for the benefit 
 * of _g_chkpad. Fail if we're too small.
 */
    if(size < PLX_QUANTUM) return -1;
    DEP(bufsiz) = ((size)/PLX_QUANTUM)*PLX_QUANTUM;
    DEP(buffr) = buffer;

    return do_open();
}

static void _plx_nod_aspect(xsize,ysize)
     Coord *xsize,*ysize;
{
    *xsize = DEP(hostdev_x);
    *ysize = DEP(hostdev_y);

    return;
}

static void _plx_nod_raspect(xsize, ysize)
     Coord *xsize, *ysize;
{
    *xsize = DEP(hostdev_x);
    *ysize = DEP(hostdev_y);

    return;
}


static int _plx_nod_setval(option, val)
     char *option;
     long val;
{
    return gplxfilt(PL_SETOPT, option, val, (char *)0, 0);
}

static int _plx_nod_getval(option, pval)
     char *option;
     long *pval;
{
    return gplxfilt(PL_GETOPT, option, (long)0, pval, 0);
}

static int _plx_nod_setstr(option, str)
     char *option, *str;
{
    return gplxfilt(PL_SETSTR, option, (long)0, str, 0);
}

static int _plx_nod_getstr(option, str, maxlen)
     char *option, *str;
     int maxlen;
{
    return gplxfilt(PL_GETSTR, option, (long)0, str, (long)maxlen);
}

static int gplx_procnum;
static void gplx_setup(), send_to_plot();
static int gplx_init = 0;

static int gplxfilt(option, buffer, olength, reply, ilength)
     int option;		/* PLOTIX command option */
     char *buffer;		/* Buffer containing data to plot */
     int olength;		/* Number of bytes of data */
     char *reply;		/* Data back from the host */
     int ilength;		/* Amount of data to get back */
{
    register char *p;
    int display, type, *ibuf, mode;
    cbx32_t dummy;

    if(gplx_init == 0) gplx_setup();

    p = cmdp;
    setpacket(p, cmd_t, CBX_PLOT);
    setpacket(p, cbx32_t, option);
    setpacket(p, cbx32_t, _xnproc);

/*
 * Now pack any additional data into the buffer.
 */
    if(option == PL_SWDISP) {
	setpacket(p, cbx32_t, olength);
    }
    if(option == PL_SETOPT) {
	setpacket(p, cbx32_t, olength);
	setstring(p, buffer);
    }
    if(option == PL_GETOPT || option == PL_GETSTR || 
       option == PL_MKDISP || option == PL_SZTEXT) {
	setstring(p, buffer);
    }
    if(option == PL_SETSTR) {
	setstring(p, buffer);
	setstring(p, reply);
    }
    if(option == PL_UPLOT) {
	p += expackdom(p);		/* Send domain info too */
    }
    if(option == PL_INTEXT) {
	while(olength--) *p++ = *buffer++;
    }
/*
 * Now send the request to the host.
 */
    mode = (IND(_plx_mode) == PLX_SYNC) ? cbx_syncmode : MODE_ASYNC;
    IND(_plx_mode) = PLX_SYNC;	/* For next time round */

    _output(cmdp, (int)(p-cmdp), mode, cbx_display);	/* Send request */
/*
 * At this point deal with any additional messages which might
 * need to come back.
 */
    switch(option) {
    case PL_TRKON:
    case PL_TRKOFF:
    case PL_TRACK:
    case PL_OPEN:
    case PL_CLOSE:
    case PL_ERASE:
    case PL_GIN:
    case PL_SETOPT:
    case PL_GETOPT:
    case PL_INTEXT:
	break;

    case PL_PLOT:
	if(gplx_procnum == 0) send_to_plot(buffer, olength);
	break;

    case PL_UPLOT:
	type = xservmode(CBX_IOTYPE, cbx_display);
	display = cbx_display;
	exread(&dummy, sizeof(dummy), &display, &type);
	send_to_plot(buffer, olength);
	break;

    case PL_APLOT:
	type = xservmode(CBX_IOTYPE, cbx_display);
	display = cbx_display;
	exread(&dummy, sizeof(dummy), &display, &type);
	send_to_plot(buffer, olength);
	break;
    }

    _input(ackp, PK_SIZE_, mode, cbx_display);	/* Get system ACK */

/* Process acknowlegement */

    p = ackp;
    if(syserror(p)) {
	errno = xpacket(p, error_t);
	return -1;
    }
    else {
	ibuf = (int *)reply;
	switch(option) {

	case PL_SZTEXT:
	    ibuf[0] = xpacket(p, cbx32_t);
	    ibuf[1] = xpacket(p, cbx32_t);
	    break;

	case PL_GIN:
	case PL_TRACK:
	    ibuf[0] = xpacket(p, cbx32_t);
	    ibuf[1] = xpacket(p, cbx32_t);
	    ibuf[2] = xpacket(p, cbx32_t);
	    break;

	case PL_OPEN:
	    ibuf[0] = xpacket(p, cbx32_t);
	    ibuf[1] = xpacket(p, cbx32_t);
	    ibuf[2] = xpacket(p, cbx32_t);
	    break;

	case PL_SWDISP:
	case PL_MKDISP:
	case PL_GETOPT:
	    ibuf[0] = xpacket(p, cbx32_t);
	    break;

	case PL_GETSTR:		/* Copy up to "ilength" chars back */
	case PL_INTEXT:
	case PL_LSDISP:
	    while(--ilength && *p) *reply++ = *p++;
	    *reply = '\0';
	    break;
	}
    }
    return 0;
}

static void gplx_setup()
{
    struct nodenv ndata;

    exparam(&ndata);
    gplx_procnum = ndata.procnum;
    gplx_init = 1;
    return;
}

/* This function is responsible for sending out the data to the display
   to be plotted. We packetise ourselves and also use types to say when
   we're done. Eventually this should be converted to a simple 
   double-buffered strategy.
*/
static void send_to_plot(buffer, length)
     char *buffer;
     int length;
{
    int send_len, display = cbx_display, type;

    if(length == 0) {
	type = xservmode(CBX_IODONE, cbx_display);
	exwrite(buffer, 0, &display, &type);
    }
    else {
	while(length > 0) {
	    if(length > PLX_QUANTUM) {
		send_len = PLX_QUANTUM;
		type = xservmode(CBX_IOTYPE, cbx_display);
	    }
	    else {
		send_len = length;
		type = xservmode(CBX_IODONE, cbx_display);
	    }
	    exwrite(buffer, send_len, &display, &type);
	    buffer += send_len;
	    length -= send_len;
	}
    }
    return;
}

/*
 * The functions from here down are for non-local graphics
 * - i.e., those operations in which the graphics output 
 * device is not directly attached to this host.
 *
 * They are predominantly used in the nodes.
 */
int _plx_remote_mkdisp(choice)
     char *choice;
{
    int vals[8];

    if(choice == (char *)0) choice = "";
    gplxfilt(PL_MKDISP, (char *)choice, 0, (char *)vals, 0);
    return vals[0];
}

int _plx_remote_swdisp(which)
     int which;
{
    int vals[8];

    gplxfilt(PL_SWDISP, (char *)0, which, (char *)vals, 0);
    return vals[0];
}

void _plx_remote_lsdisp(buffer)
     char *buffer;
{
    gplxfilt(PL_LSDISP, (char *)0, 0, (char *)buffer, 64);
}

/*
 * Assign the function pointers that describe a particular driver.
 */
static int set_fn_ptrs(ptr)
     struct _plx_ind *ptr;
{
    if(ptr == (struct _plx_ind *)0) return -1;

    ptr->setclip    = 0;
    ptr->endclip    = 0;
    ptr->closepl    = _plx_nod_closepl;
    ptr->color      = _plx_nod_color;
    ptr->cont       = _plx_nod_cont;
    ptr->dotext     = _plx_nod_dotext;
    ptr->intext     = _plx_nod_intext;
    ptr->sztext     = _plx_nod_sztext;
    ptr->erase      = _plx_nod_erase;
    ptr->gin        = _plx_nod_gin;
    ptr->gtrack_on  = _plx_nod_gtrack_on;
    ptr->gtrack     = _plx_nod_gtrack;
    ptr->gtrack_off = _plx_nod_gtrack_off;
    ptr->aspect     = _plx_nod_aspect;
    ptr->raspect    = _plx_nod_raspect;
    ptr->sendplot   = _plx_nod_sendplot;
    ptr->asendplot  = _plx_nod_asendplot;
    ptr->usendplot  = _plx_nod_usendplot;
    ptr->plothwm    = _plx_nod_plothwm;
    ptr->point      = _plx_nod_point;
    ptr->linemod    = _plx_nod_linemod;
    ptr->move       = _plx_nod_move;
    ptr->openpl     = _plx_nod_openpl;
    ptr->kopenp     = _plx_nod_kopenp;
    ptr->rainbow    = _plx_nod_rainbow;
    ptr->greyscale  = _plx_nod_greyscale;
    ptr->initpanel  = _plx_nod_initpanel;
    ptr->panelpoint = _plx_nod_panelpoint;
    ptr->endpanel   = _plx_nod_endpanel;
    ptr->polgn      = 0;
    ptr->box        = 0;
    ptr->setval     = _plx_nod_setval;
    ptr->getval     = _plx_nod_getval;
    ptr->setstr     = _plx_nod_setstr;
    ptr->getstr     = _plx_nod_getstr;
    ptr->setcurs    = 0;
    return 0;
}

/*
 * This routine serves a VERY IMPORTANT purpose. It must allocate
 * sapce for the nexly created deivces "dependent" and "independent"
 * structures (or allocate them statically) and fill in any entries
 * that must have sensible defaults.
 */
/*
 * HOW SHOULD THIS BE INVOKED? It's either portable (and always called
 * mk_inst (sic!) or it isn't in which case the user has to know
 * what to call.
 */
static struct _plx_ind inds;
static struct _plx_dep indd;

int _plx_nod_ctor(pdev)
     struct _plx_ind **pdev;
{
    extern char *calloc();
    struct _plx_ind *noddev = &inds;	
    struct _plx_dep *noddep = &indd;

#if 1
    noddev = (struct _plx_ind *)calloc(1, sizeof(*noddev));
    noddep = (struct _plx_dep *)calloc(1, sizeof(*noddep));
    if(noddep == (struct _plx_dep *)0 || noddev == (struct _plx_ind *)0)
	return -1;
#endif

    *pdev = noddev;
    _plx_curdev = noddev;
    IND(devdep) = noddep;

    set_fn_ptrs(noddev);
/*
 * Set sensible defaults for the new device.
 */
#define DEV_x	(65535)
#define DEV_y	(65535)
    IND(devx) = DEV_x;
    IND(devy) = DEV_y;

    IND(_plx_mode) = PLX_SYNC;	/* sync/async mode flag: */
    DEP(bufsiz) = 0;		/* Size of I/O buffer. */
    DEP(bufptr) = 0;		/* Pointer to top of data */
    DEP(bufhwm) = 0;		/* High water mark in I/O buffer */

    return 0;
}
