/*********
* plotix *              J.W.Flower,  ParaSoft Corp., Feb 1988
**********

Header file for the various system parameters used in the PLOTIX
system.
*/

#ifndef _sysPLOTIXdotH
#define _sysPLOTIXdotH
/*
 * You MUST include "plotix.h" before this file in order to get the
 * Coord typedef.
 */
struct _plx_ind {

    Coord off[2], mul[2];       /* Coord transform factors */
    Coord xold, yold;           /* Point of last "move" */
    Coord devx, devy;           /* Hardware resolution */

    int rain_lo, rain_hi;       /* Low/hi range for "rainbow" */
    int grey_lo, grey_hi;       /* Low/hi range for "greyscale" */

    int _plx_mode;              /* (A)Synchronous mode flag */
    int open;                   /* Has openpl been called? */
    int clip, softclip;         /* Hardware and software clipping states */
    int curs_index;		/* Current cursor index */
/*
 * Manipuations for viewports.
 */
#define NVPORT  (8)             /* Max number of active vports */
    struct _vptab {
        Coord V0[2], V1[2];         /* Viewport position */
        Coord W0[2], W1[2];         /* User coordinate system */
        int clip;                   /* Clipping state */
        Coord clipwin[4];           /* Cliiping window */
    } vptab[NVPORT];

    int vpinited;        /* Flag: Have we set the vports ? */
    int cvport;          /* Currently active vport */
    int nvport;          /* Number of active veiewports */
/*
 * Now come the function pointers which represent the driver
 * functions.
 */
    void (*setclip)();
    void (*endclip)();
    void (*closepl)();
    void (*color)();
    void (*cont)();
    void (*intext)();
    void (*dotext)();
    void (*sztext)();
    void (*erase)();
    int  (*gin)();
    int  (*gtrack_on)();
    int  (*gtrack)();
    void (*gtrack_off)();
    void (*aspect)();
    void (*raspect)();
    void (*sendplot)();
    void (*asendplot)();
    void (*usendplot)();
    int  (*plothwm)();
    void (*point)();
    void (*linemod)();
    void (*move)();
    int  (*openpl)();
    int  (*kopenp)();
    void (*rainbow)();
    void (*greyscale)();
    void (*initpanel)();
    void (*panelpoint)();
    void (*endpanel)();
    void (*polgn)();
    void (*box)();
    int  (*setval)();
    int  (*getval)();
    int  (*setstr)();
    int  (*getstr)();
    int  (*setcurs)();
/*
 * Any hardware dependent parameters are contained in this 
 * structure whose definition varies from device
 * to device.
 */
    void *devdep;
};

/*
 * This variable contains a pointer to the current graphics device.
 * It is the only "visible" handle to the system.
 */
extern struct _plx_ind *_plx_curdev;   

/*
 * These macros are used to extract/store things from/into the structures.
 */
#define IND(field)      ((_plx_curdev)->field)
#define DEP(field)      (((struct _plx_dep *)(_plx_curdev->devdep))->field)

#ifndef OLD_FN_CALLS
#define INVOKE(fn)	_plx_curdev->fn
#else
#define INVOKE(fn)	(*_plx_curdev->fn)
#endif
/*
 * Mode flags used internally.
 */
#define PLX_SYNC                (0)             /* Synchronous mode command */
#define PLX_ASYNC               (1)             /* Asynchronous mode command */

/* 
 * The nodes and the host deal in quanta of size .... 
 */
#define PLX_QUANTUM     (1024)
#endif				/* _sysPLOTIXdotH */
