/****************************************************/
/*    DIME: alloc.c                                 */
/*    This code used to allocate and free           */
/*    the various structures used by DIME.          */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/
#define DIMECODE
#include "dime.h"
int ntype, untype, etype, uetype, btype, ubtype, atype, nntype;

NODEPTR node_alloc()
/*
	Allocate memory for a new node, with nneigh neighnode structures attached
	in a list. The new node is also put in the circular list of all nodes.
*/
{
	NODEPTR this;
	int in;
	static int first = 1;
	if(first){
		ntype = init_smalloc(sizeof(_NODE));
		untype = init_smalloc(uns);
		first = 0;
#ifdef DEBUG
	fprintf(stdsys, "node type %d size %d\n", ntype, sizeof(_NODE));
	fprintf(stdsys, "node user type %d size %d\n", untype, uns);
#endif
	}
	this = (NODEPTR)smalloc(ntype);
	if(this == (NODEPTR)NULL)
		abort(100);					/* malloc failed for node */
#ifdef DEBUG
printf(">>> node = %x, %d rem 8 = %d\n", this, this, ((long)this)%8);
#endif
	this->b = (BDYPTR)NULL;
	this->n = (NODENEIGHPTR)NULL;
	if(the_nodes == (NODEPTR)NULL)
		the_nodes = this->last = this->next = this;
	else {
		this->next = the_nodes;
		this->last = the_nodes->last;
		the_nodes->last->next = this;
		the_nodes->last = this;
	}

	if(uns > 0)
		if((this->user = smalloc(untype)) == (char *)NULL)
			abort(101);				/* malloc failed for user_node */
	return(this);
}

NODENEIGHPTR nodeneigh_alloc()
/*
	Allocate space for the structure which points to a elmt from a node.
*/
{
	NODENEIGHPTR n;
	static int first = 1;
	if(first){
		nntype = init_smalloc(sizeof(NODENEIGH));
		first = 0;
#ifdef DEBUG
	fprintf(stdsys, "nodeneigh type %d size %d\n", nntype, sizeof(NODENEIGH));
#endif
	}
	n = (NODENEIGHPTR )smalloc(nntype);
	if(n == (NODENEIGHPTR)NULL)
		abort(102);					/* malloc failed for nodeneigh */
	return(n);
}

BDYPTR b_alloc()
/*
	Allocate space for the structure which points to a elmt from a node.
*/
{
	BDYPTR n;
	static int first = 1;
	if(first){
		btype = init_smalloc(sizeof(_BDY));
		ubtype = init_smalloc(ubs);
		first = 0;
#ifdef DEBUG
	fprintf(stdsys, "bdy type %d size %d\n", btype, sizeof(_BDY));
	fprintf(stdsys, "bdy user type %d size %d\n", ubtype, ubs);
#endif
	}
	n = (BDYPTR )smalloc(btype);
	if(n == (BDYPTR)NULL)
		abort(103);					/* malloc failed for bdy */
	n->alias = (ALIASPTR)NULL;
	n->oppalias = (ALIASPTR)NULL;
	n->is_on_bdy = 0;
	if(ubs > 0)
		if((n->user = smalloc(ubtype)) == (char *)NULL)
			abort(104);				/* malloc failed for bdy_user */
	return(n);
}

ELMTPTR elmt_alloc()
/*
	Allocate space for a elmt, and put it in the circular list of all elmts
*/
{
	ELMTPTR this;
	static int first = 1;
	if(first){
		etype = init_smalloc(sizeof(_ELMT));
		uetype = init_smalloc(ues);
		first = 0;
#ifdef DEBUG
	fprintf(stdsys, "elmt type %d size %d\n", etype, sizeof(_ELMT));
	fprintf(stdsys, "elmt user type %d size %d\n", uetype, ues);
#endif
	}
	this = (ELMTPTR )smalloc(etype);
	if(this == (ELMTPTR)NULL)
		abort(105);					/* malloc failed for elmt */
	if(the_elmts == (ELMTPTR)NULL)
		the_elmts = this->last = this->next = this;
	else {
		this->next = the_elmts;
		this->last = the_elmts->last;
		the_elmts->last->next = this;
		the_elmts->last = this;
	}
	if(ues > 0)
		if((this->user = smalloc(uetype)) == (char *)NULL)
			abort(106);				/* malloc failed for user_elmt */
	return(this);
}

ALIASPTR alias_alloc()
/*
	Allocate space for an alias-structure
*/
{
	ALIASPTR new;
	static int first = 1;
	if(first){
		atype = init_smalloc(sizeof(ALIAS));
		first = 0;
#ifdef DEBUG
	fprintf(stdsys, "alias type %d size %d\n", atype, sizeof(ALIAS));
#endif
	}
	new = (ALIASPTR )smalloc(atype);
	if(new == (ALIASPTR)NULL)
		abort(107);					/* malloc failed for alias */
	return(new);
}

freemesh()
/*
	Frees everything associated with the current mesh and sets the_nodes and
	the_elmts to NULL.
*/
{
	NODEPTR node;
	ELMTPTR elmt;
	if((node = the_nodes) != (NODEPTR)NULL){
		node = node->next;
		do {
			freenode(node);
		} while((node = node->next) != the_nodes);
		freenode(node);
	}
	if((elmt = the_elmts) != (ELMTPTR)NULL){
		elmt = elmt->next;
		do {
			freeelmt(elmt);
		} while((elmt = elmt->next) != the_elmts);
		freeelmt(elmt);
	}
	the_nodes = (NODEPTR)NULL;
	the_elmts = (ELMTPTR)NULL;
}

freeelmt(elmt)
/*
	Frees a elmt and removes it from the circular list.
*/
ELMTPTR elmt;
{
	if(elmt == the_elmts){
		the_elmts = elmt->next;
		if(the_elmts->next == the_elmts)
			the_elmts = (ELMTPTR)NULL;
	}
	elmt->next->last = elmt->last;
	elmt->last->next = elmt->next;
	if(ues > 0)
		sfree(uetype, elmt->user);
	sfree(etype, elmt);
}

freenode(node)
/*
	Frees a node and its attendent nodeneighs, and removes it 
	from the circular list.
*/
NODEPTR node;
{
	NODENEIGHPTR n, nstart, n1;
	ALIASPTR ali, alj, astart;
	n = nstart = node->n;
	if(n != (NODENEIGHPTR)NULL){
		do {
			n1 = n->next;
			sfree(nntype, n);
		} while((n = n1) != nstart);
	}
	if(node == the_nodes){
		the_nodes = node->next;
		if(the_nodes->next == the_nodes)
			the_nodes = (NODEPTR)NULL;
	}
	node->next->last = node->last;
	node->last->next = node->next;
	if(node->b != (BDYPTR)NULL)
		freeb(node->b);
	if(uns > 0)
		sfree(untype, node->user);
	sfree(ntype, node);
}

freeb(bs)
BDYPTR bs;
/*
	Frees a boundary structure and any aliases that come along with it.
*/
{
	ALIASPTR ali, astart, alj;
	if(bs->alias != (ALIASPTR)NULL){
		ali = astart = bs->alias;
		if(astart == (ALIASPTR)NULL)
			abort(108);			/* Should have an alias but doesn't */
		do {
			alj = ali->next;
			sfree(atype, ali);
		} while((ali = alj) != astart);
	}
	if(bs->oppalias != (ALIASPTR)NULL)
		sfree(atype, bs->oppalias);
	if(ubs > 0)
		sfree(ubtype, bs->user);
	sfree(btype, bs);
}
