/****************************************************/
/*    DIME: dime_util.c                             */
/*    These are functions used by many              */
/*    different DIME utilities.                     */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/
#define DIMECODE
#include "dime.h"
#ifdef CUBIX
#ifdef DEBUG
#define STDSYS_SIZE			32768
#else
#define STDSYS_SIZE			128
#endif
#endif
int procnum, curr_doc, max_doc, debug, *allprocs;
char *buffer;
char psbuf[PS_BUF];
extern int pixport, menuport, mouseport, batchport, monitorport;
extern double menu_width;
int pixport, menuport, mouseport, batchport, monitorport;
int mesh_is_distributed;
int (*distort)() = NULL;
double (*sqdist)();
int (*shall_we_swop)();
int (*refine_func)();
int (*swops_func)();

NODEPTR the_nodes;
ELMTPTR the_elmts;

int uns, ues, ubs;

MNUPTR root;

dime_setup(argc, argv, pass_uns, pass_ues, pass_ubs, pass_mesh)
int argc;
char **argv;
int pass_uns, pass_ues, pass_ubs;
int pass_mesh;
/*
	Sets up cube parameters, crystal router buffers, opens plotting and
	makes the screen, then triangulates the users boundary.
*/
{
	int num;
	int plotmesh(), picture(), debug_toggle(), balance(), swops();
	int buffers(), intact_refine(), exit(), postscript(), disp_erase();
	int extents(), magnify(), exp_zoom(), dimen(), readmesh(), writemesh();
	int default_shall_we_swop();
	double default_sqdist();
	MNUPTR dime_menu, drawtools, meshtools, zoom;

#ifdef CUBIX
	char stdsys_buf[STDSYS_SIZE];
	struct cubenv cube;
	cparam(&cube);
	procnum = cube.procnum;
	max_doc = cube.doc;
/* we're not using this and need more memory, so ...
	set_cr_buf(CR_IN_BUF, 0);
	set_cr_buf(CR_OUT_BUF, 1);
*/

	setvbuf(stdsys, stdsys_buf, _IOCBF, STDSYS_SIZE);
#else
	procnum = 0;
	max_doc = 0;
#endif
	set_cr_doc(curr_doc = max_doc);

	mesh_is_distributed = pass_mesh;

	swops_func = NULL;
	refine_func = NULL;
	shall_we_swop = default_shall_we_swop;

	uns = pass_uns;
	ues = pass_ues;
	ubs = pass_ubs;

	num = 	sizeof(_ELMT) + ues + 
			3*(sizeof(_NODE) + uns + sizeof(_BDY) + ubs) +
			16*(4 + sizeof(ALIAS));
#ifdef EBUG
	printf("malloc %d\n", num);
#endif
	buffer = malloc(num);


#ifdef EBUG
	printf("malloc %d\n", (1<<max_doc)*sizeof(int));
#endif
	allprocs = (int *)malloc((1<<max_doc)*sizeof(int));
	for(num=0; num<(1<<max_doc); num++) allprocs[num] = num;

	the_nodes = (NODEPTR)NULL;
	the_elmts = (ELMTPTR)NULL;

	root = menu_make(ROOT);
	dime_menu = menu_make(BRANCH);

		drawtools = menu_make(BRANCH);
		add_item(dime_menu, "READ_MESH", LEAF, readmesh);
		add_item(dime_menu, "WRITE_MESH", LEAF, writemesh);

		add_item(dime_menu, "DRAWTOOLS", BRANCH, drawtools);
			add_item(drawtools, "MESH", LEAF, plotmesh);
			add_item(drawtools, "PICTURE", LEAF, picture);
			add_item(drawtools, "POSTSCRIPT", LEAF, postscript);

		meshtools = menu_make(BRANCH);
		add_item(dime_menu, "MESHTOOLS", BRANCH, meshtools);
			add_item(meshtools, "REFINE", LEAF, intact_refine);
			add_item(meshtools, "TOPO_RELAX", LEAF, swops);

		if(mesh_is_distributed){
		add_item(dime_menu, "BALANCE", LEAF, balance);
		add_item(dime_menu, "DIMENSION", LEAF, dimen);
		}
		add_item(dime_menu, "MEMORY", LEAF, buffers);

	add_item(root, "DIME", BRANCH, dime_menu);
	add_item(root, "TERSE", LEAF, debug_toggle);
	add_item(root, "ERASE", LEAF, disp_erase);

	zoom = menu_make(BRANCH);
	add_item(root, "ZOOM", BRANCH, zoom);
		add_item(zoom, "EXTENTS", LEAF, extents);
		add_item(zoom, "MAGNIFY", LEAF, magnify);
		add_item(zoom, "EXPLICIT", LEAF, exp_zoom);

	add_item(root, "QUIT", LEAF, exit);

	fsingl(stdout);
	fsingl(stdin);
	debug = 1;
	sqdist = default_sqdist;

	printf("**************************************************\n");
	printf("*           This is DIME version 2.0             *\n");
	printf("**************************************************\n");
	if(max_doc == 0)
	printf("*           Running on 1 processor               *\n");
	else
	printf("*           Running on %4d processors           *\n", 1<<max_doc);
	printf("**************************************************\n");
	printf("*            Copyright Caltech 1988              *\n");
	printf("**************************************************\n");
	printf("User_node size %d bytes\n", uns);
	printf("User_elmt size %d bytes\n", ues);
	printf("User_bdy  size %d bytes\n", ubs);
	fflush(stdout);

	parse(argc, argv);
	if(graphics){
#ifdef CUBIX
		openpl(PLOTIX_BUF, (FILE *)0);
#else
		openpl(PLOTIX_BUF, stderr);
#endif
		greyscale(WHITE, BLACK);
		rainbow(RAINSTART, RAINEND);
		menu_width = 0.2;
		menuport = vport(0.0, 0.0, menu_width, 1.0);
		ortho_space(0.0, 0.0, menu_width, 1.0, -1);
		monitorport = vport(0.95, 0.0, 1.0, 1.0); 
		if(graphics == 1)
			pixport = vport(0.05, 0.0, 0.95, 1.0);
		else
			pixport = vport(menu_width, 0.0, 0.95, 1.0);
		setvport(pixport);
	}
}

dimen()
{
	printf("New Cube Dimension?\n"); fflush(stdout);
	get_int(&curr_doc);
	if(curr_doc > max_doc)
		curr_doc = max_doc;
	set_cr_doc(curr_doc);
}
    
parse(argc, argv)
     int argc;
     char *argv[];
{
    int c;
    char includefile[50];
    
    graphics = -1;
    
    while(--argc && **(++argv) == '-') {
		switch(c = *((*argv)+1)) {
		case 'g':
			graphics = atoi((*argv)+2);
			break;
		}
	}

	if(graphics == -1){
		fprintf(stderr, "Graphics? (0 none, 1 batch, 2 mouse)"); 
		fflush(stderr);
		scanf("%d", &graphics);
	}

	if(graphics == 0 && argc == 0){
		fprintf(stderr, "Program file? ");
		fflush(stderr);
		scanf("%s", includefile);
		menu_includefile(includefile);
	}
	if(argc)
		menu_includefile(argv[0]);

    return;
}
