#include <stdio.h>
#include <math.h>
#include "menu.h"
#include "gr_prim.h"

extern int pixport, menuport, monitorport;
extern double menu_width, z_xmin, z_ymin, z_xmax, z_ymax;
extern int graphics;

#define PI	3.1415927

MNUPTR Root, current_menu = (MNUPTR)NULL;
FILE *logfp = (FILE *)NULL, *infp = stdin;

add_item(menu, s, type, act)
MNUPTR menu;
char *s;
int type;
char *act;
{
	int nitem;
	MITEMPTR item;
	nitem = menu->nitem++;
#ifdef EBUG
	printf("malloc %d\n", sizeof(MITEM));
#endif
	item = menu->item[nitem] = (MITEMPTR)malloc(sizeof(MITEM));
	if(item == (MITEMPTR)NULL)
		abort(901);				/* cannot make menu item */
	strcpy(item->item_label, s);
	item->type = type;
	if(type == LEAF)
		item->to_do.action = (PFI)act;
	else
		item->to_do.next = (MNUPTR)act;
}

MNUPTR menu_make(this_is_Root)
int this_is_Root;
{
	MNUPTR new;
	int toRoot(), menu_record(), menu_end_record(), menu_include();
#ifdef EBUG
	printf("malloc %d\n", sizeof(MNU));
#endif
	new = (MNUPTR)malloc(sizeof(MNU));
	if(new == (MNUPTR)NULL)
		abort(902);				/* cannot make menu */
	new->nitem = 0;
	if(this_is_Root){
		Root = new;
		add_item(Root, "RECORD", LEAF, menu_record);
		add_item(Root, "END_RECORD", LEAF, menu_end_record);
		add_item(Root, "PROGRAM", LEAF, menu_include);
	}
	else
		add_item(new, "TO_ROOT", LEAF, toRoot);
	return(new);
}

toRoot()
{
	if(infp == stdin)
		menu_screen(Root);
	current_menu = Root;
}

menu_include()
{
	char batchfile[50];
	printf("Include file? "); fflush(stdout);
	fscanf(infp, "%s", batchfile);
	menu_includefile(batchfile);
}

menu_includefile(batchfile)
char *batchfile;
{
	infp = fopen(batchfile, "r");
	if(infp == (FILE *)NULL)
		printf("Cannot open %s\n", batchfile);fflush(stdout);
}

menu_record()
{
	char file[50];
	if(logfp != NULL){
		printf("Log file already activated\n");
		return;
	}
	printf("Log file name? "); fflush(stdout);
	get_string(file);
	if(infp != stdin){
		printf("Cannot log from command file\n");
		return;
	}
	logfp = fopen(file, "w");
	if(logfp == (FILE *)NULL)
		printf("Cannot open %s\n", file);fflush(stdout);
}

menu_end_record()
{
	if(logfp == (FILE *)NULL){
		printf("No log file open\n");
		return;
	}
	fclose(logfp);
	logfp = (FILE *)NULL;
}

menu_pick()
{
	int key, result, i;
	double x, y;
	char command[MAXLEN];
	PFI func;
	MITEMPTR item;
again:
	if(infp == stdin){
		setvport(menuport);
		do {
			while(!gin(&key, &x, &y)){
				printf("Pick menu item\n"); fflush(stdout);
			}
			result = (1-y)*20;
		} while(result >= current_menu->nitem);
		item = current_menu->item[result];
		printf("%s\n", item->item_label); fflush(stdout);
		setvport(pixport);
	} else {
		if(fscanf(infp, "%s", command) == EOF){
			fclose(infp);
			if(graphics == 0)
				exit(0);
			graphics = 2;
			infp = stdin;
			menu_screen(current_menu);
			goto again;
		}
		printf("%s", command);fflush(stdout);
		for(i=0; i<current_menu->nitem; i++){
			item = current_menu->item[i];
			if(strcmp(item->item_label, command) == 0){
				printf("\n");
				goto gotit;
			}
		}
		printf(" (Ignored)\n");
		goto again;
	}
gotit:
	if(logfp != (FILE *)NULL && 
			strcmp(item->item_label, "END_RECORD") != 0 &&
			strcmp(item->item_label, "PROGRAM") != 0){
		fprintf(logfp, "%s\n", item->item_label);fflush(logfp);
		fflush(logfp);
	}
	if(item->type == LEAF){
		func = item->to_do.action;
		(*func)();
		return;
	} else {
		current_menu = item->to_do.next;
		if(infp == stdin)
			menu_screen(current_menu);
	}
}

menu_input()
{
	current_menu = Root;
	menu_screen(Root);
	while(1)
		menu_pick();
}

menu_screen(menu)
MNUPTR menu;
{
	int i, nitem;
	if(graphics == 2){
		color(1);
		nitem = menu->nitem;
		setvport(menuport);
		box(0.0, 0.0, menu_width, 1.0, 0, 1);
		for(i=0; i<nitem; i++)
			label(menu->item[i]->item_label, 0.02, 1.-(i+0.75)/20.);
		sendplot();
		setvport(pixport);
	}
}

disp_erase()
/*
	Erases the interesting part of the screen but leaves the menu boxes.
*/
{
	erase();
	if(graphics && current_menu != (MNUPTR)NULL)
		menu_screen(current_menu);
}

get_double(f)
double *f;
{
	fscanf(infp, "%lf", f);
	printf("%f\n", *f);fflush(stdout);
	if(logfp != (FILE *)NULL){
		fprintf(logfp, "%f\n", *f);
		fflush(logfp);
	}
}

get_int(f)
int *f;
{
	fscanf(infp, "%d", f);
	printf("%d\n", *f);fflush(stdout);
	if(logfp != (FILE *)NULL){
		fprintf(logfp, "%d\n", *f);
		fflush(logfp);
	}
}

get_string(s)
char *s;
{
	fscanf(infp, "%s", s);
	if(logfp != (FILE *)NULL){
		fprintf(logfp, "%s\n", s);
		fflush(logfp);
	}
}

get_gin(key, x, y)
int *key;
double *x, *y;
{
	if(infp == stdin)
		gin(key, x, y);
	else
		fscanf(infp, "%d %lf %lf", key, x, y);
	if(logfp != (FILE *)NULL){
		fprintf(logfp, "%d %f %f\n", *key, *x, *y);
		fflush(logfp);
	}
}

get_polars(r, th)
double *r, *th;
{
	double rad, x, y;
	double xx, yy;
	int i, key;
	if(infp == stdin){
		rad = 0.45*menu_width;
		color(1);
		setvport(menuport);
		move(rad, 2*rad);
		cont(rad, 0.0);
		move(0.0, rad);
		cont(2*rad, rad);
		for(i=1; i<=32; i++)
			cont(rad+rad*cos(i*PI/16), rad+rad*sin(i*PI/16));
		sendplot();
		gin(&key, &xx, &yy);
		x = (xx-rad)/rad;
		y = (yy-rad)/rad;
		*th = atan2(y, x);
		*r = sqrt(x*x+y*y);
		setvport(pixport);
	} else
		fscanf(infp, "%lf %lf", r, th);
	if(logfp != (FILE *)NULL){
		fprintf(logfp, "%f %f\n", *r, *th);
		fflush(logfp);
	}
}

menu_monitor(min, max)
int min, max;
{
	int i;
	char num[3];
	color(2);
	setvport(monitorport);
	Space(0.0, (double)min, 1.0, (double)max);
	box(0.0, (double)min, 1.0, (double)max, 0, 1);
	for(i=min+1; i<max; i++){
		sprintf(num, "%d", i);
		label(num, 0.3, (double)i);
	}
	sendplot();
	setvport(pixport);
}

put_monitor(value)
double value;
{
	double y;
	setvport(monitorport);
	move(0.0, value);
	cont(1.0, value);
	sendplot();
	setvport(pixport);
}
