#include "dime.h"
#define CODE_3D
#include "3d.h"
/*
#include "../string.h"
*/

extern MNUPTR root;
static double currextent = 1.;

setup_3dmenu()
{
	int render(), rotate(), get_range();
	int reset_3d(), write_taac(), curvature();
	MNUPTR menu3d;

	menu3d = menu_make(BRANCH);
	add_item(root, "3D", BRANCH, menu3d);
	add_item(menu3d, "RESET", LEAF, reset_3d);
	add_item(menu3d, "RENDER", LEAF, render);
	add_item(menu3d, "TAAC_FILE", LEAF, write_taac);
	add_item(menu3d, "ROTATE", LEAF, rotate);
	add_item(menu3d, "RANGE", LEAF, get_range);
	add_item(menu3d, "CURVATURE", LEAF, curvature);
}

reset_3d()
{
	ELMTPTR elmt, list_sort();
	NODEPTR node;
	double w, rcm[3], extent, maxextent;
	int i, compar(), change, nnode = 0;

	rcm[0] = rcm[1] = rcm[2] = 0;
	FORALLNODES(node)
		nnode++;
		get_3d_coords(node, node->user->node3d.coord);
		for(i=0; i<3; i++)
			rcm[i] += node->user->node3d.coord[i];
	NEXTNODE(node)

	for(i=0; i<3; i++)
		rcm[i] /= nnode;

	maxextent = 0;
	FORALLNODES(node)
		for(i=0; i<3; i++) {
			node->user->node3d.coord[i] -= rcm[i];
			if (node->user->node3d.coord[i] > maxextent)
				maxextent = node->user->node3d.coord[i];
		}
		for(i=0; i<3; i++)
			node->user->node3d.coord[i] /= maxextent;
	NEXTNODE(node)

	FORALLELMTS(elmt)
		get_area(elmt);
	NEXTELMT(elmt)

	do_rotation();

	extent = 0;
	FORALLNODES(node)
		if(fabs(node->XV) > extent) extent = fabs(node->XV);
		if(fabs(node->YV) > extent) extent = fabs(node->YV);
	NEXTNODE(node)

	if(extent > currextent){
		currextent = 1.1*extent;
		ortho_space(-currextent,-currextent,currextent,currextent, 0);
		printf("Zoomout to %f\n", currextent);
	}
	else if(extent < 0.001*currextent){
		currextent *= 0.5;
		ortho_space(-currextent,-currextent,currextent,currextent, 0);
		printf("Zoom in to %f\n", currextent);
	}

	the_elmts = list_sort(the_elmts, compar);
	return;
}

get_area(elmt)
ELMTPTR elmt;
{
	double ax, ay, az, bx, by, bz, cx, cy, cz, c;
	int i;
	ax = elmt->neigh[1]->XG - elmt->neigh[0]->XG;
	bx = elmt->neigh[2]->XG - elmt->neigh[0]->XG;
	ay = elmt->neigh[1]->YG - elmt->neigh[0]->YG;
	by = elmt->neigh[2]->YG - elmt->neigh[0]->YG;
	az = elmt->neigh[1]->ZG - elmt->neigh[0]->ZG;
	bz = elmt->neigh[2]->ZG - elmt->neigh[0]->ZG;
	cx = ay*bz - az*by;
	cy = az*bx - ax*bz;
	cz = ax*by - ay*bx;
	c = sqrt(cx*cx + cy*cy + cz*cz);
	if(c > 1.e-12){
		elmt->user->elmt3d.area[0] = cx/c;
		elmt->user->elmt3d.area[1] = cy/c;
		elmt->user->elmt3d.area[2] = cz/c;
		elmt->user->elmt3d.absarea = 0.5*c;
	}
}

plotmeasure(array, n, max)
double *array, max;
int n;
{
#define MAXN		5
#define MAXH		10
#define SCAL(h)		((-1-(h)*0.4/MAXH)*currextent)
#define RAT			0.7
	int i, h;
	double bottom, top, height;
	static double hist[MAXH][MAXN];
	static int first = 1;

	if(first){
		for(h=0; h<MAXH; h++)
			for(i=0; i<=n; i++)
				hist[h][i] = 0;
		first = 0;
	}
	for(h=MAXH-1; h>0; h--)
		for(i=0; i<=n; i++)
			hist[h][i] = hist[h-1][i];
		
	hist[0][0] = bottom = 0;
	for(i=0; i<n; i++){
		height = array[i]/max;
		hist[0][i+1] = bottom + height;
		bottom += height;
	}
	for(h=0; h<MAXH; h++){
		for(i=0; i<n; i++){
			box(SCAL(h  ), RAT*(2*hist[h][i]-1)*currextent,
				SCAL(h+1), RAT*(2*hist[h][i+1]-1)*currextent,
				i+2, 0);
		}
	}
}

thermometer(val, ptval)
double val, ptval;
{
	box(currextent, -RAT*currextent,
		1.3*currextent, RAT*currextent, 7, 0);
	box(currextent, -RAT*currextent,
		1.3*currextent, RAT*(2*val-1)*currextent, 6, 0);
	initpanel(5, 0);
	panelpoint(currextent, RAT*(2*ptval-1)*currextent);
	panelpoint(0.9*currextent, RAT*(2*ptval-0.9)*currextent);
	panelpoint(0.9*currextent, RAT*(2*ptval-1.1)*currextent);
	endpanel();
	
}
