/*********************************************************************

STRING

C.F. Baillie, D.A. Johnston and R.D. Williams

Caltech, 1989

Jan 1992, Leping Han began to simplify the codes to be good for
some special purpose.
Jan 9, now it is cut to everything in one directory and no graphics,
and with one makefile to compile them all.

*********************************************************************/


/* definitions */

#include "dime.h" 

#define MAXDIM		4
#define MAXNEIGH	128
#define NOBS		6

#define LARGE		1.e25
#define SMALL		1.e-25

struct user_node {
	double r[MAXDIM];
	double gauss_action;
	double area_action;
	double edge_action;
	int nneigh;
};

struct user_bdy	{
	double unused;
};

struct user_elmt {
	double area;
	double edge_sq[3];
	double alt_sq[3];
	double h[MAXDIM][3];
};

extern int dim;
extern int nnode;
extern double temp, alpha, lambdaA, lambdaE;

#define GAUSS		act[0]
#define LN_W		act[1]
#define AREA_CURVE	act[2]
#define EDGE_CURVE	act[3]
#define X2		act[4]
#define NNEIGH		act[5]

#define GAUSS2		act2[0]
#define LN_W2		act2[1]
#define AREA_CURVE2	act2[2]
#define EDGE_CURVE2	act2[3]
#define X22		act2[4]
#define NNEIGH2		act2[5]

#define BINGAUSS		bin_ave[0]
#define BINLN_W			bin_ave[1]
#define BINAREA_CURVE		bin_ave[2]
#define BINEDGE_CURVE		bin_ave[3]
#define BINX2			bin_ave[4]
#define BINNNEIGH		bin_ave[5]

#define BINGAUSS2		bin_ave2[0]
#define BINLN_W2		bin_ave2[1]
#define BINAREA_CURVE2		bin_ave2[2]
#define BINEDGE_CURVE2		bin_ave2[3]
#define BINX22			bin_ave2[4]
#define BINNNEIGH2		bin_ave2[5]

#define AVEGAUSS		tot_ave[0]
#define AVELN_W			tot_ave[1]
#define AVEAREA_CURVE		tot_ave[2]
#define AVEEDGE_CURVE		tot_ave[3]
#define AVEX2			tot_ave[4]
#define AVENNEIGH		tot_ave[5]

#define AVEGAUSS2		tot_ave2[0]
#define AVELN_W2		tot_ave2[1]
#define AVEAREA_CURVE2		tot_ave2[2]
#define AVEEDGE_CURVE2		tot_ave2[3]
#define AVEX22			tot_ave2[4]
#define AVENNEIGH2		tot_ave2[5]


/* string */

extern double pranf();
#define RAN		((double)pranf())

#define SWEEPSPERWRITE 100

#define BINSIZE        10	/* SWEEPSPERPRINT */

int nnn = 1;
double bin_ave[NOBS+2];
double bin_ave2[NOBS+2];
double tot_ave[NOBS+2];
double tot_ave2[NOBS+2];
	double act[NOBS+2];
	double act2[NOBS+2];

extern MNUPTR root;
int dim = 3;
extern int nnode;
extern int procnum;
long nproc = 1;
double lambdaA, lambdaE, tot_action, hit_size = 0.5;
double metro_success, flip_success;
FILE *out = NULL;

main(argc, argv)
int argc;
char **argv;
{
	MNUPTR mymenu;
	int reset(), ascii_print_action(), sweeps();
	int singmult(), dmpseed();
	int outfile(), add_nodes(), check_mesh(), read_string();

	int string_distort(), string_read(), string_write();
	double string_sqdist();

#ifdef CUBIX
	struct cubenv cube;
	cparam(&cube);
	procnum = cube.procnum;
	nproc = cube.nproc;
#endif

pranset(42);

	DIME_SETUP(argc, argv, MULTI);
	user_read = string_read;
	user_write = string_write;
	distort = string_distort;
	sqdist = string_sqdist;

/*	mymenu = menu_make(BRANCH);
	add_item(root, "STRING", BRANCH, mymenu);
	add_item(mymenu, "RESET", LEAF, reset);
	add_item(mymenu, "READ_STRING", LEAF, read_string);
	add_item(mymenu, "OUTFILE", LEAF, outfile);
	add_item(mymenu, "ADD_NODES", LEAF, add_nodes);
	add_item(mymenu, "CHECK_MESH", LEAF, check_mesh);
	add_item(mymenu, "SWEEPS", LEAF, sweeps);
	add_item(mymenu, "ACTION", LEAF, ascii_print_action);
	add_item(mymenu, "SINGMULT", LEAF, singmult);
	add_item(mymenu, "DMPSEED", LEAF, dmpseed);
*/
	read_string();
	add_nodes();
	reset();
	sweeps();

/*	menu_input(); 
*/

}

singmult()
{
	if(mesh_is_distributed = 1-mesh_is_distributed)
		printf("Mesh is MULTI\n");
	else
		printf("Mesh is SINGL\n");
}

read_string()
{
	char string[100];
	double lambda, action();
	int i=0, islash, ibar=0;
	FILE *in;

    printf("Input Filename (without suffix)?\n");
    get_string(string);

	islash = -1;
	while(string[i] != '\0'){
		if(string[i] == '/')
			islash = i;
		i++;
	}
	while(string[ibar++] != '_');
	sscanf(string+ibar, "%d", &dim);
	sscanf(string+islash+6, "%lf", &lambda);
	if(string[islash+5] == 'A'){
		lambdaA = lambda;
		lambdaE = 0.0;
	} else {
		lambdaA = 0.0;
		lambdaE = lambda;
	}
/*
	printf("Dim = %d, lambda%c = %lf\n", dim, string[islash+5], lambda);
*/
    strcat(string, ".ct");

    if((in = fopen(string, "r")) == (FILE *)NULL){
        printf("Cannot open %s\n", string);
		exit(0);
	}
    else
        readct(in);

	fill_user_structs();
/*
	tot_action = action();
*/
}

write_string()
{
	char filename[100];
	FILE *write;

	if(lambdaE == 0.0)
		sprintf(filename, "N%d/MeshA%4.3f_%d.ct", nnode, lambdaA, dim);
	else
		sprintf(filename, "N%d/MeshE%4.3f_%d.ct", nnode, lambdaE, dim);

	printf("Writing mesh to %s\n", filename);
	if((write = fopen(filename, "w")) == NULL){
		perror(filename);
		return;
	} else
		writect(write);
}

read_aves(mes_done)
int *mes_done;
{
    char filename[100];
    FILE *in;

    if(lambdaE == 0.0)
	sprintf(filename, "N%d/avesA%4.3f_%d.ct", nnode, lambdaA, dim);
    else
	sprintf(filename, "N%d/avesE%4.3f_%d.ct", nnode, lambdaE, dim);

    printf("Reading aves from %s\n", filename);
    if((in = fopen(filename, "r")) == (FILE *)NULL){
        printf("Cannot open aves %s\n", filename);
	exit(0);
    } else {
#ifdef CUBIX
	fmulti(in);
#endif
	fread(mes_done, sizeof(int), 1, in); 
	fread(tot_ave, sizeof(double), NOBS+2, in); 
	fread(tot_ave2, sizeof(double), NOBS+2, in); 
	fclose(in);
    }
    printf("Sweeps so far %d\n", *mes_done);
    fflush(stdout);
}

write_aves(mes_done)
int *mes_done;
{
	char filename[100];
	FILE *write;

	if(lambdaE == 0.0)
	    sprintf(filename, "N%d/avesA%4.3f_%d.ct", nnode, lambdaA, dim);
	else
	    sprintf(filename, "N%d/avesE%4.3f_%d.ct", nnode, lambdaE, dim);

	printf("Writing aves to %s\n", filename);
	if((write = fopen(filename, "w")) == NULL){
	    perror(filename);
	    return;
	} else {
#ifdef CUBIX
	    fmulti(write);
#endif
	    fwrite(mes_done, sizeof(int), 1, write); 
	    fflush(write);
	    fwrite(tot_ave, sizeof(double), NOBS+2, write); 
	    fflush(write);
	    fwrite(tot_ave2, sizeof(double), NOBS+2, write); 
	    fclose(write);
	}
}

reset()
{
	double action();

	fill_user_structs();

	printf("Dimension (current = %d)? ", dim); fflush(stdout);
	get_int(&dim);

	printf("Area action coefficient? "); fflush(stdout);
	get_double(&lambdaA);

	printf("Edge action coefficient? "); fflush(stdout);
	get_double(&lambdaE);

	printf("Hit Size? "); fflush(stdout);
	get_double(&hit_size);

	tot_action = action();

setseed();
}

outfile()
{
	char filename[100];

	if(out != NULL)
		fclose(out);

	if(lambdaE == 0.0)
		sprintf(filename, "N%d/A%4.3f_%d", nnode, lambdaA, dim);
	else
		sprintf(filename, "N%d/E%4.3f_%d", nnode, lambdaE, dim);

	/*
	printf("Sending data to %s\n", filename);
	*/

	if((out = fopen(filename, "w")) == NULL){
		perror(filename);
	}
#ifdef CUBIX
	fmulti(out);
#endif
}

sweeps()
{
	int sweep, sweeps, mes_steps, mes_done, mes_i, i;
	int first_time;
	double metro_move(), flip_move();

	printf("How many sweeps per measurement? "); fflush(stdout);
	get_int(&sweeps);
	printf("How many measurements\n" ); fflush(stdout);
	get_int(&mes_steps);
	printf("First Time ???\n" ); fflush(stdout);
	get_int(&first_time);
	printf("BINSIZE = %d\n\n",BINSIZE); fflush(stdout);
	for(i=0; i<NOBS+2; i++) {
		bin_ave[i] = 0.;
		bin_ave2[i] = 0.;
	}
if (first_time) {
	mes_done = 0;
	for(i=0; i<NOBS+2; i++) {
		tot_ave[i] = 0.;
		tot_ave2[i] = 0.;
	}
} else {
	read_aves(&mes_done);
}

	for(mes_i=0; mes_i<mes_steps; mes_i++){
		metro_success = flip_success = 0;

		for(sweep=0; sweep<sweeps; sweep++){
			metro_success += metro_move();
			flip_success  += flip_move();
		}

		metro_success /= sweeps;
		flip_success  /= sweeps;

	avg_measure(act, act2);
	for(i=0; i<NOBS+2; i++) {
		bin_ave[i] += act[i];
		bin_ave2[i] += act2[i];
		tot_ave[i] += act[i];
		tot_ave2[i] += act2[i];
	}

	    if(mes_done % BINSIZE == BINSIZE - 1) {
		for(i=0; i<NOBS+2; i++) {
			bin_ave[i] /= (double)BINSIZE;
			bin_ave2[i] /= (double)BINSIZE;
		}
		nnn = mes_done+1;
		if(out != NULL)
			binary_print_action();
		ascii_print_action();
		for(i=0; i<NOBS+2; i++) {
			bin_ave[i] = 0.;
			bin_ave2[i] = 0.;
		}
	    }

		if(mes_done % SWEEPSPERWRITE == SWEEPSPERWRITE - 1) {
			write_string();
			nnn = mes_done+1;
			write_aves(&nnn);
		}
	mes_done++;
	}

#ifndef CUBIX
	nnn = mes_done;
	if(out != NULL)
		binary_print_action();
	ascii_print_action();
		printf("Accumulated action %f\n", tot_action);
#endif
	if(out != NULL) {
		write_string();
		write_aves(&mes_done);
	}
}

double metro_move()
{
	NODEPTR node;
	double newr[MAXDIM];
	double metro_delta();
	double dact;
	int idim, try=0, succeed=0;

	FORALLNODES(node)
		for(idim=0; idim<dim; idim++)
			newr[idim] = node->user->r[idim] + hit_size*(RAN-0.5);

		dact = metro_delta(node, newr);
		try++;
		if(dact < 0.0 || exp(-dact) > RAN){
			tot_action += dact;
			metro_accept();
			succeed++;
		} else
			metro_reject();
	NEXTNODE(node)
	return((double)succeed/try);
}

double flip_move()
{
	ELMTPTR elmt;
	int i, try=0, succeed=0;
	double flip_delta();
	double dact;

	FORALLELMTS(elmt)
		i = 3*RAN;
		dact = flip_delta(elmt, i);
		try++;
/*
printf("||| %f  :   ",dact);
*/
		if(dact > 1.e20)
			continue;
		if(dact < 0.0 || exp(-dact) > RAN){
			tot_action += dact; 
			flip_accept();
			succeed++;
		} else 
			flip_reject();
	NEXTELMT(elmt)
	return((double)succeed/try);
}

double action()
{
	double act[NOBS];
	measure(act);
	return(GAUSS + 0.5*dim*LN_W + lambdaA*AREA_CURVE + lambdaE*EDGE_CURVE);
}

ascii_print_action()
{
	int i;
	if (nnn == 1) avg_measure(act, act2);
#ifdef CUBIX
	fmulti(stdout);
#endif
	/*
	if(procnum == 0) {
	*/
		printf("Proc %d (%d):\n",procnum,nnn);
		/*
		printf("Gauss, Ln W               %10.5f %10.5f\n", GAUSS, LN_W);
		printf("Curve, Area Edge          %10.5f %10.5f\n", AREA_CURVE, EDGE_CURVE);
		printf("x2                        %10.5f\n", X2);
		printf("Success, Metro Flip       %10.5f %10.5f\n", metro_success, flip_success);
		printf("Max neigh                 %10.5f\n", NNEIGH);
		printf("Total Action              %10.5f\n", 
GAUSS + 0.5*dim*LN_W + lambdaA*AREA_CURVE + lambdaE*EDGE_CURVE);
		printf("Square Gauss, Ln W        %10.5f %10.5f\n", GAUSS2, LN_W2);
		printf("Square Curve, Area Edge   %10.5f %10.5f\n", AREA_CURVE2, EDGE_CURVE2);
		printf("Square x2                 %10.5f\n", X22);
		*/

		printf("%10.3f %10.3f\n", BINGAUSS, BINLN_W);
		printf("%10.3f %10.3f\n", BINAREA_CURVE, BINEDGE_CURVE);
		printf("%10.3f\n", BINX2);
		printf("%10.3f %10.3f\n", BINGAUSS2, BINLN_W2);
		printf("%10.3f %10.3f\n", BINAREA_CURVE2, BINEDGE_CURVE2);
		printf("%10.3f\n", BINX22);

		printf("%10.3f %10.3f\n", AVEGAUSS/nnn, AVELN_W/nnn);
		printf("%10.3f %10.3f\n", AVEAREA_CURVE/nnn, AVEEDGE_CURVE/nnn);
		printf("%10.3f\n", AVEX2/nnn);
		printf("%10.3f %10.3f\n", AVEGAUSS2/nnn, AVELN_W2/nnn);
		printf("%10.3f %10.3f\n", AVEAREA_CURVE2/nnn, AVEEDGE_CURVE2/nnn);
		printf("%10.3f\n\n", AVEX22/nnn);
	/*
	}
	*/
#ifdef CUBIX
	fsingl(stdout);
#endif
}

void _ex_swaw(from, to, N)
    char *from, *to;
    int N;
{
    char q[4];
    int i;
	
    for(i=0 ; i<N/4 ; i++) {
	q[0] = *(from+3); q[1] = *(from+2);  q[2] = *(from+1); q[3] = *from;
	memcpy(to, q, 4);
	from += 4;
	to += 4;
    }
}

binary_print_action()
{
/*
file out contains records with fields: TOT AVE
							0. Gaussian action
							1. Log W
							2. Area curvature
							3. Edge Curvature
							4. x2
							5. Maximum # neighbors
							6. metropolis success rate
							7. flip success rate
							+ squares of all these
*/
	float temp[NOBS+2];
	char *ptr;
	int i;
	long itemp;
	if(out == NULL)
		return;

	/* close and open file - i.e. rewind */
	outfile();

	/* write number of processors i.e. bins */
	if(procnum == 0) {
		itemp = nproc;
#ifdef CUBIX /* swap the f**king bytes */
#ifndef A2010 /* not on Ametek */
	_ex_swaw(&itemp,&itemp,4);
#endif
#endif
		fwrite(&itemp, sizeof(long), 1, out); 
	}

	for(i=0; i<NOBS+2; i++)
		temp[i] = tot_ave[i]/nnn;
#ifdef CUBIX /* swap the f**king bytes */
#ifndef A2010 /* not on Ametek */
	_ex_swaw(temp,temp,(NOBS+2)*4);
#endif
#endif
	fwrite(temp, sizeof(float), NOBS+2, out); 

	for(i=0; i<NOBS+2; i++)
		temp[i] = tot_ave2[i]/nnn;
#ifdef CUBIX /* swap the f**king bytes */
#ifndef A2010 /* not on Ametek */
	_ex_swaw(temp,temp,(NOBS+2)*4);
#endif
#endif
	fwrite(temp, sizeof(float), NOBS+2, out); 
	fflush(out);
}

avg_measure(act, act2)
double act[NOBS+2];
double act2[NOBS+2];
{
	int i;
	measure(act);
	act[NOBS] = metro_success;
	act[NOBS+1] = flip_success;
	for(i=0; i<NOBS+2; i++)
		act2[i] = act[i]*act[i];

/* DO NOT COMBINE !!!
#ifdef CUBIX
	combine(act, fsum, sizeof(double), NOBS-1);
	combine(act+NOBS-1, fmax, sizeof(double), 1);
	combine(act+NOBS, fsum, sizeof(double), 2);

	for(i=0; i<NOBS+2; i++)
		if(i != NOBS-1)
			act[i] /= nproc;

	combine(act2, fsum, sizeof(double), NOBS-1);
	combine(act2+NOBS-1, fmax, sizeof(double), 1);
	combine(act2+NOBS, fsum, sizeof(double), 2);

	for(i=0; i<NOBS+2; i++)
		if(i != NOBS-1)
			act2[i] /= nproc;
#endif
*/

}

add_nodes()
{
	NODEPTR node;
	ELMTPTR elmt;
	int addem;

	nnode = 0;
	FORALLNODES(node)
		nnode++;
	NEXTNODE(node);
	printf("Currently %d nodes; add how many?", nnode);
	fflush(stdout);
	get_int(&addem);

	FORALLELMTS(elmt)
		if(addem-- == 0)
			break;
		add_node(elmt, 0);
	NEXTELMT(elmt)

	nnode = 0;
	FORALLNODES(node)
		nnode++;
	NEXTNODE(node);
	printf("Now %d nodes\n", nnode);
}


/* dime_string */

NODEPTR metro_victim;

NODEPTR A, B, C, D;
ELMTPTR ABC, ACD;

struct user_node oldnode;
struct user_elmt oldelmtbuf[MAXNEIGH];
struct user_node oldnodebuf[MAXNEIGH];

int nnode;

fill_user_structs()
{
	NODEPTR node;
	ELMTPTR elmt;

	FORALLELMTS(elmt)
		get_elmt_data(elmt);
	NEXTELMT(elmt)

	nnode = 0;
	FORALLNODES(node)
		get_node_data(node);
		nnode++;
	NEXTNODE(node)
	printf("%d nodes in fill_user_structs\n", nnode);
}

get_elmt_data(elmt)
ELMTPTR elmt;
{
	NODEPTR n0, n1, n2;
	int idim, i, in;
	double dot, rr, temp, r[3][MAXDIM];

	n0 = elmt->neigh[0];
	n1 = elmt->neigh[1];
	n2 = elmt->neigh[2];
	for(idim=0; idim<dim; idim++){
		r[0][idim] = n1->user->r[idim] - n2->user->r[idim];
		r[1][idim] = n2->user->r[idim] - n0->user->r[idim];
		r[2][idim] = n0->user->r[idim] - n1->user->r[idim];
	}
	for(i=0; i<3; i++){
		in = (i+1)%3;
		rr = dot = 0;
		for(idim=0; idim<dim; idim++){
			rr += r[i][idim] * r[i][idim];
			dot += r[i][idim] * r[in][idim];
		}
		elmt->user->edge_sq[i] = rr;
		dot /= rr;
		rr = 0;
		for(idim=0; idim<dim; idim++){
			temp = elmt->user->h[idim][i] = r[in][idim] - dot * r[i][idim];
			rr += temp * temp;
		}
		elmt->user->alt_sq[i] = rr;
	}
	elmt->user->area = 0.5*sqrt(elmt->user->edge_sq[0]*elmt->user->alt_sq[0]);
}

get_node_data(node)
NODEPTR node;
{
	ELMTPTR elmt, elmtn;
	NODEPTR noden;
	NODEPTR A, B, C, D;
	NODENEIGHPTR n, nstart;
	int i, idim, in;
	double node_area, node_length, dr[MAXNEIGH], one_dr[MAXNEIGH];
	double edge_curve();
	double hB2, hC2, hD2, AB_AC, BC_BA, DC_DA, AD_AC;
	double AC2, BC2, BA2, DA2, DC2, cota, cotb, cotd, cotan;
	double AB[MAXDIM], BC[MAXDIM], AC[MAXDIM], DA[MAXDIM], DC[MAXDIM];

	node->user->nneigh = 0;
	node->user->edge_action = 0;
	node->user->area_action = 0;
	node->user->gauss_action = 0;

	node_area = 0;
	node_length = 0;
	for(idim=0; idim<dim; idim++)
		dr[idim] = 0.0;

	n = nstart = node->n;
	do {
		node->user->nneigh++;
		elmt = n->elmt;
		elmtn = n->next->elmt;
		i = (indexc(elmt, node)+1)%3;
		in = (indexc(elmtn, node)+2)%3;

		node->user->edge_action += edge_curve(elmt, i, elmtn, in);
		node->user->gauss_action += 0.5*elmt->user->edge_sq[i];
		noden = elmt->neigh[(i+1)%3];
		for(idim=0; idim<dim; idim++)
			one_dr[idim] = noden->user->r[idim] - node->user->r[idim];
#ifndef DUAL
		node_area += elmt->user->area;
#else
/* dual area and length terms */
		AC2 = elmt->user->edge_sq[i];
		BC2 = elmt->user->edge_sq[(i+1)%3];
		BA2 = elmt->user->edge_sq[(i+2)%3];
		DA2 = elmtn->user->edge_sq[(in+1)%3];
		DC2 = elmtn->user->edge_sq[(in+2)%3];
	C = node;
	B = elmt->neigh[i];
	A = noden;
	D = elmtn->neigh[in];
	hB2 = elmt->user->alt_sq[i];
	hC2 = elmt->user->alt_sq[(i+2)%3];
	hD2 = elmtn->user->alt_sq[in];
	for(idim=0; idim<dim; idim++) {
		AB[idim] = B->user->r[idim] - A->user->r[idim];
		BC[idim] = C->user->r[idim] - B->user->r[idim];
		AC[idim] = C->user->r[idim] - A->user->r[idim];
		DA[idim] = A->user->r[idim] - D->user->r[idim];
		DC[idim] = C->user->r[idim] - D->user->r[idim];
	}
	AB_AC = 0.;
	BC_BA = 0.;
	DC_DA = 0.;
	AD_AC = 0.;
	for(idim=0; idim<dim; idim++) {
		AB_AC += AB[idim] * AC[idim];
		BC_BA += BC[idim] * (-AB[idim]);
		DC_DA += DC[idim] * DA[idim];
		AD_AC += (-DA[idim]) * AC[idim];
	}
	cota = AB_AC / ( sqrt(AC2 * hB2) );
	cotb = BC_BA / ( sqrt(BA2 * hC2) );
	cotd = DC_DA / ( sqrt(DA2 * hC2) );
	cotan = AD_AC / ( sqrt(AC2 * hD2) );
		node_area += 0.125 * (BC2*cota + AC2*cotb);
		node_length = 0.5 * (cotb + cotd);
	for(idim=0; idim<dim; idim++)
		one_dr[idim] *= node_length;
#endif
	for(idim=0; idim<dim; idim++)
		dr[idim] += one_dr[idim];
	} while((n = n->next) != nstart);

	for(idim=0; idim<dim; idim++) {
		node->user->area_action += dr[idim] * dr[idim];
	}
	node->user->area_action /= node_area;
/*
printf("|| node_area = %f action = %f AC = %f node_length = %f dr = %f %f %f\n",
node_area, node->user->area_action, sqrt(AC2), node_length, dr[0], dr[1], dr[2]);
*/
}

double edge_curve(elmt, i, elmtn, in)
ELMTPTR elmt, elmtn;
int i, in;
{
	double dot = 0, hh;
	int idim;
	for(idim=0; idim<dim; idim++)
		dot += elmt->user->h[idim][i] * elmtn->user->h[idim][in];
	hh = elmt->user->alt_sq[i] * elmtn->user->alt_sq[in];
	if(hh < SMALL){
/*
#ifndef CUBIX
		printf("warning -- collinear\n");
#endif
*/
		return(0.0);
	}
	else
		return(0.5*(1 + dot / sqrt(hh)));
}

measure(act)
double *act;
{
	NODEPTR node, n1, n2;
	ELMTPTR elmt;
	double dd, dr;
	int idim;

/***************************GAUSSIAN ACTION **************/
	GAUSS = 0;
	FORALLNODES(node)
		GAUSS += node->user->gauss_action;
	NEXTNODE(node)

/***************************LOG PRODUCT NNEIGH ***********/
	LN_W = 0;
	FORALLNODES(node)
		LN_W += log((double)node->user->nneigh);
	NEXTNODE(node)

/***************************AREA CURVATURE ***************/
	AREA_CURVE = 0;
	FORALLNODES(node)
		AREA_CURVE += node->user->area_action;
	NEXTNODE(node)

/***************************EDGE CURVATURE ***************/
	EDGE_CURVE = 0;
	FORALLNODES(node)
		EDGE_CURVE += node->user->edge_action;
	NEXTNODE(node)
/***************************GYRATION RADIUS***************/
	X2 = 0;
	FORALLNODES(n1)
		FORALLNODES(n2)
			if(n1 >= n2)
				continue;
			dd = 0;
			for(idim=0; idim<dim; idim++){
				dr = n1->user->r[idim] - n2->user->r[idim];
				dd += dr*dr;
			}
			X2 += dd*n1->user->nneigh*n2->user->nneigh;
		NEXTNODE(n2)
	NEXTNODE(n1)
	X2 /= 9*nnode*(nnode-1);
/************************MAXNEIGH***************************/
	NNEIGH = 0;
	FORALLNODES(node)
		if(node->user->nneigh > NNEIGH)
			NNEIGH = node->user->nneigh;
	NEXTNODE(node)
}

double metro_delta(node, newr)
NODEPTR node;
double newr[MAXDIM];
{
	ELMTPTR elmt;
	NODEPTR noden;
	int i, idim, ineigh;
	double gaussdiff, areadiff, edgediff;
/*
		Copy existing node, neignbor elmts and neighbor nodes
		into temporary storage.
*/
	metro_victim = node;
	oldnode = *(node->user);
	ineigh = 0;
	FORALLNEIGH(node, elmt)
		oldelmtbuf[ineigh] = *(elmt->user);
		i = (indexc(elmt, node) + 1)%3;
		oldnodebuf[ineigh] = *(elmt->neigh[i]->user);
		ineigh++;
	NEXTNEIGH(node, elmt)
/*
		Copy proposed coordinates of victim node into
		existing structure.
*/
	for(idim=0; idim<dim; idim++){
		node->user->r[idim] = newr[idim];
	}
/*
		Get new data for neighboring elmts and nodes
*/
	FORALLNEIGH(node, elmt)
		get_elmt_data(elmt);

	NEXTNEIGH(node, elmt)
	get_node_data(node);
	FORALLNEIGH(node, elmt)
		i = (indexc(elmt, node) + 1)%3;
		noden = elmt->neigh[i];
		get_node_data(noden);
	NEXTNEIGH(node, elmt)
/*
		Calculate differences from current (proposed) configuration
		from stored (old) configuration;
*/
	ineigh = 0;
	gaussdiff = node->user->gauss_action - oldnode.gauss_action;
	areadiff = node->user->area_action - oldnode.area_action;
	edgediff = node->user->edge_action - oldnode.edge_action;
	FORALLNEIGH(node, elmt)
		i = (indexc(elmt, node) + 1)%3;
		noden = elmt->neigh[i];
		gaussdiff += noden->user->gauss_action - oldnodebuf[ineigh].gauss_action;
		areadiff += noden->user->area_action - oldnodebuf[ineigh].area_action;
		edgediff += noden->user->edge_action - oldnodebuf[ineigh].edge_action;
		ineigh++;
	NEXTNEIGH(node, elmt)

	return( gaussdiff + lambdaA*areadiff + lambdaE*edgediff);
}

metro_accept()
{
}

metro_reject()
{
	NODEPTR node;
	ELMTPTR elmt;
	int ineigh, i;

	node = metro_victim;
	*(node->user) = oldnode;
	ineigh = 0;
	FORALLNEIGH(node, elmt)
		*(elmt->user) = oldelmtbuf[ineigh];
		i = (indexc(elmt, node) + 1)%3;
		*(elmt->neigh[i]->user) = oldnodebuf[ineigh];
		ineigh++;
	NEXTNEIGH(node, elmt)
}

double flip_delta(elmt, i)
ELMTPTR elmt;
int i;
{
	NODENEIGHPTR indexnn(), remove();
	double gaussdiff, areadiff, edgediff, lnwdiff;

	ABC = elmt;
	B = ABC->neigh[i];
	C = ABC->neigh[(i+1)%3];
	A = ABC->neigh[(i+2)%3];

	ACD = indexnn(A, ABC)->next->elmt;
	D = ACD->neigh[(indexc(ACD, A) + 2)%3];

	if(A->user->nneigh <= 3 || C->user->nneigh <= 3
		|| B->user->nneigh >= MAXNEIGH || D->user->nneigh >= MAXNEIGH
		|| connected(B, D))
			return(LARGE);

	oldelmtbuf[0] = *(ABC->user);
	oldelmtbuf[1] = *(ACD->user);
	oldnodebuf[0] = *(A->user);
	oldnodebuf[1] = *(B->user);
	oldnodebuf[2] = *(C->user);
	oldnodebuf[3] = *(D->user);

	ABC->neigh[indexc(ABC, A)] = D;
	ACD->neigh[indexc(ACD, C)] = B;	
	insert(B, ABC, remove(C, ACD), ACD);
	insert(D, ACD, remove(A, ABC), ABC);

	get_elmt_data(ABC);
	get_elmt_data(ACD);
	get_node_data(A);
	get_node_data(B);
	get_node_data(C);
	get_node_data(D);

	gaussdiff = A->user->gauss_action - oldnodebuf[0].gauss_action;
	areadiff = A->user->area_action - oldnodebuf[0].area_action;
	edgediff = A->user->edge_action - oldnodebuf[0].edge_action;
	lnwdiff = log((double)A->user->nneigh/oldnodebuf[0].nneigh);

	gaussdiff += B->user->gauss_action - oldnodebuf[1].gauss_action;
	areadiff += B->user->area_action - oldnodebuf[1].area_action;
	edgediff += B->user->edge_action - oldnodebuf[1].edge_action;
	lnwdiff += log((double)B->user->nneigh/oldnodebuf[1].nneigh);

	gaussdiff += C->user->gauss_action - oldnodebuf[2].gauss_action;
	areadiff += C->user->area_action - oldnodebuf[2].area_action;
	edgediff += C->user->edge_action - oldnodebuf[2].edge_action;
	lnwdiff += log((double)C->user->nneigh/oldnodebuf[2].nneigh);

	gaussdiff += D->user->gauss_action - oldnodebuf[3].gauss_action;
	areadiff += D->user->area_action - oldnodebuf[3].area_action;
	edgediff += D->user->edge_action - oldnodebuf[3].edge_action;
	lnwdiff += log((double)D->user->nneigh/oldnodebuf[3].nneigh);

/*
printf(">>> %f %f %f %f  :   ",gaussdiff, 0.5*dim*lnwdiff, lambdaA*areadiff, lambdaE*edgediff);
*/
	return(gaussdiff + 0.5*dim*lnwdiff + lambdaA*areadiff + lambdaE*edgediff);

}

flip_accept()
{
}

flip_reject()
{
	NODENEIGHPTR indexnn(), remove();
	int i;

	ABC->neigh[indexc(ABC, D)] = A;
	ACD->neigh[indexc(ACD, B)] = C;	
	insert_before(C, ABC, remove(B, ACD), ACD);
	insert_before(A, ACD, remove(D, ABC), ABC);

	*(ABC->user) = oldelmtbuf[0];
	*(ACD->user) = oldelmtbuf[1];
	*(A->user) = oldnodebuf[0];
	*(B->user) = oldnodebuf[1];
	*(C->user) = oldnodebuf[2];
	*(D->user) = oldnodebuf[3];
}

add_node(elmt, i)
ELMTPTR elmt;
int i;
{
	NODEPTR A, B, C, D, E, node_alloc();
	ELMTPTR ABC, CDE, ACD, ABE, elmt_alloc();
	NODENEIGHPTR indexnn(), nodeneigh_alloc();
	int idim, old_abc_a, old_acd_c;

	ABC = elmt;
	B = ABC->neigh[i];
	C = ABC->neigh[(i+1)%3];
	A = ABC->neigh[(i+2)%3];

	ACD = indexnn(A, ABC)->next->elmt;
	D = ACD->neigh[(indexc(ACD, A) + 2)%3];

	old_abc_a = indexc(ABC, A);
	old_acd_c = indexc(ACD, C);

	E = node_alloc();
	ABE = elmt_alloc();
	CDE = elmt_alloc();

	ACD->neigh[old_acd_c] = E;
	indexnn(C, ACD)->elmt = CDE;

	ABC->neigh[old_abc_a] = E;
	indexnn(A, ABC)->elmt = ABE;

	ABE->neigh[0] = A;
	ABE->neigh[1] = B;
	ABE->neigh[2] = E;

	CDE->neigh[0] = C;
	CDE->neigh[1] = D;
	CDE->neigh[2] = E;

	insert(B, ABC, nodeneigh_alloc(), ABE);
	insert(D, ACD, nodeneigh_alloc(), CDE);

	add_neigh(ABC, E);
	add_neigh(CDE, E);
	add_neigh(ACD, E);
	add_neigh(ABE, E);
	sortout(E);

	for(idim=0; idim<dim; idim++)
		E->user->r[idim] = 0.5*(A->user->r[idim] + C->user->r[idim]);
}

check_mesh()
{
	NODEPTR node, p;
	ELMTPTR elmt, e;
	NODENEIGHPTR n, indexnn();
	int i, k;

	FORALLNODES(node)
		if(node->user->nneigh >= MAXNEIGH){
			printf("Too many neighbors\n");
			return(-1);
		}
		k = 0;
		FORALLNEIGH(node, elmt)
			if(!antik(n, n->next, node)){
				printf("mesh buggered\n");
				return(-1);
			}
			k++;
		NEXTNEIGH(node, elmt)
		if(k != node->user->nneigh){
			printf("nneigh buggered\n");
			return(-1);
		}
		if(node->user->gauss_action < 0.0){
			printf("gauss_action buggered\n");
			return(-1);
		}
		if(node->user->area_action < 0.0){
			printf("area_action buggered\n");
			return(-1);
		}
	NEXTNODE(node)

	FORALLELMTS(elmt)
		if(elmt->user->area < 0.0){
			printf("elmt area buggered\n");
			return(-1);
		}
		for(i=0; i<3; i++){
			n = indexnn(elmt->neigh[i], elmt);
			if(elmt->user->edge_sq[i] < 0.0){
				printf("elmt edge_sq buggered\n");
				return(-1);
			}
			if(elmt->user->alt_sq[i] < 0.0){
				printf("elmt alt_sq buggered\n");
				return(-1);
			}
		}
	NEXTELMT(elmt)
	return(0);
}

string_distort(node, ax, ay)
NODEPTR node;
double *ax, *ay;
{
	*ax = node->user->r[0];
	*ay = node->user->r[1];
}

double string_sqdist(n1, n2)
NODEPTR n1, n2;
{
	int idim;
	double dx, d = 0;
	for(idim=0; idim<dim; idim++){
		dx = n1->user->r[idim] - n2->user->r[idim];
		d += dx*dx;
	}
	return (d);
}

string_read(fp, type, ptr)
FILE *fp;
int type;
char *ptr;
{
    NODEPTR n;
    ELMTPTR e;
    BDYPTR b;
	int i;

    switch(type){
		case MISC:
			return;
        case ELMT:
            e = (ELMTPTR)ptr;
            return;
        case NODE:
            n = (NODEPTR)ptr;
#ifdef DEBUG
printf("(%d)", procnum);
#endif
			for(i=0; i<dim; i++){
				fscanf(fp, "%lf", &n->user->r[i]);
#ifdef DEBUG
printf(" %f ", n->user->r[i]);
#endif
			}
#ifdef DEBUG
printf("\n");
fflush(stdout);
#endif
            return;
        case BDY:
            b = (BDYPTR)ptr;
            return;
    }
}

string_write(fp, type, ptr)
FILE *fp;
int type;
char *ptr;
{
    NODEPTR n;
    ELMTPTR e;
    BDYPTR b;
	int i;

    switch(type){
		case MISC:
			fprintf(fp, "\n");
			return;
        case ELMT:
            e = (ELMTPTR)ptr;
			fprintf(fp, "\n");
            return;
        case NODE:
            n = (NODEPTR)ptr;
			for(i=0; i<dim; i++)
				fprintf(fp, "%f ", n->user->r[i]);
			fprintf(fp, "\n");
            return;
        case BDY:
            b = (BDYPTR)ptr;
			fprintf(fp, "\n");
            return;
    }
}


/* random numbers */

/*
#include <math.h>
*/
#include <fcntl.h>
#include <stdio.h>

#define	MULT	1103515245
#define	ADD 	12345
#define	MASK	( 0x7fffffff )
#define	TWOTO31	( (double)MASK + 1.0 )
long	AAA;	/* multiplier */
long	BBB;	/* summand    */
long	randx;	/* last value */

extern double lambdaE, lambdaA;
extern int nnode, dim;

pranset(_s_eed)
long   _s_eed;
{

	AAA = MULT & MASK ;
	BBB = ADD  & MASK ;

	randx = (AAA*_s_eed + BBB) & MASK;
	
}

double pranf()
/*
	Return a random double in [0, 1.0)
*/
{
	double retvalue;

	retvalue=randx / TWOTO31;
	randx = (AAA*randx + BBB)& MASK;
	return( retvalue );
}

setseed()
{
  FILE *fp,*fopen() ;
  int seed,mode,i ;
	char filename[50];
	if(lambdaE == 0.0)
		sprintf(filename, "N%d/seedA%4.3f_%d", nnode, lambdaA, dim);
	else
		sprintf(filename, "N%d/seedE%4.3f_%d", nnode, lambdaE, dim);

  printf("Select mode of seed setting: \n");
  printf("0  <->  by hand\n");
  printf("1  <->  read in from file called '%s'\n",filename);
  /*
  scanf("%d",&mode);
  */
  get_int(&mode);

  if ( mode == 0 ) {
    printf("Input seed ");
	/*
    scanf("%d",&seed);
    printf("%d\n",seed);
	*/
	get_int(&seed);
    pranset(seed);
  }  
  else {
    if( (fp = fopen(filename,"r")) == NULL) {
      printf("Couldn't find file called '%s'\n",filename) ;
      exit(1);
    }
    fscanf(fp,"%d",&seed);
    fclose(fp) ; 
    printf("%d\n",seed);
    pranset(seed);
  }
  printf("seed set\n");
  return(0) ;
}

dmpseed()
{
  FILE *fp,*fopen() ;
	char filename[50];
	if(lambdaE == 0.0)
		sprintf(filename, "N%d/seedA%4.3f_%d", nnode, lambdaA, dim);
	else
		sprintf(filename, "N%d/seedE%4.3f_%d", nnode, lambdaE, dim);

  fp = fopen(filename,"w") ;
    fprintf(fp,"%d",randx);
  fclose(fp) ;
  printf("Dumped randx to file called '%s'\n",filename) ;
  return(0) ;
}


/* Need to add this because plotix expects it */
_ex_error(){}
