#include <stdio.h>
#define MINSIZE			8
#define MAXSIZE			400
#define BLOCKSIZE		4000
#define BITSPERINT		32
#define MAXINTSPERBLOCK	(BLOCKSIZE/(BITSPERINT*MINSIZE) + 1)
#define BNULL			(BLOCKPTR)0
#define MAXTYPES		12
extern char *malloc();

typedef struct block {
	unsigned int vacant[MAXINTSPERBLOCK];
	struct block *next;
/* MAXINTSPERBLOCK is 16 so this declaration makes number of ints before data 18 rather than 17 */
int align_dataBLOCK;
	char data[BLOCKSIZE];
} BLOCK, *BLOCKPTR;

BLOCKPTR block_alloc();
BLOCKPTR	curr_block,	known_blk_ptrs[MAXTYPES];;
int			curr_iv,	iv[MAXTYPES];
int			curr_ib,	ib[MAXTYPES];
int			curr_size,	known_sizes[MAXTYPES];
int			curr_flags,	known_flags[MAXTYPES];
int			knowntypes = 0;

unsigned int mask[BITSPERINT];

int init_smalloc(size)
int size;
{
	int ibs;

#ifdef EBUG
printf("size = %d\t", size);
#endif
/* Change size to be a multiple of 8 (if it isn't already) */
	if (size%8) size = size + (8 - (size%8));
#ifdef EBUG
printf("-> %d<<<\n", size);
#endif

	if(size < MINSIZE){
		fprintf(stderr, 
		"smalloc: size is %d: less than minimum size %d\n",
		size, MINSIZE);
		abort(8005);
	}
	for(ibs=0; ibs<BITSPERINT; ibs++)
		mask[ibs] = 1<<ibs;
	curr_size = known_sizes[knowntypes] = size;
	curr_block = BNULL;
	iv[knowntypes] = ib[knowntypes] = 0;
	if(size < MAXSIZE){
		curr_flags = known_flags[knowntypes] = BLOCKSIZE/(BITSPERINT*size) + 1;
		if((known_blk_ptrs[knowntypes] = block_alloc()) == BNULL)
			return(-1);
	} else {
		known_flags[knowntypes] = 0;
	}
	knowntypes++;
	return(knowntypes-1);
}

char *smalloc(type)
int type;
{
	BLOCKPTR start, block_alloc();
	int ibs, ivs;
	char *ptr;

	curr_size =				known_sizes[type];

	if(curr_size >= MAXSIZE){
#ifdef EBUG
		printf("malloc %d\n", curr_size);
#endif
		ptr = malloc(curr_size);
		if(ptr == NULL){
			fprintf(stderr, "malloc failed: size %d", curr_size);
			abort(999);
		}
		known_flags[type]++;
		return(ptr);
	}
	curr_block = start =	known_blk_ptrs[type];
	curr_iv =				iv[type];
	curr_ib =				ib[type];
	curr_flags =			known_flags[type];

	do {
		ivs = curr_iv;
		do {
			if(curr_block->vacant[curr_iv] != 0){
				ibs = curr_ib;
				do {
					if((curr_block->vacant[curr_iv]) & (mask[curr_ib]))
						goto foundit;
				} while((curr_ib = (curr_ib+1)%BITSPERINT) != ibs);
				fprintf(stderr, "error in smalloc: size %d", curr_size);
				abort(801);
			}
		} while((curr_iv = (curr_iv+1)%curr_flags) != ivs);
	} while((curr_block = curr_block->next) != start);

	curr_block = block_alloc();

foundit:
	iv[type] = curr_iv;
	ib[type] = curr_ib;
	curr_block->vacant[curr_iv] &= ~mask[curr_ib];
	ptr = curr_block->data + curr_size*(BITSPERINT*curr_iv + curr_ib);
	return(ptr);
}

BLOCKPTR block_alloc()
{
	BLOCKPTR new;
	int iv, ib;

#ifdef EBUG
	printf("malloc %d\n", sizeof(BLOCK));
#endif
	new = (BLOCKPTR)malloc(sizeof(BLOCK));
	if(new == BNULL){
#ifdef CUBIX
		fprintf(stdsys, "block_alloc failed %d bytes in use\n", s_inventory());
#endif
		abort(s_inventory()/1000);
	}

	for(iv=0; iv<curr_flags; iv++){
		new->vacant[iv] = 0;
		for(ib=0; ib < BITSPERINT; ib++){
			if((iv*BITSPERINT + ib + 1)*curr_size > BLOCKSIZE)
				continue;
			new->vacant[iv] |= mask[ib];
		}
	}

	if(curr_block == BNULL){
		curr_block = new;
		curr_block->next = curr_block;
	} else {
		new->next = curr_block->next;
		curr_block->next = new;
	}
	curr_iv = curr_ib = 0;
	return(new);
}

sfree(type, ptr)
int type;
char *ptr;
{
	BLOCKPTR start;
	int iv, ib, posit;
	curr_block = start = known_blk_ptrs[type];
	curr_size = known_sizes[type];
	if(curr_size >= MAXSIZE){
		free(ptr);
		known_flags[type]--;
		return;
	}
	if(start == BNULL)
		goto fail;
	do {
		posit = ptr - curr_block->data;
		if(posit >= 0 && posit < BLOCKSIZE){
			posit /= curr_size;
			iv = posit/BITSPERINT;
			ib = posit%BITSPERINT;
			if(curr_block->vacant[iv] & mask[ib]){
				abort(802);
			}
			curr_block->vacant[iv] |= mask[ib];
			return;
		}
	} while((curr_block = curr_block->next) != start);
fail:
#ifdef CUBIX
	fprintf(stdsys, "sfree: not mine: type %d size %d\n", type, curr_size);
#endif
	abort(803);
}

int s_inventory()
{
	
	BLOCKPTR start, loop;
	int type, nblock = 0, mallocked = 0;
	for(type=0; type<knowntypes; type++){
		if(known_sizes[type] >= MAXSIZE){
			mallocked += known_sizes[type] * known_flags[type];
		} else {
			start = loop = known_blk_ptrs[type];
			do {
				nblock++;
			} while((loop = loop->next) != start);
		}
	}
	return(nblock*sizeof(BLOCK) + mallocked);
}
