#include "dime.h" 
#include "string.h"

/* If ALPHA is defined then program reads in alpha; otherwise its 0.5*dim */
#define ALPHA

extern double pranf();
#define RAN		((double)pranf())

#define SWEEPSPERWRITE 100

#define BINSIZE        100	/* SWEEPSPERPRINT */

int nnn = 1;
double bin_ave[NOBS+2];
double bin_ave2[NOBS+2];
double tot_ave[NOBS+2];
double tot_ave2[NOBS+2];
	double act[NOBS+2];
	double act2[NOBS+2];

extern MNUPTR root;
int dim = 3;
extern int nnode;
extern int procnum;
long nproc = 1;
double lambdaA, lambdaE, tot_action, hit_size = 0.5;
double metro_success, flip_success;
double alpha = 1.5;
FILE *out = NULL;

main(argc, argv)
int argc;
char **argv;
{
	MNUPTR mymenu;
	int reset(), ascii_print_action(), sweeps();
	int singmult(), dmpseed();
	int outfile(), add_nodes(), check_mesh(), read_string();

	int string_distort(), string_read(), string_write();
	double string_sqdist();

#ifdef CUBIX
	struct cubenv cube;
	cparam(&cube);
	procnum = cube.procnum;
	nproc = cube.nproc;
#endif

pranset(42);

	DIME_SETUP(argc, argv, MULTI);
	user_read = string_read;
	user_write = string_write;
	distort = string_distort;
	sqdist = string_sqdist;

	mymenu = menu_make(BRANCH);
	add_item(root, "STRING", BRANCH, mymenu);
	add_item(mymenu, "RESET", LEAF, reset);
	add_item(mymenu, "READ_STRING", LEAF, read_string);
	add_item(mymenu, "OUTFILE", LEAF, outfile);
	add_item(mymenu, "ADD_NODES", LEAF, add_nodes);
	add_item(mymenu, "CHECK_MESH", LEAF, check_mesh);
	add_item(mymenu, "SWEEPS", LEAF, sweeps);
	add_item(mymenu, "ACTION", LEAF, ascii_print_action);
	add_item(mymenu, "SINGMULT", LEAF, singmult);
	add_item(mymenu, "DMPSEED", LEAF, dmpseed);

	menu_input();

}

singmult()
{
	if(mesh_is_distributed = 1-mesh_is_distributed)
		printf("Mesh is MULTI\n");
	else
		printf("Mesh is SINGL\n");
}

read_string()
{
	char string[100];
	double lambda, action();
	int i=0, islash, ibar=0;
	FILE *in;

    printf("Input Filename (without suffix)?\n");
    get_string(string);

	islash = -1;
	while(string[i] != '\0'){
		if(string[i] == '/')
			islash = i;
		i++;
	}
	while(string[ibar++] != '_');
	sscanf(string+ibar, "%d", &dim);
	sscanf(string+islash+6, "%lf", &lambda);
	if(string[islash+5] == 'A'){
		lambdaA = lambda;
		lambdaE = 0.0;
	} else {
		lambdaA = 0.0;
		lambdaE = lambda;
	}
	printf("Dim = %d, lambda%c = %f\n", dim, string[islash+5], lambda);
    strcat(string, ".ct");

    if((in = fopen(string, "r")) == (FILE *)NULL){
        printf("Cannot open %s\n", string);
		exit(0);
	}
    else
        readct(in);

	fill_user_structs();
	tot_action = action();
}

write_string()
{
	char filename[100];
	FILE *write;

	if(lambdaE == 0.0)
		sprintf(filename, "N%d/MeshA%4.3f_%d.ct", nnode, lambdaA, dim);
	else
		sprintf(filename, "N%d/MeshE%4.3f_%d.ct", nnode, lambdaE, dim);

	printf("Writing mesh to %s\n", filename);
	if((write = fopen(filename, "w")) == NULL){
		perror(filename);
		return;
	} else
		writect(write);
}

read_aves(mes_done)
int *mes_done;
{
    char filename[100];
    FILE *in;

    if(lambdaE == 0.0)
	sprintf(filename, "N%d/avesA%4.3f_%d.ct", nnode, lambdaA, dim);
    else
	sprintf(filename, "N%d/avesE%4.3f_%d.ct", nnode, lambdaE, dim);

    printf("Reading aves from %s\n", filename);
    if((in = fopen(filename, "r")) == (FILE *)NULL){
        printf("Cannot open aves %s\n", filename);
	exit(0);
    } else {
#ifdef CUBIX
	fmulti(in);
#endif
	fread(mes_done, sizeof(int), 1, in); 
	fread(tot_ave, sizeof(double), NOBS+2, in); 
	fread(tot_ave2, sizeof(double), NOBS+2, in); 
	fclose(in);
    }
    printf("Sweeps so far %d\n", *mes_done);
    fflush(stdout);
}

write_aves(mes_done)
int *mes_done;
{
	char filename[100];
	FILE *write;

	if(lambdaE == 0.0)
	    sprintf(filename, "N%d/avesA%4.3f_%d.ct", nnode, lambdaA, dim);
	else
	    sprintf(filename, "N%d/avesE%4.3f_%d.ct", nnode, lambdaE, dim);

	printf("Writing aves to %s\n", filename);
	if((write = fopen(filename, "w")) == NULL){
	    perror(filename);
	    return;
	} else {
#ifdef CUBIX
	    fmulti(write);
#endif
	    fwrite(mes_done, sizeof(int), 1, write); 
	    fflush(write);
	    fwrite(tot_ave, sizeof(double), NOBS+2, write); 
	    fflush(write);
	    fwrite(tot_ave2, sizeof(double), NOBS+2, write); 
	    fclose(write);
	}
}

reset()
{
	double action();

	fill_user_structs();

	printf("Dimension (current = %d)? ", dim); fflush(stdout);
	get_int(&dim);

#ifdef ALPHA
	printf("Alpha (current = %f)? ", alpha); fflush(stdout);
	get_double(&alpha);
#endif

	printf("Area action coefficient? "); fflush(stdout);
	get_double(&lambdaA);

	printf("Edge action coefficient? "); fflush(stdout);
	get_double(&lambdaE);

	printf("Hit Size? "); fflush(stdout);
	get_double(&hit_size);

	tot_action = action();

setseed();
}

outfile()
{
	char filename[100];

	if(out != NULL)
		fclose(out);

	if(lambdaE == 0.0)
		sprintf(filename, "N%d/A%4.3f_%d", nnode, lambdaA, dim);
	else
		sprintf(filename, "N%d/E%4.3f_%d", nnode, lambdaE, dim);

	/*
	printf("Sending data to %s\n", filename);
	*/

	if((out = fopen(filename, "w")) == NULL){
		perror(filename);
	}
#ifdef CUBIX
	fmulti(out);
#endif
}

sweeps()
{
	int sweep, sweeps, mes_steps, mes_done, mes_i, i;
	int first_time;
	double metro_move(), flip_move();

	printf("How many sweeps per measurement? "); fflush(stdout);
	get_int(&sweeps);
	printf("How many measurements\n" ); fflush(stdout);
	get_int(&mes_steps);
	printf("First Time ???\n" ); fflush(stdout);
	get_int(&first_time);
	printf("BINSIZE = %d\n\n",BINSIZE); fflush(stdout);
	for(i=0; i<NOBS+2; i++) {
		bin_ave[i] = 0.;
		bin_ave2[i] = 0.;
	}
if (first_time) {
	mes_done = 0;
	for(i=0; i<NOBS+2; i++) {
		tot_ave[i] = 0.;
		tot_ave2[i] = 0.;
	}
} else {
	read_aves(&mes_done);
}

	for(mes_i=0; mes_i<mes_steps; mes_i++){
		metro_success = flip_success = 0;

		for(sweep=0; sweep<sweeps; sweep++){
			metro_success += metro_move();
			flip_success  += flip_move();
		}

		metro_success /= sweeps;
		flip_success  /= sweeps;

	avg_measure(act, act2);
	for(i=0; i<NOBS+2; i++) {
		bin_ave[i] += act[i];
		bin_ave2[i] += act2[i];
		tot_ave[i] += act[i];
		tot_ave2[i] += act2[i];
	}

	    if(mes_done % BINSIZE == BINSIZE - 1) {
		for(i=0; i<NOBS+2; i++) {
			bin_ave[i] /= (double)BINSIZE;
			bin_ave2[i] /= (double)BINSIZE;
		}
		nnn = mes_done+1;
		if(out != NULL)
			binary_print_action();
		ascii_print_action();
		for(i=0; i<NOBS+2; i++) {
			bin_ave[i] = 0.;
			bin_ave2[i] = 0.;
		}
	    }

		if(mes_done % SWEEPSPERWRITE == SWEEPSPERWRITE - 1) {
			write_string();
			nnn = mes_done+1;
			write_aves(&nnn);
		}
	mes_done++;
	}

#ifndef CUBIX
	nnn = mes_done;
	if(out != NULL)
		binary_print_action();
	ascii_print_action();
		printf("Accumulated action %f\n", tot_action);
#endif
	if(out != NULL) {
		write_string();
		write_aves(&mes_done);
	}
}

double metro_move()
{
	NODEPTR node;
	double newr[MAXDIM];
	double metro_delta();
	double dact;
	int idim, try=0, succeed=0;

	FORALLNODES(node)
		for(idim=0; idim<dim; idim++)
			newr[idim] = node->user->r[idim] + hit_size*(RAN-0.5);

		dact = metro_delta(node, newr);
		try++;
		if(dact < 0.0 || exp(-dact) > RAN){
			tot_action += dact;
			metro_accept();
			succeed++;
		} else
			metro_reject();
	NEXTNODE(node)
	return((double)succeed/try);
}

double flip_move()
{
	ELMTPTR elmt;
	int i, try=0, succeed=0;
	double flip_delta();
	double dact;

	FORALLELMTS(elmt)
		i = 3*RAN;
		dact = flip_delta(elmt, i);
		try++;
/*
printf("||| %f  :   ",dact);
*/
		if(dact > 1.e20)
			continue;
		if(dact < 0.0 || exp(-dact) > RAN){
			tot_action += dact; 
			flip_accept();
			succeed++;
		} else 
			flip_reject();
	NEXTELMT(elmt)
	return((double)succeed/try);
}

double action()
{
	double act[NOBS];
	measure(act);
#ifdef ALPHA
	return(GAUSS - alpha*LN_W + lambdaA*AREA_CURVE + lambdaE*EDGE_CURVE);
#else
	return(GAUSS - 0.5*dim*LN_W + lambdaA*AREA_CURVE + lambdaE*EDGE_CURVE);
#endif
}

ascii_print_action()
{
	int i;
	if (nnn == 1) avg_measure(act, act2);
#ifdef CUBIX
	fmulti(stdout);
#endif
	/*
	if(procnum == 0) {
	*/
		printf("Proc %d (%d):\n",procnum,nnn);
		/*
		printf("Gauss, Ln W               %10.5f %10.5f\n", GAUSS, LN_W);
		printf("Curve, Area Edge          %10.5f %10.5f\n", AREA_CURVE, EDGE_CURVE);
		printf("x2                        %10.5f\n", X2);
		printf("Success, Metro Flip       %10.5f %10.5f\n", metro_success, flip_success);
		printf("Max neigh                 %10.5f\n", NNEIGH);
		printf("Total Action              %10.5f\n", 
#ifdef ALPHA
GAUSS - alpha*LN_W + lambdaA*AREA_CURVE + lambdaE*EDGE_CURVE);
#else
GAUSS - 0.5*dim*LN_W + lambdaA*AREA_CURVE + lambdaE*EDGE_CURVE);
#endif
		printf("Square Gauss, Ln W        %10.5f %10.5f\n", GAUSS2, LN_W2);
		printf("Square Curve, Area Edge   %10.5f %10.5f\n", AREA_CURVE2, EDGE_CURVE2);
		printf("Square x2                 %10.5f\n", X22);
		*/

		printf("%10.3f %10.3f\n", BINGAUSS, BINLN_W);
		printf("%10.3f %10.3f\n", BINAREA_CURVE, BINEDGE_CURVE);
		printf("%10.3f\n", BINX2);
		printf("%10.3f %10.3f\n", BINGAUSS2, BINLN_W2);
		printf("%10.3f %10.3f\n", BINAREA_CURVE2, BINEDGE_CURVE2);
		printf("%10.3f\n", BINX22);

		printf("%10.3f %10.3f\n", AVEGAUSS/nnn, AVELN_W/nnn);
		printf("%10.3f %10.3f\n", AVEAREA_CURVE/nnn, AVEEDGE_CURVE/nnn);
		printf("%10.3f\n", AVEX2/nnn);
		printf("%10.3f %10.3f\n", AVEGAUSS2/nnn, AVELN_W2/nnn);
		printf("%10.3f %10.3f\n", AVEAREA_CURVE2/nnn, AVEEDGE_CURVE2/nnn);
		printf("%10.3f\n\n", AVEX22/nnn);
	/*
	}
	*/
#ifdef CUBIX
	fsingl(stdout);
#endif
	fflush(stdout);
}

void _ex_swaw(from, to, N)
    char *from, *to;
    int N;
{
    char q[4];
    int i;
	
    for(i=0 ; i<N/4 ; i++) {
	q[0] = *(from+3); q[1] = *(from+2);  q[2] = *(from+1); q[3] = *from;
	memcpy(to, q, 4);
	from += 4;
	to += 4;
    }
}

binary_print_action()
{
/*
file out contains records with fields: TOT AVE
							0. Gaussian action
							1. Log W
							2. Area curvature
							3. Edge Curvature
							4. x2
							5. Maximum # neighbors
							6. metropolis success rate
							7. flip success rate
							+ squares of all these
*/
	float temp[NOBS+2];
	char *ptr;
	int i;
	long itemp;
	if(out == NULL)
		return;

	/* close and open file - i.e. rewind */
	outfile();

	/* write number of processors i.e. bins */
	if(procnum == 0) {
		itemp = nproc;
#ifdef CUBIX /* swap the f**king bytes */
#ifndef A2010 /* not on Ametek */
	_ex_swaw(&itemp,&itemp,4);
#endif
#endif
		fwrite(&itemp, sizeof(long), 1, out); 
	}

	for(i=0; i<NOBS+2; i++)
		temp[i] = tot_ave[i]/nnn;
#ifdef CUBIX /* swap the f**king bytes */
#ifndef A2010 /* not on Ametek */
	_ex_swaw(temp,temp,(NOBS+2)*4);
#endif
#endif
	fwrite(temp, sizeof(float), NOBS+2, out); 

	for(i=0; i<NOBS+2; i++)
		temp[i] = tot_ave2[i]/nnn;
#ifdef CUBIX /* swap the f**king bytes */
#ifndef A2010 /* not on Ametek */
	_ex_swaw(temp,temp,(NOBS+2)*4);
#endif
#endif
	fwrite(temp, sizeof(float), NOBS+2, out); 
	fflush(out);
}

avg_measure(act, act2)
double act[NOBS+2];
double act2[NOBS+2];
{
	int i;
	measure(act);
	act[NOBS] = metro_success;
	act[NOBS+1] = flip_success;
	for(i=0; i<NOBS+2; i++)
		act2[i] = act[i]*act[i];

/* DO NOT COMBINE !!!
#ifdef CUBIX
	combine(act, fsum, sizeof(double), NOBS-1);
	combine(act+NOBS-1, fmax, sizeof(double), 1);
	combine(act+NOBS, fsum, sizeof(double), 2);

	for(i=0; i<NOBS+2; i++)
		if(i != NOBS-1)
			act[i] /= nproc;

	combine(act2, fsum, sizeof(double), NOBS-1);
	combine(act2+NOBS-1, fmax, sizeof(double), 1);
	combine(act2+NOBS, fsum, sizeof(double), 2);

	for(i=0; i<NOBS+2; i++)
		if(i != NOBS-1)
			act2[i] /= nproc;
#endif
*/

}

add_nodes()
{
	NODEPTR node;
	ELMTPTR elmt;
	int addem;

	nnode = 0;
	FORALLNODES(node)
		nnode++;
	NEXTNODE(node);
	printf("Currently %d nodes; add how many?", nnode);
	fflush(stdout);
	get_int(&addem);

	FORALLELMTS(elmt)
		if(addem-- == 0)
			break;
		add_node(elmt, 0);
	NEXTELMT(elmt)

	nnode = 0;
	FORALLNODES(node)
		nnode++;
	NEXTNODE(node);
	printf("Now %d nodes\n", nnode);
}
