/*********
* _panel *               Roy Williams  1986
**********

Routines to draw panels, polygons etc.........
*/
#include <stdio.h>
#include "plotix.h"
#include "sys/plotix.h"
#include "xplot.h"

static unsigned char hatch[]={
    0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* SUN hatch 1 */
    0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0,  /* SUN hatch 2 */
    0x81, 0x03, 0x06, 0x0c, 0x18, 0x30, 0x60, 0xc0,
    0x81, 0xc0, 0x60, 0x30, 0x18, 0x0c, 0x06, 0x03,
    0x18, 0x30, 0x60, 0xc0, 0x81, 0xc0, 0x60, 0x30,
    0x08, 0x00, 0x00, 0x80, 0xc1, 0x62, 0x36, 0x1c,
    0xf0, 0xf0, 0xf0, 0xf0, 0x0f, 0x0f, 0x0f, 0x0f,
    0x01, 0x01, 0x01, 0xff, 0x10, 0x10, 0x10, 0xff,
    0x3c, 0x66, 0xc3, 0x81, 0x81, 0xc3, 0x66, 0x3c,
    0x88, 0x88, 0x88, 0xff, 0x88, 0x88, 0x88, 0xff   /* SUN hatch 10 */
};

#ifndef MAXVERTS
#define MAXVERTS        (1000)
#endif
#define NHATCH          (sizeof(hatch)/8)
static int sv_shade, sv_edge;
static int Npts;
static int polyverts[MAXVERTS*2], *polys;
extern int _g_grey_hi,_g_grey_lo,_g_rain_hi,_g_rain_lo;
static XPoint polyarray[MAXVERTS];

void _initpanel(shade,edge)
int shade,edge;
{
    sv_shade = shade;
    sv_edge  = edge;
    polys = polyverts;
    Npts = 0;
    return;
}

void _panelpoint(x, y)
Coord x, y;
{
    int x_curr, y_curr;
    char text[80];

    x_curr = (int) USER_to_X(x);
    y_curr = (int) USER_to_Y(y);

    if(++Npts > MAXVERTS) {
        sprintf(text,"Too many polygon vertices, max = %d", MAXVERTS);
        if (_ex_error(text))
            exit(1);
        Npts = 0;
        return;
    }
    *polys++ = x_curr;
    *polys++ = y_curr;

    return;
}

void _endpanel()
{
    int i, shade, index;
    char text[80];
    Pixmap stipple=NULL;

    /* First we have to close the polygon, if necessary */

    if(*(polys-2) != polyverts[0] || *(polys-1) != polyverts[1]) {
        if(++Npts > MAXVERTS) {
            sprintf(text,"Too many polygon vertices, max = %d", MAXVERTS);
            if (_ex_error(text)) exit(1);
            Npts = 0;
            return;
        }
        *polys++ = polyverts[0];        /* Close polygon for filler */
        *polys++ = polyverts[1];
    }

    /* Now select the proper fill pattern and color */
    if (_g_mono) sv_shade *= -1;
    if(sv_shade < 0) {
        index = -sv_shade-1;
        shade = (index - (index % NHATCH))/NHATCH;
        shade %= 7;
        shade++;
	while (shade > 7) shade -= 7;
	stipple = XCreateBitmapFromData(display,RootWindow(display,screen), 
		&(hatch[8*(index%NHATCH)]),8,8);
	XSetStipple(display,gc,stipple);
	XSetForeground(display,gc,_g_cmap[shade]);
	XSetFillStyle(display,gc,FillOpaqueStippled);
    }
    else {
        if ( _g_grey_lo <= sv_shade && sv_shade <= _g_grey_hi) 
	    XSetForeground(display,gc,_g_cmap[sv_shade]);
        else if (_g_rain_lo <= sv_shade && sv_shade <= _g_rain_hi) 
	    XSetForeground(display,gc,_g_cmap[sv_shade]);
        else      /* Undefined color --> map into original 8 */
            XSetForeground(display,gc,_g_cmap[sv_shade % 8]); 
    }

    /* convert formats */
    for (i = 0; i < Npts; i++) {
        polyarray[i].x = polyverts[i << 1];
        polyarray[i].y = polyverts[(i << 1) + 1];
    }
    XFillPolygon(display,_g_win,gc,polyarray,Npts,Complex,CoordModeOrigin);
    if (stipple != NULL) {
	XFreePixmap(display,stipple);
	XSetFillStyle(display,gc,FillSolid);
    }
    XSetForeground(display,gc,_g_cmap[_g_fgcolor]);
    if(sv_edge) 
	XDrawLines(display,_g_win,gc,polyarray,Npts,CoordModeOrigin);
    Npts = 0;
}


void polgn(num,x,y,color,edge)
int num,color,edge ;
Coord *x,*y ;
{
    int i;
    initpanel(color, edge);
    for(i=0; i<num; i++) panelpoint(x[i], y[i]);
    endpanel();
}

void box(x0, y0, x1, y1, shade, edge)
int shade, edge;
Coord x0, y0, x1, y1;
{
    initpanel(shade, edge);
    panelpoint(x0, y0);
    panelpoint(x0, y1);
    panelpoint(x1, y1);
    panelpoint(x1, y0);
    endpanel();
}
