/*********
* openpl *                Wendell T. Hicken Sept 1989
**********

This routine serves several purposes.
i)      All declarations of variables takes place here.
ii)     This routine opens an output file and copies in the header file
*/
#include <stdio.h>
#include "plotix.h"
#include "sys/plotix.h"
#include "xplot.h"
extern int _g_grey_lo, _g_grey_hi;
extern int _g_rain_lo, _g_rain_hi;
int _g_plotix = 0;
#define MAX (unsigned int) 65535

int _ex_error() {}
int _ex_fatal() {}

int openpl(bufsiz, grafp)
int bufsiz;
FILE *grafp;
{
    char *display_name = NULL;
    int display_width,display_height;
    XSizeHints size_hints;
    XGCValues values;
    unsigned char map[3];
    int depth;

    if (!_g_plotix) {
	if ( (display = XOpenDisplay(display_name)) == NULL ) {
	    return -1;
	}
	screen = DefaultScreen(display);
	display_width  = DisplayWidth(display,screen);
	display_height = DisplayHeight(display,screen);
	if(_g_wishx == 0) _g_wishx = 2*display_width/3;
	if(_g_wishy == 0) _g_wishy = display_height/2;
	_g_win = XCreateSimpleWindow(display,RootWindow(display,screen),
		0,0,_g_wishx,_g_wishy,1,
		BlackPixel(display,screen),
		WhitePixel(display,screen) );	
	size_hints.flags = PPosition | PSize | PMinSize;
/* Center on display */
	if(_g_atx == 0) size_hints.x = _g_wishx/6;
	else            size_hints.x = _g_atx;
	if(_g_aty == 0) size_hints.y = _g_wishy/4;
	else            size_hints.y = _g_aty;

	size_hints.width = _g_wishx;
	size_hints.height = _g_wishy;
	size_hints.min_width = 350;
	size_hints.min_height = 250;
	XSetStandardProperties(display,_g_win,_g_winlab,_g_winlab,
		None, NULL, 0, &size_hints);
	XSelectInput(display,_g_win,ExposureMask | KeyPressMask |
		ButtonPressMask | StructureNotifyMask );
	_g_font = XLoadQueryFont(display,"8x13");
	if (_g_font == NULL) 
	    return -1;
	gc = XCreateGC(display,_g_win,0L,&values);
	XSetFont(display,gc,_g_font->fid);
	XSetForeground(display,gc,BlackPixel(display,screen));
	XSetBackground(display,gc,WhitePixel(display,screen));
	XMapWindow(display,_g_win);
	_g_cursor = XCreateFontCursor(display,132);
	XDefineCursor(display,_g_win,_g_cursor);
        _g_mouse = XGetPointerMapping(display,map,3);  /* Get # of buttons */	
	depth = DisplayPlanes(display,screen);
	if (depth == 1) _g_mono = 1;
	_g_xcmap = DefaultColormap(display,screen);
	setcmap(MAX,MAX,MAX,0);		/* White (background) */
	setcmap(  0,  0,  0,1);	        /* Black (foreground) */
	setcmap(MAX,  0,  0,2);	        /* red */
	setcmap(  0,MAX,  0,3);	        /* green */
	setcmap(  0,  0,MAX,4);	        /* blue */
	setcmap(  0,MAX,MAX,5);	        /* cyan */
	setcmap(MAX,  0,MAX,6);	        /* magenta */
	setcmap(MAX,MAX,  0,7);	        /* yellow */
	_g_plotix = 1;
	while(_g_devx == 0 || _g_devy == 0)
	    doEvent(1);
        vport((Coord) 0,(Coord) 0, (Coord) 1, (Coord) 1);
        space((Coord) 0,(Coord) 0, (Coord) 1, (Coord) 1);
        box(0.,0.,1.,1.,0,0);   /* clear device */
    }
    return 1;
}

void greyscale(from, to)
int from, to;
{
    int i;
    unsigned int rgb;

    if (to < from) {
	i = from; from = to; to = i;
    }
    if (to > 255) to = 255;
    _g_grey_lo = from;
    _g_grey_hi = to;
    for (i = from; i < to; i++) {
        rgb = (unsigned int) ((double) (MAX * (i - _g_grey_lo))/ (double)(_g_grey_hi - _g_grey_lo));
	setcmap(rgb,rgb,rgb,i);
    }
    return;
}

void rainbow(from, to)
int from, to;
{
    int i,j;
    unsigned int r,g,b;
    double h,f,n;

    if (to < from) {
	i = from; from = to; to = i;
    }
    if (to > 255) to = 255;
    _g_rain_lo = from;
    _g_rain_hi = to;

    for (j = from; j < to; j++) {
        h = (double) (_g_rain_hi - j) / (_g_rain_hi - _g_rain_lo);
        if (h > .999) h = 0.;
        h *= 6.;
        i = (int) h;
        f = h - (double) i;
		f -= 2*f*(f-.5)*(f-1);
        n = 1. - f;
        switch(i) {
        case 0: r = MAX; g=(unsigned int) (f * MAX); b = 0; break;
        case 1: r = (unsigned int) (n * MAX); g = MAX; b = 0; break;
        case 2: r = 0; g = MAX; b = (unsigned int) (f * MAX); break;
        case 3: r = 0; g = (unsigned int) (n * MAX); b = MAX; break;
        case 4: r = (unsigned int) (f * MAX); g = 0; b = MAX; break;
        case 5: r = MAX; g = 0; b=(unsigned int) (n * MAX); break;
        }
	setcmap(r,g,b,j);
    }
    return;
}

setcmap(r,g,b,pos)
    unsigned int r,g,b;
    int pos;
{
    XColor colorcelldef;

    colorcelldef.red = r;
    colorcelldef.blue = b;
    colorcelldef.green = g;
    colorcelldef.flags = DoRed | DoGreen | DoBlue;
    if (XAllocColor(display,_g_xcmap,&colorcelldef) == 0) 
	_g_cmap[pos] = pos;
    else
        _g_cmap[pos] = colorcelldef.pixel;
}
