/********
* space *
*********

*/
#include "plotix.h"
#include "sys/plotix.h"
#include "xplot.h"

static int vpinited = 0;        /* Flag: Have we set the vports ? */
static int nvport = 0;          /* Number of active veiewports */
extern Coord _g_V0[2],_g_V1[2],_g_W0[2],_g_W1[2];
extern int _g_clip;

void _vpinit()
{
    int i;

    for(i=0; i<NVPORT; i++) {
        vptab[i].V0[0] = _g_V0[0];
        vptab[i].V0[1] = _g_V0[1];
        vptab[i].V1[0] = _g_V1[0];
        vptab[i].V1[1] = _g_V1[1];
        vptab[i].W0[0] = _g_W0[0];
        vptab[i].W0[1] = _g_W0[1];
        vptab[i].W1[0] = _g_W1[0];
        vptab[i].W1[1] = _g_W1[1];
        vptab[i].clip = 0;
    }
    vpinited = 1;
}

/**************
* ortho_space * Roy Williams
***************
Given a user rectangle, finds a larger rectangle which has no
circle-to-ellipse distortion and centers the user's rectangle
within it.
*/
void ortho_space(x0, y0, x1, y1, position)
Coord x0, y0, x1, y1;
int position;
{
    Coord xp0, yp0, xp1, yp1;
    double portasp, size, surfasp, xtemp, ytemp;

    if(!vpinited) _vpinit();
    aspect(&xtemp, &ytemp);
    surfasp = xtemp/ytemp;

    portasp = (surfasp*(vptab[cvport].V1[0]-vptab[cvport].V0[0]))/
        (vptab[cvport].V1[1]-vptab[cvport].V0[1]);
    if(portasp*(y1-y0) > (x1-x0)){
        size = (y1-y0)*portasp;
        if(position == 0){
            xp0 = 0.5*(x0+x1) - 0.5*size;
            xp1 = 0.5*(x0+x1) + 0.5*size;
        }
        else if(position < 0){
            xp0 = x0;
            xp1 = xp0 + size;
        }
        else {
            xp1 = x1;
            xp0 = xp1 - size;
        }
        yp0 = y0;
        yp1 = y1;
    }
    else {
        xp0 = x0;
        xp1 = x1;
        size = (x1-x0)/portasp;
        if(position == 0){
            yp0 = 0.5*(y0+y1) - 0.5*size;
            yp1 = 0.5*(y0+y1) + 0.5*size;
        } 
        else if(position < 0){
            yp0 = y0;
            yp1 = yp0 + size;
        }
        else {
            yp1 = y1;
            yp0 = yp1 - size;
        }
    }
    space(xp0, yp0, xp1, yp1);
}

/********
* space *               J.W.Flower
*********

Sets the world coordinates defined by the entire viewport
*/
void space(x0,y0,x1,y1)
Coord x0,y0,x1,y1;
{
    if(!vpinited) _vpinit();

    vptab[cvport].W0[0] = x0;
    vptab[cvport].W1[0] = x1;
    vptab[cvport].W0[1] = y0;
    vptab[cvport].W1[1] = y1;

    _g_mul[0] = _g_devx*(vptab[cvport].V1[0]-vptab[cvport].V0[0])/
        (vptab[cvport].W1[0]-vptab[cvport].W0[0]);
    _g_mul[1] = _g_devy*(vptab[cvport].V1[1]-vptab[cvport].V0[1])/
        (vptab[cvport].W1[1]-vptab[cvport].W0[1]);

    _g_off[0] = _g_devx*(vptab[cvport].W1[0]*vptab[cvport].V0[0]-
        vptab[cvport].W0[0]*vptab[cvport].V1[0])/
        (vptab[cvport].W1[0]-vptab[cvport].W0[0]);
    _g_off[1] = _g_devy*(vptab[cvport].W1[1]*vptab[cvport].V0[1]-
        vptab[cvport].W0[1]*vptab[cvport].V1[1])/
        (vptab[cvport].W1[1]-vptab[cvport].W0[1]);
    return;
}

/********
* vport *               J.W.Flower
*********

Sets the viewport in coords relative to the whole screen (0 to 1).
*/

int vport(x0,y0,x1,y1)
Coord x0,x1,y0,y1;
{
    int portnum;

    if(!vpinited) _vpinit();

    if(nvport >= NVPORT) return -1;

    portnum = nvport++;

    vptab[portnum].V0[0] = x0;
    vptab[portnum].V1[0] = x1;
    vptab[portnum].V0[1] = y0;
    vptab[portnum].V1[1] = y1;

    vptab[portnum].clip = 0;

    setvport(portnum);               /* Make new viewport current */
    return cvport; /* same as portnum after setvport() */
}

/***********
* setvport *
************

Switch to a new viewport. Index numbers are returned by vport.
*/
int setvport(vpnum)
int vpnum;
{
    if(!vpinited) _vpinit();

    if(vpnum < 0 || vpnum >= nvport) return -1;

    /* Save the relevant information about the current vport */
    vptab[cvport].clip = _g_clip;   /* Save old clipping state */

    /* Now switch to the new vport */

    cvport = vpnum;
    _g_clip = vptab[cvport].clip;

    _g_mul[0] = _g_devx*(vptab[cvport].V1[0]-vptab[cvport].V0[0])/
        (vptab[cvport].W1[0]-vptab[cvport].W0[0]);
    _g_mul[1] = _g_devy*(vptab[cvport].V1[1]-vptab[cvport].V0[1])/
        (vptab[cvport].W1[1]-vptab[cvport].W0[1]);

    _g_off[0] = _g_devx*(vptab[cvport].W1[0]*vptab[cvport].V0[0]-
        vptab[cvport].W0[0]*vptab[cvport].V1[0])/
        (vptab[cvport].W1[0]-vptab[cvport].W0[0]);
    _g_off[1] = _g_devy*(vptab[cvport].W1[1]*vptab[cvport].V0[1]-
        vptab[cvport].W0[1]*vptab[cvport].V1[1])/
        (vptab[cvport].W1[1]-vptab[cvport].W0[1]);

    if(vptab[cvport].clip) {
        setclip(vptab[cvport].clipwin[0], vptab[cvport].clipwin[1],
        vptab[cvport].clipwin[2], vptab[cvport].clipwin[3]);
    }
    else endclip();
    setupclipping();
    return 0;
}

int _inwindow(x, y)
Coord x, y;
{
    if(!vpinited) _vpinit();

    if(x < vptab[cvport].W0[0]) return 0;
    if(x > vptab[cvport].W1[0]) return 0;
    if(y < vptab[cvport].W0[1]) return 0;
    if(y > vptab[cvport].W1[1]) return 0;
    return 1;
}

int clipvp()
/* Set Clipping rectangle to be current viewport limits */
/* February 1990 */
{
	if(!vpinited) _vpinit();

	setclip(
		vptab[cvport].W0[0],
		vptab[cvport].W0[1],
		vptab[cvport].W1[0],
		vptab[cvport].W1[1]
	);
	return 1;
}

/* This function sets up the clipping region for XWindows */
setupclipping() {
    int l,b,r,t; /* Clipping coordinates left,bottom,right,top */
    XRectangle rect[1];

    if (_g_clip) {
        l = (int) USER_to_X(vptab[cvport].clipwin[0]);
        b = (int) USER_to_Y(vptab[cvport].clipwin[1]);
        r = (int) USER_to_X(vptab[cvport].clipwin[2]);
        t = (int) USER_to_Y(vptab[cvport].clipwin[3]);
    }
    else {
        l = (int) (vptab[cvport].V0[0] * _g_devx); /* Convert */
        b = (int) ( ((Coord) 1 - vptab[cvport].V0[1]) * _g_devy);
        r = (int) (vptab[cvport].V1[0] * _g_devx);
        t = (int) ( ((Coord) 1 - vptab[cvport].V1[1]) * _g_devy);
    }
    rect[0].x = 0; rect[0].y = 0;
    rect[0].width = (r - l + 1);
    rect[0].height = (b - t + 1);
    XSetClipRectangles(display,gc,l,t,rect,1,YXSorted);
}

#if _3L
kortho(x0, y0, x1, y1, position)
float *x0, *y0, *x1, *y1;
int *position;
{
    ortho_space((Coord)*x0, (Coord)*y0, (Coord)*x1, (Coord)*y1, *position);
}

kclpvp()
{
    clipvp();
}

kspac(x0,y0,x1,y1)
float *x0, *y0, *x1, *y1;
{
    space((Coord)*x0, (Coord)*y0, (Coord)*x1, (Coord)*y1);
}

kvport(x0,y0,x1,y1)
float *x0, *y0, *x1, *y1;
{
    vport((Coord)*x0, (Coord)*y0, (Coord)*x1, (Coord)*y1);
}

ksetvp(vpnum)
int *vpnum;
{
    setvport(*vpnum);
}
#endif                  /* _3L */
