#include "errno.h" 
#include "stdio.h" 
#ifdef CUBIX 
#include "cros.h" 
#else
#define stdsys stderr
#endif

extern int debug;
 
#define IN      0 
#define OUT     1 
#define CURRENT 0 
#define OLD     1 
#define IN_BUFSIZ  10000 
#define OUT_BUFSIZ 10000 
#define ANY         999 
 
double hwm_in=0.0, hwm_out=0.0; 
int cr_doc, cr_procnum, cr_nproc, cr_error, in_count, first_call=1; 
int in_bytes, out_bytes, in_buf_size, out_buf_size; 
int com_bytes, com_buf_size; 
char *com_start_ptr, *com_buf_ptr; 
char *in_buf_ptr, *out_buf_ptr, *in_start_ptr, *out_start_ptr; 
char in_default_buf[IN_BUFSIZ]; 
char out_default_buf[OUT_BUFSIZ]; 
int isize, header_size; 

int set_cr_doc(new_doc)
int new_doc;
{
	cr_doc = new_doc;
	cr_nproc = 1 << cr_doc;
}
 
 
int add_to_out_buf(buf,nbytes,ndest,dest) 
char *buf; 
int nbytes,ndest,*dest; 
{ 
	int valid_dest,destination,i,resp,temp_bytes,message_size; 
	double tempx; 
 
	resp = 0; 
 
	if(first_call) cr_init(); 
 
	if (nbytes<0){ 
		set_bit(2); 
		return -1; 
	} 
	if(ndest<0){ 
		set_bit(0); 
		return -1; 
	} 
 
	valid_dest = ndest; 
	for(i=0;i<ndest;++i){ 
		destination = dest[i]; 
		if(destination<0 || destination>(cr_nproc-1)){ 
			set_bit(1); 
			resp = -1; 
			valid_dest--; 
		} 
	} 
	if(valid_dest==0) return 0; 
 
	header_size = (valid_dest + 4)*isize; 
	message_size = nbytes + header_size; 
	temp_bytes = out_bytes + message_size; 
	if(temp_bytes<=out_buf_size){ 
		*((int *)out_buf_ptr)           = valid_dest; 
		*((int *)(out_buf_ptr+isize))   = cr_procnum; 
		*((int *)(out_buf_ptr+2*isize)) = nbytes; 
		*((int *)(out_buf_ptr+3*isize)) = CURRENT; 
		out_buf_ptr += 4*isize; 
		for(i=0;i<valid_dest;++i){ 
			destination = dest[i]; 
			if(destination>=0 && destination<cr_nproc){ 
				*((int *)out_buf_ptr) = destination; 
				out_buf_ptr += isize; 
			} 
		} 
		for(i=0;i<nbytes;++i) *out_buf_ptr++ = *buf++; 
		out_bytes = temp_bytes; 
		tempx = ((double)out_bytes)/((double)out_buf_size); 
		hwm_out = (tempx > hwm_out) ? tempx : hwm_out; 
		return resp; 
	} 
	else{ 
		hwm_out   = 1.0; 
		set_bit(3); 
		return -1; 
	} 
} 
 
/*********************** end of routine add_to_out_buf **************/ 
 
int add_to_com_buf(buf,nbytes) 
char *buf; 
int nbytes; 
{ 
	int i,temp_bytes; 
	double tempx; 
 
	temp_bytes = com_bytes + nbytes; 
	if(temp_bytes<=com_buf_size){ 
		for(i=0;i<nbytes;++i) *com_buf_ptr++ = *buf++; 
		com_bytes = temp_bytes; 
		tempx = (double)(out_bytes+com_bytes)/((double)out_buf_size); 
		hwm_out = (tempx > hwm_out) ? tempx : hwm_out; 
		return 0; 
	} 
	else{ 
		hwm_out = 1.0; 
		set_bit(3); 
		return -1; 
	} 
} 
 
/*********************** end of routine add_to_com_buf ********************/ 
 
int compress_out_buf() 
{ 
	int old_bytes; 
	char *compress(); 
 
	old_bytes    = out_bytes; 
 
	out_buf_ptr = compress(out_start_ptr,&out_bytes); 
 
	return (old_bytes - out_bytes); 
 
} 
 
/********************** end of routine compress_out_buf *************/ 
 
int compress_in_buf() 
{ 
	int old_bytes; 
	char *compress(); 
 
	old_bytes  = in_bytes; 
 
	in_buf_ptr = compress(in_start_ptr,&in_bytes); 
 
	return (old_bytes - in_bytes); 
 
} 
 
/********************** end of routine compress_in_buf *************/ 
 
char *compress(start_ptr,total_bytes) 
int *total_bytes; 
char *start_ptr; 
{ 
	int j,ndest,nbytes,message_size,status,tot_bytes; 
	int bytes; 
	char *buf_ptr,*cur_buf_ptr; 
 
	buf_ptr = cur_buf_ptr = start_ptr; 
	tot_bytes     = 0; 
	bytes         = 0; 
	while(bytes< *total_bytes){ 
		ndest  = *((int *)buf_ptr); 
		nbytes = *((int *)(buf_ptr+2*isize)); 
		status = *((int *)(buf_ptr+3*isize)); 
		message_size = nbytes + (4+ndest)*isize; 
		if(status==CURRENT){ 
			tot_bytes += message_size; 
			if(cur_buf_ptr != buf_ptr) 
				for(j=0;j<message_size;++j) 
					*cur_buf_ptr++ = *buf_ptr++; 
			else{ 
				cur_buf_ptr += message_size; 
				buf_ptr     += message_size; 
			} 
		} 
		else buf_ptr += message_size; 
		bytes += message_size; 
	} 
 
	*total_bytes = tot_bytes; 
 
	return cur_buf_ptr; 
} 
 
/************************ end of routine compress **************************/ 
 
int check_buffer(dim) 
int dim; 
{ 
	int p4,ndest,status,bit,mask,message_size,dest,nbytes,resp,bytes; 
	int cur_dest,i,*dest_ptr,sent,c_bytes,s4,err; 
	char *start_ptr,*c_start; 
 
	mask = 1<<dim; 
	bit  = cr_procnum&mask; 
	s4   = 4*isize; 
 
	com_start_ptr = out_buf_ptr; 
	out_buf_ptr   = out_start_ptr; 
	com_buf_size  = out_buf_size - out_bytes; 
	com_buf_ptr   = com_start_ptr; 
	com_bytes     = 0; 
 
	err   = 0; 
 
	bytes = 0; 
	while(bytes<out_bytes){ 
		resp   = 0; 
		ndest  = *((int *)out_buf_ptr); 
		nbytes = *((int *)(out_buf_ptr+2*isize)); 
		status = *((int *)(out_buf_ptr+3*isize)); 
		header_size  = (4+ndest)*isize; 
		message_size = nbytes + header_size; 
		if(status==OLD){ 
			out_buf_ptr += message_size; 
			bytes       += message_size; 
			continue; 
		} 
		dest_ptr = (int *)(out_buf_ptr+4*isize); 
		start_ptr= out_buf_ptr; 
		cur_dest = ndest; 
		sent     = 0; 
		c_bytes  = com_bytes; 
		c_start  = com_buf_ptr; 
		for(i=0;i<ndest;++i){ 
			dest = *dest_ptr++; 
			if(dest<0) cur_dest--; 
			else if((dest&mask)^bit){  
				p4 = (4+i)*isize; 
				if(!sent){ 
					resp = add_to_com_buf(out_buf_ptr,s4); 
					err += resp; 
					if(resp<0) break; 
				} 
				resp = add_to_com_buf(&dest,isize); 
				err += resp; 
				if(resp==0){ 
					cur_dest--; 
					*((int *)(start_ptr+p4)) = -99;  
					sent++; 
				} 
				else break; 
			} 
		} 
		if(!cur_dest) *((int *)(start_ptr+3*isize)) = OLD; 
		if(sent && resp==0){ 
			resp = add_to_com_buf(out_buf_ptr+header_size,nbytes); 
			err += resp; 
			*((int *)c_start) = sent; 
		} 
		if(resp != 0){ 
			start_ptr += 4*isize; 
			for(i=0;i<sent;++i){ 
				dest = *((int *)(c_start+(4+i)*isize)); 
				while(*((int *)start_ptr) != -99) 
					start_ptr += isize; 
				*((int *)start_ptr) = dest; 
			} 
			com_bytes   = c_bytes; 
			com_buf_ptr = c_start; 
		} 
		out_buf_ptr += message_size; 
		bytes       += message_size; 
	} 
	return (err<0 ? -1 : 0); 
} 
 
/*********************** end of routine check_buffer *********************/ 
 
int crystal_router() 
{ 
	int rec_bytes,d,resp,chan,read_bytes; 
	double tempx; 
 
	if(first_call) cr_init(); 
 
	resp = 0; 
 
	compress_in_buf(); 
	resp += copy_to_in_buf(out_start_ptr,out_bytes); 
 
 
	for(d=0;d<cr_doc;++d){ 
		chan = 1<<d; 
		if(d>0) resp += compress_out_buf(); 
		resp += check_buffer(d); 
		com_buf_size = out_buf_size - out_bytes; 
		read_bytes = cshift(com_start_ptr,chan,com_buf_size, 
				    com_start_ptr,chan,com_bytes); 
		if(read_bytes<0){ 
			hwm_out = 1.0; 
			rec_bytes = rescue_buf(com_start_ptr,com_buf_size); 
			resp = -1; 
			set_bit(4); 
		} 
		else{ 
		  tempx = ((double)(out_bytes+read_bytes))/((double)out_buf_size); 
		  hwm_out = (tempx > hwm_out) ? tempx:hwm_out; 
			 rec_bytes = read_bytes; 
		} 
		resp += copy_to_in_buf(com_start_ptr,rec_bytes); 
		out_bytes += rec_bytes; 
		out_buf_ptr = out_start_ptr + out_bytes; 
	} 
	in_count = in_bytes; 
 
	compress_out_buf(); 
 
	return (resp<0 ? -1 : 0); 
} 
 
/*********************** end of routine crystal_router ***************/ 
 
int rescue_buf(start_ptr,max_bytes) 
char *start_ptr; 
int max_bytes; 
{ 
	int nbytes,ndest,temp,bytes; 
	char *buf_ptr; 
 
	buf_ptr = start_ptr; 
	bytes = 0; 
	while(bytes<max_bytes){ 
		temp = bytes + 4*isize; 
		if(temp>max_bytes) break; 
		ndest  = *((int *)buf_ptr); 
		nbytes = *((int *)(buf_ptr+2*isize)); 
		temp  += nbytes + ndest*isize; 
		if(temp>max_bytes) break; 
		bytes = temp; 
		buf_ptr += nbytes + (ndest+4)*isize; 
	} 
 
	return bytes; 
} 
 
/************************ end of routine rescue_buf ****************/ 
 
int copy_to_in_buf(start_ptr,total_bytes) 
char *start_ptr; 
int total_bytes; 
{ 
	int temp_bytes,cur_dest,ndest,i,dest,nbytes,status,message_size,bytes; 
	int j,p4,cpy_size,resp,source,*dest_ptr; 
	double tempx; 
	char *cpy_ptr,*buf_ptr; 
 
	resp = 0; 
 
	buf_ptr = start_ptr; 
 
	bytes    = 0; 
	while(bytes<total_bytes){ 
		ndest  = *((int *)buf_ptr); 
		source = *((int *)(buf_ptr+isize)); 
		nbytes = *((int *)(buf_ptr+2*isize)); 
		status = *((int *)(buf_ptr+3*isize)); 
		header_size  = ( 4+ndest )*isize; 
		message_size = nbytes + header_size; 
		if(status==OLD){ 
			buf_ptr += message_size; 
			bytes   += message_size; 
			continue; 
		} 
		cpy_size = 5*isize + nbytes; 
		cur_dest = ndest; 
		dest_ptr = (int *)(buf_ptr + 4*isize); 
		for(i=0;i<ndest;++i){ 
			cpy_ptr = buf_ptr + header_size; 
			dest = *dest_ptr++; 
			if(dest==cr_procnum){  
		  		temp_bytes = in_bytes + cpy_size; 
		  		if(temp_bytes<=in_buf_size){ 
					p4 = (4+i)*isize; 
					*((int *)in_buf_ptr)          = 1; 
					*((int *)(in_buf_ptr+isize))  = source; 
					*((int *)(in_buf_ptr+2*isize))= nbytes; 
					*((int *)(in_buf_ptr+3*isize))= status; 
					*((int *)(in_buf_ptr+4*isize))= dest; 
					in_buf_ptr += 5*isize; 
		    			for(j=0;j<nbytes;++j)  
				 		*in_buf_ptr++ = *cpy_ptr++; 
		    			*((int *)(buf_ptr+p4)) = -99; 
					cur_dest--; 
		    			in_bytes = temp_bytes; 
				 tempx = ((double)in_bytes)/((double)in_buf_size); 
				 hwm_in = (tempx>hwm_in) ? tempx : hwm_in; 
		  		} 
		  		else{ 
					hwm_in = 1.0; 
		    			set_bit(6); 
					resp = -1; 
		  		} 
			} 
			else if(dest<0) cur_dest--; 
		} 
		if(!cur_dest) *((int *)(buf_ptr+3*isize)) = OLD; 
		bytes   += message_size; 
		buf_ptr += message_size; 
	} 
 
	return resp; 
} 
 
/********************* end of routine copy_to_in_buf **************/ 
 
int error_messages(stream) 
FILE *stream; 
{ 
	int i,bit,multi_on_entry; 
 
	if(cr_error==0) return 0; 
 
	for(i=0;i<11;++i){ 
		bit = (cr_error>>i)&1; 
		if(bit){ 
	fprintf(stdsys,"\n\n**** Crystal_router Error in processor %d ****\n", 
			cr_procnum); 
			switch(i){ 
				case 0: 
   fprintf(stdsys," Invalid number of destination processors passed to add_to_out_buf\n"); 
   fprintf(stdsys," Check arguments to add_to_out_buf\n"); 
					break; 
				case 1: 
   fprintf(stdsys," Invalid destination processor passed to add_to_out_buf\n"); 
   fprintf(stdsys," Check arguments to add_to_out_buf\n"); 
					break; 
				case 2: 
   fprintf(stdsys," Negative number of nbytes passed to add_to_out_buf\n"); 
   fprintf(stdsys," Check arguments to add_to_out_buf\n"); 
					break; 
				case 3: 
   fprintf(stdsys," Outgoing message buffer has overflowed while queueing\n"); 
   fprintf(stdsys," Use set_cr_buf to increase size of buffer\n"); 
					break; 
				case 4: 
   fprintf(stdsys," Communications buffer has become full while copying\n"); 
   fprintf(stdsys," Use set_cr_buf to increase size of buffer\n"); 
					break; 
				case 5: 
   fprintf(stdsys," Communications buffer has become full while reading\n"); 
   fprintf(stdsys," Use set_cr_buf to increase size of buffer\n"); 
					break; 
				case 6: 
   fprintf(stdsys," Received message buffer has overflowed\n"); 
   fprintf(stdsys," Use set_cr_buf to increase size of buffer\n"); 
					break; 
				case 7: 
   fprintf(stdsys," Unable to allocate storage for outgoing message buffer\n"); 
   fprintf(stdsys," Use set_cr_buf to decrease size of buffer\n"); 
					break; 
				case 8: 
   fprintf(stdsys," Unable to allocate storage for received message buffer\n"); 
   fprintf(stdsys," Use set_cr_buf to decrease size of buffer\n"); 
					break; 
				case 9: 
   fprintf(stdsys," Unable to copy all the messages to the new outgoing message buffer\n"); 
   fprintf(stdsys," Use set_cr_buf to increase size of buffer\n"); 
					break; 
				case 10: 
   fprintf(stdsys," Unable to copy all the messages to the new received message buffer\n"); 
   fprintf(stdsys," Use set_cr_buf to increase size of buffer\n"); 
					break; 
			} 
		} 
	} 
	cr_error = 0; 
	return 0; 
} 
 
/************************ end of routine error_messages ***************/ 
 
int get_message(buf,source,nbytes) 
char *buf; 
int *source,*nbytes; 
{ 
	int ncur,stemp,message_size,src,bytes,found,i,nb,dest,status; 
 
	if(first_call) cr_init(); 
 
	if(in_count>=in_bytes){ 
		in_count = 0; 
		in_buf_ptr = in_start_ptr; 
	} 
 
	if(in_bytes==0) return 1; 
 
	bytes = 0; 
	src = *source; 
	ncur  = 0; 
	found = 0; 
	header_size = 5*isize; 
	while(bytes<in_bytes && !found){ 
		stemp  = *((int *)(in_buf_ptr+isize)); 
		nb     = *((int *)(in_buf_ptr+2*isize)); 
		status = *((int *)(in_buf_ptr+3*isize)); 
		dest   = *((int *)(in_buf_ptr+4*isize)); 
		message_size = nb + header_size; 
		if(status==CURRENT){ 
		  ncur++; 
		  if(dest==cr_procnum && (src==stemp || src==ANY)) found = 1; 
		} 
		if(!found){ 
			bytes += message_size; 
			in_buf_ptr += message_size; 
			in_count += message_size; 
			if(in_count>=in_bytes) { 
				in_count = 0; 
				in_buf_ptr  = in_start_ptr; 
			} 
		} 
	} 
	if(!ncur) return 3; 
	if(!found) return 2; 
 
	*source = stemp; 
	*nbytes = nb; 
	*((int *)(in_buf_ptr+3*isize)) = OLD; 
	in_buf_ptr += header_size; 
	for(i=0;i<nb;++i) *buf++ = *in_buf_ptr++; 
	in_count += message_size; 
 
	return 0; 
} 
 
/******************** end of routine get_message ************************/ 
 
cr_rewind() 
{ 
	int bytes = 0, nb, *mess_stat; 
	char *cptr; 
	cptr = in_start_ptr; 
	while(bytes<in_bytes){ 
		nb     = *((int *)(cptr+2*isize)); 
		mess_stat = ((int *)(cptr+3*isize)); 
		if(*mess_stat == OLD) *mess_stat = CURRENT; 
		cptr  += nb+header_size; 
		bytes += nb+header_size; 
	} 
} 
 
 
/******************** end of routine rewind************************/ 
int set_cr_buf(buf_size,in_out) 
int buf_size,in_out; 
{ 
	char *new_ptr,*malloc(); 
	int i,resp; 
 
	if(first_call) cr_init(); 
 
	resp = 0; 
 
#ifdef EBUG
	printf("malloc %d\n", buf_size);
#endif
	new_ptr = malloc(buf_size); 
	if(in_out==OUT){ 
		if(new_ptr==(char *)NULL){ 
		   set_bit(7); 
		   return -1; 
		} 
		else{ 
		   if(out_bytes>buf_size){ 
		       set_bit(9); 
		       resp = -1; 
		       out_bytes= rescue_buf(out_start_ptr,buf_size); 
		   } 
		   out_buf_ptr = new_ptr; 
		   for(i=0;i<out_bytes;++i)  
			*out_buf_ptr++ = *out_start_ptr++; 
		   if(out_start_ptr != out_default_buf) free(out_start_ptr); 
		   out_start_ptr = new_ptr; 
		   out_buf_size = buf_size; 
		} 
	} 
	else if(in_out==IN){ 
		if(new_ptr==(char *)NULL){ 
		   set_bit(8); 
		   return -1; 
		} 
		else{ 
		   if(in_bytes>buf_size){ 
		       set_bit(10); 
		       resp = -1; 
		       in_bytes= rescue_buf(in_start_ptr,buf_size); 
		       if(in_count>buf_size) in_count = 0; 
		   } 
		   in_buf_ptr = new_ptr; 
		   for(i=0;i<in_bytes;++i)  
			*in_buf_ptr++ = *in_start_ptr++; 
		   if(in_start_ptr != in_default_buf) free(in_start_ptr); 
		   in_start_ptr = new_ptr; 
		   in_buf_size  = buf_size; 
		} 
	} 
 
	return resp; 
} 
 
/*********************** end of routine set_cr_buf *********************/ 
 
set_bit(n) 
int n; 
{ 
	if(!((cr_error>>n)&1)) cr_error += 1<<n; 
} 
 
/********************* end of routine set_bit *************************/ 
	 
cr_init() 
{ 
#ifdef CUBIX
	struct cubenv crenv; 
	cparam(&crenv); 
	cr_doc   = crenv.doc; 
	cr_procnum = crenv.procnum; 
	cr_nproc = crenv.nproc; 
#else
	cr_doc = cr_procnum = 0;
	cr_nproc = 1;
#endif
 
	out_start_ptr = out_default_buf; 
	in_start_ptr  = in_default_buf; 
	out_buf_size  = OUT_BUFSIZ; 
	in_buf_size   = IN_BUFSIZ; 
	isize         = sizeof(int); 
	out_buf_ptr   = out_start_ptr; 
	in_buf_ptr    = in_start_ptr; 
	out_bytes     = 0; 
	in_bytes      = 0; 
	in_count      = 0; 
	cr_error      = 0; 
	first_call    = 0; 
 
} 
 
/************************ end of routine cr_init **********************/ 
 
int clear_cr_buf(in_or_out) 
int in_or_out; 
{ 
	if(first_call) cr_init(); 
 
	if(in_or_out==IN){ 
		in_buf_ptr = in_start_ptr; 
		in_bytes   = 0; 
		in_count   = 0; 
	} 
	else if(in_or_out==OUT){ 
		out_buf_ptr = out_start_ptr; 
		out_bytes   = 0; 
	} 
 
	return 0; 
} 
 
/********************* end of routine clear_buf ***********************/ 
 
double show_buffer_usage(stream,in_out) 
FILE *stream; 
int in_out; 
{ 
	double usage; 
 
	if (in_out==IN)      usage = hwm_in; 
        else if(in_out==OUT) usage = hwm_out; 
        else                 usage = -1.0; 
 
	if(stream != (FILE *)NULL){ 
	   fmulti(stream); 
	   fprintf(stream,"\n"); 
	   if (in_out==IN)  
	      fprintf(stream,"\nInput buffer usage for processor %d : %7.3f", 
				cr_procnum,usage); 
	   else if(in_out==OUT) 
	      fprintf(stream,"\nOutput buffer usage for processor %d : %7.3f", 
				cr_procnum,usage); 
	   else  
       	     fprintf(stream,"\nInvalid buffer ID passed to show_buffer_usage"); 
	   fprintf(stream,"\n"); 
	   fsingl(stream); 
	} 
	return usage; 
} 
 
/********************** end of routine show_buffer_usage *******************/ 
 
int zero_hwm(in_out) 
int in_out; 
{ 
	if (in_out==IN){ 
		hwm_in = 0.0; 
		return 0; 
	} else if(in_out==OUT){ 
		hwm_out = 0.0; 
		return 0; 
	} else    return -1; 
} 
 
/*********************** end of routine zero_hwm **************************/ 
