/****************************************************/
/*    DIME: grafix.c                                */
/*    Various graphics functions for dime.          */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/
#define DIMECODE
#include "dime.h"
#define 	COLOR(k)	((k+7)%7 + 1)
#define xold		work_space.nf[0]		/* For picture() */
#define yold		work_space.nf[1]

FILE *ps_file;
int plotting = 0;
int graphics = 0;
extern double z_xmin, z_ymin, z_xmax, z_ymax;

plotmesh()
/*
	Draw the edges of each elmt, but only do each edge once.
*/
{
	ELMTPTR this;
	NODEPTR n0, n1, node;
	double x, y, x1, y1;
	int in, bedge;
	fmulti(stdout);
	color(4);
/*
	Do internal edges
*/
	FORALLELMTS(this)
		for(in=0; in<3; in++){
			n0 = this->neigh[in];
			n1 = this->neigh[(in+1)%3];
			bedge = (n0->b != (BDYPTR)NULL) && (n1->b != (BDYPTR)NULL) && 
					((n0->b->clock == n1) || (n1->b->clock == n0));
			if(!bedge && (long)n0 < (long)n1){
				DISTORT(n0, x, y);
				move(x, y);
				DISTORT(n1, x, y);
				cont(x, y);
			}
		}
	NEXTELMT(this)
/*
	Do boundaries, with real boundaries and processor boundaries different
*/
	FORALLNODES(node)
		if(node->b != (BDYPTR)NULL){
			DISTORT(node, x, y);
			if(node->b->oppalias == (ALIASPTR)NULL)
				color(3);
			else if(node->b->oppalias->proc & HIBITS)
				color(2);
			else
				color(5);
			move(x, y);
			DISTORT(node->b->clock, x1, y1);
			cont(x1, y1);
		}
	NEXTNODE(node)
	if(mesh_is_distributed)
		usendplot();
	else
		sendplot();
	fsingl(stdout);
}

drawelmt(this, shade)
/*
	Using color 'shade', draw the elmt 'this' as a filled triangle a little
	smaller than the whole elmt.
*/
ELMTPTR this;
int shade;
{
	int i;
	double x[3], y[3], xc, yc;
	for(i=0; i<3; i++)
		DISTORT(this->neigh[i], x[i], y[i]);
	xc = (x[0] + x[1] + x[2])/3;
	yc = (y[0] + y[1] + y[2])/3;
	initpanel(COLOR(shade), 0);
	for(i=0; i<3; i++){
		panelpoint(0.2*xc + 0.8*x[i], 0.2*yc + 0.8*y[i]);
	}
	endpanel();
}

distortold(node, axold, ayold)
NODEPTR node;
double *axold, *ayold;
{
	*axold = node->xold;
	*ayold = node->yold;
}

drawoldelmt(this, shade)
ELMTPTR this;
int shade;
{
	int i, distortold(), (*keep)();
	double x[3], y[3], xc, yc;
	keep = distort;
	distort = distortold;
	for(i=0; i<3; i++)
		DISTORT(this->neigh[i], x[i], y[i]);
	xc = (x[0] + x[1] + x[2])/3;
	yc = (y[0] + y[1] + y[2])/3;
	initpanel(COLOR(shade), 0);
	for(i=0; i<3; i++){
		panelpoint(0.5*xc + 0.5*x[i], 0.5*yc + 0.5*y[i]);
	}
	endpanel();
	distort = keep;
}

drawsmallelmt(this, shade)
ELMTPTR this;
int shade;
{
	int i;
	double x[3], y[3], xc, yc;
	for(i=0; i<3; i++)
		DISTORT(this->neigh[i], x[i], y[i]);
	xc = (x[0] + x[1] + x[2])/3;
	yc = (y[0] + y[1] + y[2])/3;
	initpanel(COLOR(shade), 0);
	for(i=0; i<3; i++){
		panelpoint(0.75*xc + 0.25*x[i], 0.75*yc + 0.25*y[i]);
	}
	endpanel();
}

draw(A, B)
NODEPTR A, B;
/*
	Draw a line from node A to node B with current color
*/
{
	double x, y;
	DISTORT(A, x, y);
	move(x, y);
	DISTORT(B, x, y);
	cont(x, y);
}

debug_toggle()
{
	if(debug)
		printf("Debug Off\n");
	else
		printf("Debug On\n");
	debug = 1-debug;
}

postscript()
{
#define HDSIZ	256
	char string[30], hdbuf[HDSIZ];
	int n;
	FILE *pshead;
	if(plotting){
		fflush(ps_file);
		fsingl(ps_file);
		fprintf(ps_file, "s\nshowpage\n");
		fclose(ps_file);
		printf("Plot finished\n"); fflush(stdout);
	} else {
		if((pshead = fopen(PSHEAD , "r")) == (FILE *)NULL){
			printf("Cannot find Postscript header %s\n", PSHEAD);
			fflush(stdout);
		}

		do {
			printf("Postscript file?\n"); fflush(stdout);
			fsingl(stdin);
			get_string(string);
		} while((ps_file = fopen(string, "w")) == (FILE *)NULL);
/* Cubix2 doesn't need this
		setbuffer(ps_file, psbuf, PS_BUF);
*/
		if(pshead != NULL){
			while((n = fread(hdbuf, 1, HDSIZ, pshead)) != 0)
					  fwrite(hdbuf, 1, n, ps_file);
			fclose(pshead);
		}
		fprintf(ps_file, "%f %f %f z\n", z_xmin, z_ymin, z_xmax-z_xmin);
		fflush(ps_file);
		fmulti(ps_file);
	}
	plotting = !plotting;
	return(1);
}

balance()
{
#define NEWPROC		work_space.ei[0]
#ifdef CUBIX
	int send_elmts(), *nextelmt();
	static int first = 1;
	ELMTPTR elmt;
	if(!mesh_is_distributed)
		return;
	FORALLELMTS(elmt)
		elmt->NEWPROC = procnum;
	NEXTELMT(elmt)
	orb(send_elmts, nextelmt, first);
	first = 0;
#endif
	return(1);
}

int *nextelmt(ax, ay, aflag)
double *ax, *ay;
int *aflag;
{
	static ELMTPTR elmt;
	double x[3], y[3];
	int i;
	if(*aflag){
		elmt = the_elmts;
		if(elmt == (ELMTPTR)NULL)
			return((int *)NULL);
	}
	if(*aflag == 0 && (elmt = elmt->next) == the_elmts)
		return((int *)NULL);
	*aflag = 0;
	for(i=0; i<3; i++)
		DISTORT(elmt->neigh[i], x[i], y[i]);
	*ax = (x[0] + x[1] + x[2])/3;
	*ay = (y[0] + y[1] + y[2])/3;
	return(&(elmt->NEWPROC));
}

intact_refine()
{
	double tol, xmin, ymin, xmax, ymax, dx, dy, xc, yc;
	int keyy, yes, i;
	ELMTPTR elmt;
	printf("Max length?");fflush(stdout);
	fsingl(stdin);
	get_double(&tol);
	printf("Click lower left corner\n"); fflush(stdout);
	get_gin(&keyy, &xmin, &ymin);
	printf("Click upper right corner\n"); fflush(stdout);
	get_gin(&keyy, &xmax, &ymax);
	rect_refine(tol, xmin, ymin, xmax, ymax);
}

rect_refine(tol, xmin, ymin, xmax, ymax)
double tol, xmin, ymin, xmax, ymax;
{
	double xc, yc, tolsq, d2;
	double x[3], y[3];
	int yes, i, nref;
	ELMTPTR elmt;
	tolsq = tol*tol;
	do {
		nref = 0;
		FORALLELMTS(elmt)
			for(i=0; i<3; i++)
				DISTORT(elmt->neigh[i], x[i], y[i]);
			xc = (x[0] + x[1] + x[2])/3;
			yc = (y[0] + y[1] + y[2])/3;
			if(xc < xmin || xc > xmax || yc < ymin || yc > ymax){
				set_refine(elmt, 0);
				continue;
			}
			yes = 0;
			for(i=0; i<3; i++){
				d2 = (*sqdist)(elmt->neigh[i], elmt->neigh[(i+1)%3]);
				yes |= (d2 > tolsq);
			}
			set_refine(elmt, yes);
			nref += yes;
		NEXTELMT(elmt)
		refine();
		combine(&nref, isum, sizeof(int), 1);
		if(nref)
			swops();
	} while(nref);
}

picture()
/*
	To draw a picture demonstating the entire structural part of the mesh.
*/
{
#define OPPALIAS		1
	int source, i, nbytes, nneigh, aliastype, aliasproc, dest;
	int distortold(), (*keep)();
	double x[3], y[3];
	NODEPTR node, mailnode;
	ELMTPTR elmt;
	NODENEIGHPTR n, nstart;
	ALIASPTR ali, astart;
	double xc, yc;
	char *mailbag;
	FORALLNODES(node)
		if(node->b != (BDYPTR)NULL && (node->b->alias != (ALIASPTR)NULL
				|| node->b->oppalias != (ALIASPTR)NULL)){
			node->xold = node->yold = 0.0;
			nneigh = 0;
			n = nstart = node->n;
			do {
				elmt = n->elmt;
				for(i=0; i<3; i++)
					DISTORT(elmt->neigh[i], x[i], y[i]);
				xc = (x[0] + x[1] + x[2])/3;
				yc = (y[0] + y[1] + y[2])/3;
				++nneigh;
				node->xold += xc;
				node->yold += yc;
			} while((n = n->next) != nstart);
			DISTORT(node, x[0], y[0]);
			node->xold = 0.5*node->xold/nneigh + 0.5*x[0];
			node->yold = 0.5*node->yold/nneigh + 0.5*y[0];
		} else {
			DISTORT(node, node->xold, node->yold);
		}
	NEXTNODE(node)

	fmulti(stdout);
	color(4);

	keep = distort;
	distort = distortold;
	FORALLNODES(node)
		n = nstart = node->n;
		do {
			elmt = n->elmt;
			for(i=0; i<3; i++)
				DISTORT(elmt->neigh[i], x[i], y[i]);
			xc = (x[0] + x[1] + x[2])/3;
			yc = (y[0] + y[1] + y[2])/3;

			if(node->b != (BDYPTR)NULL && n == nstart)
				color(6);

			DISTORT(node, x[0], y[0]);
			move(x[0], y[0]);
			cont((x[0] + xc)/2, (y[0] + yc)/2);

			if(node->b != (BDYPTR)NULL && n == nstart)
				color(4);

		} while((n = n->next) != nstart);
	NEXTNODE(node)
	if(mesh_is_distributed)
		usendplot();
	else
		sendplot();

	FORALLNODES(node)
		if(node->b == (BDYPTR)NULL)
			continue;
		DISTORT(node, xc, yc);
		if(B_NODE(node)){
			color(5);
			marker(1, xc, yc, 0.01*(z_xmax-z_xmin));
		}
		if(B_OPP(node)){
			color(2);
			marker(3, xc, yc, 0.01*(z_xmax-z_xmin));
		}
		if(node->b->oppalias == (ALIASPTR)NULL)
			color(3);
		else if(node->b->oppalias->proc & HIBITS)
			color(2);
		else
			color(5);
		move(xc, yc);
		DISTORT(node->b->clock, xc, yc);
		cont(xc, yc);

		DISTORT(node, xc, yc);
		if(node->b->alias != (ALIASPTR)NULL){
			ali = astart = node->b->alias;
			do {
				mailbag = buffer;
				xpacket(mailbag, NODEPTR ) = ali->node;
				xpacket(mailbag, double) = xc;
				xpacket(mailbag, double) = yc;
				xpacket(mailbag, int) = ali->proc;
				xpacket(mailbag, int) = !OPPALIAS;
				dest = (ali->proc) & LOWBITS;
				if(add_to_out_buf(buffer, (int)(mailbag - buffer), 1, &dest) == -1)
					abort(301);			/* AOB alias failed in picture */
			} while((ali = ali->next) != astart);
		}
		if(node->b->oppalias != (ALIASPTR)NULL){
			mailbag = buffer;
			xpacket(mailbag, NODEPTR ) = node->b->oppalias->node;
			xpacket(mailbag, double) = xc;
			xpacket(mailbag, double) = yc;
			xpacket(mailbag, int) = node->b->oppalias->proc;
			xpacket(mailbag, int) = OPPALIAS;
			dest = (node->b->oppalias->proc) & LOWBITS;
			if(add_to_out_buf(buffer, (int)(mailbag - buffer), 1, &dest) == -1)
				abort(302);			/* AOB oppalias failed in picture */
		}
	NEXTNODE(node)
	if(mesh_is_distributed)
		usendplot();
	else
		sendplot();

	FORALLELMTS(elmt)
		drawoldelmt(elmt, procnum);
	NEXTELMT(elmt)
	distort = distortold;
	if(mesh_is_distributed)
		usendplot();
	else
		sendplot();

	if(crystal_router() == -1)
		abort(303);						/* CR failed in picture */

	while(source = ANY, get_message(buffer, &source, &nbytes) == 0){
		mailbag = buffer;
		mailnode = xpacket(mailbag, NODEPTR );
		xc = xpacket(mailbag, double);
		yc = xpacket(mailbag, double);
		aliasproc = xpacket(mailbag, int);
		aliastype = xpacket(mailbag, int);

		if(aliasproc & HIBITS)
			color(2);
		else
			color(5);
		if(aliastype == OPPALIAS)
			linemod(1);
		else
			linemod(0);
		DISTORT(mailnode, x[0], y[0]);
		move((x[0] + xc)/2, (y[0] + yc)/2);
		cont(xc, yc);
	}
	linemod(0);

	if(mesh_is_distributed)
		usendplot();
	else
		sendplot();

	fsingl(stdout);
	distort = keep;
	return(1);
}

elmt_contor(func, nlevel, min, max)
/*
	To put a color number on each element corresponding to 
	the function func.
*/
double (*func)();
int nlevel;
double min, max;
{
	ELMTPTR elmt;
	NODEPTR node;
#define LARGE		1.e12
	int i, shade;
	double temp, x, y, Max = -LARGE, Min = LARGE;
	if(nlevel < 1)
		return;
	if(Min >= Max){
		if(the_elmts != (ELMTPTR)NULL)
			Max = Min = (*func)(the_elmts);
		FORALLELMTS(elmt)
			temp = (*func)(elmt);
			if(temp > Max)	Max = temp;
			if(temp < Min)	Min = temp;
		NEXTELMT(elmt);
		combine(&Max, fmax, sizeof(double), 1);
		combine(&Min, fmin, sizeof(double), 1);
	} else {
		Min = min;
		Max = max;
	}
	printf("Min %f Max %f\n", Min, Max);fflush(stdout);
	Max -= (Max-Min)/nlevel;
	if(fabs(Max - Min) < 1.e-12){
		printf("Contor: flat function\n");fflush(stdout);
		return(1);
	}
	fmulti(stdout);
	FORALLELMTS(elmt)
		shade = BLACK + (WHITE-BLACK)*(((*func)(elmt) - Min)/(Max - Min));
		initpanel(shade, 0);
		for(i=0; i<3; i++){
			DISTORT(elmt->neigh[i], x, y);
			panelpoint(x, y);
		}
		endpanel();
	NEXTELMT(elmt)

	color(1);
	FORALLNODES(node)
		if(B_EDGE(node))
			draw(node, node->b->clock);
	NEXTNODE(node)
	if(mesh_is_distributed)
		usendplot();
	else
		sendplot();
	fsingl(stdout);
}

arrow(x, y, vx, vy)
double x, y, vx, vy;
{
	double dx, dy;
	move(x,  y);
	cont( x = x + vx, y = y + vy);
	dx = -0.2*vx;
	dy = -0.2*vy;
	vx = 0.6*dx + 0.4*dy;
	vy = -0.4*dx + 0.6*dy;
	cont(x+vx, y+vy);
	move(x, y);
	vx = 0.6*dx - 0.4*dy;
	vy = 0.4*dx + 0.6*dy;
	cont(x+vx, y+vy);
}
