/****************************************************/
/*    DIME: send_elmts.c                            */
/*    This code used to send a some elements        */
/*    from processor to processor for load          */
/*    balancing, download, and topological changes. */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/
#define DIMECODE
#include "dime.h"

#define DELETE 0					/* Three Message Types */
#define APPEND 1
#define OPPROC 2

#define RUBOUT		work_space.ni[0]	/* A node with this flag is dead */
#define OLDADDRESS	work_space.na[1]	/* Where from in sending processor */
#define NEWPROC		work_space.ei[0]	/* Processor to go to */

#define DELETEFLAG -1				
/*
	DELETEFLAG signifies that a node is almost dead, and could be freeed
	except that someone from another processor may reference it, and we 
	do not want a memory fault. It can also be used as the 'proc' member
	of an alias structure to mean that the alias is to an almost dead
	node.
*/
char coabuffer[COA_BUF], *coamailbag;

/*
	This code passes a set of triangles or elmts from processors to
	other processors. General plan:
		1. Make message buffer of elmts and nodes to be sent away
		2. Exchange messages
		3. Examine elmts and nodes which have arrived, and either
			hook-up to existing structure or malloc for them.
		   Sort elmts to anticlockwise order around nodes, and make
			pointers along boundary.
		   Make Change of Address message Buffer.
		4. Exchange Change of Address messages.
		5. Change addresses and delete defunct nodes.

	The message sent in 2. is a elmt plus its 3 nodes represented by a
	message with format:

	# aliases for node0, # aliases for node1, # aliases for node2
		for node0: 	node0, node0->x,  node0->y,
				 	alias0->node,  alias0->proc;
					alias1->node,  alias1->proc;
							....
		for node1: node1, node1->x,  node1->y,
					alias0->node,  alias0->proc;
							....
		for node2: node2, node2->x,  node2->y,
					alias0->node,  alias0->proc;
							....
*/
#define MAXSEND 4
send_elmts()
{
	ELMTPTR elmt;
	int too_many, totsend, or();
	do {
		totsend = 0;
		FORALLELMTS(elmt)
			totsend += (elmt->NEWPROC != procnum);
		NEXTELMT(elmt)
		if(too_many = (totsend >= MAXSEND))
			totsend = MAXSEND;
		combine(&too_many, or, sizeof(int), 1);
		send_some_elmts(totsend);
	} while(too_many);
}

send_some_elmts(number)
int number;
/*
	Manager function for load-balancing. It finds elmts to be moved,
	calls package_elmt() repeatedly to make message buffer, then uses
	crystal_router() to send them. process_mailbag() puts the elmts in their
	new homes, then uses crystal_router() to send change of address messages,
	which are sent to ALL processors. process_coa() deals with the
	changes of address.
*/
{
	NODEPTR node;
/*
	Set all workspace to reasonable values
*/
	FORALLNODES(node)
		node->OLDADDRESS = node;
		node->RUBOUT = 0;
	NEXTNODE(node)

	if(debug)
		fmulti(stdout);
	package_victims(number);
	if(debug)
		usendplot();

	if(crystal_router() == -1)
		abort(601);					/* CR1 failed in send_elmts */

	process_mailbag();

	if(crystal_router() == -1)
		abort(602);					/* CR2 failed in send_elmts */

	process_coa();
	clean_house();

	if(crystal_router() == -1)
		abort(603);					/* CR3 failed in send_elmts */

	process_coa();
	do_opposites();

	if(crystal_router() == -1)
		abort(604);					/* CR4 failed in send_elmts */

	process_opp();
	if(debug)
		usendplot();
	fsingl(stdout);
	printf("* ");
	fflush(stdout);
}

package_victims(number)
int number;
/*
	Elmts are supposed to be marked in NEWPROC as to their destination.
	This function removes them from the list of elmts and puts them in
	message buffer. Going thru a linked list deleting is SUBTLE.
*/
{
	ELMTPTR elmt, elmtnext;
	if(the_elmts != (ELMTPTR)NULL){
		elmtnext = the_elmts->next;
		do {
			elmt = elmtnext->last;
			if(elmt->NEWPROC != procnum && number > 0){
				package_elmt(elmt, elmt->NEWPROC);
				--number;
			}
			if(the_elmts == (ELMTPTR)NULL)
				return;
		} while((elmtnext = elmtnext->next) != the_elmts);

		if(the_elmts->last->NEWPROC != procnum && number > 0)
			package_elmt(the_elmts->last, the_elmts->last->NEWPROC);
	}
}

package_elmt(elmt, receiver)
ELMTPTR elmt;
int receiver;
/*
	To take a elmt and wrap it up with its 3 nodes, and put it in the
	mailbag to be delivered to processor 'receiver'.
*/
{
	int i, nalias;
	NODEPTR node;
	ALIASPTR ali, astart;
	NODENEIGHPTR remove();
	char *mailbag, *numproc[9];
	double xlab, ylab;
	if(receiver == procnum)
		return;
	if(debug)
		drawsmallelmt(elmt, receiver);

	mailbag = buffer;
/*
	Count the aliases of the 3 nodes and put these in as a message-header
*/
	for(i=0; i<3; i++){
		node = elmt->neigh[i];
		nalias = 1;				/* Include an alias back to parent node */
		if(node->b != (BDYPTR)NULL)
			if(node->b->alias != (ALIASPTR)NULL){
				ali = astart = node->b->alias;
				do {
					nalias++;
				} while((ali = ali->next) != astart);
			}
		xpacket(mailbag, int) = (B_NODE(node)) ? -nalias : nalias;
	}
	for(i=0; i<3; i++){					/* LOOP OVER 3 NODES */
		node = elmt->neigh[i];
/*
	Put in node address, position, boundary information
*/
		xpacket(mailbag, NODEPTR) = node;
		xpacket(mailbag, double) = node->x;
		xpacket(mailbag, double) = node->y;
/*
	Put in aliases for this node
*/
		if(node->b != (BDYPTR)NULL && node->b->alias != (ALIASPTR)NULL){
			ali = astart = node->b->alias;
			do {
				xpacket(mailbag, NODEPTR) = ali->node;
				xpacket(mailbag, int) = ali->proc;
			} while((ali = ali->next) != astart);
		}
/*
	Remember alias back to parent node
*/
		xpacket(mailbag, NODEPTR) = node;
		xpacket(mailbag, int) = procnum;
/*
	Sever connecting nodeneigh to the elmt which is being removed
*/
		sfree(nntype, remove(node, elmt));
/*
	If node has no elmts attached, mark it for deletion at the end of the day
*/
		if(node->n == (NODENEIGHPTR)NULL)
			node->RUBOUT = DELETEFLAG;
	}
/*
	Send user information last
*/
	for(i=0; i<3; i++){
		node = elmt->neigh[i];
		copybuf(node->user, mailbag, uns);
		mailbag += uns;
		if(B_NODE(node)){
			copybuf(node->b->user, mailbag, ubs);
			mailbag += ubs;
		}
	}
	copybuf(elmt->user, mailbag, ues);
	mailbag += ues;
/*
	Finally kill the elmt and put the assembled message into the mail
*/
	freeelmt(elmt);
	if(add_to_out_buf(buffer, (int)(mailbag - buffer), 
			1, &receiver) == -1)
		abort(605);						/* AOB failed in package_elmt */
}

process_mailbag()
/*
	Takes a set of packaged 3node-elmt combinations, and for each one,
	tries to hookup nodes with existing mesh, otherwise malloc for them.
	Also puts messages into the change of address mailbag.
*/
{
#define MAXALIAS		20
	int source, nbytes, i, ialias, nalias[3], mailproc[MAXALIAS], mailbagloop;
	int bdyorcut[3];
	double mailbagx, mailbagy, mailbagboun, mailbagp;
	NODEPTR mailbagnode, resident[3], node_alloc(), find_brother();
	NODEPTR mailalias[MAXALIAS], node, nextnode;
	ELMTPTR elmt, elmt_alloc();
	BDYPTR b_alloc();
	char *mailbag;
/*
	Get all the messages --
*/
	while(source = ANY, get_message(buffer, &source, &nbytes) == 0){
		mailbag = buffer;
/*
	Find out how many aliases the 3 nodes come with
*/
		for(i=0; i<3; i++){
			nalias[i] = xpacket(mailbag, int);
			if(bdyorcut[i] = (nalias[i] < 0))
				nalias[i] = -nalias[i];
		}
/*
	Now get the node information
*/
		for(i=0; i<3; i++){
			mailbagnode = xpacket(mailbag, NODEPTR);
			mailbagx = xpacket(mailbag, double);
			mailbagy = xpacket(mailbag, double);
/*
	resident[i] is the destination node for the incoming node, and may be
	either created anew, or resident[i] may have an alias which is
	the incoming node, or the incoming node may have an alias in this
	processor.
*/
			resident[i] = (NODEPTR)NULL;
			for(ialias=0; ialias<nalias[i]; ialias++){
				mailalias[ialias] = xpacket(mailbag, NODEPTR);
				mailproc[ialias] = xpacket(mailbag, int);
/*
	Does the incoming node have an alias in here? Lucky we didn't delete
	any nodes yet or there may be a memory fault.
*/
				if(mailproc[ialias] == procnum && 
								mailalias[ialias]->RUBOUT != DELETEFLAG)
					resident[i] = mailalias[ialias];
			}
/*
	Does anyone here have an alias which is the incoming node?
*/
			if(resident[i] == (NODEPTR)NULL)
				resident[i] = find_brother(mailbagnode, source);
/*
	I have no node corresponding to the incoming node. Must make a
	new struct node for it.
*/
			if(resident[i] == (NODEPTR)NULL){
				resident[i] = node_alloc();
				resident[i]->x = mailbagx;
				resident[i]->y = mailbagy;
				resident[i]->OLDADDRESS = mailbagnode;
				resident[i]->RUBOUT = 0;
				if(bdyorcut[i])
					resident[i]->b = b_alloc();
				else
					resident[i]->b = (BDYPTR)NULL;
/*
	Send in Change of Address mail: If you have an alias to the old address
	of the incoming node, append an alias to this node.
*/
				coamailbag = coabuffer;
				xpacket(coamailbag, int) = APPEND;
				xpacket(coamailbag, NODEPTR) = mailbagnode;
				xpacket(coamailbag, int) = source;
				xpacket(coamailbag, NODEPTR) = resident[i];
				if(add_to_out_buf(coabuffer, 
						(int)(coamailbag-coabuffer), 1<<curr_doc, allprocs) == -1)
					abort(606);			/* AOB APPEND in process_mailbag */
			}
/*
	Add to resident[i] any aliases which came with the incoming node,
	including the alias back to its parent.
*/
			for(ialias=0; ialias<nalias[i]; ialias++)
				add_alias(resident[i], mailalias[ialias], mailproc[ialias]);
		}
/*
	Make space for the elmt, and connect it to the resident[] array, and
	connect them to the elmt.
*/
		elmt = elmt_alloc();
		for(i=0; i<3; i++){
			elmt->neigh[i] = resident[i];
			add_nodeneigh(resident[i], elmt);
		}
		elmt->NEWPROC = procnum;
		for(i=0; i<3; i++){
			node = elmt->neigh[i];
			copybuf(mailbag, node->user, uns);
			mailbag += uns;
			if(bdyorcut[i]){
				node->b->is_on_bdy = 1;
				copybuf(mailbag, node->b->user, ubs);
				mailbag += ubs;
			}
		}
		copybuf(mailbag, elmt->user, ues);
		mailbag += ues;
		if(debug)
			drawelmt(elmt, procnum);
	}
/*
	If anyone is flagged to be deleted, put message in Change of Address
	mail that if you have an alias to this node, then delete the alias.
	We still don't actually delete the node, because we may still want
	to append aliases to it and not get memory fault.
*/
	FORALLNODES(node)
		if(node->RUBOUT == DELETEFLAG){
			coamailbag = coabuffer;
			xpacket(coamailbag, int) = DELETE;
			xpacket(coamailbag, NODEPTR) = node;
			if(add_to_out_buf(coabuffer, 
					(int)(coamailbag-coabuffer), 1<<curr_doc, allprocs) == -1)
				abort(607);				/* AOB DELETE in process_mailbag */
		}
	NEXTNODE(node)
}

clean_house()
/*
	To sort out the elmts of each node into anticlockwise order and establish
	the pointers along the boundary.
*/
{
	NODEPTR node, noden;
	NODENEIGHPTR nn;
	BDYPTR b_alloc();
	ALIASPTR ali, a, astart, an, anstart;
	char *mailbag;
	int i;
/*
	First, any nodes aliased to themselves are joined into one
*/
	again:
	FORALLNODES(node)
		if(node->RUBOUT == DELETEFLAG)
			continue;
		if(node->b != (BDYPTR)NULL && node->b->alias != (ALIASPTR)NULL){
			ali = astart = node->b->alias;
			do {
				if(ali->proc == procnum){
					kill_self_alias(node, ali->node);
/*
	We don't want this pair picked up again, so mark the alias as almost dead
*/
					ali->proc = DELETEFLAG;
					goto again;
				}
			} while((ali = ali->next) != astart);
		}
	NEXTNODE(node)
/*
	Now sort the elmts around each node into anticlockwise order, making a
	new node if necessary which will be aliased to the original.
*/
	FORALLNODES(node)
		if(node->RUBOUT == DELETEFLAG)
			continue;
		node->OLDADDRESS = node;
		sortout(node);
	NEXTNODE(node)
/*
	Now set the boundary pointers. Sortout() should have put the boundary
	part between node->n and node->n->next.
*/
	FORALLNODES(node)
		if(node->RUBOUT == DELETEFLAG)
			continue;
		if(!antik(node->n, node->n->next, node) || node->n == node->n->next){
			if(node->b == (BDYPTR)NULL)
				node->b = b_alloc();
			i = (indexc(node->n->elmt, node) + 2)%3;
			node->b->clock = node->n->elmt->neigh[i];
			node->n->elmt->neigh[i]->b->antik = node;
		} else 
			if(node->b != (BDYPTR)NULL){
				freeb(node->b);
				node->b = (BDYPTR)NULL;
			}
	NEXTNODE(node)
}

kill_self_alias(node1, node2)
/*
	node1 and node2 are in the same processor and aliases to each other.
	This function joins them up and frees node2.
*/
NODEPTR node1, node2;
{
	NODENEIGHPTR n, nstart;
	n = nstart = node2->n;
	do {
		n->elmt->neigh[indexc(n->elmt, node2)] = node1;
	} while((n = n->next) != nstart);

	n = node1->n->next;
	node1->n->next = node2->n->next;
	node2->n->next = n;
	node2->n = (NODENEIGHPTR)NULL;
	node2->RUBOUT = DELETEFLAG;
	coamailbag = coabuffer;
	xpacket(coamailbag, int) = DELETE;
	xpacket(coamailbag, NODEPTR) = node2;
	if(add_to_out_buf(coabuffer, (int)(coamailbag-coabuffer),
						1<<curr_doc, allprocs) == -1)
		abort(608);					/* AOB in kill_self_alias */
}

process_coa()
/*
	Process Change of Address mail. First deal with APPEND messages:
	If you have an alias to the old address, append an alias to the 
		new address.
	Next do DELETE messages:
	If you have an alias to this node, delete it.
*/
{
	int source, nbytes, coaflag, sendproc;
	NODEPTR oldnode, find_alias(), newnode;
	char *mailbag;
	while(source = ANY, get_message(buffer, &source, &nbytes) == 0){
		mailbag = buffer;
		coaflag = xpacket(mailbag, int);
		if(coaflag != APPEND)
			continue;

		oldnode = xpacket(mailbag, NODEPTR);
		sendproc = xpacket(mailbag, int);
		newnode = xpacket(mailbag, NODEPTR);
			append_alias(oldnode, sendproc, newnode, source);
	}
/*
	Have dealt with the APPEND messages. Now rewind the messages and get the
	DELETE messages.
*/
	cr_rewind();
	while(source = ANY, get_message(buffer, &source, &nbytes) == 0){
		mailbag = buffer;
		coaflag = xpacket(mailbag, int);
		if(coaflag != DELETE)
			continue;

		oldnode = xpacket(mailbag, NODEPTR);
		delete_alias(oldnode, source);
		delete_alias(oldnode, source | HIBITS);
/*
	Now we can actually delete a node with no fear that anyone will try
	to reference it.
*/
		if((source&LOWBITS) == procnum)
			freenode(oldnode);
	}
}

do_opposites()
{
/*
	For getting the opproc information: send our identity to anyone who may
	be interested.
*/
	int dest;
	NODEPTR node, noden;
	ALIASPTR a, astart, an, anstart;
	char *mailbag;
	FORALLNODES(node)
		if(node->b != (BDYPTR)NULL && node->b->oppalias != (ALIASPTR)NULL){
			sfree(atype, node->b->oppalias);
			node->b->oppalias = (ALIASPTR)NULL;
		}
	NEXTNODE(node)

	FORALLNODES(node)
		if(node->b == (BDYPTR)NULL)
			continue;
		noden = node->b->clock;
		if(node->b->alias == (ALIASPTR)NULL)
			continue;
		a = astart = node->b->alias;
		do {
			if(a->proc&HIBITS)
				node->b->is_on_bdy |= 2;
			if(noden->b->alias == (ALIASPTR)NULL)
				continue;
			an = anstart = noden->b->alias;
			do {
				if(a->proc == an->proc){
					mailbag = buffer;
					xpacket(mailbag, NODEPTR) = an->node;
					xpacket(mailbag, NODEPTR) = a->node;
					xpacket(mailbag, int) = (a->proc & HIBITS)?
						procnum | HIBITS : procnum;
					xpacket(mailbag, NODEPTR) = node;
					dest = (a->proc) & LOWBITS;
					if(add_to_out_buf(buffer, (int)(mailbag - buffer),
							1, &dest) == -1)
						abort(609);		/* AOB in do_opposites */
				}
			} while((an = an->next) != anstart);
		} while((a = a->next) != astart);
	NEXTNODE(node)
}

process_opp()
/*
	To get the messages about opposite elmts.
*/
{
	char *mailbag;
	int source, nbytes;
	NODEPTR n, nn, hisaddress;
	ALIASPTR alias_alloc();
	while(source = ANY, get_message(buffer, &source, &nbytes) == 0){
		mailbag = buffer;
		n = xpacket(mailbag, NODEPTR);
		nn = xpacket(mailbag, NODEPTR);
		source = xpacket(mailbag, int);
		hisaddress = xpacket(mailbag, NODEPTR);
		if(n->b != (BDYPTR)NULL && n->b->clock == nn){
			n->b->oppalias = alias_alloc();
			n->b->oppalias->proc = source;
			n->b->oppalias->node = hisaddress;
		}
	}
}

add_nodeneigh(node, elmt)
NODEPTR node;
ELMTPTR elmt;
/*
	Create a link from a node to a elmt
*/
{
	NODENEIGHPTR nodeneigh_alloc(), new;
	new = nodeneigh_alloc();
	new->elmt = elmt;
	if(node->n == (NODENEIGHPTR)NULL){
		node->n = new;
		new->next = new;
	} else {
		new->next = node->n->next;
		node->n->next = new;
	}
}

append_alias(oldnode, sendproc, newnode, receiver)
NODEPTR oldnode, newnode;
int sendproc, receiver;
/*
	Change of Address: sendproc::oldnode has spawned receiver::newnode
	1. Ignore nodes which are to be deleted
	2. If I am the new node, don't make an alias to myself
	3. If I am the old node, make the alias to the new node
	4. If I have an alias to the old node, append an alias to the new
*/
{
	ALIASPTR newa, alias_alloc(), ali, astart;
	NODEPTR node;
	FORALLNODES(node)
		if(node->RUBOUT == DELETEFLAG)
			continue;
		if(receiver == procnum && newnode == node)
			continue;

/* Here a parent gets the new address of its child  */
		if(sendproc == procnum && oldnode == node){
			add_alias(node, newnode, receiver);
			continue;
		}
		if(node->b == (BDYPTR)NULL || node->b->alias == (ALIASPTR)NULL)
			continue;

		ali = astart = node->b->alias;
		do {
			if(ali->node == oldnode && (ali->proc & LOWBITS) == sendproc)
				add_alias(node, newnode, (ali->proc & HIBITS) ? 
					receiver | HIBITS : receiver);
		} while((ali = ali->next) != astart);
	NEXTNODE(node)
}

NODEPTR find_brother(newnode, receiver)
NODEPTR newnode;
int receiver;
/*
	Asks if some node in this processor has an alias which points to
	receiver::newnode and returns one of the candidates.
*/
{
	NODEPTR node;
	ALIASPTR ali, astart;
	FORALLNODES(node)
		if(node->RUBOUT == DELETEFLAG ||
				 node->b == (BDYPTR)NULL || node->b->alias == (ALIASPTR)NULL)
			continue;
		ali = astart = node->b->alias;
		do {
			if(ali->node == newnode && ali->proc == receiver)
				return(node);
		} while((ali = ali->next) != astart);
	NEXTNODE(node)
	return((NODEPTR)NULL);
}

delete_alias(anode, aproc)
NODEPTR anode;
int aproc;
/*
	Removes all references to aproc::anode from the nodes in this processor.
	If there are then no aliases at all, either node->alias is NULL, or
	the entire b-structure is free'ed
*/
{
	NODEPTR node;
	ALIASPTR alast, anext, astart;
	FORALLNODES(node)
		if(node->b == (BDYPTR)NULL || node->b->alias == (ALIASPTR)NULL)
			continue;
		again : alast = astart = node->b->alias;
		do {
			anext = alast->next;
			if(anext == (ALIASPTR)NULL)
				abort(610);		/* Null alias in delete_alias */
			if((anext->node == anode && anext->proc == aproc) 
							|| anext->proc == DELETEFLAG){
				if(alast == anext){
					sfree(atype, anext);
					node->b->alias = (ALIASPTR)NULL;
					if(!B_NODE(node)){
						freeb(node->b);
						node->b = (BDYPTR)NULL;
					}
					break;
				} else {
					alast->next = anext->next;
/*
	If we destroy the alias that the node points to, make it point elsewhere
*/
					if(anext == astart)
						node->b->alias = alast;
					sfree(atype, anext);
					goto again;
				}
			}
		} while((alast = anext) != astart);
	NEXTNODE(node)
}
