/****************************************************/
/*    DIME: smooth.c                                */
/*    To apply a greedy-algorithm smoothing of      */
/*    processor boundaries after balancing.         */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/
#define DIMECODE
#include "dime.h"
#define NEWPROC		work_space.ei[0]	/* Processor to go to */

smooth()
/*
	To look at each elmt in turn and find how much boundary is with its
	own kind and how much is with other processors. If the boundary with
	some other processor is greater, send the elmt to that processor.
	To prevent parallel collisions, this is done in two stages, first
	where the procnum of an elmt may increase, then when it may decrease.
*/
{
	ELMTPTR elmt;
	NODEPTR node;
	int maxproc, procee[3], ican, i, ncan, candidate[3], on_boundary, flipper;
	double edge[3], meedge, d, maxedge;
/*
	flipper determines whether elmts may move up or down in procnum
*/
	for(flipper=0; flipper<2; flipper++){
		FORALLELMTS(elmt)
			on_boundary = 0;
			for(i=0; i<3; i++){
				node = elmt->neigh[i];
				candidate[i] = procnum;
				if(node->b == (BDYPTR)NULL || node->b->clock != elmt->neigh[(i+2)%3])
					continue;
				candidate[i] = -1;
				if(node->b->oppalias == (ALIASPTR)NULL)		continue;
				candidate[i] = node->b->oppalias->proc;
				on_boundary = 1;
			}
/*
	If the elmt has no border with another processor, it will not
	be considered for sending away.
*/
			if(!on_boundary)
				continue;
			meedge = 0.0;
			ncan = 0;
			for(i=0; i<3; i++){
				d = sqrt((*sqdist)(elmt->neigh[i], elmt->neigh[(i+1)%3]));
/*
	d is edge length, which can be added to the total for length bordering
	this processor or to the total for some other processor,
	which may already have some credit for this elmt, in which case it
	is in procee[].
*/
				if(candidate[i] == -1) 
					continue;
				for(ican=0; ican<ncan; ican++)
				if(candidate[i] == procee[ican]){
					edge[ican] += d;
					continue;
				}
				procee[ncan] = candidate[i];
				edge[ncan] = d;
				ncan++;
			}
			if(ncan == 0)
				continue;
			maxedge = 0.0;
			for(ican=0; ican<ncan; ican++){
				if(edge[ican] > maxedge){
					maxproc = procee[ican];
					maxedge = edge[ican];
				}
			}
		
			if(flipper ^ (procnum > maxproc))
				elmt->NEWPROC = maxproc;
			else
				elmt->NEWPROC = procnum;
		NEXTELMT(elmt);
/*
	Finally send away the elmts we have chosen. This sending happens twice
	for flipper being TRUE and FALSE.
*/
		send_elmts();
	}
}
