/*
Some routines to do parallel random number generation.  The idea is
to take a standard linear congruential algorithm, and have every
processor compute the Pth iterate of that algorithm (where P is the
number of processors.)  If the seeds are set up properly (staggered)
the processors leapfrog over one another, and it is just as good as
having used the basic algorithm on a sequential machine. See Chap 12
of the Book.
*/

#include "string.h"
#include <fcntl.h>

#define MULT	1103515245
#define ADD	12345
#define MASK	( 0x7fffffff )

static double _twoto31 = 2147483648.0 ;
static long AAA,BBB;
long randx;

pranset(seed)
/*
Unlike the sequential algorithm, you MUST call pranset, even if
you want a seed of 1.  Otherwise all processors will produce the
same sequence.
*/
long seed;
{
	int proc;

	AAA= 1;
	BBB = 0;
	for (proc=0;proc<nproc;++proc) {
	   AAA = (MULT * AAA) & MASK;
	   BBB = (MULT * BBB + ADD) & MASK;
	   /* 
	   if ((first) && (proc == procnum)) randx = (AAA*seed + BBB) & MASK;
	   */
	   if (proc == procnum) randx = (AAA*seed + BBB) & MASK;
	}
	/* else just use randx values from previous run */
	/* if (!first) randx = seed; */
}

float ranf()
/*
	Return a random float in [0, 1.0)
*/
{
	float retvalue;

	retvalue = randx / _twoto31;
	randx = (AAA*randx+BBB)&MASK;
	return(retvalue);
}

iranf()
/*
	Return a random int in [0, 2^31)
*/
{

	randx = (AAA*randx+BBB)&MASK;
	return(randx);
}

