/****************************************************/
/*    DIME: sort.c                                  */
/*    To sort the neighbors of a node into          */
/*    anticlockwise order, establishing aliases     */
/*    from itself to itself if necessary.           */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/

#define DIMECODE
#include "string.h"

#define DELETE 0		/* Three Message Types */
#define APPEND 1
#define OPPROC 2

#define RUBOUT	     work_space.ni[0]    /* A node with this flag is dead */
#define OLDADDRESS   work_space.na[1]    /* Where from in sending processor */
#define NEWPROC	     work_space.ei[0]    /* Processor to go to */

#define DELETEFLAG -1				

sortout(node)
/*
	To sort out the elmts of a node into anticlockwise order and establish
	a pointer along the boundary if necessary.
*/
NODEPTR node;
{
	NODENEIGHPTR n, nstart, ncurr, nn, before(); 
	BDYPTR b_alloc();
	NODEPTR node_alloc(), newnode, newlast, newnext;
	ALIASPTR newi, alias_alloc(), ali, astart;
	int i, found_bound, found_last;
	char coabuffer[COA_BUF], *coamailbag;
/*
	First move anticlockwise from nstart adding elmts which are 
	consecutively anticlockwise.
*/
	ncurr = nstart = node->n;
	do {
		n = ncurr;
		do {
			found_bound = 1;
			nn = n->next;
			if(antik(ncurr, nn, node)){
				if(n != ncurr){
					n->next = nn->next;
					nn->next = ncurr->next;
					ncurr->next = nn;
				}
				found_bound = 0;
				break;
			}
		} while((n = n->next) != nstart);
		if(found_bound)
			goto boundary;
	} while((ncurr = ncurr->next) != nstart);
	return;
/*
	We did not come all the way round. We must have hit a boundary, and will
	now have to go clockwise from nstart.
*/
boundary:
	do {
		n = ncurr;
		found_last = 1;
		do {
			nn = n->next;
			if(antik(nn, nstart, node)){
				if(nn->next != nstart){ 
					n->next = nn->next;
					before(nstart)->next = nn;
					nn->next = nstart;
				}
				nstart = nn;
				found_last = 0;
				break;
			}
		} while((n = n->next) != nstart);
	} while(!found_last);
/*
	We have the first and last (nstart to ncurr) neighbors of the node,
	and can thus make the boundary connection. 
*/
	node->n = ncurr;
	if(ncurr->next == nstart){
		return;
	}
#ifdef DISTRIBUTED
/*
	We still have not come full circle.
	Now for the case when the neighbor elmt of 'node' do not form a single
	consecutive set: must make a new node.
*/
	newnode = node_alloc();
	newnode->x = node->x;
	newnode->y = node->y;
	copybuf(node->user, newnode->user, uns);
	newnode->RUBOUT = 0;
	if(node->b != (BDYPTR)NULL){
		newnode->b = b_alloc();
		if(B_NODE(node)){
			newnode->b->is_on_bdy = node->b->is_on_bdy;
			copybuf(node->b->user, newnode->b->user, ubs);
		}
		if(node->b->alias != (ALIASPTR)NULL){
			ali = astart = node->b->alias;
			do {
				if(ali == astart)
					newnode->b->alias = newi = alias_alloc();
				else
					newi = (newi->next = alias_alloc());
				newi->node = ali->node;
				newi->proc = ali->proc;
			} while((ali = ali->next) != astart);
			newi->next = newnode->b->alias;
		}
	}
/*
	Process_coa expects a child to be aliased back to its parent, and so 
	this alias will not be added in...we must do it now.
*/
	add_alias(newnode, node, procnum);
/*
	Now to send the message to append 'newnode' to your aliases if you
	already have an alias to 'node'.
*/
	coamailbag = coabuffer;
	xpacket(coamailbag, int) = APPEND;
	xpacket(coamailbag, NODEPTR) = node->OLDADDRESS;
	xpacket(coamailbag, int) = procnum;
	xpacket(coamailbag, NODEPTR) = newnode;
	if(add_to_out_buf(coabuffer, (int)(coamailbag-coabuffer),
						1<<curr_doc, allprocs) == -1)
		abort(700);		/* AOB APPEND in sortout */
/*
	We split the elmt into the consecutively anticlockwise part that we 
	already found, for 'node', and all the rest for 'newnode'.
	Since 'newnode' has been put immediately before 'the_nodes', we can be
	sure that it will be processed.
*/
	newnode->n = ncurr->next;
	before(nstart)->next = newnode->n;
	ncurr->next = nstart;
/*
	Any elmt pointing to node should point to newnode
*/
	n = nstart = newnode->n;
	do {
		n->elmt->neigh[indexc(n->elmt, node)] = newnode;
	} while((n = n->next) != nstart);
#else
	abort(701);		/* Disjoint boundary in sortout */
#endif DISTRIBUTED
}


antik(n1, n2, node)
NODENEIGHPTR n1, n2;
NODEPTR node;
/*
	Returns 1 if c2 is the next elmt anticlockwise after c1 for node 'node.
*/
{
	int i1, i2;
	i1 = (indexc(n1->elmt, node) + 2)%3;
	i2 = (indexc(n2->elmt, node) + 1)%3;
	return(n1->elmt->neigh[i1] == n2->elmt->neigh[i2]);
}

