/* Read in parameters and initialize a bunch of stuff */

#include "strings.h"
#include <string.h>

char datatype[128];

init()
{
   int mode;

   /* Print out data in ASCII or BINARY form? */
   printf("Output data type (ASCII or BINARY)? ");
   get_string(datatype);
   if (strncmp("BINARY",datatype,1) == 0) {
     binary = TRUE;
   }
   else if (strncmp("ASCII",datatype,1) == 0) {
     binary = FALSE;
   }
   else {
     perror("Input data type not specified");
     exit(-2);
   }

   /* SINGL mesh (same for each processor) or
    * MULTI mesh (different for each processor)?
    */
   printf("Initial mesh (MULTI or SINGL)? ");
   get_string(datatype);
   if (strncmp("MULTI",datatype,1) == 0) {
     mode = MULTI;
   }
   else if (strncmp("SINGL",datatype,1) == 0) {
     mode = SINGL;
   }
   else {
     perror("Input mesh type not specified");
     exit(-2);
   }

   return mode;
}


reset()
{
   char datafile[128], ndir[128], sys_command[128];
   double action();
   int num_bins;

   fill_user_structs();

   printf("Dimension (current = %d)? ", dim); fflush(stdout);
   get_int(&dim);

#ifdef ALPHA
   printf("Alpha (current = %f)? ", alpha); fflush(stdout);
   get_double(&alpha);
#endif ALPHA

   printf("Area action coefficient? "); fflush(stdout);
   get_double(&lambdaA);

   printf("Edge action coefficient? "); fflush(stdout);
   get_double(&lambdaE);

   printf("Hit Size? "); fflush(stdout);
   get_double(&hit_size);

   tot_action = action();
#ifdef CUBIX
   fmulti(stdout);
#endif CUBIX
   printf("Action = %f\n", tot_action); fflush(stdout);
#ifdef CUBIX
   fsingl(stdout);
#endif CUBIX

   /* Initialize the random number generator */
   setseed();

   printf("Print out every measurement (Y or N)? ");
   get_string(datatype);
   if (strncmp("YES",datatype,1) == 0) {
      printout = TRUE;
   }
   else if (strncmp("NO",datatype,1) == 0) {
      printout = FALSE;
   }
   else {
      perror("Printout option not specified");
      exit(-3);
   }
   fflush(stdout);

   /* Read in the run parameters */
   printf("How many sweeps per measurement? "); fflush(stdout);
   get_int(&mes_sweeps);

   printf("Bin size? "); fflush(stdout);
   get_int(&binsize);

   printf("Number of bins? "); fflush(stdout);
   get_int(&num_bins);
   mes_steps = num_bins * binsize;

   printf("First Time ? " ); fflush(stdout);
   get_int(&first_time);

   printf("Back up mesh (Y or N)? ");
   get_string(datatype);
   if (strncmp("YES",datatype,1) == 0) {
      backup = TRUE;
   }
   else if (strncmp("NO",datatype,1) == 0) {
      backup = FALSE;
   }
   else {
      perror("Backup option not specified");
      exit(-2);
   }
   printf("\n");

   /* Initialize some other stuff */
   data_bytes = (NOBS+2)*sizeof(float);

   /* Name of directory to put the data in */
   sprintf(ndir, "N%d", nnode);
   sprintf(datafile, "%s/test_dir", ndir);
   /* Check that the directory is there */
   if((fdata = fopen(datafile, "w")) == (FILE *)NULL) {
      /* If not, try to make it */
      printf("Cannot find directory %s\n", ndir);
      printf("Trying to make new directory %s...", ndir); fflush(stdout);
      sprintf(sys_command, "mkdir N%d", nnode);
      system(sys_command); 
      /* Check again */
      if((fdata = fopen(datafile, "w")) == (FILE *)NULL) {
	 /* You're hosed - do it manually */
         printf("didn't work!\n");
         printf("Please create directory %s and try again\n", ndir);
	 fflush(stdout);
         exit(0);
      }
      else {
         printf("made directory OK\n\n"); fflush(stdout);
      }
   }
   fclose(fdata);

   if (printout) {
      /* Open the output data file */
      if(lambdaE == 0.0) {
         sprintf(datafile, "N%d/dataA%4.3f_%d", nnode, lambdaA, dim);
      }
      else {
         sprintf(datafile, "N%d/dataE%4.3f_%d", nnode, lambdaE, dim);
      }

#ifdef CFS
	sprintf(datafile, "/cfs/paulc/strings/%s", datafile);
#endif CFS

      if((fdata = fopen(datafile, "w")) == (FILE *)NULL) {
         printf("Cannot open datafile %s\n", datafile);
         exit(0);
      }
#ifdef CUBIX
      fmulti(fdata);
#endif CUBIX
   }

}

