/****************************************************/
/*    DIME: dime_util.c                             */
/*    These are functions used by many              */
/*    different DIME utilities.                     */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/

#define DIMECODE
#include "dime.h"

extern int procnum;
extern int nproc;

int *allprocs;
char *buffer;
int mesh_is_distributed;
int (*distort)() = NULL;
double (*sqdist)();

NODEPTR the_nodes;
ELMTPTR the_elmts;

int uns, ues, ubs;

/* This is originally from dime_setup.c */

dime_setup(argc, argv, pass_uns, pass_ues, pass_ubs, pass_mesh)
int argc;
char **argv;
int pass_uns, pass_ues, pass_ubs;
int pass_mesh;
/*
	Sets up some cube parameters and triangulates the users boundary.
*/
{
	int num,max_doc;
	int default_shall_we_swop();
	double default_sqdist();

	mesh_is_distributed = pass_mesh;

	uns = pass_uns;
	ues = pass_ues;
	ubs = pass_ubs;

	/* Find max_doc = log_{base 2} nproc */
	max_doc = 0;
	num = nproc;
	while (num > 1) {
	   num /= 2;
	   max_doc++;
	}

	num = 	sizeof(_ELMT) + ues + 
			3*(sizeof(_NODE) + uns + sizeof(_BDY) + ubs) +
			16*(4 + sizeof(ALIAS));
	buffer = malloc(num);
	allprocs = (int *)malloc((1<<max_doc)*sizeof(int));
	for(num=0; num<(1<<max_doc); num++) allprocs[num] = num;

	the_nodes = (NODEPTR)NULL;
	the_elmts = (ELMTPTR)NULL;

	sqdist = default_sqdist;

	printf("**************************************************\n");
	printf("*           This is DIME version 2.0             *\n");
	printf("**************************************************\n");
	if(max_doc == 0)
	printf("*           Running on 1 processor               *\n");
	else
	printf("*           Running on %4d processors           *\n", 1<<max_doc);
	printf("**************************************************\n");
	printf("*            Copyright Caltech 1988              *\n");
	printf("**************************************************\n");
	printf("User_node size %d bytes\n", uns);
	printf("User_elmt size %d bytes\n", ues);
	printf("User_bdy  size %d bytes\n", ubs);
	fflush(stdout);

}


NODENEIGHPTR indexnn(node, elmt)
/*
	Finds the nodeneigh around 'node' which points to 'elmt'
*/
ELMTPTR elmt;
NODEPTR node;
{
	NODENEIGHPTR n, nstart;
	n = nstart = node->n;
	do {
		if(n->elmt == elmt)
			return(n);
	} while((n = n->next) != nstart);
	abort(206);					/* indexnn failed */
}


add_neigh(elmt, this)
ELMTPTR elmt;
NODEPTR this;
/*
	Puts a nodeneigh into node pointing to elmt. The place is arbitrary
	so the node will need sorting eventually.
*/
{
	NODENEIGHPTR nodeneigh_alloc(), n;
	n = nodeneigh_alloc();
	n->elmt = elmt;
	if(this->n == (NODENEIGHPTR)NULL){
		n->next = n;
		this->n = n;
	} else {
		n->next = this->n->next;
		this->n->next = n;
	}
}


NODENEIGHPTR before(nn)
NODENEIGHPTR nn;
/*
	Returns the nodeneigh immediately BEFORE the argument.
*/
{
	NODENEIGHPTR n = nn;
	do {
		if(n->next == nn)
			return(n);
	} while((n = n->next) != nn);
}


indexc(elmt, node)
/*
	Returns the subscript of 'node' from 'elmt' -- that is
	we solve for 'in' the equation "elmt->neigh[in] = node".
*/
ELMTPTR elmt;
NODEPTR node;
{
	int in;
	NODEPTR n;
	for(in=0; in<3; in++)
		if(elmt->neigh[in] == node)
			return(in);
	printf("Error in Indexc (%f %f)\n", node->x, node->y);fflush(stdout);
/*	in=0;
	in=1/in;
*/
}


double area(elmt)
ELMTPTR elmt;
/*
	Returns the area of the elmt.
*/
{
	NODEPTR n0, n1, n2;
	double ax, ay, bx, by;
	n0 = elmt->neigh[0];
	n1 = elmt->neigh[1];
	n2 = elmt->neigh[2];
	ax = n0->x - n1->x;			ay = n0->y - n1->y;
	bx = n1->x - n2->x;			by = n1->y - n2->y;
	return(0.5*(ax*by - ay*bx));
}


insert(node, elmt, newneigh, newelmt)
/*
    In node 'node', insert the nodeneigh 'new' in position after
    the nodeneigh which points to the elmt 'elmt', and make it point
    to the elmt 'newelmt'.
*/
NODEPTR node;
ELMTPTR elmt, newelmt;
NODENEIGHPTR newneigh;
{
    NODENEIGHPTR nstart, n;
    n = nstart = node->n;
    do {
		if(n->elmt == elmt){
            newneigh->next = n->next;
            n->next = newneigh;
            newneigh->elmt = newelmt;
            return;
        }
    } while((n = n->next) != nstart);
    abort(207);						/* insert failed */
}


insert_before(node, elmt, newneigh, newelmt)
/*
    In node 'node', insert the nodeneigh 'new' in position before
    the nodeneigh which points to the elmt 'elmt', and make it point
    to the elmt 'newelmt'.
*/
NODEPTR node;
ELMTPTR elmt, newelmt;
NODENEIGHPTR newneigh;
{
    NODENEIGHPTR nstart, n;
    n = nstart = node->n;
    do {
		if(n->next->elmt == elmt){
            newneigh->next = n->next;
            n->next = newneigh;
            newneigh->elmt = newelmt;
            return;
        }
    } while((n = n->next) != nstart);
    abort(207);						/* insert failed */
}


NODENEIGHPTR remove(node, victim_elmt)
/*
    In node 'node', removes the nodeneigh which points to 'elmt', and
    returns a pointer to it for recycling or free()ing.
*/
NODEPTR node;
ELMTPTR victim_elmt;
{
    NODENEIGHPTR nstart, n, return_n;
    n = nstart = node->n;
    do {
        if(n->next->elmt == victim_elmt){
            if((return_n = n->next) == n)
				node->n = (NODENEIGHPTR)NULL;
			else {
				n->next = n->next->next;
				node->n = n;  	/* Mustn't remove the one 
						   the node points to */
			}
			return(return_n);
        }
    } while((n = n->next) != nstart);
    abort(208);					/* remove failed */
}


get_bdy_nn(node)
NODEPTR node;
/* 
	To make sure that if node is on the boundary, then node->n points
	to the elmt just before the actual boundary.
*/
{
	NODENEIGHPTR n, nstart;
	if(node->b == (BDYPTR)NULL)
		return;
	n = nstart = node->n;
	do {
		if(!antik(n, n->next, node)){
			node->n = n;
			return;
		}
	} while((n = n->next) != nstart);
	abort(209);			/* On boundary but not on boundary */
}


add_alias(node, newnode, receiver)
NODEPTR node, newnode;
int receiver;
/*
	Add an alias to this node which points to receiver::newnode, unless
	we already have such an alias.
*/
{
	ALIASPTR newa, alias_alloc(), ali, astart;
	BDYPTR b_alloc();
	if(newnode == node && receiver == procnum)
		return;
	if(node->b == (BDYPTR)NULL)
		node->b = b_alloc();
	else if(node->b->alias != (ALIASPTR)NULL){
		ali = astart = node->b->alias;
		do {
			if(ali->node == newnode && ali->proc == receiver)
				return;
		} while((ali = ali->next) != astart);
	}
	newa = alias_alloc();
	newa->node = newnode;
	newa->proc = receiver;
	if(!(node->b->alias != (ALIASPTR)NULL)){
		node->b->alias = newa;
		newa->next = newa;
	} else {
		newa->next = node->b->alias->next;
		node->b->alias->next = newa;
	}
}


double default_sqdist(n1, n2)
NODEPTR n1, n2;
{
	double dx, dy;
	dx = n1->x - n2->x;
	dy = n1->y - n2->y;
	return(dx*dx + dy*dy);
}


connected(n1, n2)
NODEPTR n1, n2;
{
	ELMTPTR elmt;
	int i;
	if(n1 == n2)
		return(1);

	FORALLNEIGH(n1, elmt)
		for(i=0; i<3; i++)
			if(elmt->neigh[i] == n2)
				return(1);
	NEXTNEIGH(n1, elmt)
	return(0);
}


/* The two routines sortout() and antik() are originally from sort.c */

sortout(node)
/*
	To sort out the elmts of a node into anticlockwise order and 
	establish a pointer along the boundary if necessary.
*/
NODEPTR node;
{
	NODENEIGHPTR n, nstart, ncurr, nn, before(); 
	BDYPTR b_alloc();
	NODEPTR node_alloc(), newnode, newlast, newnext;
	int i, found_bound, found_last;
/*
	First move anticlockwise from nstart adding elmts which are 
	consecutively anticlockwise.
*/
	ncurr = nstart = node->n;
	do {
		n = ncurr;
		do {
			found_bound = 1;
			nn = n->next;
			if(antik(ncurr, nn, node)){
				if(n != ncurr){
					n->next = nn->next;
					nn->next = ncurr->next;
					ncurr->next = nn;
				}
				found_bound = 0;
				break;
			}
		} while((n = n->next) != nstart);
		if(found_bound)
			goto boundary;
	} while((ncurr = ncurr->next) != nstart);
	return;
/*
	We did not come all the way round. We must have hit a boundary, 
	and will now have to go clockwise from nstart.
*/
boundary:
	do {
		n = ncurr;
		found_last = 1;
		do {
			nn = n->next;
			if(antik(nn, nstart, node)){
				if(nn->next != nstart){ 
					n->next = nn->next;
					before(nstart)->next = nn;
					nn->next = nstart;
				}
				nstart = nn;
				found_last = 0;
				break;
			}
		} while((n = n->next) != nstart);
	} while(!found_last);
/*
	We have the first and last (nstart to ncurr) neighbors of the node,
	and can thus make the boundary connection. 
*/
	node->n = ncurr;
	if(ncurr->next == nstart){
		return;
	}
	abort(701);		/* Disjoint boundary in sortout */
}


antik(n1, n2, node)
NODENEIGHPTR n1, n2;
NODEPTR node;
/*
	Returns 1 if c2 is the next elmt anticlockwise after c1 for node 'node'.
*/
{
	int i1, i2;
	i1 = (indexc(n1->elmt, node) + 2)%3;
	i2 = (indexc(n2->elmt, node) + 1)%3;
	return(n1->elmt->neigh[i1] == n2->elmt->neigh[i2]);
}


/* These are i/o routines originally from menu.c */

FILE *logfp = (FILE *)NULL, *infp = stdin;

get_double(f)
double *f;
{
	fscanf(infp, "%lf", f);
	printf("%f\n", *f);fflush(stdout);
	if(logfp != (FILE *)NULL){
		fprintf(logfp, "%f\n", *f);
		fflush(logfp);
	}
}

get_int(f)
int *f;
{
	fscanf(infp, "%d", f);
	printf("%d\n", *f);fflush(stdout);
	if(logfp != (FILE *)NULL){
		fprintf(logfp, "%d\n", *f);
		fflush(logfp);
	}
}

get_string(s)
char *s;
{
	fscanf(infp, "%s", s);
	if(logfp != (FILE *)NULL){
		fprintf(logfp, "%s\n", s);
		fflush(logfp);
	}
}


/*
Here are the standard combining functions declared in dime.h.
These are used in combining results from different processors.
*/

fmax(a1, a2, size)
double *a1, *a2; int size; { if(*a1 < *a2) *a1 = *a2; return(0); }

imax(a1, a2, size)
int *a1, *a2, size; { if(*a1 < *a2) *a1 = *a2; return(0); }

fmin(a1, a2, size)
double *a1, *a2; int size; { if(*a1 > *a2) *a1 = *a2; return(0); }

imin(a1, a2, size)
int *a1, *a2, size; { if(*a1 > *a2) *a1 = *a2; return(0); }

fsum(a1, a2, size)
double *a1, *a2; int size; { *a1 = *a2 = *a1 + *a2; return(0); }

isum(a1, a2, size)
int *a1, *a2, size; { *a1 = *a2 = *a1 + *a2; return(0); }

or(a1, a2, size)
int *a1, *a2, size; { *a1 = *a2 = (*a1) | (*a2); return(0); }

