
/* This program converts the raw data for the strings runs to 
   averages and errors, and works out the autocorrelations.
   The input can be multiple files of data.
*/

#include <math.h>
#include <string.h>
#include <stdio.h>

#define MAXPROCS 65    /* max number of processors (plus one for averages) */
#define NOBS  15       /* number of observables */
#define NAUTO 6        /* number of observables to print out 
			  autocorrelations for */
#define MAXNA 1000     /* max distance of autocorr measurements */
#define M     NOBS

double lastNEA[NOBS][MAXNA][MAXPROCS];
double EA[NOBS][MAXNA][MAXPROCS], EA2[NOBS][MAXNA][MAXPROCS]; 
double errEA[NOBS][MAXNA]; 

float energy[NOBS];
float maxenergy[NOBS][MAXPROCS], minenergy[NOBS][MAXPROCS];
double spheat[NOBS][MAXPROCS];
double errener[NOBS][MAXPROCS], errspheat[NOBS][MAXPROCS];
double avenergy[NOBS][MAXPROCS], avenersq[NOBS][MAXPROCS];
double avspheat[NOBS][MAXPROCS], avspheatsq[NOBS][MAXPROCS];
double avbinenersq[NOBS][MAXPROCS];
double partenergy[NOBS][MAXPROCS], partenersq[NOBS][MAXPROCS];
double errenersq[NOBS], errspheatsq[NOBS];
double binmeansq;

float lambda, mass;
int i,j,m,n, binsz, numbins, sites, nprocs;
int cntenergy, NA, sweeps, ndata;

FILE *infp, *fp, *fopen(), *fpax, *fpav;

main()
{
char *tmp_ptr, fileerror[128],filename[128];
char outname[128],dataname[128];
short next;

    /* Read in the input parameters */

    if ((infp = fopen("anal.in","r")) == NULL) {
       fprintf(stderr,"File anal.in not found\n");
       exit(-1);
    }
    fscanf(infp,"%d",&nprocs);   /* number of processors */
    fscanf(infp,"%d",&sites);    /* number of nodes in the mesh */
    fscanf(infp,"%d",&binsz);    /* size of bins for error estimates */
    fscanf(infp,"%d",&NA);       /* number of autocorrs measured */
    fscanf(infp,"%d",&sweeps);   /* sweeps per measurement */
    fscanf(infp,"%s",outname);   /* name of output file extension */

    /* initialize everything */

    cntenergy=0; 

for (m=0;m<NOBS;m++) {
for (n=0;n<MAXPROCS;n++) {

    maxenergy[m][n] = -100000.0; 
    minenergy[m][n] =  100000.0;

    avenergy[m][n]=0.;   avenersq[m][n]=0.; 
    avspheat[m][n]=0.;   avspheatsq[m][n]=0.; 
    partenergy[m][n]=0.; partenersq[m][n]=0.; 
    avbinenersq[m][n]=0; 
    errenersq[m]=0; 
    errspheatsq[m]=0; 
    
    for(i=0; i<NA; i++) {
       lastNEA[m][i][n] = 0.0;
       EA[m][i][n] = 0.0; 
       EA2[m][i][n] = 0.0;
    }
}}

    /* Open output file for averages */
    sprintf(dataname, "anal.%s", outname); 
    fpav = fopen(dataname,"w");

/* Loop over the data files */
while (fscanf(infp,"%s",filename) != EOF) {

    if ((fp = fopen(filename,"r")) == NULL) {
       fprintf(fpav, "Data file %s not found\n", filename);
       fflush(fpav);
       sprintf(fileerror, "Data file %s not found\n", filename);
       fprintf(stderr,fileerror);
       exit(-1);
    }

    /* Read in the data */
    for (n=1;n<=nprocs;n++) {
       /* check for EOF */
       while (fscanf(fp,"%f", &energy[0]) != EOF) {
          for (m=1;m<NOBS;m++) {
             fscanf(fp,"%f", &energy[m]);
          }
          if (n == 1) cntenergy++;
          /* Process the data */
          analyze_data(n);
       }
    }

} /* end of loop over input files */

     /* Work out averages and errors */
     numbins = cntenergy/binsz;
     for (m=0;m<M;m++) {
     for (n=1;n<=nprocs;n++) {
        avenergy[m][n] /= cntenergy;
        avenersq[m][n] /= cntenergy;
        avbinenersq[m][n] /= numbins;
        spheat[m][n] = avbinenersq[m][n] - avenergy[m][n]*avenergy[m][n];
        errener[m][n] = sqrt( spheat[m][n] / numbins );
        avspheat[m][n] /= numbins;
        avspheatsq[m][n] /= numbins;
        spheat[m][n] = avspheatsq[m][n] - avspheat[m][n]*avspheat[m][n];
        errspheat[m][n] = sqrt( spheat[m][n] / numbins );
        spheat[m][n] = avenersq[m][n] - avenergy[m][n]*avenergy[m][n];

        if (cntenergy >= NA) { 	 /* finish energy autocorr */
           finish_autocorr(cntenergy,m,n);
	}
     }}

     /* Average data over processors */
     if (nprocs > 1) {
        merge_data();
     }


/* Print out results */

for (n=0; n<=nprocs; n++) {

/* Open output file */
if (n == 0) {
   fpax = fpav;
}
else {
   sprintf(dataname, "anal.%s.N%d", outname, n); 
   fpax = fopen(dataname,"w");
}

/* Don't do averages for 1 processor */
if (nprocs == 1) n = 1;

if (n == 0) {
   fprintf(fpax,"average over %d processors\n", nprocs);
   ndata = cntenergy * nprocs;
}
else {
   fprintf(fpax,"processor %d of %d\n", n,nprocs);
   ndata = cntenergy;
}
fprintf(fpax,"nodes = %d\n",sites);
fprintf(fpax,"bin size = %d\n",binsz);
fprintf(fpax,"NA = %d\n",NA);
fprintf(fpax,"sweeps per measurement = %d\n",sweeps);
fflush(fpax);

fprintf(fpax, "Number of measurements = %d\n", ndata);
fprintf(fpax, "Number of bins = %d\n", numbins);
fprintf(fpax, "\n");
fflush(fpax);

/* Average of energies */
for (m=0;m<M;m++) {
fprintf(fpax,"ave (%2d) %12.5f  +/- %9.5f\n", m,avenergy[m][n],errener[m][n]);
}
fprintf(fpax,"\n");
fflush(fpax);

/* Specific heats */
for (m=0;m<M;m++) {
fprintf(fpax,"var (%2d) %12.5f  :  %12.5f  +/- %9.5f\n", 
	m,sites*spheat[m][n],sites*avspheat[m][n],sites*errspheat[m][n]);
}
fprintf(fpax,"\n");
fflush(fpax);

/* Maxima and minima of energies */
for (m=0;m<M;m++) {
fprintf(fpax,"%3d   min =%12.6f    max =%12.6f\n", 
	      m, minenergy[m][n],maxenergy[m][n]);
}
fprintf(fpax,"\n");
fflush(fpax);

/* Autocorrelations in the energies */
if (cntenergy >= NA) {
   /* Print out autocorrelation functions */
   print_autocorrs(0,NAUTO,n);
   print_autocorrs(NAUTO,2*NAUTO,n);
   /* Print out autocorrelation lengths */
   print_autolen(0,NAUTO,n);
   print_autolen(NAUTO,2*NAUTO,n);
}

  /* close output file */
  fclose(fpax);

} /* end of loop over processors */

  exit(0);

} /* end of main */



/* FUNCTIONS */


analyze_data(n)
int n;
{
   for (m=0;m<M;m++) {
      partenergy[m][n] += energy[m];
      partenersq[m][n] += energy[m] * energy[m];
      if (energy[m] < minenergy[m][n]) minenergy[m][n] = energy[m];
      if (energy[m] > maxenergy[m][n]) maxenergy[m][n] = energy[m];
      if (cntenergy % binsz == 0) {
	 binmeansq = (partenergy[m][n]*partenergy[m][n])/(binsz*binsz);
	 spheat[m][n] = partenersq[m][n]/binsz - binmeansq;
	 avspheat[m][n] += spheat[m][n];
	 avspheatsq[m][n] += spheat[m][n] * spheat[m][n];
         avenergy[m][n] += partenergy[m][n];
         avenersq[m][n] += partenersq[m][n];
	 avbinenersq[m][n] += binmeansq;
         partenergy[m][n] = 0.;
         partenersq[m][n] = 0.;
      }
      /* Auto-correlations */
      autocorrs(m,n);
   }
}


autocorrs(m,n)
int m,n;
{
   for (i=1;i<NA;i++) 
      lastNEA[m][i-1][n] = lastNEA[m][i][n];
   lastNEA[m][NA-1][n] = energy[m];
   for (i=0;i<NA;i++) {
      EA[m][i][n] += lastNEA[m][0][n] * lastNEA[m][i][n];
   }
}


finish_autocorr(count,m,n)
int count, m,n;
{
   for (j=0;j<NA;j++) {
      for (i=1;i<NA;i++)
	 lastNEA[m][i-1][n] = lastNEA[m][i][n];
      lastNEA[m][NA-1][n] = 0;
      for (i=0;i<NA;i++)
	 EA[m][i][n] += lastNEA[m][0][n] * lastNEA[m][i][n];
   }
   for (i=0;i<NA;i++) {
      EA[m][i][n] /= (count-i);
      EA[m][i][n] -= (avenergy[m][n] * avenergy[m][n]);
   }
}


print_autocorrs(first,last,n)
int first, last, n;
{
  fprintf(fpax,"Auto correlations\n");
  fprintf(fpax,"\n");
  for (i=0;i<NA;i++) {
     fprintf(fpax,"%3d  ", i); 
     for (m=first;m<last;m++) {
	if (EA[m][0][n] != 0.0)
           fprintf(fpax,"%9.5f  ", EA[m][i][n]/EA[m][0][n]);
	else
           fprintf(fpax,"  0.00000  ");
     }
     fprintf(fpax,"\n");
  }
  fprintf(fpax,"\n");
  fflush(fpax);
}


print_autolen(first,last,n)
int first, last, n;
{

  /* Autocorrelation lengths */
  fprintf(fpax,"Auto correlation length\n");
  fprintf(fpax,"\n");
  for (i=1;i<NA;i++) {
     fprintf(fpax,"%3d  ", i);
     for (m=first;m<last;m++) {
	if (EA[m][i][n] != 0.0)
           mass = EA[m][i-1][n]/EA[m][i][n];
	else
           mass = 0.0;
        if (mass > 0) mass = 1./log(mass);
        else          mass = 0.;
	if (mass < 0.0) mass = 0.0;
        fprintf(fpax,"%9.5f  ", mass);
     }
     fprintf(fpax,"\n");
  }
  fprintf(fpax,"\n");
  fflush(fpax);
}


merge_data()
{
   merge(avenergy,avenersq,errener,errenersq,nprocs);
   merge(avspheat,avspheatsq,errspheat,errspheatsq,nprocs);
   merge_autocorrs(nprocs);

   analyze(avenergy,avenersq,errenersq,errener,nprocs);
   analyze(avspheat,avspheatsq,errspheatsq,errspheat,nprocs);
   analyze_autocorrs(nprocs);
}


merge(average,avgesq,error,errsq,nprocs)
double average[NOBS][MAXPROCS],avgesq[NOBS][MAXPROCS];
double error[NOBS][MAXPROCS];
double errsq[NOBS];
int nprocs;
{
   int m,n;

   for (m=0;m<M;m++) {
   for (n=1;n<=nprocs;n++) {
      average[m][0] += average[m][n];
      avgesq[m][0] += average[m][n] * average[m][n];
      errsq[m] += error[m][n] * error[m][n];
   }}
}


merge_autocorrs(nprocs)
int nprocs;
{
   int i,m,n;

   for (m=0;m<M;m++) {
   for (i=0;i<NA;i++) {
   for (n=1;n<=nprocs;n++) {
      EA[m][i][0]  += EA[m][i][n];
      EA2[m][i][0] += EA[m][i][n] * EA[m][i][n];
   }}}
}


analyze(average,avgesq,errsq,error,nprocs)
double average[NOBS][MAXPROCS],avgesq[NOBS][MAXPROCS];
double error[NOBS][MAXPROCS];
double errsq[NOBS];
int nprocs;
{
   double diff;
   int m;

   for (m=0;m<M;m++) {
      average[m][0] /= nprocs;
      avgesq[m][0] /= nprocs;
      diff = avgesq[m][0] - average[m][0]*average[m][0];
      if (diff == 0.0) 
	 error[m][0] = 0.0;
      else
         error[m][0] = sqrt( diff / nprocs );
      errsq[m] /= (nprocs * nprocs);
   }
}


analyze_autocorrs(nprocs)
int nprocs;
{
   int i,m;

   for (m=0;m<M;m++) {
   for (i=0;i<NA;i++) {
      EA[m][i][0] /= nprocs;
      EA2[m][i][0] /= nprocs;
      errEA[m][i] = sqrt( (EA2[m][i][0] - EA[m][i][0]*EA[m][i][0]) / nprocs );
   }}
   for (m=0;m<M;m++) {
   for (i=0;i<NA;i++) {
      EA[m][i][0] /= EA[m][0][0];
      errEA[m][i] /= EA[m][0][0];
   }}
}


