/****************************************************/
/*    DIME: dime.h                                  */
/*    Header file containing structure definitions  */
/*    externs and macros.                           */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/

#include <stdio.h>
#include <math.h>

#ifdef CUBIX
#include <express.h>
#else
#define stdsys stderr
#endif

typedef struct node { 
	double x, y;			/* Node position */
	struct bdy *b;			/* Boundary structure or NULL */
	struct nodeneigh *n;		/* Pointer to list of neighbors */
	struct node *last, *next;	/* Double list of all nodes */
#ifndef DIMECODE
	struct user_node *user;
#else
	char *user;			/* Pointer to user data */
#endif
#define	NODE_WORK_SPACE	2
	union {
		double		 nf[NODE_WORK_SPACE];
		int		 ni[NODE_WORK_SPACE];
		struct node	*na[NODE_WORK_SPACE];
	} work_space;			/* 8 bytes for work space */
} _NODE, *NODEPTR;

typedef struct bdy {
	int is_on_bdy;
	struct node *clock, *antik;	/* Next along boundary (downstream) */
	struct alias *alias;		/* If processor boundary, here are aliases */
	struct alias *oppalias;		/* Pointer to opposite elmt */
#ifndef DIMECODE
	struct user_bdy *user;
#else
	char *user;			/* Pointer to user data */
#endif
#define	BDY_WORK_SPACE 2
	union {
		int		 bi[BDY_WORK_SPACE];
		struct node 	*ba[BDY_WORK_SPACE];
	} work_space;			/* 8 bytes of working space */
} _BDY, *BDYPTR;

typedef struct alias {
	int proc;			/* Procnum of sibling node */
	struct node *node;		/* Address of sibling in other processor */
	struct alias *next;		/* Next alias in linked list */
} ALIAS, *ALIASPTR;

typedef struct nodeneigh {
	struct elmt *elmt;		/* Pointer to this elmt */
	struct nodeneigh *next;		/* Next in neighbor list */
} NODENEIGH, *NODENEIGHPTR;

typedef struct elmt {
	struct node *neigh[3];		/* 3 neighbor nodes of this elmt */
	struct elmt *last, *next;	/* Double list of all elmts */
#ifndef DIMECODE
	struct user_elmt *user;
#else
	char *user;			/* Pointer to user data */
#endif

#define	ELMT_WORK_SPACE	1
	union {
		double	ef[ELMT_WORK_SPACE];
		int	ei[ELMT_WORK_SPACE];
	} work_space;			/* 8 bytes of work space */
} _ELMT, *ELMTPTR;

extern NODEPTR	the_nodes;
extern ELMTPTR	the_elmts;

extern int *allprocs;
extern char *buffer;
extern int uns, ues, ubs;
/*
extern char *malloc(), *smalloc();
*/
extern char *smalloc();
extern int or(), fsum(), isum(), fmax(), imax(), fmin(), imin();
extern int nntype, atype;		/* used for sfree'ing nodeneigh's */
extern int mesh_is_distributed;

extern int (*user_read)();
extern int (*user_write)();
extern double (*sqdist)();
extern int (*distort)();

#define ANY		999
#define LOWBITS		0xffff
#define HIBITS		0x10000

#define ELMT		1
#define NODE		2
#define BDY		3
#define MISC		4

#define MULTI		1
#define SINGL		0

#define remove	Remove		/* The name 'remove' conflicts with cubix */

#define DIME_SETUP(argc, argv, distributed) \
	dime_setup(argc, argv, sizeof(struct user_node),\
	sizeof(struct user_elmt), sizeof(struct user_bdy), distributed)

#define FORALLNODES(node) \
	if((node = the_nodes) != (NODEPTR)NULL){ do {

#define NEXTNODE(node) \
	} while((node = node->next) != the_nodes); }

#define FORALLELMTS(elmt) \
	if((elmt = the_elmts) != (ELMTPTR)NULL){ do {

#define NEXTELMT(elmt) \
	} while((elmt = elmt->next) != the_elmts); }

#define NODE_COMBINE(member, func, number) \
	{struct user_node dummy; \
	node_combine((char *)(&dummy.member) - (char *)(&dummy), \
	sizeof(dummy.member), number, func);}

#define BDY_COMBINE(member, func, number) \
	{struct user_bdy dummy; \
	bdy_combine((char *)(&dummy.member) - (char *)(&dummy), \
	sizeof(dummy.member), number, func);}

#define ELMT_EXCHANGE(from, to, number) \
	{struct user_elmt dummy; \
	elmt_exchange((char *)(dummy.from) - (char *)(&dummy), \
	(char *)(dummy.to  ) - (char *)(&dummy), \
	sizeof(dummy.from[0])*(number));}

#define B_EDGE(node) \
	(node->b != (BDYPTR)NULL && node->b->oppalias == (ALIASPTR)NULL)

#define B_NODE(node) \
	(node->b != (BDYPTR)NULL && (node->b->is_on_bdy&1))

#define B_OPP(node) \
	(node->b != (BDYPTR)NULL && (node->b->is_on_bdy&2))

#define FORALLNEIGH(node, elmt) \
	{NODENEIGHPTR n, nstart;	\
	n = nstart = node->n; do {	\
	elmt = n->elmt;

#define NEXTNEIGH(node, elmt)	\
	} while((n = n->next) != nstart);}

#define ELMT_DATA(c, member, type) \
	{struct user_elmt dummy; \
	elmt_data(c, (char *)(&dummy.member) - (char *)(&dummy), type);}

#define NODE_DATA(c, member, type) \
	{struct user_node dummy; \
	node_data(c, (char *)(&dummy.member) - (char *)(&dummy), type);}

#define BDY_DATA(c, member, type) \
	{struct user_bdy dummy; \
	bdy_data(c, (char *)(&dummy.member) - (char *)(&dummy), type);}

