/****************************************************/
/*    DIME: meshio.c                                */
/*    To read and write meshes & their data to disk */
/****************************************************/
/*        Roy Williams Copyright Caltech 1987       */
/****************************************************/

#define DIMECODE
#include "strings.h"

int (*user_read)();
int (*user_write)();

#define ELMTNO	work_space.ei[0]	/* Element number from input file */
#define NODENO	work_space.ni[0]	/* Node number from input file */
#define BDYNO	work_space.bi[0]

#define MAXSTRUCT	10000
#define MYSTUFF		(!mesh_is_distributed || currproc == procnum)

#define FRESH		4
#define NEWLINE		5

FILE *in;
#define isMPW	(1)

readmesh()
{
	char string[30];
	FILE *in;

	printf("Input Filename (without suffix)?\n");
	get_string(string);
	strcat(string, ".ct");
	if((in = fopen(string, "r")) == (FILE *)NULL)
		printf("Cannot open %s\n", string);
	else
		readct(in);
}

readct(in)
FILE *in;
/*
	To read in a definition of a coarse mesh to feed to dime. The tool
	coarse_tool will produce such a file with the extension .ct.
*/
{
	ELMTPTR elmt, elmt_alloc(), find_elmt();
	NODEPTR node, node_alloc(), find_node(), othernode;
	BDYPTR b_alloc();
	NODENEIGHPTR n, nstart;
	ALIASPTR ali, astart, alias_alloc();
	int i, ic, ielmt, state, ialias, oldstate, iuser, num, currproc;
	int dest, source, nbytes, type, othernum;
	int thisisbdynode;
	_NODE dummy_node;
	_ELMT dummy_elmt;
	_BDY  dummy_bdy;
	dummy_elmt.user = buffer;
	dummy_node.user = buffer;
	dummy_bdy.user = buffer;
/* 
	The input file should have a .ct extension
*/
	freemesh();

#ifdef CUBIX
	fsingl(in);
#endif CUBIX

	if(mesh_is_distributed)
		printf("Mesh is MULTI\n");
	else
		printf("Mesh is SINGL\n");

#ifdef CUBIX
#ifdef DEBUG
fmulti(stdout);
#endif DEBUG
#endif CUBIX

	(*user_read)(in, MISC, NULL);
	while((ic = getc(in)) != '\n');

	state = FRESH;
	while((ic = getc(in)) != EOF){
		switch(state){
		case NEWLINE:
			if(ic == 'E' || ic == 'N' || ic == '%')
				state = FRESH;
			else {
				state = oldstate;
				break;
			}
		case FRESH:
			switch(ic){
				case 'E':
					state = ELMT;
					fscanf(in, "%d", &num);
					currproc = num/MAXSTRUCT;
					if(MYSTUFF){
						elmt = elmt_alloc();
						elmt->ELMTNO = num%MAXSTRUCT;
#ifdef DEBUG
printf("(%d)E%d ", procnum, elmt->ELMTNO);
#endif
					}
#ifdef DEBUG
fflush(stdout);
#endif
					i = 0;
					break;
				case 'N':
					state = NODE;
					fscanf(in, "%d", &num);
					currproc = num/MAXSTRUCT;
					thisisbdynode = 0;
					if(MYSTUFF){
						node = node_alloc();
						node->NODENO = num%MAXSTRUCT;
#ifdef DEBUG
printf("(%d)N%d ", procnum, node->NODENO);
#endif
					}
#ifdef DEBUG
fflush(stdout);
#endif
					break;
				case '%':
					while((ic = getc(in)) != '\n');
					break;
				default:
					abort(911);	/* linestart without N, E or % */
			} break;
		case ELMT:
			switch(ic){
				case 'N':
					fscanf(in, "%d", &num);
					if(MYSTUFF){
						elmt->neigh[i] = (NODEPTR)(num%MAXSTRUCT);
#ifdef DEBUG
printf("(%d)N%d ", procnum, (int)elmt->neigh[i]);
#endif
						i++;
					}
#ifdef DEBUG
fflush(stdout);
#endif
					break;
				case '\n':
					(*user_read)(in, ELMT, 
						MYSTUFF ? (char *)elmt:(char *)(&dummy_elmt));
					while((ic = getc(in)) != '\n');
					oldstate = state;
					state = NEWLINE;
#ifdef DEBUG
printf("\n");
fflush(stdout);
#endif
					break;
			} break;
		case NODE:
			switch(ic){
				case 'B':
					thisisbdynode = 1;
					fscanf(in, "%d", &num);
					if(MYSTUFF){
						if(node->b == (BDYPTR)NULL)
							node->b = b_alloc();
						node->b->BDYNO = num%MAXSTRUCT;
#ifdef DEBUG
printf("(%d)B%d ", procnum, node->b->BDYNO);
#endif
					}
#ifdef DEBUG
fflush(stdout);
#endif
					break;
				case 'E':
					fscanf(in, "%d", &ielmt);
					if(MYSTUFF){
						add_neigh(ielmt%MAXSTRUCT, node);
#ifdef DEBUG
printf("(%d)E%d ", procnum, ielmt);
#endif
					}
#ifdef DEBUG
fflush(stdout);
#endif
					break;
				case 'a':
					thisisbdynode = 1;
					fscanf(in, "%d", &ialias);
					if(MYSTUFF){
						if(node->b == (BDYPTR)NULL)
							node->b = b_alloc();
						add_alias(node, ialias%MAXSTRUCT, ialias/MAXSTRUCT);
#ifdef DEBUG
printf("(%d)a%d ", procnum, ialias);
#endif
					}
#ifdef DEBUG
fflush(stdout);
#endif
					break;
				case 'o':
					thisisbdynode = 1;
					fscanf(in, "%d", &ialias);
					if(MYSTUFF){
						if(node->b == (BDYPTR)NULL)
							node->b = b_alloc();
						node->b->oppalias = alias_alloc();
						node->b->oppalias->node = (struct node *)(ialias%MAXSTRUCT);
						node->b->oppalias->proc = ialias/MAXSTRUCT;
#ifdef DEBUG
printf("(%d)o%d ", procnum, ialias);
#endif
					}
#ifdef DEBUG
fflush(stdout);
#endif
					break;
				case 'A':
					thisisbdynode = 1;
					fscanf(in, "%d", &ialias);
					if(MYSTUFF){
						if(node->b == (BDYPTR)NULL)
							node->b = b_alloc();
						add_alias(node, ialias%MAXSTRUCT, procnum|HIBITS);
#ifdef DEBUG
printf("(%d)A%d ", procnum, ialias);
#endif
					}
#ifdef DEBUG
fflush(stdout);
#endif
					break;
				case 'O':
					thisisbdynode = 1;
					fscanf(in, "%d", &ialias);
					if(MYSTUFF){
						if(node->b == (BDYPTR)NULL)
							node->b = b_alloc();
						node->b->oppalias = alias_alloc();
						node->b->oppalias->node = (struct node *)(ialias%MAXSTRUCT);
						node->b->oppalias->proc = procnum|HIBITS;
#ifdef DEBUG
printf("(%d)O%d ", procnum, ialias);
#endif
					}
#ifdef DEBUG
fflush(stdout);
#endif
					break;
				case '\n':
					(*user_read)(in, NODE, 
						MYSTUFF ? (char *)node:(char *)(&dummy_node));
					while((ic = getc(in)) != '\n');
					if(thisisbdynode){
						(*user_read)(in, BDY, 
							MYSTUFF ? (char *)node->b:(char *)(&dummy_bdy));
						while((ic = getc(in)) != '\n');
					}
					oldstate = state;
					state = NEWLINE;
#ifdef DEBUG
printf("\n");
fflush(stdout);
#endif
			} break;
		}
	}
#ifdef DEBUG
printf("\n(%d)End of file\n", procnum);
fflush(stdout);
#ifdef CUBIX
fsingl(stdout);
#endif CUBIX
#endif
	fclose(in);
	FORALLELMTS(elmt)
		for(i=0; i<3; i++)
			elmt->neigh[i] = find_node((int)elmt->neigh[i]);
	NEXTELMT(elmt)

#define OPPMESS		0
#define ALIMESS		1

	FORALLNODES(node)
		n = nstart = node->n;
		do {
			n->elmt = find_elmt((int)n->elmt);
		} while((n = n->next) != nstart);

		if(node->b == (BDYPTR)NULL)
			continue;

		node->b->clock = find_node(node->b->BDYNO);
	NEXTNODE(node)

	FORALLNODES(node)
		if(node->b != (BDYPTR)NULL)
			node->b->clock->b->antik = node;
		sortout(node);
	NEXTNODE(node)

	FORALLNODES(node)
		if(node->b != (BDYPTR)NULL){
			if(B_EDGE(node) || B_EDGE(node->b->antik))
				node->b->is_on_bdy |= 1;
			if(node->b->alias != (ALIASPTR)NULL)
				node->b->is_on_bdy |= 2;
		}
	NEXTNODE(node)

	fflush(stdout);
}


NODEPTR find_node(inode)
int inode;
/*
	Returns the address of the node with number inode
*/
{
	NODEPTR node;
	FORALLNODES(node)
		if(node->NODENO == inode)
			return(node);
	NEXTNODE(node)
	printf("Cannot find node %d\n", inode);
	abort(401);		/* find_node failed in readct */
}


ELMTPTR find_elmt(ielmt)
int ielmt;
{
	ELMTPTR elmt;
	FORALLELMTS(elmt)
		if(elmt->ELMTNO == ielmt)
			return(elmt);
	NEXTELMT(elmt)
	printf("Cannot find elmt %d\n", ielmt);
	abort(402);		/* find_elmt failed in readct */
}


writemesh()
{
	char string[30];
	FILE *out;
	printf("Filename (without suffix)?\n");
	get_string(string);
	strcat(string, ".ct");
	if((out = fopen(string, "w")) == (FILE *)NULL){
		printf("Cannot open %s\n", string);
		fflush(stdout);
	} else
		writect(out);
}


writect(out)
FILE *out;
{
	NODEPTR node;
	ELMTPTR elmt;
	ALIASPTR ali, astart;
	int i, ibdy, dest, num;

	(*user_write)(out, MISC, NULL);
	fflush(out);

#ifdef CUBIX
	if(mesh_is_distributed)
		fmulti(out);
#endif CUBIX

	i = MAXSTRUCT*procnum;
	FORALLELMTS(elmt)
		elmt->ELMTNO = i++;
	NEXTELMT(elmt)

	i = MAXSTRUCT*procnum;
	FORALLNODES(node)
		node->NODENO = i++;

		if(node->b == NULL)
			continue;
	NEXTNODE(node)

	FORALLELMTS(elmt)
		fprintf(out, "E%d: ", elmt->ELMTNO);
		for(i=0; i<3; i++)
			fprintf(out, "N%d ", elmt->neigh[i]->NODENO);
		fprintf(out, "\n");
		(*user_write)(out, ELMT, (char *)elmt);
	NEXTELMT(elmt)
	fflush(out);

	FORALLNODES(node)
		fprintf(out, "N%d: ", node->NODENO);

		if(node->b != (BDYPTR)NULL){
			fprintf(out, "B%d ", node->b->clock->NODENO);
		}

		FORALLNEIGH(node, elmt)
			fprintf(out, "E%d ", elmt->ELMTNO);
		NEXTNEIGH(node, elmt)

		fprintf(out, "\n");
		(*user_write)(out, NODE, (char *)node);
		if(node->b != (BDYPTR)NULL)
			(*user_write)(out, BDY, (char *)node->b);
	NEXTNODE(node)
	fclose(out);
}

