/*
 *  Do the Monte Carlo iterations. 
 *
 *  Measurements are taken every 'mes_sweeps' iterations. 
 *  The number of measurements is given by 'mes_steps', so the
 *  total number of Monte Carlo sweeps is mes_steps * mes_sweeps.
 *  After every 'binsize' measurements, the data is written to a file,
 *  and the mesh is backed up (if the 'backup' flag is set TRUE).
 *  This would also be the real binsize if we added averages and
 *  error analysis to the code, rather than just printing out the 
 *  data to be analysed later.
 */

#include "strings.h"

extern FILE *out;

sweeps()
{
   int sweep, mes_i, i;
   double metro_move(), flip_move();

   /* Initialize the averages */
   for (i=0; i<NOBS+2; i++) {
   	bin_ave[i] = 0.;
   	bin_ave2[i] = 0.;
   }
   if (first_time) {
   	mes_done = 0;
   	for(i=0; i<NOBS+2; i++) {
   		tot_ave[i] = 0.;
   		tot_ave2[i] = 0.;
   	}
   } else {
   	read_aves(&mes_done);
   }

   /* Do the Monte Carlo run */
   for(mes_i=0; mes_i<mes_steps; mes_i++){

	metro_success = flip_success = 0;

	/* Do the update */
	for(sweep=0; sweep<mes_sweeps; sweep++){
		metro_success += metro_move();
		flip_success  += flip_move();
	}
	metro_success /= mes_sweeps;
	flip_success  /= mes_sweeps;

	/* Do the measurements */
	avg_measure(act, act2);
	for(i=0; i<NOBS+2; i++) {
		bin_ave[i] += act[i];
		bin_ave2[i] += act2[i];
		tot_ave[i] += act[i];
		tot_ave2[i] += act2[i];
	}

	/* Write the data to a file */
	if (printout) {
	   print_data(act);
	}
 
	/* Bin the data and print out averages */
	if (((mes_done+1) % binsize) == 0) {
		for(i=0; i<NOBS+2; i++) {
			bin_ave[i] /= (double)binsize;
			bin_ave2[i] /= (double)binsize;
		}
		nnn = mes_done+1;
		if (binary) {
			binary_print_action();
		}
		else {
			ascii_print_action();
		}
		for(i=0; i<NOBS+2; i++) {
			bin_ave[i] = 0.;
			bin_ave2[i] = 0.;
		}

		/* Back up the mesh and the data and the rng seeds */
    		if (backup) {
			write_string();
               		write_seeds(seedsname,ptsname);
			nnn = mes_done+1;
			write_aves(&nnn);
    		}
	}
	mes_done++;
   }

}

