/*-------------------------------------------------------------------*
 *                                                                   *
 *  Strings simulation on dynamically triangulated random surfaces.  *  
 *                                                                   *
 *  Original code by Clive Baillie and Roy Williams, Caltech, 1989,  *
 *  using DIME (Distributed Irregular Mesh Environment) created by   *
 *  Roy Williams, Caltech, 1987.                                     *
 *                                                                   *
 *  Code revised by Paul Coddington and Leping Han,                  *
 *  Syracuse University, 1992.                                       *
 *                                                                   *
 *-------------------------------------------------------------------*/


#define MAIN

#include "strings.h"

int nnn = 1;
int dim = 3;
double hit_size = 0.5;
FILE *out = NULL;


main(argc, argv)
int argc;
char **argv;
{
   int mode;
   int reset(), init(), sweeps();
   int add_nodes(), read_string();
   int string_distort(), string_read(), string_write();
   double string_sqdist();

#ifdef CUBIX
   /* Set up the Express parameters: 
    * nproc   = number of processors
    * procnum = processor number 
    */
   struct nodenv nodedata;
   exparam(&nodedata);
   procnum = nodedata.procnum;
   nproc = nodedata.nprocs;
#else
   /* We only have one processor */
   nproc = 1;
   procnum = 0;
#endif

   /* Initialize some DIME things */
   mode = init();

   /* Initialize the DIME (triangular mesh) environment */
   DIME_SETUP(argc, argv, mode);
	
   /* Initialize DIME pointers to user functions */
   user_read = string_read;
   user_write = string_write;
   distort = string_distort;
   sqdist = string_sqdist;

   /* Read in and set up the triangular mesh */
   read_string();
   add_nodes();

   /* Read in various parameters and initialize some stuff */
   reset();

   /* Run the Metropolis update */
   sweeps();

   exit(0);
}

