
/* This program converts the raw data for the strings runs to 
   averages and errors, and works out the autocorrelations.
   The input can be multiple files of data.
*/

#include <math.h>
#include <string.h>
#include <stdio.h>

#define NOBS  15       /* number of observables */
#define NAUTO 6        /* number of observables to print out 
			  autocorrelations for */
#define MAXNA 1000     /* max distance of autocorr measurements */
#define M     NOBS

double lastNEA[NOBS][MAXNA], EA[NOBS][MAXNA], EA2[NOBS][MAXNA]; 

float energy[NOBS], spheat[NOBS];
float errener[NOBS], errspheat[NOBS];
float maxenergy[NOBS], minenergy[NOBS];
double avenergy[NOBS], avenersq[NOBS];
double avspheat[NOBS], avspheatsq[NOBS], avbinenersq[NOBS];
double partenergy[NOBS], partenersq[NOBS];
double binmeansq;

float lambda, mass;
int i,j,m,n, binsz, numbins, sites;
int cntenergy, NA, sweeps;

FILE *infp, *fp, *fopen(), *fpax;

main()
{
char *tmp_ptr, fileerror[128],filename[128];
char outname[128],dataname[128];
short next;

    if ((infp = fopen("anal.in","r")) == NULL) {
       fprintf(stderr,"File anal.in not found\n");
       exit(-1);
    }

    fscanf(infp,"%d",&sites);    /* size of nodes */
    fscanf(infp,"%d",&binsz);    /* size of bins for error estimates */
    fscanf(infp,"%d",&NA);       /* number of autocorrs measured */
    fscanf(infp,"%d",&sweeps);   /* sweeps per measurement */
    fscanf(infp,"%s",outname);   /* name of output file extension */

    /* initialize everything */

    cntenergy=0; 

for (m=0;m<NOBS;m++) {

    maxenergy[m] = -100000.0; 
    minenergy[m] =  100000.0;

    avenergy[m]=0.; avenersq[m]=0.; 
    avspheat[m]=0.; avspheatsq[m]=0.; 
    partenergy[m]=0.; partenersq[m]=0.; 
    avbinenersq[m]=0; 
    
    for(i=0; i<NA; i++) {
       lastNEA[m][i] = 0.0;
       EA[m][i] = 0.0; 
       EA2[m][i] = 0.0;
    }
}

    /* Open output file */
    sprintf(dataname, "analysis.%s", outname); 
    fpax = fopen(dataname,"w");

    fprintf(fpax,"nodes = %d\n",sites);
    fprintf(fpax,"bin size = %d\n",binsz);
    fprintf(fpax,"NA = %d\n",NA);
    fprintf(fpax,"sweeps per measurement = %d\n",sweeps);
    fprintf(fpax,"output file = analysis.%s\n\n",outname);
    fflush(fpax);

while (fscanf(infp,"%s",filename) != EOF) {

    if ((fp = fopen(filename,"r")) == NULL) {
       fprintf(fpax, "Data file %s not found\n", filename);
       fflush(fpax);
       sprintf(fileerror, "Data file %s not found\n", filename);
       fprintf(stderr,fileerror);
       exit(-1);
    }

    /* check for EOF */
    while (fscanf(fp,"%f", &energy[0]) != EOF) {
       /* Read in and process the data */
       for (m=1;m<NOBS;m++) {
          fscanf(fp,"%f", &energy[m]);
       }
       cntenergy++;
       analyze();
    }

} /* end of loop over input files */


     numbins = cntenergy/binsz;
     for (m=0;m<M;m++) {
        avenergy[m] /= cntenergy;
        avenersq[m] /= cntenergy;
        avbinenersq[m] /= numbins;
        spheat[m] = avbinenersq[m] - avenergy[m]*avenergy[m];
        errener[m] = sqrt( spheat[m] / numbins );
        avspheat[m] /= numbins;
        avspheatsq[m] /= numbins;
        spheat[m] = avspheatsq[m] - avspheat[m]*avspheat[m];
        errspheat[m] = sqrt( spheat[m] / numbins );
        spheat[m] = avenersq[m] - avenergy[m]*avenergy[m];

        if (cntenergy >= NA) { 	 /* finish energy autocorr */
           finish_autocorr(lastNEA,EA,avenergy,cntenergy,m);
	}
     }


/* Print out results */

fprintf(fpax, "Number of measurements = %d\n", cntenergy);
fprintf(fpax, "Number of bins = %d\n", numbins);
fprintf(fpax, "\n");

/* Average of energies */
for (m=0;m<M;m++) {
fprintf(fpax,"ave (%2d) %12.5f  +/- %9.5f\n", m,avenergy[m],errener[m]);
}
fprintf(fpax,"\n");
fflush(fpax);

/* Specific heats */
for (m=0;m<M;m++) {
fprintf(fpax,"var (%2d) %12.5f  :  %12.5f  +/- %9.5f\n", 
	m,sites*spheat[m],sites*avspheat[m],sites*errspheat[m]);
}
fprintf(fpax,"\n");
fflush(fpax);

/* Maxima and minima of energies */
for (m=0;m<M;m++) {
fprintf(fpax,"%3d   min =%12.6f    max =%12.6f\n", 
	      m, minenergy[m],maxenergy[m]);
}
fprintf(fpax,"\n");
fflush(fpax);

/* Autocorrelations in the energies */
if (cntenergy >= NA) {
   /* Print out autocorrelation functions */
   print_autocorrs(0,NAUTO);
   print_autocorrs(NAUTO,2*NAUTO);
   /* Print out autocorrelation lengths */
   print_autolen(0,NAUTO);
   print_autolen(NAUTO,2*NAUTO);
}

  /* close output file */
  fclose(fpax);

  exit(0);

} /* end of main */



/* FUNCTIONS */


analyze()
{
   for (m=0;m<M;m++) {
      partenergy[m] += energy[m];
      partenersq[m] += energy[m] * energy[m];
      if (energy[m] < minenergy[m]) minenergy[m] = energy[m];
      if (energy[m] > maxenergy[m]) maxenergy[m] = energy[m];
      if (cntenergy % binsz == 0) {
	 binmeansq = (partenergy[m]*partenergy[m])/(binsz*binsz);
	 spheat[m] = partenersq[m]/binsz - binmeansq;
	 avspheat[m] += spheat[m];
	 avspheatsq[m] += spheat[m] * spheat[m];
         avenergy[m] += partenergy[m];
         avenersq[m] += partenersq[m];
	 avbinenersq[m] += binmeansq;
         partenergy[m] = 0.;
         partenersq[m] = 0.;
      }
      /* Energy auto-correlations */
      /* autocorrs(lastNEA,EA,EA2,energy,m); */
      autocorrs(lastNEA,EA,energy,m);
   }
}


autocorrs(lastNA,A,value,m)
double lastNA[M][MAXNA], A[M][MAXNA];
float value[M];
int m;
{
   for (i=1;i<NA;i++) lastNA[m][i-1] = lastNA[m][i];
   lastNA[m][NA-1] = value[m];
   for (i=0;i<NA;i++) {
      A[m][i] += lastNA[m][0] * lastNA[m][i];
   }
}


finish_autocorr(lastNA,A,average,count,m)
double lastNA[M][MAXNA], A[M][MAXNA];
double average[M];
int count, m;
{
   for (j=0;j<NA;j++) {
      for (i=1;i<NA;i++)
	 lastNA[m][i-1] = lastNA[m][i];
      lastNA[m][NA-1] = 0;
      for (i=0;i<NA;i++)
	 A[m][i] += lastNA[m][0] * lastNA[m][i];
   }
   for (i=0;i<NA;i++) {
      A[m][i] /= (count-i);
      A[m][i] -= (average[m] * average[m]);
   }
}


print_autocorrs(first,last)
int first, last;
{
  fprintf(fpax,"Auto correlations\n");
  fprintf(fpax,"\n");
  for (i=0;i<NA;i++) {
     fprintf(fpax,"%3d  ", i); 
     for (m=first;m<last;m++) {
	if (EA[m][0] != 0.0)
           fprintf(fpax,"%9.5f  ", EA[m][i]/EA[m][0]);
	else
           fprintf(fpax,"  0.00000  ");
     }
     fprintf(fpax,"\n");
  }
  fprintf(fpax,"\n");
  fflush(fpax);
}


print_autolen(first,last)
int first, last;
{

  /* Autocorrelation lengths */
  fprintf(fpax,"Auto correlation length\n");
  fprintf(fpax,"\n");
  for (i=1;i<NA;i++) {
     fprintf(fpax,"%3d  ", i);
     for (m=first;m<last;m++) {
	if (EA[m][i] != 0.0)
           mass = EA[m][i-1]/EA[m][i];
	else
           mass = 0.0;
        if (mass > 0) mass = 1./log(mass);
        else          mass = 0.;
	if (mass < 0.0) mass = 0.0;
        fprintf(fpax,"%9.5f  ", mass);
     }
     fprintf(fpax,"\n");
  }
  fprintf(fpax,"\n");
  fflush(fpax);
}
